/************************************************************************/
/*									*/
/*	File:	malloc.c						*/
/*									*/
/*	Allocates a specified amount of heap memory.			*/
/*									*/
/*	Uses an optimized boundary tag algorithm with an in-use		*/
/*	overhead of one word per block.  This file is also included by	*/
/*	the other components of the memory allocation system (with	*/
/*	NOTMALLOC defined), to get the type and macro definitions.	*/
/*									*/
/*	Copyright (C) 1988-1990 by Logical Systems.			*/
/*									*/
/*	Last update:	2/20/90						*/
/*									*/
/************************************************************************/

#define		_MEM_ROUTINE_	1	/* Let "stdlib" know not to redefine */

#include	<stdlib.h>
#include	<conc.h>

#define	ALIGNMENT	4		/* Must be at least 4 */

/*
 *	Bit positions within "size" block overhead.
 */

#define	THISINUSE	1		/* Current block in use if TRUE */
#define	PREVINUSE	2		/* Previous block in use */

struct	header				/* Linkage structure for free memory */
	{
	size_t	size;			/* Size of block in bytes */
	struct	header	*left;		/* Previous block in list */
	struct	header	*right;		/* Next block in list */
	};

extern	struct	header	*_freelist;	/* Root of free list */
extern	struct	header	*_heapstart;	/* Points to start of heap */
extern	struct	header	*_heapend;	/* Points to the end of the heap */

#ifndef		NOTMALLOC		/* i.e. free, realloc, ... */

void	*
malloc(size)
	register size_t	size;
	{
	register size_t	bsize,diff;
	register struct	header	*hp;
	int	old_priority;

	if(size == 0)
		return (NULL);

	if(size <= ((size_t) (sizeof(struct header) + sizeof(struct header *) -
		sizeof(size_t))))
		{
		size = sizeof(struct header) + sizeof(struct header *);
		}
	else
		size += sizeof(size_t);	/* Size and tag overhead */

	if(size & (ALIGNMENT - 1))
		size = (size & (~(ALIGNMENT - 1))) + ALIGNMENT;

	old_priority = ProcToHigh();	/* Protect critical section */
	hp = _freelist->right;		/* Starting point in freelist search */
	do
		{
		if((bsize = hp->size & (~(THISINUSE | PREVINUSE))) >= size)
			{
			diff = bsize - size;
			if(diff < (size_t) (sizeof(struct header) +
				sizeof(struct header *)))
/*
 *	Allocate whole block.
 */
				{
				(hp->left)->right = hp->right;
				(hp->right)->left = hp->left;
				hp->size |= (THISINUSE | PREVINUSE);
				*((size_t *) ((char *) hp + bsize)) |=
					(THISINUSE | PREVINUSE);
				_freelist = hp->left;
				}
			else
/*
 *	Allocate upper part of block.
 */
				{
				hp->size = diff | PREVINUSE;
				*((struct header **) ((char *) hp + diff -
					sizeof(struct header *))) = hp;
				_freelist = hp;
				hp = (struct header *) ((char *) hp + diff);
				hp->size = size | THISINUSE;
				*((size_t *) ((char *) hp + size)) |=
					(THISINUSE | PREVINUSE);
				}
			if(old_priority)
				ProcToLow();	/* End of critical section */
			return((void *) (&hp->left));
			}
		hp = hp->right;
		} while(hp != _freelist->right);
/*
 *	We couldn't find anything in the free list, try extending the heap
 *	boundary (like "sbrk" for you traditionalists).
 */
	hp = _heapstart;
	_heapstart = (struct header *) ((char *) _heapstart + size);
	if(_heapstart >= _heapend)
		{
		_heapstart = hp;	/* Restore to original */
		hp = NULL;
		}
	else
		{
		_heapstart->size = (THISINUSE | PREVINUSE);
		hp->size &= PREVINUSE;
		hp->size |= (size | THISINUSE);
		hp = (struct header *) (&(hp->left));
		}
	if(old_priority)
		ProcToLow();		/* End of critical section */
	return ((void *) hp);
	}
#endif		/* NOTMALLOC */
