/*
 *	t4fp32.c
 *
 *	Contains the T800 floating point instruction emulation routines for
 *	use with 32 bit software floating point on the T414.
 *
 *	NOTES:
 *
 *	If both high and low priority processes use floating point, then the
 *	high priority process(es) must save and restore the simulated floating
 *	point evaluation stack via savefp() and restorefp().
 *
 *	All of the instruction emulation procedures are atomic with respect to
 *	timeslicing.  This means that one copy of the soft registers will work
 *	for all low priority processes without the need to save and restore for
 *	each context switch.
 *
 *	Most of the core 32 bit math code is based on INMOS supplied code.
 *	Remaining routines provided by Logical Systems and others.  Particular
 *	thanks to Scott Lloyd of CSA and Robert Eifrig of Tektronix.
 */

#include	<conc.h>

#ifndef	STATIC_LINK
FPstate	_FPstate = 0;			/* Current registers for FP */
#endif	/* !STATIC_LINK */

void
savefp(fp)
	FPstate *fp;
	{
	#pragma	asm
	.ldc	_FPstate
	ldl	[fp]
	.ldc	24
	move				/* Get a copy of current contents */
	#pragma	endasm
	}

void
restorefp(fp)
	FPstate	*fp;
	{
	#pragma	asm
	ldl	[fp]
	.ldc	_FPstate
	.ldc	24
	move				/* Restore previous copy of FP regs */
	ret

#ifdef	STATIC_LINK
	.align
	.pub	_FPstate
_FPstate
	.ds	24			;Simulated registers in code space
#endif	/* STATIC_LINK */
	
	.pub	?FPLDZEROSN
?FPLDZEROSN:				;Push 0.0
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	2			;FBreg
	ldl	0
	stnl	4			;FCreg
	ldl	0
	ldnl	0			;FAreg
	ldl	0
	stnl	2			;FBreg
	.ldc	0
	ldl	0
	stnl	0			;FAreg
	ajw	1
	ldl	3			;Restore integer evaluation stack
	ldl	2
	ldl	1
	ret

	.pub	?FPDUP
?FPDUP:					;Duplicate FAreg
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	2			;FBreg
	ldl	0
	stnl	4			;FCreg
	ldl	0
	ldnl	0			;FAreg
	ldl	0
	stnl	2			;FBreg
	ajw	1
	ldl	3			;Restore integer evaluation stack
	ldl	2
	ldl	1
	ret

	.pub	?FPREV
?FPREV:					;Exchange FAreg and FBreg
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	0			;FAreg
	ldl	0
	ldnl	2			;FBreg
	ldl	0
	stnl	0			;FAreg
	ldl	0
	stnl	2			;FBreg
	ajw	1
	ldl	3			;Restore integer evaluation stack
	ldl	2
	ldl	1
	ret

	.pub	?SN32TOFAREG
?SN32TOFAREG:				;Load 32 bit fp in A into FAreg
	.ldc	_FPstate
	stl	2
	ldl	1			;Number to convert
	ldl	2
	stnl	0			;Store into FAreg
	.ldc	0
	ldl	2
	stnl	6			;Set round mode to "Round to nearest"
	ret

	.pub	?FPLDNLSN
?FPLDNLSN:				;Load single precision from addr in A
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	2			;FBreg
	ldl	0
	stnl	4			;FCreg
	ldl	0
	ldnl	0			;FAreg
	ldl	0
	stnl	2			;FBreg
	ldl	2
	ldnl	0			;FAreg
	ldl	0
	stnl	0			;FAreg
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	4
	ldl	3
	ajw	1
	ret

	.pub	?FPSTNLSN
?FPSTNLSN:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	0			;FAreg
	ldl	2			;Store into address in A
	stnl	0
	ldl	0
	ldnl	2			;FBreg
	ldl	0
	stnl	0			;FAreg
	ldl	0
	ldnl	4			;FCreg
	ldl	0
	stnl	2			;FBreg
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	4			;C ==> B
	ldl	3			;B ==> A
	ajw	1
	ret

	.pub	?FPLDNLADDSN
?FPLDNLADDSN:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	3
	ldnl	0			;Contents pointed to by A
	stl	0
	ldl	1
	ldnl	6			;Add (==0) plus rounding mode
	ldl	1
	ldnl	0			;FAreg
	ldl	1			;&FAreg
	call	@r32op
	.ldc	0
	ldl	1
	stnl	6			;Set round mode to "Round to nearest"
	ldl	5			;C ==> B
	ldl	4			;B ==> A
	ajw	2
	ret

	.pub	?FPLDNLMULSN
?FPLDNLMULSN:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	3
	ldnl	0
	stl	0			;Contents of A
	ldl	1
	ldnl	6
	adc	2			;Multiply
	ldl	1
	ldnl	0			;FAreg
	ldl	1			;&FAreg
	call	@r32op
	.ldc	0
	ldl	1
	stnl	6			;Set round mode to "Round to nearest"
	ldl	5
	ldl	4
	ajw	2
	ret

	.pub	?FPUDIVBY2
?FPUDIVBY2:
	ajw	-2
	.ldc	_FPstate
	stl	1
	.ldc	0x40000000		;2.0
	stl	0
	ldl	1
	ldnl	6
	adc	3			;Divide
	ldl	1
	ldnl	0			;FAreg
	ldl	1
	call	@r32op
	.ldc	0
	ldl	1
	stnl	6			;Set round mode to "Round to nearest"
	ldl	5
	ldl	4			;Restore integer evaluation stack
	ldl	3
	ajw	2
	ret

	.pub	?FPUMULBY2
?FPUMULBY2:
	ajw	-2
	.ldc	_FPstate
	stl	1
	.ldc	0x40000000		;2.0
	stl	0
	ldl	1
	ldnl	6
	adc	2			;Multiply
	ldl	1
	ldnl	0			;FAreg
	ldl	1
	call	@r32op
	.ldc	0
	ldl	1
	stnl	6			;Set round mode to "Round to nearest"
	ldl	5
	ldl	4			;Restore integer evaluation stack
	ldl	3
	ajw	2
	ret

	.pub	?FPADD
?FPADD:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	1
	ldnl	0			;FAreg
	stl	0
	ldl	1
	ldnl	6			;Add code (==0) plus rouding mode
	ldl	1
	ldnl	2			;FBreg
	ldl	1
	call	@r32op
	.ldc	0
	cj	@ret2

	.pub	?FPSUB
?FPSUB:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	1
	ldnl	0			;FAreg
	stl	0
	ldl	1
	ldnl	6
	adc	1			;Subtract
	ldl	1
	ldnl	2			;FBreg
	ldl	1
	call	@r32op
	.ldc	0
	cj	@ret2

	.pub	?FPMUL
?FPMUL:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	1
	ldnl	0			;FAreg
	stl	0
	ldl	1
	ldnl	6
	adc	2			;Multiply
	ldl	1
	ldnl	2			;FBreg
	ldl	1
	call	@r32op
	.ldc	0
	cj	@ret2

	.pub	?FPDIV
?FPDIV:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	1
	ldnl	0			;FAreg
	stl	0
	ldl	1
	ldnl	6
	adc	3			;Divide
	ldl	1
	ldnl	2			;FBreg
	ldl	1
	call	@r32op

ret2:
	ldl	1
	ldnl	4			;FCreg
	ldl	1
	stnl	2			;FBreg
	.ldc	0
	ldl	1
	stnl	6			;Set round mode to "Round to nearest"
	ajw	2
	ldl	3
	ldl	2			;Restore integer evaluation stack
	ldl	1
	ret

;
;	32-bit floating-point operations for the T414.
;
;	r32op performs the operation "*r = x op y;".
;	op is formed by op = function + round_mode.
;
;	function = 0  +
;	function = 1  -
;	function = 2  *
;	function = 3  /
;
;	round_mode =  0  Round to Nearest
;	round_mode =  4  Round to Zero
;	round_mode =  8  Round to Plus Infinity
;	round_mode = 12  Round to Minus Infinity
;
;	r32op(r, x, op, y)
;	float	*r;
;	float	x;
;	int	op;
;	float	y;
;
;	Based on INMOS supplied code, decompiled by Scott Lloyd of CSA.
;	Modified for atomicity by Robert Eifrig of Tektronix.
;
	.align
L016:
	.dw	L036,L019,L019,L019
	.dw	L027,L037,L019,L019
	.dw	L027,L027,L038,L019
	.dw	L027,L027,L027,L039
	.dw	L036,L019,L019,L019
	.dw	L027,L037,L019,L019
	.dw	L027,L027,L038,L019
	.dw	L027,L027,L027,L039
	.dw	L040,L040,L041,L019
	.dw	L040,L046,L042,L019
	.dw	L041,L042,L042,L019
	.dw	L027,L027,L027,L039
	.dw	L043,L042,L042,L019
	.dw	L040,L044,L042,L019
	.dw	L040,L040,L045,L019
	.dw	L027,L027,L027,L039
r32op:	ajw	-9
	ldl	12			;op
	ldl	13			;y
	unpacksn
	stl	6
	stl	5
	ldl	11			;x
	unpacksn
	stl	7
	stl	0
	.ldc	L016
	wsub
	ldnl	0
	gcall
L046:
	ldl	0
	ldl	5
	add
	adc	-126
	stl	0
	.ldc	0
	ldl	7
	ldl	6
	lmul
	norm
	postnormsn
	roundsn
	ldl	11
	ldl	13
	xor
	mint
	and
	or
	ldl	10
	stnl	0
	ajw	9
	ret
L036:
	ldl	12
	eqc	1
	cj	@L017
	mint
L017:
	ldl	13
	xor
	ldl	11
	and
	mint
	and
	ldl	10
	stnl	0
	ajw	9
	ret
L044:
	ldl	7
	.ldc	0
	norm
	stl	3
	stl	7
	stl	2
	ldl	6
	.ldc	0
	norm
	stl	3
	stl	6
	adc	127
	ldl	0
	ldl	5
	sub
	add
	ldl	2
	sub
	stl	0
	ldl	7
	.ldc	1
	shr
	.ldc	0
	ldl	6
	ldiv
	rev
	eqc	0
	eqc	0
	norm
	postnormsn
	roundsn
	ldl	11
	ldl	13
	xor
	mint
	and
	or
	ldl	10
	stnl	0
	ajw	9
	ret
L038:
	ldl	12
	eqc	1
	cj	@L018
	mint
L018:
	ldl	13
	xor
	ldl	11
	xor
	mint
	and
	cj	@L019
	.ldc	524288
	stl	1
	.ldc	0			;Atomic branch
	cj	@L021
L039:
	ldl	6
	ldl	7
	gt
	cj	@L019
L027:
	ldl	12
	eqc	1
	cj	@L022
	mint
	ldl	13
	xor
	ldl	10
	stnl	0
	ajw	9
	ret
L022:
	ldl	13
	.ldc	0
	cj	@L020
L019:
	ldl	11
	.ldc	0
L020:
	rev
	ldl	12
	.ldc	1
	gt
	cj	@L023
	mint
	not
	and
	ldl	11
	ldl	13
L021:
	xor
	mint
	and
L023:
	or
	ldl	10
	stnl	0
	ajw	9
	ret
L040:
	ldl	11
	ldl	13
	xor
	mint
	and
	ldl	10
	stnl	0
	ajw	9
	ret
L042:
	ldl	11
	ldl	13
	xor
	mint
	and
	stl	1
	.ldc	0			;Atomic branch
	cj	@L024
L043:
	.ldc	4194304
	stl	1
	.ldc	0			;Atomic branch
	cj	@L024
L045:
	.ldc	2097152
	stl	1
	.ldc	0			;Atomic branch
	cj	@L024
L041:
	.ldc	1048576
	stl	1
L024:
	ldl	1
	.ldc	2139095040
	or
	ldl	10
	stnl	0
	ajw	9
	ret
L037:
	ldl	0
	ldl	5
	sub
	stl	4
	ldl	4
	.ldc	0
	gt
	cj	@L025
	.ldc	26
	ldl	4
	gt
	cj	@L019
	ldl	6
	.ldc	0
	ldl	4
	lshr
	stl	3
	stl	6
	.ldc	0			;Atomic branch
	cj	@L026
L025:
	ldl	4
	.ldc	-26
	gt
	cj	@L027
	ldl	7
	.ldc	0
	ldl	4
	sub
	.ldc	0
	rev
	lshr
	stl	3
	stl	7
	ldl	5
	stl	0
L026:
	ldl	12
	eqc	1
	cj	@L028
	mint
L028:
	ldl	13
	xor
	ldl	11
	xor
	mint
	and
	cj	@L029
	.ldc	0
	.ldc	0
	ldl	3
	ldiff
	stl	3
	stl	2
	ldl	2
	ldl	7
	ldl	6
	ldiff
	stl	4
	cj	@L030
	ldl	2
	ldl	6
	ldl	7
	ldiff
	ldl	3
	norm
	postnormsn
	roundsn
	ldl	12
	eqc	1
	cj	@L031
	mint
L031:
	ldl	13
	xor
	mint
	and
	or
	ldl	10
	stnl	0
	ajw	9
	ret
L030:
	ldl	4
	cj	@L032
	ldl	4
	ldl	3
	norm
	postnormsn
	roundsn
	ldl	11
	mint
	and
	or
L032:
	ldl	10
	stnl	0
	ajw	9
	ret
L029:
	.ldc	0
	ldl	6
	ldl	7
	lsum
	stl	7
	stl	2
	ldl	2
	cj	@L033
	ldl	0
	adc	1
	stl	0
	ldl	7
	ldl	3
	.ldc	1
	lshr
	stl	3
	mint
	or
	stl	7
	.ldc	0			;Atomic branch
	cj	@L034
L033:
	ldl	0
	eqc	1
	cj	9
	ldl	7
	mint
	and
	eqc	0
	cj	@L034
	.ldc	0
	stl	0
L034:
	ldl	0
	ldl	7
	ldl	3
	roundsn
	ldl	11
	mint
	and
	or
	ldl	10
	stnl	0
	ajw	9
	ret

	.pub	?FPGT
?FPGT:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	0			;FAreg
	ldl	0
	ldnl	2			;FBreg
	ldlp	4
	call	@r32gt
retbool:
	ldl	0
	ldnl	4			;FCreg
	ldl	0
	stnl	0			;FAreg
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	3			;B ==> A
	ldl	2			;A ==> B
	ldl	4			;Results
	ajw	1
	ret


;
;	32-bit floating-point comparison for the T414.
;
;	r32gt performs the operation "*r = x > y;".
;
;	r32gt(r, x, y)
;	int	*r;
;	float	x, y;
;
;	Based on INMOS supplied code, decompiled by Scott Lloyd of CSA.
;	Modified for atomicity by Robert Eifrig of Tektronix.
;
	.align
L215:
	.dw	0x7fffffff
r32gt:	ajw	-3
	.ldc	L215
	stl	2
	ldl	5
	ldl	2
	ldnl	0
	and
	stl	1
	ldl	6
	ldl	2
	ldnl	0
	and
	stl	0
	.ldc	0
	ldl	5
	ldl	6
	xor
	gt
	cj	@L202
	.ldc	0
	ldl	6
	gt
	cj	@L201
	ldl	1
	ldl	0
	sum
	eqc	0
	eqc	0
L201:	ldl	4
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L204
L202:	.ldc	0
	ldl	5
	gt
	cj	@L203
	ldl	0
	ldl	1
	gt
	ldl	4
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L204
L203:	ldl	1
	ldl	0
	gt
	ldl	4
	stnl	0
L204:	ajw	3
	ret

	.pub	?FPEQ
?FPEQ:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	0			;FAreg
	ldl	0
	ldnl	2			;FBreg
	ldlp	4
	call	@r32eq
	.ldc	0			;Atomic branch
	cj	@retbool


;
;	32-bit floating-point comparison for the T414.
;
;	r32eq performs the operation "*r = x == y;".
;
;	r32eq(r, x, y)
;	int	*r;
;	float	x, y;
;
;	Based on INMOS supplied code, decompiled by Scott Lloyd of CSA.
;	Modified for atomicity by Robert Eifrig of Tektronix.
;
	.align
L314:
	.dw	0x7fffffff
r32eq:	ajw	-3
	.ldc	L314
	stl	2
	ldl	5
	ldl	2
	ldnl	0
	and
	stl	1
	ldl	6
	ldl	2
	ldnl	0
	and
	stl	0
	ldl	5
	ldl	6
	diff
	eqc	0
	eqc	0
	cj	@L305
	ldl	1
	ldl	0
	sum
	eqc	0
	eqc	0
L305:	eqc	0
	ldl	4
	stnl	0
	ajw	3
	ret

	.pub	??U32TOR32
??U32TOR32:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	2			;FBreg
	ldl	0
	stnl	4			;FCreg
	ldl	0
	ldnl	0			;FAreg
	ldl	0
	stnl	2			;FBreg
	ldl	2
	ldnl	0			;*(int *)A
	.ldc	1			;Round
	ldl	0
	call	@r32fmu
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	4			;C ==> B
	ldl	3			;B ==> A
	ajw	1
	ret

	.pub	?FPI32TOR32
?FPI32TOR32:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	2			;FBreg
	ldl	0
	stnl	4			;FCreg
	ldl	0
	ldnl	0			;FAreg
	ldl	0
	stnl	2			;FBreg
	ldl	2
	ldnl	0			;*(unsigned int *)A
	.ldc	1			;Round
	ldl	0
	call	@r32fmi
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	4			;C ==> B
	ldl	3			;B ==> A
	ajw	1
	ret

;
;	32-bit floating-point conversion for the T414.
;
;	r32fmi converts an int to a float "*r = i;".
;	md specifies the ROUND or TRUNC mode.
;
;	md = 1  ROUND
;	md = 0  TRUNC
;
;	r32fmi(r, md, i)
;	float	*r;
;	int	md;
;	int	i;
;
;	Based on INMOS supplied code, decompiled by Scott Lloyd of CSA.
;	Modified for atomicity by Robert Eifrig of Tektronix.
;
r32fmi:	ajw	-6
	ldl	9
	eqc	0
	cj	@L407
	.ldc	0
	ldl	7
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L406
L407:
	.ldc	0
	ldl	9
	gt
	cj	@L408
	mint
	stl	0
	.ldc	0
	ldl	9
	diff
	stl	1
	.ldc	0			;Atomic branch
	cj	@L409
r32fmu:					;Entry point for unsigned conversion
	ajw	-6
L408:
	.ldc	0
	stl	0
	ldl	9
	stl	1
L409:
	ldl	1
	.ldc	0
	norm
	stl	5
	stl	1
	stl	4
	.ldc	32
	ldl	4
	sub
	adc	126
	stl	3
	ldl	3
	ldl	1
	.ldc	1
	shl
	.ldc	9
	lshr
	stl	2
	stl	3
	ldl	4
	.ldc	8
	gt
	cj	@L410
	.ldc	0			;Atomic branch
	cj	@L411
L410:
	ldl	1
	.ldc	255
	and
	stl	1
	ldl	8
	cj	@L411
	ldl	1
	.ldc	128
	and
	eqc	0
	cj	@L412
	.ldc	0			;Atomic branch
	cj	@L411
L412:
	ldl	1
	.ldc	127
	and
	ldl	2
	.ldc	1
	and
	or
	eqc	0
	cj	@L413
	.ldc	0			;Atomic branch
	cj	@L411
L413:
	ldl	2
	adc	1
	stl	2
L411:
	ldl	0
	ldl	2
	or
	ldl	7
	stnl	0
L406:
	ajw	6
	ret

	.pub	?FPSTNLI32
?FPSTNLI32:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl     0
	ldnl	0			;FAreg
	.ldc	0			;Truncate
	ldl	2			;A
	call	@r32toi
	ldl	0
	ldnl	2			;FBreg
	ldl	0
	stnl	0
	ldl	0
	ldnl	4			;FCreg
	ldl	0
	stnl	2			;FBreg
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	4			;C ==> B
	ldl	3			;B ==> A
	ajw	1
	ret

;
;	32-bit floating-point conversion for the T414.
;
;	r32toi converts a float to an int "*i = r;".
;	md specifies the ROUND or TRUNC mode.
;
;	md = 1  ROUND
;	md = 0  TRUNC
;
;	r32toi(i, md, r)
;	int	*i;
;	int	md;
;	float	r;
;
;	Based on INMOS supplied code, decompiled by Scott Lloyd of CSA.
;	Modified for atomicity by Robert Eifrig of Tektronix.
;
r32toi:	ajw	-4
	mint
	stl	0
	.ldc	0
	ldl	7
	ldl	0
	not
	and
	.ldc	9
	lshl
	stl	1
	stl	2
	ldl	2
	eqc	255
	cj	@L546
	seterr
	.ldc	0			;Atomic branch
	cj	@L547
L546:
	.ldc	126
	ldl	2
	gt
	cj	@L548
	.ldc	0
	ldl	5
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L547
L548:
	.ldc	127
	ldl	2
	gt
	cj	@L549
	ldl	6
	cj	@L550
	ldl	1
	eqc	0
	cj	@L551
L550:
	.ldc	0
	ldl	5
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L547
L551:
	ldl	7
	ldl	0
	and
	eqc	0
	cj	@L552
	.ldc	1
	ldl	5
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L547
L552:
	.ldc	-1
	ldl	5
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L547
L549:
	ldl	7
	eqc	-822083584
	cj	@L553
	ldl	0
	ldl	5
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L547
L553:
	ldl	2
	adc	-127
	stl	2
	.ldc	1
	stl	3
	ldl	2
	.ldc	30
	gt
	cj	@L554
	ldl	7
	eqc	-822083584
	cj	@L556
	.ldc	0			;Atomic branch
	cj	@L555
L556:
	seterr
L555:
	ldl	3
	ldl	1
	ldl	2
	adc	-32
	lshl
	stl	1
	stl	3
	.ldc	0			;Atomic branch
	cj	@L557
L554:
	ldl	3
	ldl	1
	ldl	2
	lshl
	stl	1
	stl	3
L557:
	ldl	6
	cj	@L558
	ldl	1
	ldl	0
	and
	eqc	0
	cj	@L559
	.ldc	0			;Atomic branch
	cj	@L558
L559:
	ldl	1
	ldl	0
	not
	and
	ldl	3
	.ldc	1
	and
	or
	eqc	0
	cj	@L560
	.ldc	0			;Atomic branch
	cj	@L558
L560:
	ldl	3
	adc	1
	stl	3
	ldl	3
	ldl	0
	and
	eqc	0
	cj	@L562
	.ldc	0			;Atomic branch
	cj	@L558
L562:
	seterr
L558:
	ldl	7
	ldl	0
	and
	eqc	0
	cj	@L561
	ldl	3
	ldl	0
	not
	and
	ldl	5
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L547
L561:
	.ldc	0
	ldl	3
	ldl	0
	not
	and
	diff
	ldl	5
	stnl	0
L547:
	ajw	4
	ret
	#pragma	endasm
	}

