/*
 *	t4fp64.c
 *
 *	Contains the T800 floating point instruction emulation routines for
 *	use with 64 bit software floating point on the T414.
 *
 *	NOTES:
 *
 *	If both high and low priority processes use floating point, then the
 *	high priority process(es) must save and restore the simulated floating
 *	point evaluation stack via savefp() and restorefp().
 *
 *	All of the instruction emulation procedures are atomic with respect to
 *	timeslicing.  This means that one copy of the soft registers will work
 *	for all low priority processes without the need to save and restore for
 *	each context switch.
 *
 *	Most of the core 64 bit math code is based on INMOS supplied code.
 *	Remaining routines provided by Logical Systems and others.  Particular
 *	thanks to Scott Lloyd of CSA and Robert Eifrig of Tektronix.
 */

#include	<conc.h>

#ifndef	STATIC_LINK
FPstate	_FPstate = 0;			/* Current registers for FP */
#endif	/* !STATIC_LINK */

static	double	two	=	2.0;

void
savefp(fp)
	FPstate *fp;
	{
	#pragma	asm
	.ldc	_FPstate
	ldl	[fp]
	.ldc	24
	move				/* Get a copy of current contents */
	#pragma	endasm
	}

void
restorefp(fp)
	FPstate	*fp;
	{
	#pragma	asm
	ldl	[fp]
	.ldc	_FPstate
	.ldc	24
	move				/* Restore previous copy of FP regs */
	ret

#ifdef	STATIC_LINK
	.align
	.pub	_FPstate
_FPstate
	.ds	24			;Simulated registers in code space
#endif	/* STATIC_LINK */

	.pub	?FPLDZEROSN
	.pub	?FPLDZERODB
?FPLDZEROSN:				;Push 0.0
?FPLDZERODB:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnlp	2			;src=&FBreg
	ldl	0
	ldnlp	4			;dst=&FCreg
	.ldc	8
	move
	ldl	0			;src=&FAreg
	ldl	0
	ldnlp	2			;dst=&FBreg
	.ldc	8
	move
	.ldc	0
	ldl	0
	stnl	0			;FAreg[0]
	.ldc	0
	ldl	0
	stnl	1			;FAreg[1]
	ajw	1
	ldl	3			;Restore integer evaluation stack
	ldl	2
	ldl	1
	ret

	.pub	?FPDUP
?FPDUP:					;Duplicate FAreg
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnlp	2			;src=&FBreg
	ldl	0
	ldnlp	4			;dst=&FCreg
	.ldc	8
	move
	ldl	0			;src=&FAreg
	ldl	0
	ldnlp	2			;dst=&FBreg
	.ldc	8
	move
	ajw	1
	ldl	3			;Restore integer evaluation stack
	ldl	2
	ldl	1
	ret

	.pub	?FPREV
?FPREV:					;Exchange FAreg and FBreg
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	0			;FAreg[0]
	ldl	0
	ldnl	2			;FBreg[0]
	ldl	0
	stnl	0			;FAreg[0]
	ldl	0
	stnl	2			;FBreg[0]
	ldl	0
	ldnl	1			;FAreg[1]
	ldl	0
	ldnl	3			;FBreg[1]
	ldl	0
	stnl	1			;FAreg[1]
	ldl	0
	stnl	3			;FBreg[1]
	ajw	1
	ldl	3			;Restore integer evaluation stack
	ldl	2
	ldl	1
	ret

	.pub	?FPLDNLDB
?FPLDNLDB:				;Load double precision from addr in A
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnlp	2			;src=&FBreg
	ldl	0
	ldnlp	4			;dst=&FCreg
	.ldc	8
	move
	ldl	0			;src=&FAreg
	ldl	0
	ldnlp	2			;dst=&FBreg
	.ldc	8
	move
	ldl	2			;src=A
	ldl	0			;dst=&FAreg
	.ldc	8
	move
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	4
	ldl	4
	ldl	3
	ajw	1
	ret

	.pub	?SN32TOFAREG
?SN32TOFAREG:				;Load 32 bit fp in A into FAreg
	.ldc	_FPstate
	stl	2
	.ldc	0
	ldl	2
	stnl	0			;Initialize FAreg[0] to be zero
	ldl	1			;Number to convert
	mint
	not
	and				;Mask off the sign bit
	stl	3			;Keep a version without sign bit
	ldl	3
	cj	@ldsnzero		;If zero
	.ldc	0x7F800000		;See if NaN or Inf
	ldl	3
	gt
	cj	@ldsnnan		;Jump if so
	ldl	3
	.ldc	(1024-128) << 20	;Exponent conversion factor
	stl	3
	.ldc	0
ldsnalign:
	.ldc	3
	lshr
	ldl	2
	stnl	0			;FAreg[0]
	ldl	3
	add
	ldl	1
	mint
	and				;Add the sign bit
	or
ldsnzero:
	ldl	2
	stnl	1			;FAreg[1]
	.ldc	0
	ldl	2
	stnl	6			;Set round mode to "Round to nearest"
	ret
ldsnnan:
	ldl	3
	.ldc	(2047-255) << 20	;NaN/Inf exponent conversion factor
	stl	3
	.ldc	0
	cj	@ldsnalign

	.pub	?FPLDNLSN
?FPLDNLSN:				;Load single precision from addr in A
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnlp	2			;src=&FBreg
	ldl	0
	ldnlp	4			;dst=&FCreg
	.ldc	8
	move
	ldl	0			;src=&FAreg
	ldl	0
	ldnlp	2			;dst=&FBreg
	.ldc	8
	move
	ldl	2
	ldnl	0			;Load from address in A
	call	@?SN32TOFAREG
	ldl	4
	ldl	4
	ldl	3
	ajw	1
	ret

	.pub	?FPSTNLDB
?FPSTNLDB:				;Store double precision to addr in A
	ajw	-2
	.ldc	_FPstate
	stl	0
	ldl	0			;src=&FAreg
	ldl	3			;dst=A
	.ldc	8
	move
	.ldc	0
	cj	@fpstnl

	.pub	?FPSTNLSN
?FPSTNLSN:				;Store single precision to addr in A
	ajw	-2
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnl	1			;FAreg[1]
	mint
	not
	and
	stl	1
	ldl	1
	cj	@stsnzero
	.ldc	0x7FF00000
	ldl	1
	gt				;See if NaN ir Inf
	cj	@stsnnan		;Jump if so
	ldl	1
	adc	(128-1024) << 20	;Change exponent bias (ignore overflow)
	ldl	0
	ldnl	0			;FAreg[0]
	.ldc	3
	lshl				;Re-align exponent-mantissa boundry
	rev
stsnzero:
	ldl	0
	ldnl	1
	mint
	and				;Replace the sign bit
	or
	ldl	3			;Store into address in A
	stnl	0
fpstnl:
	ldl	0
	ldnlp	2			;src=&FBreg
	ldl	0			;dst=&FAreg
	.ldc	8
	move
	ldl	0
	ldnlp	4			;src=&FCreg
	ldl	0
	ldnlp	2			;dst=&FBreg
	.ldc	8
	move
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	5			;C ==> C
	ldl	5			;C ==> B
	ldl	4			;B ==> A
	ajw	2
	ret
stsnnan:
	ldl	1
	adc	(255-2047) << 20	;Change exponent bias (ignore overflow)
	ldl	0
	ldnl	0			;FAreg[0]
	.ldc	3
	lshl				;Re-align exponent-mantissa boundry
	rev
	ldl	0
	ldnl	1
	mint
	and				;Replace the sign bit
	or
	ldl	3			;Store into address in A
	stnl	0
	.ldc	0
	cj	@fpstnl

	.pub	?FPLDNLADDDB
?FPLDNLADDDB:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	3			;A
	stl	0
	ldl	1
	ldnl	6			;Add (==0) plus rounding mode
	ldl	1			;&FAreg
	ldl	1			;&FAreg
	call	@_r64op
	.ldc	0
	ldl	1
	stnl	6			;Set round mode to "Round to nearest"
	ldl	5			;C ==> C
	ldl	5			;C ==> B
	ldl	4			;B ==> A
	ajw	2
	ret

	.pub	?FPLDNLADDSN
?FPLDNLADDSN:
	ajw	-4
	.ldc	_FPstate
	stl	3
	ldl	3
	ldnl	0			;FAreg[0]
	stl	1			;Temp[0]
	ldl	3
	ldnl	1			;FAreg[1]
	stl	2			;Temp[1]
	ldl	5
	ldnl	0			;Contents pointed to by A
	call	@?SN32TOFAREG
	ldlp	1
	stl	0
	ldl	3
	ldnl	6			;Add (==0) plus rounding mode
	ldl	3			;&FAreg
	ldl	3			;&FAreg
	call	@_r64op
	ldl	7			;C ==> C
	ldl	7			;C ==> B
	ldl	6			;B ==> A
	ajw	4
	ret

	.pub	?FPLDNLMULDB
?FPLDNLMULDB:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	3			;A
	stl	0			
	ldl	1
	ldnl	6
	adc	2			;Multiply
	ldl	1			;&FAreg
	ldl	1			;&FAreg
	call	@_r64op
	.ldc	0
	ldl	1
	stnl	6			;Set round mode to "Round to nearest"
	ldl	5
	ldl	5
	ldl	4
	ajw	2
	ret

	.pub	?FPLDNLMULSN
?FPLDNLMULSN:
	ajw	-4
	.ldc	_FPstate
	stl	3
	ldl	3
	ldnl	0			;FAreg[0]
	stl	1			;Temp[0]
	ldl	3
	ldnl	1			;FAreg[1]
	stl	2			;Temp[1]
	ldl	5
	ldnl	0			;Contents pointed to by A
	call	@?SN32TOFAREG
	ldlp	1
	stl	0
	ldl	3
	ldnl	6
	adc	2			;Multiply
	ldl	3			;&FAreg
	ldl	3			;&FAreg
	call	@_r64op
	ldl	7
	ldl	7
	ldl	6
	ajw	4
	ret

	.pub	?FPUDIVBY2
?FPUDIVBY2:
	ajw	-2
	.ldc	_FPstate
	stl	1
	.ldc	two			;&2.0
	stl	0
	ldl	1
	ldnl	6
	adc	3			;Divide
	ldl	1			;&FAreg
	ldl	1			;&FAreg
	call	@_r64op
	.ldc	0
	ldl	1
	stnl	6			;Set round mode to "Round to nearest"
	ldl	5
	ldl	4			;Restore integer evaluation stack
	ldl	3
	ajw	2
	ret

	.pub	?FPUMULBY2
?FPUMULBY2:
	ajw	-2
	.ldc	_FPstate
	stl	1
	.ldc	two			;&2.0
	stl	0
	ldl	1
	ldnl	6
	adc	2			;Multiply
	ldl	1			;&FAreg
	ldl	1			;&FAreg
	call	@_r64op
	.ldc	0
	ldl	1
	stnl	6			;Set round mode to "Round to nearest"
	ldl	5
	ldl	4			;Restore integer evaluation stack
	ldl	3
	ajw	2
	ret

	.pub	?FPADD
?FPADD:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	1			;&FAreg
	stl	0
	ldl	1
	ldnl	6			;Add code (==0) plus rouding mode
	ldl	1
	adc	8			;&FBreg
	ldl	1			;&FAreg
	call	@_r64op
	.ldc	0
	cj	@ret2

	.pub	?FPSUB
?FPSUB:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	1
	stl	0			;&FAreg
	ldl	1
	ldnl	6
	adc	1			;Subtract
	ldl	1
	adc	8			;&FBreg
	ldl	1			;&FAreg
	call	@_r64op
	.ldc	0
	cj	@ret2

	.pub	?FPMUL
?FPMUL:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	1
	stl	0			;&FAreg
	ldl	1
	ldnl	6
	adc	2			;Multiply
	ldl	1
	adc	8			;&FBreg
	ldl	1			;&FAreg
	call	@_r64op
	.ldc	0
	cj	@ret2

	.pub	?FPDIV
?FPDIV:
	ajw	-2
	.ldc	_FPstate
	stl	1
	ldl	1
	stl	0			;&FAreg
	ldl	1
	ldnl	6
	adc	3			;Divide
	ldl	1
	adc	8			;&FBreg
	ldl	1			;&FAreg
	call	@_r64op

ret2:
	ldl	1
	ldnlp	4			;src=&FCreg
	ldl	1
	ldnlp	2			;dst=&FBreg
	.ldc	8
	move
	.ldc	0
	ldl	1
	stnl	6			;Set round mode to "Round to nearest"
	ajw	2
	ldl	3
	ldl	2			;Restore integer evaluation stack
	ldl	1
	ret

;	64-bit floating-point operations for the T414.
;
;	_r64op performs the operation "*r = *x op *y;".
;	op is formed by op = function + round_mode.
;
;	function = 0  +
;	function = 1  -
;	function = 2  *
;	function = 3  /
;
;	round_mode =  0  Round to Nearest
;	round_mode =  4  Round to Zero
;	round_mode =  8  Round to Plus Infinity
;	round_mode = 12  Round to Minus Infinity
;
;	_r64op(r, x, op, y)
;	double *r;
;	double *x;
;	int op;
;	double *y;
;
;	Based on INMOS supplied code, decompiled by Scott Lloyd of CSA.
;	Modified for atomicity by Robert Eifrig of Tektronix.
;
	.align
L0:
	.dw	0x00000000
	.dw	0x7ff80000
	.dw	0x7ff00000
	.dw	0x00000000
	.dw	0x7ff40000
	.dw	0x00000000
	.dw	0x7ff10000
	.dw	0x00000000
	.dw	0x7ff20000
	.dw	0x7fefffff
	.pub	_r64op
_r64op:	ajw	-21
	.ldc	L0
	stl	20
	.ldc	0
	stl	19
	mint
	stl	2
	ldl	24
	.ldc	3
	and
	stl	18
	ldl	24
	.ldc	2
	shr
	stl	17
	ldl	18
	eqc	1
	cj	@L01
	ldl	25
	ldnl	1
	ldl	2
	xor
	stl	4
	.ldc	0			;Atomic branch
	cj	@L02
L01:	ldl	25
	ldnl	1
	stl	4
L02:	ldl	25
	ldnl	0
	stl	3
	ldl	23
	ldnl	1
	ldl	23
	ldnl	0
	.ldc	11
	lshl
	stl	13
	stl	14
	ldl	4
	ldl	25
	ldnl	0
	.ldc	11
	lshl
	stl	11
	stl	12
	ldl	23
	ldnl	1
	.ldc	20
	shr
	.ldc	2047
	and
	stl	16
	ldl	4
	.ldc	20
	shr
	.ldc	2047
	and
	stl	15
	.ldc	0
	stl	5
	.ldc	0
	stl	10
	.ldc	0
	stl	19
	ldl	16
	eqc	2047
	eqc	0
	cj	@L032
	ldl	15
	eqc	2047
	eqc	0
	cj	@L032
	ldl	16
	cj	@L03
	ldl	15
	cj	@L03
	.ldc	1
	stl	10
	ldl	14
	ldl	2
	or
	stl	14
	ldl	12
	ldl	2
	or
	stl	12
	.ldc	0			;Atomic branch
	cj	@L065
L03:	ldl	16
	ldl	14
	or
	ldl	13
	or
	eqc	0
	cj	@L015
	ldl	15
	ldl	12
	or
	ldl	11
	or
	eqc	0
	cj	@L011
	.ldc	2
	ldl	18
	gt
	cj	@L07
	ldl	17
	eqc	3
	cj	@L05
	.ldc	0
	stl	5
	ldl	23
	ldnl	1
	ldl	4
	or
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L05:	.ldc	0
	stl	5
	ldl	23
	ldnl	1
	ldl	4
	and
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L07:	ldl	18
	eqc	2
	cj	@L08
	ldl	23
	ldnl	1
	ldl	4
	xor
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L08:	ldl	20
	ldlp	5
	.ldc	8
	move
	.ldc	1
	stl	19
	.ldc	0			;Atomic branch
	cj	@L065
L011:	.ldc	2
	ldl	18
	gt
	cj	@L012
	ldl	4
	stl	6
	ldl	3
	stl	5
	.ldc	0			;Atomic branch
	cj	@L065
L012:	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L015:	ldl	15
	ldl	12
	or
	ldl	11
	or
	eqc	0
	cj	@L020
	.ldc	2
	ldl	18
	gt
	cj	@L016
	ldl	23
	ldnl	1
	stl	6
	ldl	23
	ldnl	0
	stl	5
	.ldc	0			;Atomic branch
	cj	@L065
L016:	ldl	18
	eqc	2
	cj	@L017
	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L017:	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	ldl	20
	ldnl	2
	or
	stl	6
	.ldc	1
	stl	19
	.ldc	0			;Atomic branch
	cj	@L065
L020:	.ldc	1
	stl	10
	ldl	16
	cj	@L021
	ldl	14
	ldl	2
	or
	stl	14
	.ldc	0			;Atomic branch
	cj	@L023
L021:	ldl	18
	.ldc	1
	gt
	cj	@L022
	ldl	14
	ldl	13
	norm
	stl	13
	stl	14
	stl	7
	.ldc	1
	ldl	7
	sub
	stl	16
	.ldc	0			;Atomic branch
	cj	@L023
L022:	.ldc	1
	stl	16
L023:	ldl	15
	cj	@L025
	ldl	12
	ldl	2
	or
	stl	12
	.ldc	0			;Atomic branch
	cj	@L065
L025:	ldl	18
	.ldc	1
	gt
	cj	@L026
	ldl	12
	ldl	11
	norm
	stl	11
	stl	12
	stl	7
	.ldc	1
	ldl	7
	sub
	stl	15
	.ldc	0			;Atomic branch
	cj	@L065
L026:	.ldc	1
	stl	15
	.ldc	0			;Atomic branch
	cj	@L065
L032:	.ldc	1
	stl	19
	ldl	14
	ldl	2
	not
	and
	stl	14
	ldl	12
	ldl	2
	not
	and
	stl	12
	ldl	16
	eqc	2047
	cj	@L047
	ldl	15
	eqc	2047
	cj	@L047
	ldl	14
	ldl	13
	or
	ldl	12
	or
	ldl	11
	or
	cj	@L038
	.ldc	0
	ldl	13
	ldl	11
	ldiff
	stl	7
	stl	9
	ldl	9
	ldl	14
	ldl	12
	ldiff
	stl	7
	stl	9
	.ldc	0
	ldl	7
	gt
	cj	@L034
	ldl	25
	ldlp	5
	.ldc	8
	move
	.ldc	0			;Atomic branch
	cj	@L035
L034:	ldl	23
	ldlp	5
	.ldc	8
	move
L035:	.ldc	2
	ldl	18
	gt
	cj	@L036
	.ldc	0			;Atomic branch
	cj	@L065
L036:	ldl	6
	ldl	2
	not
	and
	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	or
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L038:	.ldc	2
	ldl	18
	gt
	cj	@L039
	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	eqc	0
	cj	@L039
	ldl	23
	ldnl	1
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L039:	ldl	18
	eqc	2
	cj	@L041
	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	ldl	20
	ldnl	2
	or
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L041:	ldl	18
	eqc	3
	cj	@L042
	ldl	20
	ldnlp	3
	ldlp	5
	.ldc	8
	move
	.ldc	0			;Atomic branch
	cj	@L065
L042:	ldl	20
	ldnlp	5
	ldlp	5
	.ldc	8
	move
	.ldc	0			;Atomic branch
	cj	@L065
L047:	ldl	16
	eqc	2047
	cj	@L057
	.ldc	2
	ldl	18
	gt
	cj	@L049
	ldl	23
	ldlp	5
	.ldc	8
	move
	.ldc	0			;Atomic branch
	cj	@L065
L049:	ldl	14
	ldl	13
	or
	cj	@L050
	ldl	23
	ldnl	0
	stl	5
	ldl	4
	ldl	2
	and
	ldl	23
	ldnl	1
	xor
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L050:	ldl	18
	eqc	2
	cj	@L053
	ldl	15
	ldl	12
	or
	ldl	11
	or
	eqc	0
	cj	@L051
	ldl	20
	ldnlp	7
	ldlp	5
	.ldc	8
	move
	.ldc	0			;Atomic branch
	cj	@L065
L051:	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	ldl	20
	ldnl	2
	or
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L053:	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	ldl	20
	ldnl	2
	or
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L057:	ldl	12
	ldl	11
	or
	cj	@L060
	.ldc	2
	ldl	18
	gt
	cj	@L058
	ldl	25
	ldlp	5
	.ldc	8
	move
	.ldc	0			;Atomic branch
	cj	@L065
L058:	ldl	23
	ldnl	1
	ldl	2
	and
	ldl	4
	xor
	stl	6
	ldl	3
	stl	5
	.ldc	0			;Atomic branch
	cj	@L065
L060:	.ldc	2
	ldl	18
	gt
	cj	@L061
	ldl	4
	stl	6
	ldl	3
	stl	5
	.ldc	0			;Atomic branch
	cj	@L065
L061:	ldl	18
	eqc	2
	cj	@L064
	ldl	16
	ldl	14
	or
	ldl	13
	or
	eqc	0
	cj	@L062
	ldl	20
	ldnlp	7
	ldlp	5
	.ldc	8
	move
	.ldc	0			;Atomic branch
	cj	@L065
L062:	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	ldl	20
	ldnl	2
	or
	stl	6
	.ldc	0			;Atomic branch
	cj	@L065
L064:	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	stl	6
L065:	ldl	10
	eqc	0
	cj	@L071
	ldl	6
	ldl	22
	stnl	1
	ldl	5
	ldl	22
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L0139
L071:	.ldc	2
	ldl	18
	gt
	cj	@L0104
	ldl	16
	ldl	15
	sub
	stl	7
	ldl	7
	eqc	0
	cj	@L072
	.ldc	0
	stl	9
	.ldc	0			;Atomic branch
	cj	@L083
L072:	ldl	7
	.ldc	0
	gt
	cj	@L078
	ldl	7
	.ldc	42
	gt
	cj	@L075
	ldl	7
	.ldc	54
	gt
	cj	@L073
	.ldc	0
	stl	12
	.ldc	0
	stl	11
	.ldc	1
	stl	9
	.ldc	0			;Atomic branch
	cj	@L083
L073:	ldl	11
	stl	10
	ldl	12
	ldl	11
	ldl	7
	adc	-32
	lshr
	stl	9
	stl	11
	ldl	9
	ldl	10
	or
	stl	9
	.ldc	0
	stl	12
	.ldc	0			;Atomic branch
	cj	@L083
L075:	ldl	11
	.ldc	0
	ldl	7
	lshr
	stl	9
	stl	10
	ldl	7
	.ldc	32
	gt
	cj	@L076
	ldl	12
	.ldc	0
	ldl	7
	adc	-32
	lshr
	stl	10
	stl	15
	ldl	9
	ldl	10
	or
	stl	9
L076:	ldl	12
	ldl	11
	ldl	7
	lshr
	stl	11
	stl	12
	.ldc	0			;Atomic branch
	cj	@L083
L078:	.ldc	-42
	ldl	7
	gt
	cj	@L081
	.ldc	-54
	ldl	7
	gt
	cj	@L079
	.ldc	0
	stl	14
	.ldc	0
	stl	13
	.ldc	1
	stl	9
	ldl	15
	stl	16
	.ldc	0			;Atomic branch
	cj	@L083
L079:	ldl	13
	stl	10
	ldl	14
	.ldc	0
	ldl	7
	sub
	adc	-32
	ldl	10
	rev
	lshr
	stl	9
	stl	13
	ldl	9
	ldl	10
	or
	stl	9
	.ldc	0
	stl	14
	ldl	15
	stl	16
	.ldc	0			;Atomic branch
	cj	@L083
L081:	ldl	15
	stl	16
	ldl	13
	.ldc	0
	ldl	7
	sub
	.ldc	0
	rev
	lshr
	stl	9
	stl	10
	.ldc	-32
	ldl	7
	gt
	cj	@L082
	ldl	14
	.ldc	0
	ldl	7
	sub
	adc	-32
	.ldc	0
	rev
	lshr
	stl	10
	stl	15
	ldl	9
	ldl	10
	or
	stl	9
L082:	ldl	14
	.ldc	0
	ldl	7
	sub
	ldl	13
	rev
	lshr
	stl	13
	stl	14
L083:	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	eqc	0
	cj	@L090
	ldl	23
	ldnl	1
	ldl	2
	and
	stl	6
	.ldc	0
	ldl	11
	ldl	13
	lsum
	stl	13
	stl	10
	ldl	10
	ldl	12
	ldl	14
	lsum
	stl	14
	stl	10
	ldl	10
	cj	@L086
	ldl	16
	adc	1
	stl	16
	ldl	13
	ldl	9
	.ldc	1
	lshr
	stl	9
	stl	10
	ldl	14
	ldl	13
	.ldc	1
	lshr
	stl	13
	stl	14
	.ldc	0			;Atomic branch
	cj	@L0115
L086:	ldl	16
	eqc	1
	cj	@L0115
	ldl	14
	ldl	2
	and
	eqc	0
	cj	@L0115
	.ldc	0
	stl	16
	.ldc	0			;Atomic branch
	cj	@L0115
L090:	.ldc	0
	.ldc	0
	ldl	9
	ldiff
	stl	9
	stl	10
	ldl	10
	ldl	13
	ldl	11
	ldiff
	stl	5
	stl	7
	ldl	7
	ldl	14
	ldl	12
	ldiff
	stl	6
	stl	7
	ldl	7
	eqc	0
	cj	@L091
	ldl	6
	stl	14
	ldl	5
	stl	13
	ldl	23
	ldnl	1
	ldl	2
	and
	stl	6
	.ldc	0			;Atomic branch
	cj	@L092
L091:	ldl	4
	ldl	2
	and
	stl	6
	ldl	10
	ldl	11
	ldl	13
	ldiff
	stl	13
	stl	10
	ldl	10
	ldl	12
	ldl	14
	ldiff
	stl	14
	stl	10
L092:	ldl	14
	ldl	2
	and
	cj	@L093
	.ldc	0			;Atomic branch
	cj	@L0115
L093:	ldl	14
	ldl	13
	or
	ldl	9
	or
	eqc	0
	cj	@L096
	ldl	17
	eqc	3
	cj	@L094
	.ldc	0
	stl	16
	ldl	2
	stl	6
	.ldc	0			;Atomic branch
	cj	@L0115
L094:	.ldc	0
	stl	16
	.ldc	0
	stl	6
	.ldc	0			;Atomic branch
	cj	@L0115
L096:	ldl	16
	.ldc	1
	gt
	cj	@L099
	ldl	14
	ldl	13
	norm
	stl	10
	stl	14
	stl	7
	ldl	16
	ldl	7
	sub
	stl	16
	ldl	16
	.ldc	0
	gt
	cj	@L097
	ldl	13
	ldl	9
	ldl	7
	lshl
	stl	9
	stl	13
	.ldc	0			;Atomic branch
	cj	@L0115
L097:	ldl	14
	.ldc	1
	ldl	16
	sub
	.ldc	0
	rev
	lshr
	stl	10
	stl	14
	ldl	7
	.ldc	1
	ldl	16
	sub
	sub
	ldl	13
	rev
	ldl	9
	rev
	lshl
	stl	9
	stl	13
	.ldc	0
	stl	16
	.ldc	0			;Atomic branch
	cj	@L0115
L099:	.ldc	0
	stl	16
	.ldc	0			;Atomic branch
	cj	@L0115
L0104:	ldl	23
	ldnl	1
	ldl	4
	xor
	ldl	2
	and
	stl	6
	ldl	18
	eqc	2
	cj	@L0105
	ldl	16
	ldl	15
	add
	adc	-1022
	stl	16
	ldl	11
	stl	0
	ldl	12
	ldlp	9
	ldlp	13
	call	@rimul
	.ldc	0			;Atomic branch
	cj	@L0107
L0105:	ldl	16
	ldl	15
	sub
	adc	1023
	stl	16
	ldl	14
	ldl	13
	.ldc	1
	lshr
	stl	13
	stl	14
	.ldc	0
	stl	9
	ldlp	11
	stl	1
	ldlp	9
	stl	0
	ldlp	13
	ldlp	14
	ldlp	14
	call	@ridiv
	.ldc	0
	stl	10
	ldlp	11
	stl	1
	ldlp	10
	stl	0
	ldlp	9
	ldlp	13
	ldlp	13
	call	@ridiv
	ldl	10
	eqc	0
	cj	@L0106
	.ldc	0			;Atomic branch
	cj	@L0107
L0106:	ldl	9
	.ldc	1
	or
	stl	9
L0107:	ldl	14
	ldl	13
	norm
	stl	10
	stl	14
	stl	7
	ldl	7
	.ldc	32
	gt
	cj	@L0109
	ldl	9
	.ldc	64
	ldl	7
	sub
	shr
	ldl	14
	or
	stl	14
L0109:	ldl	13
	ldl	9
	ldl	7
	lshl
	stl	9
	stl	13
	ldl	16
	ldl	7
	sub
	stl	16
	.ldc	2047
	ldl	16
	gt
	eqc	0
	cj	@L0110
	.ldc	2047
	stl	16
	.ldc	0			;Atomic branch
	cj	@L0115
L0110:	ldl	16
	.ldc	0
	gt
	cj	@L0111
	.ldc	0			;Atomic branch
	cj	@L0115
L0111:	.ldc	-52
	ldl	16
	gt
	cj	@L0112
	.ldc	0
	stl	13
	.ldc	0
	stl	14
	.ldc	1
	stl	9
	.ldc	0
	stl	16
	.ldc	0			;Atomic branch
	cj	@L0115
L0112:	ldl	13
	.ldc	2047
	and
	ldl	9
	or
	stl	7
	.ldc	-30
	ldl	16
	gt
	cj	@L0113
	ldl	7
	ldl	13
	or
	stl	7
	ldl	14
	.ldc	0
	ldl	16
	sub
	adc	-31
	ldl	13
	rev
	lshr
	stl	9
	stl	13
	.ldc	0
	stl	14
	.ldc	0			;Atomic branch
	cj	@L0114
L0113:	ldl	13
	.ldc	1
	ldl	16
	sub
	ldl	9
	rev
	lshr
	stl	9
	stl	10
	ldl	14
	.ldc	1
	ldl	16
	sub
	ldl	13
	rev
	lshr
	stl	13
	stl	14
L0114:	.ldc	0
	stl	16
	ldl	9
	ldl	7
	or
	stl	9
L0115:	.ldc	2047
	ldl	16
	gt
	cj	@L0133
	ldl	13
	stl	7
	ldl	14
	ldl	2
	not
	and
	ldl	13
	.ldc	11
	lshr
	stl	13
	stl	14
	ldl	16
	.ldc	0
	.ldc	12
	lshr
	stl	16
	stl	10
	ldl	14
	ldl	16
	or
	stl	14
	ldl	17
	eqc	0
	cj	@L0123
	ldl	7
	.ldc	1024
	and
	eqc	0
	cj	@L0119
	.ldc	0			;Atomic branch
	cj	@L0130
L0119:	ldl	7
	.ldc	1023
	and
	ldl	9
	or
	ldl	13
	.ldc	1
	and
	or
	eqc	0
	cj	@L0120
	.ldc	0			;Atomic branch
	cj	@L0130
L0120:	.ldc	0
	.ldc	1
	ldl	13
	lsum
	stl	13
	stl	10
	ldl	10
	.ldc	0
	ldl	14
	lsum
	stl	14
	stl	10
	.ldc	0			;Atomic branch
	cj	@L0130
L0123:	ldl	17
	eqc	3
	cj	@L0124
	ldl	6
	cj	@L0130
L0124:	ldl	17
	eqc	2
	cj	@L0125
	ldl	6
	eqc	0
	cj	@L0130
L0125:	ldl	17
	eqc	1
	cj	@L0128
	.ldc	0			;Atomic branch
	cj	@L0130
L0128:	ldl	7
	.ldc	2047
	and
	ldl	9
	or
	eqc	0
	cj	@L0129
	.ldc	0			;Atomic branch
	cj	@L0130
L0129:	.ldc	0
	.ldc	1
	ldl	13
	lsum
	stl	13
	stl	10
	ldl	10
	.ldc	0
	ldl	14
	lsum
	stl	14
	stl	10
L0130:	ldl	14
	eqc	2146435072
	stl	19
	ldl	6
	ldl	14
	or
	ldl	22
	stnl	1
	ldl	13
	ldl	22
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L0139
L0133:	ldl	17
	eqc	3
	cj	@L0134
	ldl	6
	cj	@L0136
L0134:	ldl	17
	eqc	2
	cj	@L0135
	ldl	6
	eqc	0
	cj	@L0136
L0135:	ldl	17
	eqc	1
	cj	@L0138
L0136:	ldl	6
	ldl	20
	ldnl	9
	or
	ldl	22
	stnl	1
	.ldc	-1
	ldl	22
	stnl	0
	.ldc	1
	stl	19
	.ldc	0			;Atomic branch
	cj	@L0139
L0138:	ldl	6
	ldl	20
	ldnl	2
	or
	ldl	22
	stnl	1
	.ldc	0
	ldl	22
	stnl	0
	.ldc	1
	stl	19
L0139:	ajw	21
	ret

; rimul(x, guard, hi, lo)
; int x[2];
; int *guard;
; int hi, lo;
rimul:	ajw	-5
	.ldc	0
	ldl	9
	ldl	6
	ldnl	0
	lmul
	stl	0
	stl	4
	ldl	4
	ldl	8
	ldl	6
	ldnl	0
	lmul
	stl	1
	stl	4
	.ldc	0
	ldl	9
	ldl	6
	ldnl	1
	lmul
	stl	2
	stl	3
	ldl	4
	ldl	8
	ldl	6
	ldnl	1
	lmul
	ldl	6
	stnl	0
	ldl	6
	stnl	1
	.ldc	0
	ldl	1
	ldl	2
	lsum
	ldl	7
	stnl	0
	stl	4
	ldl	4
	ldl	3
	ldl	6
	ldnl	0
	lsum
	ldl	6
	stnl	0
	stl	4
	ldl	4
	.ldc	0
	ldl	6
	ldnl	1
	lsum
	ldl	6
	stnl	1
	stl	4
	ldl	0
	eqc	0
	cj	@L0M1
	.ldc	0			;Atomic branch
	cj	@L0M2
L0M1:	ldl	7
	ldnl	0
	.ldc	1
	or
	ldl	7
	stnl	0
L0M2:	ajw	5
	ret

; ridiv(quot, hi, lo, guard, yfrac)
; int *quot, *hi, *lo, *guard;
; int yfrac[2];
ridiv:	ajw	-5
	ldl	10
	ldnl	1
	ldl	7
	ldnl	0
	diff
	eqc	0
	cj	@L0D1
	.ldc	-1
	stl	3
	.ldc	0			;Atomic branch
	cj	@L0D2
L0D1:	ldl	7
	ldnl	0
	ldl	8
	ldnl	0
	ldl	10
	ldnl	1
	ldiv
	stl	3
	stl	4
L0D2:	ldl	3
	cj	@L0D4
	.ldc	0
	ldl	10
	ldnl	0
	ldl	3
	lmul
	stl	0
	stl	1
	ldl	1
	ldl	10
	ldnl	1
	ldl	3
	lmul
	stl	1
	stl	2
	.ldc	0
	ldl	9
	ldnl	0
	ldl	0
	ldiff
	ldl	9
	stnl	0
	stl	4
	ldl	4
	ldl	8
	ldnl	0
	ldl	1
	ldiff
	ldl	8
	stnl	0
	stl	4
	ldl	4
	ldl	7
	ldnl	0
	ldl	2
	ldiff
	ldl	7
	stnl	0
	stl	4
L0D3:	ldl	7
	ldnl	0
	mint
	and
	cj	@L0D4
	.ldc	0
	ldl	3
	.ldc	1
	ldiff
	stl	3
	stl	4
	.ldc	0
	ldl	10
	ldnl	0
	ldl	9
	ldnl	0
	lsum
	ldl	9
	stnl	0
	stl	4
	ldl	4
	ldl	10
	ldnl	1
	ldl	8
	ldnl	0
	lsum
	ldl	8
	stnl	0
	stl	4
	ldl	4
	.ldc	0
	ldl	7
	ldnl	0
	lsum
	ldl	7
	stnl	0
	stl	4
	.ldc	0			;Atomic branch
	cj	@L0D3
L0D4:	ldl	3
	ldl	6
	stnl	0
	ajw	5
	ret

	.pub	?FPGT
?FPGT:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0			;&FAreg
	ldl	0
	adc	8			;&FBreg
	call	@_r64gt
	stl	4
retbool:
	ldl	0
	ldnlp	4			;src=&FCreg
	ldl	0			;dst=&FAreg
	.ldc	0
	move
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	3			;B ==> C
	ldl	2			;A ==> B
	ldl	4			;Results
	ajw	1
	ret

;
;	64-bit floating-point comparison for the T414.
;
;	_r64gt performs the operation "*x > *y;".
;
;	_r64gt(x, y)
;	double *x, *y;
;
;	Based on INMOS supplied code, decompiled by Scott Lloyd of CSA.
;	Modified for atomicity by Robert Eifrig of Tektronix.
;
	.align
L20:
	.dw	0x7fffffff
	.pub	_r64gt
_r64gt:	ajw	-3
	.ldc	L20
	stl	2
	ldl	4
	ldnl	1
	ldl	2
	ldnl	0
	and
	stl	1
	ldl	5
	ldnl	1
	ldl	2
	ldnl	0
	and
	stl	0
	.ldc	0
	ldl	4
	ldnl	1
	ldl	5
	ldnl	1
	xor
	gt
	cj	@L22
	.ldc	0
	ldl	5
	ldnl	1
	gt
	cj	@L21
	ldl	1
	ldl	0
	sum
	ldl	4
	ldnl	0
	or
	ldl	5
	ldnl	0
	or
	eqc	0
	eqc	0
L21:	ajw	3
	ret
L22:	.ldc	0
	ldl	4
	ldnl	1
	gt
	cj	@L25
	ldl	0
	ldl	1
	diff
	eqc	0
	cj	@L23
	ldl	5
	ldnl	0
	mint
	xor
	ldl	4
	ldnl	0
	mint
	xor
	gt
	ajw	3
	ret
L23:	ldl	0
	ldl	1
	gt
	ajw	3
	ret
L25:	ldl	1
	ldl	0
	diff
	eqc	0
	cj	@L26
	ldl	4
	ldnl	0
	mint
	xor
	ldl	5
	ldnl	0
	mint
	xor
	gt
	ajw	3
	ret
L26:	ldl	1
	ldl	0
	gt
	ajw	3
	ret

	.pub	?FPEQ
?FPEQ:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0			;&FAreg
	ldl	0
	adc	8			;&FBreg
	call	@_r64eq
	stl	4
	.ldc	0			;Atomic branch
	cj	@retbool

;
;	64-bit floating-point comparison for the T414.
;
;	_r64eq performs the operation "*x == *y;".
;
;	_r64eq(x, y)
;	double *x, *y;
;
;	Based on INMOS supplied code, decompiled by Scott Lloyd of CSA.
;	Modified for atomicity by Robert Eifrig of Tektronix.
;
	.align
L30:
	.dw	0x7fffffff
	.pub	_r64eq
_r64eq:	ajw	-3
	.ldc	L30
	stl	2
	ldl	4
	ldnl	1
	ldl	2
	ldnl	0
	and
	stl	1
	ldl	5
	ldnl	1
	ldl	2
	ldnl	0
	and
	stl	0
	ldl	4
	ldnl	1
	ldl	5
	ldnl	1
	diff
	eqc	0
	cj	@L31
	ldl	4
	ldnl	0
	ldl	5
	ldnl	0
	diff
	eqc	0
L31:	eqc	0
	cj	@L32
	ldl	1
	ldl	0
	sum
	ldl	4
	ldnl	0
	or
	ldl	5
	ldnl	0
	or
	eqc	0
	eqc	0
L32:	eqc	0
	ajw	3
	ret

	.pub	??U32TOR32
	.pub	?FPB32TOR64
??U32TOR32:
?FPB32TOR64:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnlp	2			;src=&FBreg
	ldl	0
	ldnlp	4			;dst=&FCreg
	.ldc	8
	move
	ldl	0			;src=&FAreg
	ldl	0
	ldnlp	2			;dst=&FBreg
	.ldc	8
	move
	ldl	2
	ldnl	0			;*(int *)A
	ldl	0			;&FAreg
	call	@r64fmu
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	4			;C ==> C
	ldl	4			;C ==> B
	ldl	3			;B ==> A
	ajw	1
	ret

	.pub	?FPI32TOR32
	.pub	?FPI32TOR64
?FPI32TOR32:
?FPI32TOR64:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl	0
	ldnlp	2			;src=&FBreg
	ldl	0
	ldnlp	4			;dst=&FCreg
	.ldc	8
	move
	ldl	0			;src=&FAreg
	ldl	0
	ldnlp	2			;dst=&FBreg
	.ldc	8
	move
	ldl	2
	ldnl	0			;*(unsigned int *)A
	ldl	0			;&FAreg
	call	@_r64fmi
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	4			;C ==> C
	ldl	4			;C ==> B
	ldl	3			;B ==> A
	ajw	1
	ret

;	64-bit floating-point conversion for the T414.
;
;	_r64fmi converts an int to a double "*r = i;".
;
;	_r64fmi(r, i)
;	double *r;
;	int i;
;
;	Based on INMOS supplied code, decompiled by Scott Lloyd of CSA.
;	Modified for atomicity by Robert Eifrig of Tektronix.
;
	.pub	_r64fmi
_r64fmi:
	ajw	-6
	ldl	8
	eqc	0
	cj	@L41
	.ldc	0
	ldl	7
	stnl	1
	.ldc	0
	ldl	7
	stnl	0
	.ldc	0			;Atomic branch
	cj	@L46
L41:	.ldc	0
	ldl	8
	gt
	cj	@L44			;jump if positive
	mint
	stl	0
	ldl	8
	mint
	diff
	eqc	0
	cj	@L42
	ldl	8
	stl	1
	.ldc	0			;Atomic branch
	cj	@L45
L42:	.ldc	0
	ldl	8
	sub
	stl	1
	.ldc	0			;Atomic branch
	cj	@L45
r64fmu:
	ajw	-6			;Unsigned int to floating point

L44:	.ldc	0
	stl	0
	ldl	8
	stl	1
L45:	ldl	1
	.ldc	0
	norm
	stl	5
	stl	1
	stl	4
	.ldc	32
	ldl	4
	sub
	adc	1022
	stl	3
	ldl	1
	.ldc	1
	shl
	stl	1
	ldl	3
	ldl	1
	.ldc	12
	lshr
	stl	2
	stl	3
	ldl	1
	.ldc	0
	.ldc	12
	lshr
	ldl	7
	stnl	0
	stl	3
	ldl	0
	ldl	2
	or
	ldl	7
	stnl	1
L46:	ajw	6
	ret

	.pub	?FPUR32TOR64
	.pub	?FPUR64TOR32
	.pub	?FPUNOROUND
?FPUR32TOR64:
?FPUR64TOR32:
?FPUNOROUND:
	ldl	3			;C
	ldl	2			;B
	ldl	1			;A
	ret

	.pub	?FPSTNLI32
?FPSTNLI32:
	ajw	-1
	.ldc	_FPstate
	stl	0
	ldl     0			;&FAreg
	.ldc	0			;Truncate
	call	@_r64toi
	ldl	2			;A
	stnl	0
	ldl	0
	ldnlp	2			;src=&FBreg
	ldl	0			;dst=&FAreg
	.ldc	8
	move
	ldl	0
	ldnlp	4			;src=&FCreg
	ldl	0
	ldnlp	2			;dst=&FBreg
	.ldc	8
	move
	.ldc	0
	ldl	0
	stnl	6			;Set round mode to "Round to nearest"
	ldl	4			;C ==> C
	ldl	4			;C ==> B
	ldl	3			;B ==> A
	ajw	1
	ret

;	64-bit floating-point conversion for the T414.
;
;	_r64toi converts a double to an int
;	md specifies the ROUND or TRUNC mode.
;
;	md = 1  ROUND
;	md = 0  TRUNC
;
;	int _r64toi(md, r)
;	int md;
;	double *r;
;
;	Based on INMOS supplied code, decompiled by Scott Lloyd of CSA.
;	Modified for atomicity by Kirk Bailey of Logical Systems.
;
	.pub	_r64toi
_r64toi:
	ajw	-9
	ldl	11
	stl	6
	mint
	stl	0
	.ldc	0
	ldl	6
	ldnl	1
	ldl	0
	not
	and
	.ldc	12
	lshl
	stl	5
	stl	3
	ldl	6
	ldnl	1
	ldl	6
	ldnl	0
	.ldc	12
	lshl
	stl	1
	stl	2
	ldl	3
	eqc	2047
	cj	@L51
	.ldc	0			;Atomic branch
	cj	@L70
L51:	ldl	6
	ldnl	1
	eqc	-1042284544
	cj	@L53
	ldl	2
	cj	@L52
	ldl	10
	eqc	0
	cj	@L53
	ldl	2
	eqc	1
	cj	@L53
L52:	ldl	0
	stl	7
	.ldc	0			;Atomic branch
	cj	@L70
L53:	.ldc	1022
	ldl	3
	gt
	cj	@L56
	.ldc	0
	stl	7
	.ldc	0			;Atomic branch
	cj	@L70
L56:	.ldc	1023
	ldl	3
	gt
	cj	@L62
	ldl	10
	cj	@L57
	ldl	2
	ldl	1
	or
	eqc	0
	cj	@L58
L57:	.ldc	0
	stl	7
	.ldc	0			;Atomic branch
	cj	@L60
L58:	ldl	6
	ldnl	1
	ldl	0
	and
	eqc	0
	cj	@L59
	.ldc	1
	stl	7
	.ldc	0			;Atomic branch
	cj	@L60
L59:	.ldc	-1
	stl	7
L60:	.ldc	0			;Atomic branch
	cj	@L70
L62:	ldl	3
	adc	-1023
	stl	3
	.ldc	1
	stl	4
	ldl	4
	ldl	2
	ldl	3
	lshl
	stl	5
	stl	4
	ldl	2
	ldl	1
	ldl	3
	lshl
	stl	1
	stl	2
	ldl	10
	cj	@L63
	ldl	2
	ldl	0
	and
	eqc	0
	cj	@L64
L63:	.ldc	0			;Atomic branch
	cj	@L66
L64:	ldl	2
	ldl	0
	not
	and
	ldl	1
	or
	ldl	4
	.ldc	1
	and
	or
	eqc	0
	cj	@L65
	.ldc	0			;Atomic branch
	cj	@L66
L65:	.ldc	0
	.ldc	1
	ldl	4
	lsum
	stl	4
	stl	5
L66:	ldl	6
	ldnl	1
	ldl	0
	and
	eqc	0
	cj	@L68
	.ldc	0			;Atomic branch
	cj	@L69
L68:	.ldc	0
	ldl	4
	diff
	stl	4
L69:	ldl	4
	stl	7
L70:	ldl	7
	ajw	9
	ret
	#pragma	endasm
	}
