/*
 *	tcio.c
 *
 *	Client (Transputer) side of "C" remote file I/O driver.  Stream I/O
 *	routines are located in "tstream.c" which uses the basic server code in
 *	this file also.
 *
 *	Copyright 1987, 1988, 1989 by Logical Systems.
 */

#include	<time.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<conc.h>
#include	<cio.h>
#include	<cioext.h>
#include	<inline.h>

#define	UCHAR	unsigned char

/*
 *	Include file consistency check for seek origin values.
 */

#if	(SEEK_CUR!=_SEEK_CUR) || (SEEK_END!=_SEEK_END) || (SEEK_SET!=_SEEK_SET)
#error	"FATAL: Seek origin inconsistency between 'cio.h' and 'stdio.h'"
#endif	/* Seek origin check */

/*
 *	In order to make FILE * pointers actually be a pointer, rather than
 *	carry around an integer in a pointer, which must get converted (with
 *	random results), we store the integer from the server in an array, and
 *	make the (FILE *) pointers point to the array item.
 *
 *	Note that nothing is really stored in this array.  We are simply using
 *	the address of the array items.  As C allows you to reference past the
 *	end of the array, there is no reason to make the array overly large.
 */

FILE	_fio[1];		/* See above comment			*/

/*
 *	Static array for "gmtime"/"localtime" use.
 */

static	struct	tm	ttemp;

/*
 *	Function definitions.
 */

void	_enc_strm(UCHAR *,FILE *); /* Convert strm index into 2 bytes	*/
void	_enc_2b(UCHAR *,int);	/* Convert int to little endian 2 bytes	*/
void	_enc_4b(UCHAR *,long);	/* Convert long to little endian 4 bytes*/
void	_get_str(UCHAR *,int);	/* Get specified # of bytes from host	*/
FILE	*_get_strm(void);	/* Get stream pointer from host		*/
void	_get_time(void);	/* Get/cnvt broken down time from host	*/
int	_get_zeof(int);		/* Get one byte worth of 0/EOF from host*/
int	_get_1b(void);		/* Get one byte from host		*/
int	_get_2b(void);		/* Get 2 bytes from host		*/
int	_get_2beof(void);	/* Get 2 bytes/EOF from host		*/
long	_get_4b(void);		/* Get 4 bytes from host		*/
void	_put_str(UCHAR *,int);	/* Send "raw" string to host		*/
void	_put_1b(int);		/* Send "raw" byte to host		*/
void	_request_errno(void);	/* Get host "errno" value		*/
void	_send_cmd(UCHAR *,int,int); /* Send command to host		*/
void	_send_cmdstr(UCHAR *,int,int,int); /* Send cmd to host w/str	*/

/*
 *	Get "argv" table and argument lists and place at specified memory
 *	address.  See the "start.c" function and the "cio.c" I/O driver
 *	source for more information.  Returns the value of "argc".
 */
long
_arg_get(UCHAR *addr)
	{
	UCHAR			buf[7];
	int			length;
	long			result;

	_enc_4b(&buf[3],(long) addr);
	_send_cmd(buf,FN_ARG_GET,7);
	length = _get_2b() - 4;		/* Total argument info length */
	result = _get_4b();		/* Ignoring "argc" value (this) */
	_get_str((UCHAR *) addr,length);
	return (result);
	}

/*
 *	Ask the "cio.c" I/O driver for the total storage that the "argv"
 *	pointer table and associated values will take (in bytes, rounded to the
 *	nearest word).  See the "start.c" function and the "cio.c" driver
 *	source for more information.
 */
long
_arg_size(void)
	{
	UCHAR			buf[5];

	_enc_2b(&buf[3],sizeof(char **));	/* Tell host bytes/word */
	_send_cmd(buf,FN_ARG_SIZE,5);
	return (_get_4b());
	}

int
_cioext(int code,void *sbuf,int sbuflen,void *rbuf,int rbuflen)
	{
	UCHAR			buf[9];
	int			rval;

	if((sbuflen > MAX_MSG_DATA) || (rbuflen > MAX_MSG_DATA))
		return (-1);
	_enc_2b(&buf[3],code);
	_enc_2b(&buf[5],sbuflen);
	_enc_2b(&buf[7],rbuflen);
	_send_cmdstr(buf,FN_CIOEXT,9,sbuflen);
	if(sbuflen > 0)
		_put_str((UCHAR *) sbuf,sbuflen);

	(void) _get_2b();		/* Ignore length */
	rval = _get_4b();		/* Get desired return value */
	errno = _get_4b();		/* Host "errno" value */
	if(rbuflen > 0)
		_get_str((UCHAR *) rbuf,rbuflen);	/* Read reply str */

	return (rval);
	}

void
clearerr(FILE *stream)
	{
	UCHAR			buf[5];

	_enc_strm(&buf[3],stream);
	_send_cmd(buf,FN_CLEARERR,5);
	_request_errno();
	}

int
close(int handle)
	{
	UCHAR			buf[5];

	_enc_2b(&buf[3],handle);
	_send_cmd(buf,FN_CLOSE,5);
	return (_get_zeof(1));
	}

int
creat(const char *pathname,int pmode)
	{
	UCHAR			buf[7];
	int			length;

	if(pathname == NULL)
		return (EOF);
	if((length = (strlen(pathname) + 1)) > MAX_MSG_DATA)
		return (EOF);
	_enc_4b(&buf[3],(long) pmode);
	_send_cmdstr(buf,FN_CREAT,7,length);
	_put_str((UCHAR *) pathname,length);
	return (_get_2beof());
	}

int
dup(int handle)
	{
	UCHAR			buf[5];

	_enc_2b(&buf[3],handle);
	_send_cmd(buf,FN_DUP,5);
	return (_get_2beof());
	}

int
dup2(int handle1,int handle2)
	{
	UCHAR			buf[7];

	_enc_2b(&buf[3],handle1);
	_enc_2b(&buf[5],handle2);
	_send_cmd(buf,FN_DUP2,7);
	return (_get_zeof(1));
	}

void
exit(int status)
	{
	UCHAR			buf[7];

	_enc_4b(&buf[3],(long) status);
	_send_cmd(buf,FN_EXIT,7);
	;
#pragma	asm
	stopp				;Kill this process off
#pragma	endasm
	}

int
fclose(FILE *stream)
	{
	UCHAR			buf[5];

	_enc_strm(&buf[3],stream);
	_send_cmd(buf,FN_FCLOSE,5);
	return (_get_zeof(1));
	}

int
fcloseall(void)
	{
	UCHAR			buf[3];

	_send_cmd(buf,FN_FCLOSEALL,3);
	return (_get_2beof());
	}

FILE	*
fdopen(int handle,const char *mode)
	{
	UCHAR			buf[5];
	int			length;

	if(mode == NULL)
		return (NULL);
	if((length = (strlen(mode) + 1)) > MAX_MSG_DATA)
		return (NULL);
	_enc_2b(&buf[3],handle);
	_send_cmdstr(buf,FN_FDOPEN,5,length);
	_put_str((UCHAR *) mode,length);
	return (_get_strm());
	}

int
feof(FILE *stream)
	{
	UCHAR			buf[5];

	_enc_strm(&buf[3],stream);
	_send_cmd(buf,FN_FEOF,5);
	return (_get_zeof(0));
	}

int
ferror(FILE *stream)
	{
	UCHAR			buf[5];

	_enc_strm(&buf[3],stream);
	_send_cmd(buf,FN_FERROR,5);
	return (_get_zeof(0));
	}

int
fflush(FILE *stream)
	{
	UCHAR			buf[5];

	_enc_strm(&buf[3],stream);
	_send_cmd(buf,FN_FFLUSH,5);
	return (_get_zeof(1));
	}

int
fileno(FILE *stream)
	{
	UCHAR			buf[5];

	_enc_strm(&buf[3],stream);
	_send_cmd(buf,FN_FILENO,5);
	return (_get_2b());
	}

FILE	*
fopen(const char *pathname,const char *mode)
	{
	UCHAR			buf[3];
	int			length1;
	int			length2;

	if((pathname == NULL) || (mode == NULL))
		return (NULL);
	if((length1 = (strlen(pathname) + 1)) > MAX_MSG_DATA)
		return (NULL);
	if((length2 = (strlen(mode) + 1)) > MAX_MSG_DATA)
		return (NULL);
	if(length1 + length2 > MAX_MSG_DATA)
		return (NULL);
	_send_cmdstr(buf,FN_FOPEN,3,length1 + length2);
	_put_str((UCHAR *) pathname,length1);
	_put_str((UCHAR *) mode,length2);
	return (_get_strm());
	}

size_t
fread(void *ptr,size_t size,size_t nmemb,FILE *stream)
	{
	UCHAR			buf[9];
	int			per_read;
	int			temp;
	size_t			total_membs;

	total_membs = 0;

	if((size == 0) || (size > MAX_MSG_DATA))
		return (0);
	per_read = MAX_MSG_DATA / size;
/*
 *	Use multiple short "fread" commands to fake one long one.
 */
	while(nmemb > per_read)
		{
		nmemb -= per_read;
		_enc_strm(&buf[3],stream);
		_enc_2b(&buf[5],(int) size);
		_enc_2b(&buf[7],(int) per_read);
		_send_cmd(buf,FN_FREAD,9);
		if((temp = _get_2b()) == 0)
			{
			if(total_membs == 0)
				_request_errno();
			return (total_membs);
			}
		_get_str((UCHAR *) ptr,temp);
		ptr = ((char *) ptr) + (per_read * size);
		total_membs += (temp /= size);	/* # of items, not length */
		if(temp < per_read)
			return (total_membs);
		}
/*
 *	Read rest of data.
 */
	if(nmemb > 0)
		{
		_enc_strm(&buf[3],stream);
		_enc_2b(&buf[5],(int) size);
		_enc_2b(&buf[7],(int) nmemb);
		_send_cmd(buf,FN_FREAD,9);
		if((temp = _get_2b()) == 0)
			{
			if(total_membs == 0)
				_request_errno();
			return (total_membs);
			}
		_get_str((UCHAR *) ptr,temp);
		return (total_membs + (temp / size));
		}
	return (total_membs);
	}

FILE	*
freopen(const char *pathname,const char *mode,FILE *stream)
	{
	UCHAR			buf[5];
	int			length1;
	int			length2;

	if((pathname == NULL) || (mode == NULL))
		return (NULL);
	if((length1 = (strlen(pathname) + 1)) > MAX_MSG_DATA)
		return (NULL);
	if((length2 = (strlen(mode) + 1)) > MAX_MSG_DATA)
		return (NULL);
	if((length1 + length2) > MAX_MSG_DATA)
		return (NULL);
	_enc_strm(&buf[3],stream);
	_send_cmdstr(buf,FN_FREOPEN,5,length1 + length2);
	_put_str((UCHAR *) pathname,length1);
	_put_str((UCHAR *) mode,length2);
	return (_get_strm());
	}

int
fseek(FILE *stream,long offset,int origin)
	{
	UCHAR			buf[10];

	_enc_strm(&buf[3],stream);
	_enc_4b(&buf[5],offset);
	buf[9] = origin;
	_send_cmd(buf,FN_FSEEK,10);
	return (_get_zeof(1));
	}

long
ftell(FILE *stream)
	{
	UCHAR			buf[5];
	long			result;

	_enc_strm(&buf[3],stream);
	_send_cmd(buf,FN_FTELL,5);
	result = _get_4b();
	if(result == RPC_EOF)
		{
		result = EOF;
		_request_errno();
		}
	return (result);
	}

size_t
fwrite(const void *ptr,size_t size,size_t nmemb,FILE *stream)
	{
	UCHAR			buf[9];
	int			per_length;
	int			per_write;
	int			temp;
	size_t			total_membs;

	total_membs = 0;

	if((size == 0) || (size > MAX_MSG_DATA))
		return (0);
	per_write = MAX_MSG_DATA / size;
	per_length = per_write * size;
/*
 *	Use multiple short "fwrite" commands to fake one long one.
 */
	while(nmemb > per_write)
		{
		nmemb -= per_write;
		_enc_strm(&buf[3],stream);
		_enc_2b(&buf[5],(int) size);
		_enc_2b(&buf[7],(int) per_write);
		_send_cmdstr(buf,FN_FWRITE,9,per_length);
		_put_str((UCHAR *) ptr,per_length);
		if(((temp = _get_2b()) == 0) && (total_membs == 0))
			_request_errno();
		ptr = ((char *) ptr) + per_length;
		total_membs += temp;	/* # of items, not length */
		if(temp < per_write)
			return (total_membs);
		}
/*
 *	Write rest of data.
 */
	if(nmemb > 0)
		{
		_enc_strm(&buf[3],stream);
		_enc_2b(&buf[5],(int) size);
		_enc_2b(&buf[7],(int) nmemb);
		_send_cmdstr(buf,FN_FWRITE,9,(per_length = nmemb * size));
		_put_str((UCHAR *) ptr,per_length);
		if(((temp = _get_2b()) == 0) && (total_membs == 0))
			_request_errno();
		return (total_membs + temp);
		}
	return (total_membs);
	}

char	*
getenv(const char *name)
	{
	UCHAR			buf[3];
	int			length;
	char			*sptr;

	if(name == NULL)
		return (NULL);
	if((length = (strlen(name) + 1)) > MAX_MSG_DATA)
		return (NULL);
	_send_cmdstr(buf,FN_GETENV,3,length);
	_put_str((UCHAR *) name,length);
	length = _get_2b() - 1;		/* Get return env string length */
	if(_get_1b() == RPC_NULL)	/* Figure out if string/NULL */
		{
		(void) _get_1b();	/* Toss dummy '\0' byte */
		return (NULL);
		}
	else				/* Real environment string, read */
		{
		if((sptr = (char *) malloc((size_t) length)) == NULL)
			{
			while(length--)
				(void) _get_1b();	/* Ignore */
			return (NULL);
			}
		else
			{
			if(length > 0)
				_get_str((UCHAR *) sptr,length);
			}
		}
	return (sptr);
	}

#ifndef	T212
struct	tm	*
gmtime(const time_t *timer)
	{
	UCHAR			buf[7];

	_enc_4b(&buf[3],(long) *timer);
	_send_cmd(buf,FN_TIMEGMT,7);
	if(_get_1b() != RPC_NULL)	/* Figure out if time/NULL */
		{
		_get_time();
		return (&ttemp);
		}
	return (NULL);
	}
#endif	/* !T212 */

#ifndef	T212
struct	tm	*
localtime(const time_t *timer)
	{
	UCHAR			buf[7];

	_enc_4b(&buf[3],(long) *timer);
	_send_cmd(buf,FN_TIMELOC,7);
	_get_time();
	return (&ttemp);
	}
#endif	/* !T212 */

long
lseek(int handle,long offset,int origin)
	{
	UCHAR			buf[10];
	long			result;

	_enc_2b(&buf[3],handle);
	_enc_4b(&buf[5],offset);
	buf[9] = origin;
	_send_cmd(buf,FN_LSEEK,10);
	result = _get_4b();
	if(result == RPC_EOF)
		{
		result = EOF;
		_request_errno();
		}
	return (result);
	}

int
open(const char *pathname,int oflag,int pmode)
	{
	UCHAR			buf[11];
	int			length;

	if(pathname == NULL)
		return (EOF);
	if((length = (strlen(pathname) + 1)) > MAX_MSG_DATA)
		return (EOF);
	_enc_4b(&buf[3],(long) oflag);
	_enc_4b(&buf[7],(long) pmode);
	_send_cmdstr(buf,FN_OPEN,11,length);
	_put_str((UCHAR *) pathname,length);
	return (_get_2beof());
	}

int
read(int handle,char *s,size_t count)
	{
	UCHAR			buf[7];
	int			temp;
	size_t			total_length;

	total_length = 0;
/*
 *	Use multiple short "read" commands to fake one long one.
 */
	while(count > MAX_MSG_DATA)
		{
		_enc_2b(&buf[3],handle);
		_enc_2b(&buf[5],(int) MAX_MSG_DATA);
		_send_cmd(buf,FN_READ,7);
		temp = _get_2b() - 1;		/* Get buffer length */
		if(_get_1b() == RPC_EOF)	/* Figure out if data/error */
			{
			_request_errno();	/* Get errno status */
			return (EOF);
			}
		if(temp > 0)
			{
			_get_str((UCHAR *) s,temp);	/* Read data */
			s += temp;
			total_length += temp;
			count -= temp;
			}
		else
			return (total_length);
		}
/*
 *	Read rest of data.
 */
	if(count > 0)
		{
		_enc_2b(&buf[3],handle);
		_enc_2b(&buf[5],(int) count);
		_send_cmd(buf,FN_READ,7);
		temp = _get_2b() - 1;		/* Get buffer length */
		if(_get_1b() == RPC_EOF)	/* Figure out if data/error */
			{
			_request_errno();	/* Get errno status */
			return (EOF);
			}
		if(temp > 0)
			_get_str((UCHAR *) s,temp);	/* Read data */
		return (total_length + temp);
		}
	return (total_length);
	}

int
remove(const char *pathname)
	{
	UCHAR			buf[3];
	int			length;

	if(pathname == NULL)
		return (EOF);
	if((length = (strlen(pathname) + 1)) > MAX_MSG_DATA)
		return (EOF);
	_send_cmdstr(buf,FN_REMOVE,3,length);
	_put_str((UCHAR *) pathname,length);
	return (_get_zeof(1));
	}

int
rename(const char *old,const char *new)
	{
	UCHAR			buf[3];
	int			length1;
	int			length2;

	if((old == NULL) || (new == NULL))
		return (EOF);
	if((length1 = (strlen(old) + 1)) > MAX_MSG_DATA)
		return (EOF);
	if((length2 = (strlen(new) + 1)) > MAX_MSG_DATA)
		return (EOF);
	if((length1 + length2) > MAX_MSG_DATA)
		return (EOF);
	_send_cmdstr(buf,FN_RENAME,3,length1 + length2);
	_put_str((UCHAR *) old,length1);
	_put_str((UCHAR *) new,length2);
	return (_get_zeof(1));
	}

void
rewind(FILE *stream)
	{
	(void) fseek(stream,0L,SEEK_SET);	/* Seek to start */
	clearerr(stream);		/* Clear possible error flag */
	}

int
system(const char *string)
	{
	UCHAR			buf[4];
	int			length;
	char			*sptr;

	if(string == NULL)
		{
		buf[3] = RPC_NULL;
		sptr = "";
		}
	else
		{
		buf[3] = RPC_NOTNULL;
		sptr = string;
		}
	if((length = (strlen(sptr) + 1)) > MAX_MSG_DATA)
		return (EOF);
	_send_cmdstr(buf,FN_SYSTEM,4,length);
	_put_str((UCHAR *) sptr,length);
	return (_get_4b());
	}

#ifndef	T212
time_t
time(time_t *timer)
	{
	UCHAR			buf[3];
	time_t			rval;

	_send_cmd(buf,FN_TIME,3);
	rval = (time_t) _get_4b();
	if(timer != NULL)
		*timer = rval;
	return (rval);
	}
#endif	/* !T212 */

FILE	*
tmpfile(void)
	{
	UCHAR			buf[3];

	_send_cmd(buf,FN_TMPFILE,3);
	return (_get_strm());
	}

char	*
tmpnam(char *string)
	{
	UCHAR			buf[3];
	int			length;
	static char		*name_ptr = NULL;

	_send_cmd(buf,FN_TMPNAM,3);
	length = _get_2b() - 1;		/* Get return temp name length */
	if(_get_1b() == RPC_NULL)	/* Figure out if string/NULL */
		{
		(void) _get_1b();	/* Toss dummy '\0' byte */
		return (NULL);
		}
	else				/* A name has been generated, get it */
		{
		if(string == NULL)	/* Store it internally? */
			{
			if(name_ptr != NULL)
				free((void *) name_ptr);
			if((name_ptr = (char *) malloc((size_t) length)) ==
				NULL)
				{
				while(length--)
					(void) _get_1b();	/* Ignore */
				return (NULL);
				}
			_get_str((UCHAR *) name_ptr,length);
			return (name_ptr);
			}
		else
			_get_str((UCHAR *) string,length);
		}
	return (string);
	}

int
ungetc(int c,FILE *stream)
	{
	UCHAR			buf[6];

	if(c == EOF)
		return (EOF);
	_enc_strm(&buf[3],stream);
	buf[5] = c;
	_send_cmd(buf,FN_UNGETC,6);
	return (_get_2beof());
	}

int
unlink(const char *pathname)
	{
	UCHAR			buf[3];
	int			length;

	if(pathname == NULL)
		return (EOF);
	if((length = (strlen(pathname) + 1)) > MAX_MSG_DATA)
		return (EOF);
	_send_cmdstr(buf,FN_REMOVE,3,length);
	_put_str((UCHAR *) pathname,length);
	return (_get_zeof(1));
	}

int
write(int handle,char *s,size_t count)
	{
	UCHAR			buf[7];
	int			temp;
	size_t			total_length;

	total_length = 0;
/*
 *	Use multiple short "write" commands to fake one long one.
 */
	while(count > MAX_MSG_DATA)
		{
		count -= MAX_MSG_DATA;
		_enc_2b(&buf[3],handle);
		_enc_2b(&buf[5],(int) MAX_MSG_DATA);
		_send_cmdstr(buf,FN_WRITE,7,(int) MAX_MSG_DATA);
		_put_str((UCHAR *) s,(int) MAX_MSG_DATA);
		s += MAX_MSG_DATA;
		if((temp = _get_2beof()) == EOF)
			return (EOF);
		total_length += temp;
		if(temp < MAX_MSG_DATA)
			return (total_length);
		}
/*
 *	Send remainder of "write" data.
 */
	if(count > 0)
		{
		_enc_2b(&buf[3],handle);
		_enc_2b(&buf[5],(int) count);
		_send_cmdstr(buf,FN_WRITE,7,(int) count);
		_put_str((UCHAR *) s,(int) count);
		if((temp = _get_2beof()) == EOF)
			return (EOF);
		return (total_length + temp);
		}
	return (total_length);
	}

/************************************************************************/
/*									*/
/*			COMMON UTILITY FUNCTIONS			*/
/*									*/
/************************************************************************/

/************************************************************************/
/*									*/
/*	Routine:	_enc_strm					*/
/*									*/
/*	Translate a stream into a two byte table index and place into	*/
/*	the specified buffer.						*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer).	*/
/*		"stream"The stream to convert.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_enc_strm(UCHAR *sptr,FILE *stream)
	{
	_enc_2b(sptr,(int) (stream - _fio));
	}

/************************************************************************/
/*									*/
/*	Routine:	_enc_2b						*/
/*									*/
/*	Place a two byte signed value into the specified buffer.	*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer).	*/
/*		"value"	The value to place in the buffer.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_enc_2b(UCHAR *sptr,int value)
	{
#ifndef	ISTC
	*sptr++ = value;
	*sptr = (value >>= 8);
#else	/* ISTC */
	#pragma	asm
	LDLP	[value]
	LB				;Low byte of "value"
	LDL	[sptr]
	SB				;Transfer byte over
	LDLP	[value]
	ADC	1
	LB				;High order byte of "value"
	LDL	[sptr]
	ADC	1
	SB				;Transfer byte over
	#pragma	endasm
#endif	/* !ISTC */
	}

/************************************************************************/
/*									*/
/*	Routine:	_enc_4b						*/
/*									*/
/*	Place a four byte signed value into the specified buffer.	*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer).	*/
/*		"value"	The value to place in the buffer.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_enc_4b(UCHAR *sptr,long value)
	{
#ifndef	ISTC
	*sptr++ = value;
	*sptr++ = (value >>= 8);
	*sptr++ = (value >>= 8);
	*sptr = (value >> 8);
#else	/* ISTC */
	#pragma	asm
	LDLP	[value]
	LB				;Low byte of "value"
	LDL	[sptr]
	SB				;Transfer byte over
	LDLP	[value]
	ADC	1
	LB				;Next higher order byte of "value"
	LDL	[sptr]
	ADC	1
	SB				;Transfer byte over
#ifdef	T212
	LDL	[value]
	MINT
	AND				;Non zero if negative
	EQC	0			;0 if negative, 1 if positive
	ADC	-1			;FFFF if negative, 0000 if positive
	STL	[value]			;Save a copy
	LDL	[value]
	LDL	[sptr]
	ADC	2
	SB				;Sign extend into next high order byte
	LDL	[value]
	LDL	[sptr]
	ADC	3
	SB				;Sign extend into high order byte
#else	/* !T212 */
	LDLP	[value]
	ADC	2
	LB				;Next higher order byte of "value"
	LDL	[sptr]
	ADC	2
	SB				;Transfer byte over
	LDLP	[value]
	ADC	3
	LB				;High order byte of "value"
	LDL	[sptr]
	ADC	3
	SB				;Transfer byte over
#endif	/* T212 */
	#pragma	endasm
#endif	/* !ISTC */
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_str					*/
/*									*/
/*	Read a string of specified length from the host system.		*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer to send).*/
/*		"length"The buffer length to read.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_get_str(UCHAR *sptr,int length)
	{
	ChanIn(_boot_chan_in,(char *) sptr,length);	/* Read data */
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_strm					*/
/*									*/
/*	Read a two byte representation of a "stream" from the host	*/
/*	system and return it.  If the return value is "RPC_NULL", get	*/
/*	a current version of the host system "errno" and return NULL.	*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer to send).*/
/*		"length"The buffer length to read.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

FILE	*
_get_strm(void)
	{
	UCHAR			buf[2];
	int			result;

	ChanIn(_boot_chan_in,(char *) buf,2);	/* Read data */
	result = buf[0] | (buf[1] << 8);
	if(result == (RPC_NULL & 0xFFFF))
		{
		_request_errno();
		return ((FILE *) NULL);
		}
	return (&_fio[(int) result]);
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_time					*/
/*									*/
/*	Read a broken down time value from the host system.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_get_time(void)
	{
	struct	rpc_time	t;

	ChanIn(_boot_chan_in,(char *) &t,RPC_TIME_DATA);	/* Read data */

	ttemp.tm_sec = t.tm_sec;
	ttemp.tm_min = t.tm_min;
	ttemp.tm_hour = t.tm_hour;
	ttemp.tm_mday = t.tm_mday;
	ttemp.tm_mon = t.tm_mon;
	ttemp.tm_year = t.tm_year;
	ttemp.tm_wday = t.tm_wday;
	ttemp.tm_yday = (t.low_tm_yday + (t.high_tm_yday << 8));
	ttemp.tm_isdst = t.tm_isdst;
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_zeof					*/
/*									*/
/*	Read a 1 byte value from the host.  If the value is non-zero,	*/
/*	return a value of EOF, otherwise just zero.  If the "error"	*/
/*	parameter is TRUE, request "errno" if non-zero has been read,	*/
/*	otherwise don't bother.						*/
/*									*/
/*	Parameters:							*/
/*		"error"	TRUE if "errno" should be requested on non-zero.*/
/*									*/
/*	Returns:	The byte value read/converted.			*/
/*									*/
/************************************************************************/

int
_get_zeof(int error)
	{
	int			result;

	result = ChanInChar(_boot_chan_in);
	if(result != 0)
		{
		if(error)
			_request_errno();
		return (EOF);
		}
	return (result);
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_1b						*/
/*									*/
/*	Read a 1 byte value from the host.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The byte value read.				*/
/*									*/
/************************************************************************/

int
_get_1b(void)
	{
	return ((int) ChanInChar(_boot_chan_in));
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_2b						*/
/*									*/
/*	Read a 2 byte value from the host.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The two byte value read (little endian).	*/
/*									*/
/************************************************************************/

int
_get_2b(void)
	{
	UCHAR			buf[2];

	ChanIn(_boot_chan_in,(char *) buf,2);	/* Read data */
	return ((int) (buf[0] | (buf[1] << 8)));
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_2beof					*/
/*									*/
/*	Read a 2 byte value from the host.  If the value is equal to	*/
/*	RPC_EOF, request "errno" and return EOF, otherwise return value.*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The two bytes read/converted (little endian).	*/
/*									*/
/************************************************************************/

int
_get_2beof(void)
	{
	UCHAR			buf[2];
	int			result;

	ChanIn(_boot_chan_in,(char *) buf,2);	/* Read data */
	result = buf[0] | (buf[1] << 8);
	if(result == (RPC_EOF & 0xFFFF))
		{
		_request_errno();
		return (EOF);
		}
	return (result);
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_4b						*/
/*									*/
/*	Read a 4 byte signed value from the host.			*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The four byte value read (little endian).  If	*/
/*			the code is running on a T2 processor the top	*/
/*			two bytes are simply thrown away.		*/
/*									*/
/************************************************************************/

long
_get_4b(void)
	{
	long			temp[2];	/* temp[1] gets excess on T2 */

	ChanIn(_boot_chan_in,(char *) temp,4);	/* Read four bytes */
	return (temp[0]);
	}

/************************************************************************/
/*									*/
/*	Routine:	_put_str					*/
/*									*/
/*	Send a string of specified length to the host system.		*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer to send).*/
/*		"length"The buffer length to send.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_put_str(UCHAR *sptr,int length)
	{
	ChanOut(_boot_chan_out,(char *) sptr,length);	/* Send data */
	}

/************************************************************************/
/*									*/
/*	Routine:	_put_1b						*/
/*									*/
/*	Send a single byte to the host system.				*/
/*									*/
/*	Parameters:							*/
/*		"c"	The character to send.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_put_1b(int c)
	{
	ChanOutChar(_boot_chan_out,c);	/* Send byte */
	}

/************************************************************************/
/*									*/
/*	Routine:	_request_errno					*/
/*									*/
/*	Request the host system value of "errno" to update the local	*/
/*	copy.								*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_request_errno(void)
	{
	UCHAR			buf[3];

	_send_cmd(buf,FN_ERRNO,3);
	errno = _get_4b();		/* Host "errno" value */
	}

/************************************************************************/
/*									*/
/*	Routine:	_send_cmd					*/
/*									*/
/*	Send a I/O command to the host system.  Note that the first	*/
/*	three bytes of the buffer being sent must be empty so that this	*/
/*	function can insert the length and function code and send the	*/
/*	whole mess as one message.					*/
/*									*/
/*	Parameters:							*/
/*		"buf"	A pointer to the buffer to send.		*/
/*		"func"	The function code to send.			*/
/*		"length"Total message length in bytes.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_send_cmd(UCHAR *buf,int func,int length)
	{
	buf[0] = length - 2;		/* Subtract length bytes */
	buf[1] = ((length - 2) >> 8);	/* Subtract length bytes */
	buf[2] = func;
	ChanOut(_boot_chan_out,(char *) buf,length);	/* Send data */
	}

/************************************************************************/
/*									*/
/*	Routine:	_send_cmdstr					*/
/*									*/
/*	Send a I/O command to the host system and indicate a trailing	*/
/*	string will be alos sent.  Note that the first three bytes of	*/
/*	the buffer being sent must be empty so that this function can	*/
/*	insert the length and function code and send the non-string	*/
/*	part of the whole mess as one message.				*/
/*									*/
/*	Parameters:							*/
/*		"buf"	A pointer to the buffer to send.		*/
/*		"func"	The function code to send.			*/
/*		"len1"	Length of fixed part of message (bytes).	*/
/*		"len2"	Length of variable part of message (bytes).	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_send_cmdstr(UCHAR *buf,int func,int len1,int len2)
	{
	len2 += (len1 - 2);		/* Subtract length bytes */
	buf[0] = len2;
	buf[1] = (len2 >> 8);
	buf[2] = func;
	ChanOut(_boot_chan_out,(char *) buf,len1);	/* Send fixed data */
	}
