/************************************************************************/
/*									*/
/*	File:								*/
/*		PP4.C	Token processing and related functions.		*/
/*									*/
/*	Name:								*/
/*		PP - A preprocessor, principally for the C language.	*/
/*									*/
/*	Written by:							*/
/*			Gary Oliver					*/
/*			3420 NW Elmwood Dr.				*/
/*			PO Box 826					*/
/*			Corvallis, Oregon 97339				*/
/*			(503)758-5549					*/
/*	Maintained by:							*/
/*			Kirk Bailey					*/
/*			Logical Systems					*/
/*			P.O. Box 1702					*/
/*			Corvallis, OR 97339				*/
/*			(503)753-9051					*/
/*									*/
/*	This program is hereby placed in the public domain.  In		*/
/*	contrast to other claims of "public domain", this means no	*/
/*	copyright is claimed and you may do anything you like with PP,	*/
/*	including selling it!  As a gesture of courtesy, please retain	*/
/*	the authorship information in the source code and		*/
/*	documentation.							*/
/*									*/
/*	Functions defined within this module:				*/
/*									*/
/*		addstr		Add a string to end of another string.	*/
/*		getnstoken	Get next non-space token.		*/
/*		gettoken	Get next token.				*/
/*		istype		Return boolean for char type mask.	*/
/*		memmov		Move bytes from place to place (MOVMEM)	*/
/*		pbcstr		Push back a copied string.		*/
/*		pbstr		Push back a string.			*/
/*		pushback	Push back a single character.		*/
/*		puttoken	Write a token to Output file.		*/
/*		type		Return type of token.			*/
/*									*/
/************************************************************************/

#include "pp.h"

/************************************************************************/
/*									*/
/*	addstr								*/
/*									*/
/*	Add a string to a buffer, test a limit and print a message	*/
/*	if the length of the new string would exceed the limit.		*/
/*									*/
/*	The error message will not be printed if a previous message	*/
/*	at the same address was printed.  A successful addstr() will	*/
/*	clear this state.  The logic will fail, however, if an error	*/
/*	is detected on one call, with a message printed, and then on	*/
/*	a next, unrelated call, the same msg is passed.  This can	*/
/*	be bypassed by insuring that the output buffer (old) is at	*/
/*	least twice (+1 byte) the size of the input (new) buffer.	*/
/*	For token building, the output buffer must be at least twice	*/
/*	(+1) the size of the token input buffer to insure that at	*/
/*	least ONE successful call to addstr can be made for each	*/
/*	time a string is to be built.					*/
/*									*/
/************************************************************************/
char *
addstr(old, limit, msg, new)
  char *old, *limit, *msg, *new;
    {
    register char *o, *n;

    static char *origmsg = 0;

    if((old + strlen(new)) >= limit)
	{
	if(msg != origmsg)
	    non_fatal(msg,"");		/* Don't print multiple messages */

	origmsg = msg;
	return (old);
	}
    else
	{
	origmsg = NULL;		/* Clear the error condition */
	o = old;
 	n = new;
	while(*o++ = *n++)
	    ;	/* Copy strings */
	return (o-1);	/* Next char position in output string */
	}
    }

/************************************************************************/
/*									*/
/*	getnstoken							*/
/*									*/
/*	Get next non-space token.					*/
/*									*/
/************************************************************************/
int
getnstoken(f)
  int f;
    {
    extern int gettoken();

    register int t;

    while(istype(t = gettoken(f), C_W) && (t != '\n'))
	;			/* Skip space tokens but not newlines */
    return (t);			/* Last token read */
    }

/************************************************************************/
/*									*/
/*	gettoken							*/
/*									*/
/*	Get alphanumeric string or single non-alpha for define		*/
/*									*/
/************************************************************************/
int
gettoken(f)
  int f;
    {
    extern int type();

    register int t, nt;
    register char *p, ch;
    register int comment_level, tmask, numstate, fail;

/*	Define numstate values:					*/
#define	F_INTPART	0	/* numstate for integer part.	*/
#define	F_FRAC		1	/* fractional part.		*/
#define	F_EXPSIGN	2	/* exponent sign test.		*/
#define	F_EXP		3	/* exponent digits.		*/
#define	F_LAST		4	/* after occurrance of f|F|l|L.	*/

#if	DEBUG
if(Debug) printf("gettoken: ");
#endif	/* DEBUG */

    p = Token;
    t = nextch();
    Token[ 0 ] = t;

    Tokenfile = Filelevel;	/* Remember file number and	*/
    Tokenline = Line;		/* Line number for this token	*/

    if(istype(t, C_L|C_D|C_W))
	{
	if(istype(t, C_L))
	    {
	    for (p = &Token[1]; p < &Token[TOKENSIZE]; p++)
		{
		t = nextch();
		*p = t;
		if(!istype(t, C_L|C_D))
		    break;
		}
	    }
	else
	if(istype(t, C_D))
	    {
	    if(t == '0')
		{
		/* Possible octal/hex numeric token */
		t = nextch();	/* Get next one */
		*++p = t;	/* Store into token buffer */

		if((t == 'x') || (t == 'X'))
		    tmask = C_X;	/* Only hex digits */
		else
		if(istype(t, C_D))
		    tmask = C_D;	/* 0..9 (8 & 9 are "octal" digits) */
		else
		    tmask = 0;		/* not part of digits */

		while((p < &Token[TOKENSIZE]) && (tmask != 0))
		    {
		    t = nextch();
		    *++p = t;
		    if(!istype(t, tmask))
			tmask = 0;
		    }

		if((t == 'l') || (t == 'L') || (t == 'u') || (t == 'U'))
		    {
		    /* Last was a numeric terminator -- look ahead one char */
		    t = nextch();
		    *++p = t;	/* Get next token char for later refetch */
		    }
		}
	    else
		{
		numstate = F_INTPART;
		fail = FALSE;
		for(p = &Token[1]; p < &Token[TOKENSIZE] && !fail; /*none*/ )
		    {
		    t = nextch();
		    *p = t;
		    switch(numstate)
			{
			case F_INTPART:
				if(istype(t, C_D))
				    break;
				else
				if(t == '.')
				    {
				    numstate = F_FRAC;
				    break;
				    }

				/* Fall thru to fraction test.	If t	*/
				/* is not 'e|E', it will bail out.	*/

			case F_FRAC:
				if(istype(t, C_D))
				    break;	/* fraction ok so far */
				else
				if((t == 'e') || (t == 'E'))
				    {
				    numstate = F_EXPSIGN;	/* tst sign */
				    break;
				    }
				else
				if((t == 'l') || (t == 'L') ||
				   (t == 'f') || (t == 'F'))
				    {
				    numstate = F_LAST;
				    break;
				    }

				fail = TRUE;	/* Failure */
				continue;

			case F_EXPSIGN:
				if((t == '+') || (t == '-') || istype(t, C_D))
				    {
				    numstate = F_EXP;
				    break;
				    }
				fail = TRUE;
				continue;	/* Error here -- loop fails */

			case F_EXP:
				if(istype(t, C_D))
				    break;	/* digits ok */

				if((t == 'f') || (t == 'F'))
				    {
				    numstate = F_LAST;
				    break;	/* Dummy state to end it */
				    }
				fail = TRUE;
				continue;	/* Error -- loop will fail */
				    
			case F_LAST:
				fail = TRUE;	/* This forces failure	*/
				continue;	/* after an 'f|F|l|L'.	*/
			}
		    ++p;	/* Advance char pointer into Token */
		    }
		}
	    }
	else
	    {
	    /* Must be some type of whitespace */
	    while(istype((t = nextch()), C_W))
		;
	    if(t == '\n')
		{
		/* Just forget we saw any of the above spaces */
		p = Token;		/* Reset token pointer */
		Token[0] = '\n';	/* a newline is all we have */
		}
	    else
		{
		p = Token + 1;		/* Leave only one in buffer */
		Token[0] = ' ';		/* Make it a single space */
		/* t = token to pushback.	*/
		}
	    }
	}
    else
	{
	if(t == '\\')
	    {
	    if((t = nextch()) == '\n')
		{
		Token[0] = ' ';		/* Generic whitespace token */
		Token[1] = '\0';
		return (' ');
		}
	    pushback(t);
	    t = '\\';		/* Return an escape char */
	    }
	else
 	if (((t == '"') || (t == '\'')) && (f & GT_STR))
	    {

#if	DEBUG
if (Debug) printf(" in quote");
#endif	/* DEBUG */

	    for(p = &Token[1]; p < &Token[TOKENSIZE]; p++)
		{
		if ((*p = nextch()) != '\\')
		    {
		    if((*p == t) || (*p == '\n'))
			break;
/*
 *	BUG: if a closing " or ' is missing from the string, the termination
 *	upon \n returns the \n as part of the string -- not as the NEXT token
 *	fetched.  This confuses readline() among others.  26-Jul-85 GO.
 */
		    }
		else
		    {
		    /* Fetched a '\\', see if next char is newline, and */
		    /* ignore both if it is.				*/
		    if((ch = nextch()) != '\n')
			*++p = ch;
		    else
			{
			/* Unrecord the '\\' already saved. */
			--p;
			}
		    }
		}
	    if(p >= &Token[TOKENSIZE])
		non_fatal("Token too long","");

	    p[1] = '\0';
	    return(t);
	    }
	else
 	if ((t == '<') && (f == GT_ANGLE))
	    {

#if	DEBUG
if (Debug) printf(" in angle bracket");
#endif	/* DEBUG */

	    for(p = &Token[1]; p < &Token[TOKENSIZE]; p++)
		{
		*p = nextch();
		if((*p == '>') || (*p == '\n'))
			break;
		}
	    if(p >= &Token[TOKENSIZE])
		non_fatal("Token too long","");

	    p[1] = '\0';
	    return(t);
	    }
	else
	    {
	    if((t == '/') && ((nt = nextch()) == '*'))
		{
		comment_level = 1;
		/* p == Token (no string yet) from above */
		t = ' ';
		Token[0] = t;
#if	DEBUG
if (Debug) printf(" in comment");
#endif	/* DEBUG */

		do
		    {
		    while ((t != '*') && (t != '/') && (t != EOF))
			t = nextch();

		    if (t == EOF)
			{
			non_fatal("EOF in comment","");
			return (EOF);
			}
		    nt = nextch();
		    if ((t == '/') && (nt == '*'))
			{
			if(! A_crecurse)
			    warning("\"/*\" found in comment","");
			else
			    comment_level++;	/* Count up */
			}
		    if ((t == '*') && (nt == '/'))
			comment_level--;
	    	    if (comment_level > 0)
			t = nt;
		    }
		while(comment_level > 0);
		t = Token[0];
		}
	    else
	    if ((t == '/') && (nt != '*'))
		pushback(nt);
	    }

	if (istype(t, C_M))
	    {
	    /* One of several meta token chars */
	    switch(t)
		{
		case LINE_TOKEN:
			sprintf(Token, "%d", Line);
			break;

		case FILE_TOKEN:
			sprintf(Token, "\"%s\"",Filestack[Filelevel]->f_name);
			break;

		case TIME_TOKEN:
			sprintf(Token, "\"%s\"",_Time);
			break;

		case DATE_TOKEN:
			sprintf(Token, "\"%s\"",Date);
			break;

		case NOW_TOKEN:
			sprintf(Token, "%u", Unique);
			break;

		case NEXT_TOKEN:
			sprintf(Token, "%u", ++Unique);
			break;

		case PREV_TOKEN:
			sprintf(Token, "%u", --Unique);
			break;

		default:
			Token[0] = t;
			Token[1] = '\0';
			break;
		}
	    return (type(Token[0] & 0xFF));
	    }
	}

    if (p >= &Token[TOKENSIZE])
	non_fatal("Token too long","");

    if(p > Token)
	{
	--p;
	pushback(t);		/* Push back previous token value */
	t = type(Token[0] & 0xFF);
	}

    p[1] = '\0';	/* Null terminated */

#if	DEBUG
if(Debug) printf(" returning: <%s> type: %c\n", Token, t);
#endif	/* DEBUG */

    return(t);
    }

#if	!PP
/************************************************************************/
/*									*/
/*	istype								*/
/*									*/
/*	Return TRUE if char is of type specified.			*/
/*									*/
/*	This function is here in case the macro capability cannot be	*/
/*	exercized. (Bootstrapping the Preprocessor.)			*/
/*									*/
/************************************************************************/
int
istype(c, v)
  int c, v;
    {
    return ((typetab+1)[c]&v);
    }
#endif	/* !PP */

#ifndef	memmov
/************************************************************************/
/*									*/
/*	memmov								*/
/*									*/
/*	Move a block of memory from one byte location to another.	*/
/*									*/
/************************************************************************/
memmov(f, t, l)
  register char *f, *t;
  register unsigned l;
    {    
    while(l--)
	*t++ = *f++;
    }
#endif	/* memmov */

/************************************************************************/
/*									*/
/*	pbcstr								*/
/*									*/
/*	Push back a copy of a string so that it is the first to be	*/
/*	refetched.  Note that we put two copies of the starting string	*/
/*	address on the stack:  The first is used as the working		*/
/*	pointer, the second serves to remember the string beginning for	*/
/*	later freeing.							*/
/*									*/
/************************************************************************/
pbcstr(s)
  char *s;
    {
    extern char	*malloc();
    extern int gchpb();

    char *cp;
    unsigned	int	length;

    if((cp = malloc((unsigned) (length = (strlen(s) + 1)))) == NULL)
	out_of_memory();
    memmov(s, cp, length);		/* Make a copy of memory */

    if (Pbbufp++ >= &Pbbuf[PUSHBACKSIZE-1])
	fatal("Pushback buffer overflow","");
    Pbbufp->pb_type = PB_STRING;
    Pbbufp->pb_val.pb_str = cp;

    if (Pbbufp++ >= &Pbbuf[PUSHBACKSIZE-1])
	fatal("Pushback buffer overflow","");
    Pbbufp->pb_type = PB_STRING;
    Pbbufp->pb_val.pb_str = cp;

    Nextch = gchpb;
    }

/************************************************************************/
/*									*/
/*	pbstr								*/
/*									*/
/*	Push back a string so it is first to be re-fetched.		*/
/*									*/
/************************************************************************/
pbstr(in)
  char *in;
    {
    register int i;

    for (i = strlen(in) - 1; i >= 0; i--)
	pushback(in[i] & 0xFF);
    }

/************************************************************************/
/*									*/
/*	pushback							*/
/*									*/
/*	Push back a character so it is first to be re-fetched.		*/
/*									*/
/************************************************************************/
pushback(c)
  int c;
    {
    extern int gchpb();

    if (Pbbufp++ >= &Pbbuf[PUSHBACKSIZE-1])
	fatal("Pushback buffer overflow","");

    Pbbufp->pb_type = PB_CHAR;
    Pbbufp->pb_val.pb_char = c;

    Nextch = gchpb;
    }

/************************************************************************/
/*									*/
/*	puttoken							*/
/*									*/
/*	Write token to output file.					*/
/*									*/
/************************************************************************/
puttoken(s)
  char s[];
    {
    register int ch;
    register char *str;
    static int lastoutc = '\n';	/* Last char written */

    str = s;

#if	DEBUG
if (Debug) printf("puttoken: <%s>\n", str);
#endif	/* DEBUG */

    if(istype(*str & 0xFF, C_N))	/* Ignore null tokens */
	return;

    if(Lineopt)
	{
	/* Get line numbers in sync before emitting token */

#if	(TARGET == ISQC) OR (TARGET == ISTC) OR (TARGET == ISQCX)
	if(!Do_asm && (((*str != '\n') && (Outline != Line)) || Do_name))
#else	/* !((TARGET == ISQC) OR (TARGET == ISTC) OR (TARGET == ISQCX)) */
	if(((*str != '\n') && (Outline != Line)) || Do_name)
#endif	/* (TARGET == ISQC) OR (TARGET == ISTC) OR (TARGET == ISQCX) */
	    {
	    do_line(lastoutc == '\n');	/* Parameter is true it at BOL	*/
	    lastoutc = '\n';		/* We're at BOL now...		*/
	    }

	while(ch = *str++)
	    {
	    if(ch == '\n')
		{
		if(lastoutc != '\n')
		    {
		    putc(lastoutc = '\n', Output);
		    Outline++;
		    }
		/* else no character written if lastoutc WAS a newline */
		}
	    else
		putc(lastoutc = ch, Output);
	    }
	}
    else
	{
	while(ch = *str++)
	    putc(ch, Output);	/* If not line mode, just output token */
	}
    }

/************************************************************************/
/*									*/
/*	type								*/
/*									*/
/*	Return token type of character.					*/
/*									*/
/************************************************************************/
int
type(c)
  int c;
    {
    if (istype(c, C_L))
	return(LETTER);
    else
    if (istype(c, C_D))
	return(NUMBER);
    else
	return(c);
    }
