/************************************************************************/
/*									*/
/*	File:	cio.c							*/
/*									*/
/*	Host side of "C" I/O driver for the INMOS Transputer.		*/
/*									*/
/*	This program must be linked with a version of the INMOS		*/
/*	"link.c" I/O interface to customize it to the associated	*/
/*	hardware!							*/
/*									*/
/*	Written by Dave Regan & Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1987-1988 by Logical Systems.		*/
/*									*/
/*				11/2/88					*/
/*									*/
/************************************************************************/

#include	<time.h>
#include	<fcntl.h>
#include	<conio.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<io.h>
#include	<stdio.h>
#include	<errno.h>
#include	<stdlib.h>
#include	<string.h>
#include	"taldef.h"
#include	"tload.h"
#include	"cio.h"

#define	MAXFILES	_NFILE		/* # of legal open files */

/************************************************************************/
/*									*/
/*		Host Library Configuration Options			*/
/*									*/
/*	TRUE to enable, FALSE to disable.  Note that these options are	*/
/*	required since the host "C" environment may not conform to ANSI	*/
/*	library specifications.  In general, the two "standard"		*/
/*	libraries at the current moment are ANSI and SYS5.  Set the	*/
/*	macros to FALSE if you have SYS5 (many "C" libraries are a	*/
/*	mixture so you will probably have to configure them		*/
/*	individually in conjunction with your "C" compiler reference	*/
/*	manual).							*/
/*									*/
/************************************************************************/

/*
 *	ANSIFSEEK should be set TRUE if the return value from "fseek" conforms
 *	to ANSI (zero if no errors, non-zero for errors).  If you set this
 *	option to FALSE the return value will be complemented before being
 *	sent to the Transputer (to simulate ANSI).
 */
#define	ANSIFSEEK	TRUE

/*
 *	ANSISYSTEM should be set to TRUE if the "system" function can be passed
 *	a NULL pointer to inquire if a command processor exists.
 */
#define	ANSISYSTEM	TRUE

/************************************************************************/
/*									*/
/*			Global Data Definitions				*/
/*									*/
/************************************************************************/

int	bytes_word	=	4;	/* # of bytes in target system word */
UCHAR	cmd_buf[MAX_MSG_DATA + MAX_MSG_OVER] = {0};	/* Buf from T */
int	LinkId		=	0;	/* LinkId from "OpenLink" */
STRING	open_str	=	NULL;	/* Link name for "OpenLink" */
UCHAR	rpy_buf[MAX_MSG_DATA + MAX_MSG_OVER] = {0};	/* Buf to T */
STREAM	strm_tbl[MAXFILES] = {NULL};	/* Stream translation table */

/************************************************************************/
/*									*/
/*			Function Type Declarations			*/
/*									*/
/************************************************************************/

VOID	banner();
VOID	close_strm();
STRING	dec_str();
STREAM	dec_strm();
int	dec_1b();
int	dec_2b();
SLONG	dec_4b();
VOID	enc_str();
VOID	enc_time();
VOID	enc_2b();
VOID	enc_4b();
VOID	err_dump();
VOID	exit();
VOID	fatal();
VOID	main();
int	open_stream();
VOID	reply_2beof();
VOID	reply_nlstr();
VOID	reply_str();
VOID	reply_zeof();
VOID	reply_1b();
VOID	reply_2b();
VOID	reply_4b();
VOID	warning();

/*
 *	The following functions may be called only for side effects.
 */
/*lint +fvr */

/*lint -fvr */

/************************************************************************/
/*									*/
/*	Routine:	main						*/
/*									*/
/*	Opens up communication with the transputer link adapter on a	*/
/*	INMOS Transputer and handles "C" file I/O requests from the	*/
/*	connected transputer.						*/
/*									*/
/*	Parameters:							*/
/*		"argc"	An integer count of the command line arguments.	*/
/*		"argv"	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
main(argc,argv)
	int			argc;
	char			*argv[];
	{
	extern	STRING		getenv();

	SLONG			args_addr;
	SLONG			args16_size;
	SLONG			args32_size;
	UCHAR			*cmd_ptr;
	int			exit_status;
	char			numbuf[20];
	UCHAR			*rpy_ptr;
	int			t_int1;
	int			t_int2;
	int			t_int3;
	SLONG			t_long1;
	STRING			t_str1;
	STREAM			t_strm1;
	time_t			t_time1;
	struct	tm		*t_tm1;

	banner();
	open_str = getenv(LINK_NAME);	/* Get desired link name if present */
	LinkId = OpenLink(open_str);	/* Open specified link channel */
	if(open_str == NULL)
		open_str = "<default>";
	if(LinkId <= 0)
		fatal(FALSE,"Unable to open link: ",open_str);
/*
 *	Initialize stream translation table.
 */
	for(t_int1 = 0; t_int1 < MAXFILES; t_int1++)
		strm_tbl[t_int1] = NULL;
	strm_tbl[0] = stdin;
	strm_tbl[1] = stdout;
	strm_tbl[2] = stderr;
/*
 *	Compute amount of Transputer workspace required to hold the "argv"
 *	array and the arguments themselves.  This data may be requested by the
 *	root Transputer if desired to initialize the "argc"/"argv" values.
 *	Note that we actually compute it two ways for either a 16 bit or 32 bit
 *	processor.
 *
 *	Note that the two loaders (LD-ONE & LD-NET), offset the "argv" array so
 *	that "argv[0]" is unused since the MSC "execvp" call clobbers it
 *	anyway.
 */
	if(argc > 1)		/* If more than one argument, ignore argv[0] */
		{
		argc--;
		argv++;
		}
	for(args16_size = 0, t_int1 = 0; t_int1 < argc; t_int1++)
		args16_size += (strlen(argv[t_int1]) + 1);
	args32_size = args16_size;
	args16_size += ((argc + 1) * 2);
	if(args16_size & (2 - 1))
		args16_size = (args16_size & (~(2L - 1L))) + 2L;
	args32_size += ((argc + 1) * 4);
	if(args32_size & (4 - 1))
		args32_size = (args32_size & (~(4L - 1L))) + 4L;
/*
 *	Handle server requests from Transputer.
 */
	FOREVER
		{
		if(ReadLink(LinkId,(STRING) cmd_buf,2,INFINITE_TIMEOUT) < 0)
			fatal(FALSE,"Error reading from link: ",open_str);
		cmd_ptr = cmd_buf;
		if(ReadLink(LinkId,(STRING) cmd_buf,dec_2b(&cmd_ptr),
			INFINITE_TIMEOUT) < 0)
			{
			fatal(FALSE,"Error reading from link: ",open_str);
			}
		cmd_ptr = cmd_buf;
		switch((int) dec_1b(&cmd_ptr))
			{
			case	FN_CIOEXT:
					t_int1 = dec_2b(&cmd_ptr);
					t_int2 = dec_2b(&cmd_ptr);
					t_int3 = dec_2b(&cmd_ptr);
					t_int1 = _cioext(t_int1,cmd_ptr,t_int2,
						(STRING) &rpy_buf[10],t_int3);
					rpy_ptr = &rpy_buf[2];
					enc_4b(&rpy_ptr,(SLONG) t_int1);
					enc_4b(&rpy_ptr,(SLONG) errno);
					reply_str(rpy_buf,t_int3 + 8);
					continue;

			case	FN_CLEARERR:
					clearerr(dec_strm(&cmd_ptr));
					continue;

			case	FN_CLOSE:
					t_int1 = dec_2b(&cmd_ptr);
					reply_zeof(close(t_int1));
					continue;

			case	FN_CREAT:
					t_int1 = dec_4b(&cmd_ptr);
					reply_2beof(creat((STRING) cmd_ptr,
						t_int1));
					continue;

			case	FN_DUP:
					reply_2beof(dup(dec_2b(&cmd_ptr)));
					continue;

			case	FN_DUP2:
					t_int1 = dec_2b(&cmd_ptr);
					reply_zeof(dup2(t_int1,
						dec_2b(&cmd_ptr)));
					continue;

			case	FN_ERRNO:
					reply_4b((SLONG) errno);
					continue;

			case	FN_FCLOSE:
					t_str1 = (STRING) cmd_ptr;
					t_strm1 = dec_strm(&cmd_ptr);
					if((t_int1 = fclose(t_strm1)) == 0)
						{
/* Backup to start of stream index */		cmd_ptr = (UCHAR *) t_str1;
						close_strm(&cmd_ptr);
						}
					reply_zeof(t_int1);
					continue;

			case	FN_FCLOSEALL:
					t_int1 = fcloseall();
					if(t_int1 != EOF)
						{
/* Close all except STDIN/STDOUT/STDERR */	for(t_int2 = 3;
							t_int2 < MAXFILES;
								t_int2++);
							{
							strm_tbl[t_int2]=NULL;
							}
						}
					reply_2beof(t_int1);
					continue;
						
			case	FN_FDOPEN:
					t_int1 = dec_2b(&cmd_ptr);
					reply_2b(open_stream(fdopen(t_int1,
						(STRING) cmd_ptr)));
					continue;

			case	FN_FEOF:
					reply_zeof(feof(dec_strm(&cmd_ptr)));
					continue;

			case	FN_FERROR:
					reply_zeof(ferror(dec_strm(&cmd_ptr)));
					continue;

			case	FN_FFLUSH:
					reply_zeof(fflush(dec_strm(&cmd_ptr)));
					continue;

			case	FN_FGETC:
					reply_2beof(fgetc(dec_strm(&cmd_ptr)));
					continue;

			case	FN_FGETS:
					t_strm1 = dec_strm(&cmd_ptr);
					if((fgets((STRING) &rpy_buf[3],
						dec_2b(&cmd_ptr),
							t_strm1)) == NULL)
						{
						rpy_buf[2] = RPC_NULL;
						rpy_buf[3] = '\0';
						}
					else
						rpy_buf[2] = RPC_NOTNULL;
					reply_str(rpy_buf,(int)
						(strlen((STRING) &rpy_buf[3]) +
							1 + 1));
					continue;

			case	FN_FILENO:
					reply_2b(fileno(dec_strm(&cmd_ptr)));
					continue;

			case	FN_FOPEN:
					t_str1 = dec_str(&cmd_ptr);
					reply_2b(open_stream(fopen(t_str1,
						(STRING) cmd_ptr)));
					continue;

			case	FN_FPUTC:
					t_strm1 = dec_strm(&cmd_ptr);
					reply_2beof(fputc(dec_1b(&cmd_ptr),
						t_strm1));
					continue;

			case	FN_FREAD:
					t_strm1 = dec_strm(&cmd_ptr);
					t_int1 = dec_2b(&cmd_ptr);
					reply_str(rpy_buf,
						t_int1 * ((int) fread(
						(STRING) &rpy_buf[2],
						(UINT) t_int1,
						(UINT) dec_2b(&cmd_ptr),
						t_strm1)));
					continue;

			case	FN_FREOPEN:
					t_strm1 = dec_strm(&cmd_ptr);
					t_str1 = dec_str(&cmd_ptr);
					reply_2b(open_stream(freopen(t_str1,
						(STRING) cmd_ptr,t_strm1)));
					continue;

			case	FN_FSEEK:
					t_strm1 = dec_strm(&cmd_ptr);
					t_long1 = dec_4b(&cmd_ptr);
					switch(t_int1 = dec_1b(&cmd_ptr))
						{
						case	_SEEK_CUR:
							t_int1 = SEEK_CUR;
							break;

						case	_SEEK_END:
							t_int1 = SEEK_END;
							break;

						case	_SEEK_SET:
							t_int1 = SEEK_SET;
							break;
						default:
							sprintf(numbuf,"0x%X",
								t_int1);
			warning("'fseek' origin value undefined: ",numbuf);
							break;
						}
#if	ANSIFSEEK
					reply_zeof(fseek(t_strm1,t_long1,
#else	/* !ANSIFSEEK */
					reply_zeof(! fseek(t_strm1,t_long1,
#endif	/* ANSIFSEEK */
						t_int1));
					continue;

			case	FN_FTELL:
					t_long1 = ftell(dec_strm(&cmd_ptr));
					if(t_long1 == EOF)
						reply_4b((SLONG) RPC_EOF);
					else
						reply_4b(t_long1);
					continue;

			case	FN_FWRITE:
					t_strm1 = dec_strm(&cmd_ptr);
					t_int1 = dec_2b(&cmd_ptr);
					t_int2 = dec_2b(&cmd_ptr);
					reply_2b((int) fwrite((STRING) cmd_ptr,
						(UINT) t_int1,(UINT) t_int2,
						t_strm1));
					continue;
#if	H_OPSYS == MS_DOS
			case	FN_GETCH:
					reply_2b(getch());
					continue;

			case	FN_GETCHE:
					reply_2b(getche());
					continue;
#endif	/* MS_DOS */					
			case	FN_GETENV:
					if((t_str1 = getenv((STRING) cmd_ptr))
						== NULL)
						{
						rpy_buf[2] = RPC_NULL;
						rpy_buf[3] = '\0';
						}
					else
						{
						rpy_buf[2] = RPC_NOTNULL;
						strcpy((STRING) &rpy_buf[3],
							t_str1);
						}
					reply_str(rpy_buf,(int)
						(strlen((STRING) &rpy_buf[3]) +
							1 + 1));
					continue;

					
			case	FN_GETS:
					if((gets((STRING) &rpy_buf[3])) ==
						NULL)
						{
						rpy_buf[2] = RPC_NULL;
						rpy_buf[3] = '\0';
						}
					else
						rpy_buf[2] = RPC_NOTNULL;
					reply_str(rpy_buf,(int)
						(strlen((STRING) &rpy_buf[3]) +
							1 + 1));
					continue;
#if	H_OPSYS == MS_DOS
			case	FN_KBHIT:
					reply_2b(kbhit());
					continue;
#endif	/* MS_DOS */					
			case	FN_LSEEK:
					t_int1 = dec_2b(&cmd_ptr);
					t_long1 = dec_4b(&cmd_ptr);
					switch(t_int2 = dec_1b(&cmd_ptr))
						{
						case	_SEEK_CUR:
							t_int2 = SEEK_CUR;
							break;

						case	_SEEK_END:
							t_int2 = SEEK_END;
							break;

						case	_SEEK_SET:
							t_int2 = SEEK_SET;
							break;
						default:
							sprintf(numbuf,"0x%X",
								t_int2);
			warning("'lseek' origin value undefined: ",numbuf);
							break;
						}
					t_long1 = lseek(t_int1,t_long1,t_int2);
					if(t_long1 == EOF)
						reply_4b((SLONG) RPC_EOF);
					else
						reply_4b(t_long1);
					continue;

			case	FN_OPEN:
					t_int1 = dec_4b(&cmd_ptr);
					t_int2 = dec_4b(&cmd_ptr);
					reply_2beof(open((STRING) cmd_ptr,
						t_int1,t_int2));
					continue;

			case	FN_READ:
					t_int1 = dec_2b(&cmd_ptr);
					t_int1 = read(t_int1,
						(STRING) &rpy_buf[3],
						(UINT) dec_2b(&cmd_ptr));
					if(t_int1 == EOF)
						{
						t_int1 = 0;
						rpy_buf[2] = RPC_EOF;
						}
					else
						rpy_buf[2] = RPC_NOTEOF;
					reply_str(rpy_buf,t_int1 + 1);
					continue;

			case	FN_REMOVE:
					reply_zeof(remove((STRING) cmd_ptr));
					continue;

			case	FN_RENAME:
					t_str1 = dec_str(&cmd_ptr);
					reply_zeof(rename(t_str1,
						(STRING) cmd_ptr));
					continue;

			case	FN_SYSTEM:
					if(dec_1b(&cmd_ptr) == RPC_NULL)
						{
#if	!ANSISYSTEM
warning("Unable to pass NULL pointer to 'system' in non-ANSI library","");
						t_str1 = "";
#else	/* ANSISYSTEM */
						t_str1 = NULL;
#endif	/* !ANSISYSTEM */
						}
					else
						t_str1 = (STRING) cmd_ptr;
					reply_4b((SLONG) system(t_str1));
					continue;

			case	FN_TIME:
					reply_4b((SLONG) time(NULL));
					continue;
/*
 *	Note if GMT/UTC time is not available on a particular host platform
 *	the following code can be changed to unconditionally do the "reply_1b"
 *	of RPC_NULL.
 */
			case	FN_TIMEGMT:
					t_time1 = (time_t) dec_4b(&cmd_ptr);
					if((t_tm1 = gmtime(&t_time1)) == NULL)
						reply_1b(RPC_NULL);
					else
						{
						rpy_buf[0] = RPC_NOTNULL;
						rpy_ptr = &rpy_buf[1];
						enc_time(&rpy_ptr,t_tm1);
						reply_nlstr(rpy_buf,
							RPC_TIME_DATA + 1);
						}
					continue;

			case	FN_TIMELOC:
					t_time1 = (time_t) dec_4b(&cmd_ptr);
					rpy_ptr = rpy_buf;
					enc_time(&rpy_ptr,localtime(&t_time1));
					reply_nlstr(rpy_buf,RPC_TIME_DATA);
					continue;

			case	FN_TMPFILE:
					reply_2b(open_stream(tmpfile()));
					continue;

			case	FN_TMPNAM:
					if((tmpnam((STRING) &rpy_buf[3])) ==
						NULL)
						{
						rpy_buf[2] = RPC_NULL;
						rpy_buf[3] = '\0';
						}
					else
						rpy_buf[2] = RPC_NOTNULL;
					reply_str(rpy_buf,(int)
						(strlen((STRING) &rpy_buf[3]) +
							1 + 1));
					continue;

			case	FN_UNGETC:
					t_strm1 = dec_strm(&cmd_ptr);
					reply_2beof(ungetc(dec_1b(&cmd_ptr),
						t_strm1));
					continue;

			case	FN_WRITE:
					t_int1 = dec_2b(&cmd_ptr);
					t_int2 = dec_2b(&cmd_ptr);
					reply_2beof(write(t_int1,
						(STRING) cmd_ptr,
							(UINT) t_int2));
					continue;

/************************************************************************/
/*									*/
/*			Low usage rate I/O routines.			*/
/*									*/
/************************************************************************/

			case FN_ARG_GET:
/*
 *	Send argc/argv info and strings to root node.
 */
/* Start of arg space */		args_addr = dec_4b(&cmd_ptr) +
						((argc + 1L) *
							((SLONG) bytes_word));
/*
 *	First fill in "argv" table.
 */
					rpy_ptr = &rpy_buf[2];
					enc_4b(&rpy_ptr,(SLONG) argc);
					for(t_int1 = 0; t_int1 < argc;
						t_int1++)
						{
						if(bytes_word == 2)
							{
							enc_2b(&rpy_ptr, (int)
								args_addr);
							}
						else
							{
							enc_4b(&rpy_ptr,
								args_addr);
							}
						args_addr +=
							(strlen(argv[t_int1]) +
								1);
						}
					if(bytes_word == 2)
						enc_2b(&rpy_ptr,0);
					else
						enc_4b(&rpy_ptr,0L);
/*
 *	Now add argument strings.
 */
					for(t_int1 = 0; t_int1 < argc;
						t_int1++)
						{
						enc_str(&rpy_ptr,argv[t_int1]);
						}
					if(bytes_word == 2)
						{
						reply_str(rpy_buf,
							((int) args16_size)+4);
						}
					else
						{
						reply_str(rpy_buf,
							((int) args32_size)+4);
						}
					continue;

			case FN_ARG_SIZE:
/*
 *	Tell root node how many bytes (rounded up to nearest word), of
 *	workspace will be required to hold the "argc"/"argv" information and
 *	the actual argument strings.  We use
 */
					if((bytes_word=dec_2b(&cmd_ptr)) == 2)
						reply_4b(args16_size);
					else if(bytes_word == 4)
						reply_4b(args32_size);
					else
						{
			fatal(FALSE,"Unexpected target system word length","");
						}
					continue;

			case	FN_EXIT:
					exit_status = (int) dec_4b(&cmd_ptr);
					break;

			default:
					cmd_ptr = cmd_buf;
					sprintf(numbuf,"0x%X",
						dec_1b(&cmd_ptr));
					fatal(TRUE,"Invalid function code: ",
						numbuf);
					break;	/* Not really needed */
			}
		break;			/* Only get here on exit */
		}
/*
 *	Note that the above loop must be modified with whatever signal
 *	catching code is needed so that the following code gets executed if
 *	the execution of CIO is aborted on a REAL computer (one which cares
 *	about matching link "open" and "close" operations...).  Some "link.c"
 *	drivers may handle the signal catching internally so all may not be
 *	lost.
 */
	if(CloseLink(LinkId) < 0)
		{
		LinkId = 0;
		fatal(FALSE,"Unable to close link: ",open_str);
		}
	exit(exit_status);		/* Head for the barn */
	}

/************************************************************************/
/*									*/
/*	Routine:	banner						*/
/*									*/
/*	Display sign-on message to user.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
banner()
	{
	fprintf(stderr,
		"CIO ('C' I/O Driver), %s [Link I/O Driver: '%s']\n",
			VERSION,IODRIVER);
	fprintf(stderr,"%s\n\n",COPYRIGHT);
	}

/************************************************************************/
/*									*/
/*	Routine:	close_strm					*/
/*									*/
/*	Read a two byte integer from a buffer and translate it into a	*/
/*	stream index table entry which is marked "closed".  Note the	*/
/*	table index has been offset by 1 to allow the RPC_NULL return	*/
/*	value to be used without conflicting with a valid stream table	*/
/*	index.								*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
close_strm(sptr)
	UCHAR			**sptr;
	{
	int			index;
	char			numbuf[20];

	index = dec_2b(sptr) - 1;	/* Stream table index */
	if((index < 0) || (index >= MAXFILES))
		{
		sprintf(numbuf,"%d",index);
		warning("Stream number is out of range: ",numbuf);
		return;
		}
	if(strm_tbl[index] == NULL)
		{
		sprintf(numbuf,"%d",index);
		warning("Operation with a closed file/stream not in use: ",
			numbuf);
		return;
		}
	strm_tbl[index] = NULL;
	}

/************************************************************************/
/*									*/
/*	Routine:	dec_str						*/
/*									*/
/*	Read a variable length zero terminated string from the		*/
/*	specified buffer and update the buffer pointer past the string	*/
/*	termination.							*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*									*/
/*	Returns:	The original buffer value.			*/
/*									*/
/************************************************************************/

STRING
dec_str(sptr)
	UCHAR			**sptr;
	{
	STRING			rval;

	rval = (STRING) *sptr;
	*sptr = *sptr + strlen((STRING) *sptr) + 1;
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	dec_strm					*/
/*									*/
/*	Read a two byte integer from a buffer and translate it into a	*/
/*	stream number taken from the stream table.  Note the table	*/
/*	index has been offset by 1 to allow the RPC_NULL return value	*/
/*	to be sent without conflicting with a valid stream table index.	*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*									*/
/*	Returns:	The translated stream pointer.			*/
/*									*/
/************************************************************************/

STREAM
dec_strm(sptr)
	UCHAR			**sptr;
	{
	int			index;
	char			numbuf[20];

	index = dec_2b(sptr) - 1;	/* Stream table index */
	if((index < 0) || (index >= MAXFILES))
		{
		sprintf(numbuf,"%d",index);
		warning("Stream number is out of range: ",numbuf);
		return (stderr);	/* Default */
		}
	if(strm_tbl[index] == NULL)
		{
		sprintf(numbuf,"%d",index);
		warning("Operation with a closed file/stream not in use: ",
			numbuf);
		return (stderr);	/* Default */
		}
	return (strm_tbl[index]);
	}

/************************************************************************/
/*									*/
/*	Routine:	dec_1b, dec_2b, dec_4b				*/
/*									*/
/*	Read a one, two, or four byte signed value from the specified	*/
/*	buffer and return the resulting value.  The buffer pointer is	*/
/*	advanced past the data which was read.				*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*									*/
/*	Returns:	The resulting value.				*/
/*									*/
/************************************************************************/

int
dec_1b(sptr)
	UCHAR			**sptr;
	{
	return ((int) *(*sptr)++);
	}

int
dec_2b(sptr)
	UCHAR			**sptr;
	{
	int			rval;

	rval = *(*sptr)++;
	rval |= (*(*sptr)++ << 8);
	return (rval);
	}

SLONG
dec_4b(sptr)
	UCHAR			**sptr;
	{
	SLONG			rval;

	rval = *(*sptr)++;
	rval |= (*(*sptr)++ << 8);
	rval |= (((SLONG) *(*sptr)++) << 16);
	rval |= (((SLONG) *(*sptr)++) << 24);
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	enc_str						*/
/*									*/
/*	Place a "\0" terminated string into the specified buffer and	*/
/*	advance the buffer pointer past the placed object.		*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*		"s"	The string to place in the buffer.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
enc_str(sptr,s)
	UCHAR			**sptr;
	register STRING		s;
	{
	register STRING		cptr;

	for(cptr = (STRING) *sptr; *s != '\0';)
		*cptr++ = *s++;
	*cptr++ = '\0';
	*sptr = (UCHAR *) cptr;
	}

/************************************************************************/
/*									*/
/*	Routine:	enc_time					*/
/*									*/
/*	Place a broken down time from either "gmtime" or "localtime"	*/
/*	into the specified buffer and advance the buffer pointer past	*/
/*	the placed object.						*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*		"tmptr"	A pointer to the broken down time to be sent.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
enc_time(sptr,tmptr)
	UCHAR			**sptr;
	struct	tm		*tmptr;
	{
	struct	rpc_time	*t;

	t = (struct rpc_time *) *sptr;
	t->tm_sec = tmptr->tm_sec;
	t->tm_min = tmptr->tm_min;
	t->tm_hour = tmptr->tm_hour;
	t->tm_mday = tmptr->tm_mday;
	t->tm_mon = tmptr->tm_mon;
	t->tm_year = tmptr->tm_year;
	t->tm_wday = tmptr->tm_wday;
	t->low_tm_yday = (tmptr->tm_yday & 0xFF);
	t->high_tm_yday = ((tmptr->tm_yday >> 8) & 0xFF);
	t->tm_isdst = tmptr->tm_isdst;

	*sptr += RPC_TIME_DATA;
	}

/************************************************************************/
/*									*/
/*	Routine:	enc_2b, enc_4b					*/
/*									*/
/*	Place a two or four byte signed value into the specified buffer	*/
/*	and advance the buffer pointer past the placed object.		*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*		"value"	The value to place in the buffer.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
enc_2b(sptr,value)
	UCHAR			**sptr;
	int			value;
	{
	*(*sptr)++ = value;
	*(*sptr)++ = (value >> 8);
	}

VOID
enc_4b(sptr,value)
	UCHAR			**sptr;
	SLONG			value;
	{
	*(*sptr)++ = value;
	*(*sptr)++ = (value >>= 8);
	*(*sptr)++ = (value >>= 8);
	*(*sptr)++ = (value >> 8);
	}

/************************************************************************/
/*									*/
/*	Routine:	err_dump					*/
/*									*/
/*	Dump the contents of the first 8 bytes of the last message from	*/
/*	the transputer to the user to aid in diagnosing problems with	*/
/*	the protocol.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
err_dump()
	{
	int			i;

	fprintf(stderr,"Message: ");
	for(i = 0; i < 8; i++)
		fprintf(stderr,"%02x ",cmd_buf[i]);
	fprintf(stderr,"Active stream table slots: ");
	for(i = 0; i < MAXFILES; i++)
		{
		if(strm_tbl[i] != NULL)
			fprintf(stderr,"%d ",i);
		}
	fprintf(stderr,"\n");
	}

/************************************************************************/
/*									*/
/*	Routine:	fatal						*/
/*									*/
/*	Display error message to user and play dead.			*/
/*									*/
/*	Parameters:							*/
/*		"dump"	TRUE if the message buffer from the Transputer	*/
/*			should be displayed.				*/
/*		"str1"	The first part of the error message to display.	*/
/*		"str2"	The second part of the error message.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fatal(dump,str1,str2)
	BOOL			dump;
	STRING			str1;
	STRING			str2;
	{
	fprintf(stderr,"FATAL: %s%s\n",str1,str2);
	if(dump)
		err_dump();
	if(LinkId > 0)
		(VOID) CloseLink(LinkId);
	exit(ERRORS);
	}

/************************************************************************/
/*									*/
/*	Routine:	open_stream					*/
/*									*/
/*	Given a host I/O system supplied stream pointer translate into	*/
/*	a integer offset into the stream translation table to preserve	*/
/*	host pointer size transportability.				*/
/*									*/
/*	Parameters:							*/
/*		"fp"	The host stream pointer to translate.		*/
/*									*/
/*	Returns:	The translated stream table index.		*/
/*									*/
/************************************************************************/

int
open_stream(fp)
	STREAM			fp;
	{
	register int		count;

	if(fp == NULL)
		return (RPC_NULL);	/* If couldn't be opened */
	for(count = 0; count < MAXFILES; count++)
		{
		if(strm_tbl[count] == NULL)
			{
			strm_tbl[count] = fp;
/*
 *	Offset the table index by 1 to avoid conflicts with RPC_NULL (0).
 */
			return (count + 1);
			}
		}
	warning("Out of file stream slots","");
	fclose(fp);
	return (RPC_NULL);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_2beof					*/
/*									*/
/*	Returns a two byte result code to the Transputer.  If the	*/
/*	parameter is EOF the value is converted to RPC_EOF, otherwise	*/
/*	the value is passed through unchanged.				*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
reply_2beof(value)
	int			value;
	{
	if(value == EOF)
		value = RPC_EOF;
	reply_2b(value);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_nlstr					*/
/*									*/
/*	Returns a variable length string to the Transputer.		*/
/*									*/
/*	Parameters:							*/
/*		"buf"	The pointer to the buffer to send.		*/
/*		"length"The # of bytes to send.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
reply_nlstr(buf,length)
	UCHAR			*buf;
	int			length;
	{
	if(WriteLink(LinkId,(STRING) buf,length,INFINITE_TIMEOUT) < 0)
		fatal(FALSE,"Error writing to link: ",open_str);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_str					*/
/*									*/
/*	Returns a variable length string to the Transputer preceeded	*/
/*	with a two byte length header.  The incoming buffer pointer	*/
/*	must have two empty bytes at the beginning to hold the inserted	*/
/*	length.								*/
/*									*/
/*	Parameters:							*/
/*		"buf"	The pointer to the buffer to send.		*/
/*		"length"The # of bytes to send (less the length).	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
reply_str(buf,length)
	UCHAR			*buf;
	int			length;
	{
	buf[0] = length;
	buf[1] = (length >> 8);

	if(WriteLink(LinkId,(STRING) buf,length + 2,INFINITE_TIMEOUT) < 0)
		fatal(FALSE,"Error writing to link: ",open_str);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_zeof					*/
/*									*/
/*	Returns a one byte result code to the Transputer.  If the	*/
/*	parameter is zero the value is sent unchanged, otherwise the	*/
/*	value is converted to a one byte version of RPC_EOF.		*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
reply_zeof(value)
	int			value;
	{
	char			c;

	if(value != 0)
		c = RPC_EOF;
	else
		c = 0;

	if(WriteLink(LinkId,&c,1,INFINITE_TIMEOUT) < 0)
		fatal(FALSE,"Error writing to link: ",open_str);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_1b					*/
/*									*/
/*	Returns a one byte result code to the Transputer.		*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
reply_1b(value)
	int			value;
	{
	char			buf[1];

	buf[0] = value;

	if(WriteLink(LinkId,buf,1,INFINITE_TIMEOUT) < 0)
		fatal(FALSE,"Error writing to link: ",open_str);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_2b					*/
/*									*/
/*	Returns a two byte result code to the Transputer.		*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
reply_2b(value)
	int			value;
	{
	char			buf[2];

	buf[0] = value;
	buf[1] = (value >> 8);

	if(WriteLink(LinkId,buf,2,INFINITE_TIMEOUT) < 0)
		fatal(FALSE,"Error writing to link: ",open_str);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_4b					*/
/*									*/
/*	Returns a four byte result code to the Transputer.		*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
reply_4b(value)
	SLONG			value;
	{
	char			buf[4];

	buf[0] = value;
	buf[1] = (value >>= 8);
	buf[2] = (value >>= 8);
	buf[3] = (value >> 8);

	if(WriteLink(LinkId,buf,4,INFINITE_TIMEOUT) < 0)
		fatal(FALSE,"Error writing to link: ",open_str);
	}

/************************************************************************/
/*									*/
/*	Routine:	warning						*/
/*									*/
/*	Display warning message to user.				*/
/*									*/
/*	Parameters:							*/
/*		"str1"	The first part of the message to display.	*/
/*		"str2"	The second part of the message.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
warning(str1,str2)
	STRING			str1;
	STRING			str2;
	{
	fprintf(stderr,"WARNING: %s%s\n",str1,str2);
	err_dump();
	}
