/************************************************************************/
/*									*/
/*	File:	ld-net.c						*/
/*									*/
/*	Transputer program downloader for use with a network of		*/
/*	Transputers connected to a INMOS Transputer system.  This	*/
/*	utility takes a script file description of a network topology	*/
/*	and downloads the associated programs.  The network must	*/
/*	contain a subgraph which forms a tree rooted at	the Transputer	*/
/*	which is connected to the host system (with the reset and	*/
/*	subsystem connections configured in the INMOS fashion).  This	*/
/*	tree must also be a strict subset of the overall link		*/
/*	connection topology for proper operation.			*/
/*									*/
/*	This program must be linked with a version of the INMOS		*/
/*	"link.c" I/O interface to customize it to the associated	*/
/*	hardware!							*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1988-1989 by Logical Systems.		*/
/*									*/
/*				12/21/89				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<ctype.h>		/* Character classification stuff */
#include	<string.h>		/* String functions */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tload.h"		/* Transputer downloading info */

/************************************************************************/
/*									*/
/*			Global Data Definitions				*/
/*									*/
/************************************************************************/

/*
 *	Configuration stuff.  Timeout values are in milliseconds.
 */

#define	DEFAULT_SERVER	"cio"		/* Default host server */

#define	DEF_DEC_TOUT	1000L		/* 1 sec load decode timeout default */
#define	MAX_DEC_TOUT	20000L		/* 20 secs maximum decode timeout */
#define	MIN_DEC_TOUT	50L		/* .5 secs minimum decode timeout */

#define	DEF_LVL_TOUT	500L		/* .5 sec/level timeout default */
#define	MAX_LVL_TOUT	1000L		/* 1 secs/level maximum timeout */
#define	MIN_LVL_TOUT	25L		/* .025 secs/level minimum timeout */

#define	LINKIN16_ADDR	0x8008		/* Base addr of input links on T4/T8 */
#define	LINKIN32_ADDR	0x80000010	/* Base addr of input links on T4/T8 */
#define	LINKOUT16_ADDR	0x8000		/* Base addr of input links on T4/T8 */
#define	LINKOUT32_ADDR	0x80000000	/* Base addr of input links on T4/T8 */
#define	MAX_LINKS	4		/* Max # of links/processor */
#define	MAX_NODES	1000		/* Maximum user node # */

#define	MAX_PACKET	255		/* Longest packet with this driver */

/*
 *	I/O definitions.
 */

char	inbuf[FNSIZE + 1] =	{0};	/* Input load filename buffer */
STREAM	in_fp		=	NULL;	/* Input load file pointer */

char	infobuf[FNSIZE + 1] =	{0};	/* Network info filename buffer */
STREAM	info_fp		=	NULL;	/* Network info file pointer */

/*
 *	General definitions.
 */

#define	FROM_SYSTEM_RESET	0x01	/* Boot is from system chain */
#define	FROM_SUBSYS_RESET	0x02	/* Boot is from sub-system chain */

struct	link_info
	{
	int	other_node;		/* The node which this connects to */
	int	link_num;		/* Link # within other node if known */
	};

struct	node_info
	{
	UCHAR	boot_status;		/* Boot system/subsystem status */
	int	parent_num;		/* Who boots me via "boot_status" */
	int	system_output;		/* System output node connection */
	int	subsys_output;		/* Sub-system output node connection */
	int	next_boot;		/* Next user node to boot */
	int	boot_id;		/* Node id during booting */
	struct	program_info	*prog;	/* Info about prog to load on node */
	SLONG	timeout;		/* Timeout value for this node */
	struct	link_info	links[MAX_LINKS];	/* Links-to info */
	};

struct	program_info
	{
	struct	program_info	*next;	/* Next program on the list */
	int	prog_num;		/* Program # in list */
	UCHAR	cpu_type;		/* Cpu type for program */
	SLONG	entry;			/* Program entrypoint */
	char	name[FNSIZE + 1];	/* Name of program being loaded */
	SLONG	stack;			/* Initial program stack pointer */
	};

UINT	buf_size	=	MAX_PACKET;	/* Maximum buffer length */
SLONG	decode_timeout	=	(DEF_DEC_TOUT * 1000L) / MICRO_LOW_TICK;
char	host_server[FNSIZE + 1] = DEFAULT_SERVER;
SLONG	host_timeout	=	0;	/* Host communication timeout */
SLONG	level_timeout	=	(DEF_LVL_TOUT * 1000L) / MICRO_LOW_TICK;
UINT	line		=	0;	/* Current input line # */
int	LinkId		=	0;	/* LinkId from "OpenLink" */
struct	node_info	nodes[MAX_NODES+1] = {0}; /* Node information table */
STRING	open_str	=	NULL;	/* Name to use for "OpenLink" call */
char	parse_buf[MAXLNSIZE + 1] = {0};	/* Net info parsing buffer */
STRING	parse_ptr	=	NULL;	/* Parsing buffer pointer */
struct	program_info	*programs = NULL;	/* List of programs to load */
int	root_node	=	-1;	/* Root node */

/************************************************************************/
/*									*/
/*			Function Type Declarations			*/
/*									*/
/************************************************************************/

struct	program_info	*add_program();
VOID	banner();
int	boot_in_help();
VOID	boot_order();
int	boot_pre_help();
VOID	bootstrap();
VOID	close_input();
VOID	copy_out();
VOID	do_response();
VOID	download();
VOID	error_scan();
VOID	exit();
VOID	exit_close();
SLONG	eval();
VOID	fatal();
BOOL	get_fname();
SLONG	hexbin();
BOOL	ishex();
BOOL	isoct();
VOID	localize_boot();
VOID	local_store();
VOID	main();
VOID	makefn();
BOOL	match();
BOOL	match_first();
BOOL	not_fname();
int	open_input();
VOID	parse();
VOID	parse_fatal();
VOID	parse_node();
VOID	parse_warning();
SLONG	pick_ichan();
VOID	skip_space();
SLONG	tin_4b();
VOID	tout_str();
VOID	tout_1b();
VOID	tout_4b();
VOID	usagerr();
VOID	warning();

/*
 *	The following functions may be called only for side effects.
 */
/*lint +fvr */

int	in_1b();
int	in_2b();
SLONG	in_4b();

/*lint -fvr */

/************************************************************************/
/*									*/
/*	Routine:	main						*/
/*									*/
/*	Opens the network information file and controls the rest of the	*/
/*	operation.  Terminates by "exec"ing the desired I/O driver	*/
/*	process.							*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
main(argc,argv)
	int			argc;
	char			*argv[];
	{
	extern	STRING		getenv();

	int			node;

	banner();			/* Tell user who we are */
	if(argc < 2) usagerr();		/* Must have at least a info file */
/*
 *	Open network information file.
 */
	makefn(argv[1],infobuf,LD_INFO_EXT);	/* Make complete filename */
	if((info_fp = (STREAM) fopen(infobuf,"r")) == NULL)
		{
		fatal("Unable to open network information file: ",infobuf);
		}
/*
 *	Initialize node information.
 */
	for(node = 0; node <= MAX_NODES; node++)
		{
		nodes[node].prog = NULL;	/* Unused initially */
		nodes[node].system_output = -1;	/* Filled in later */
		nodes[node].subsys_output = -1;	/* Filled in later */
		nodes[node].boot_id = -1;	/* Node id during bootstrap */
		}
/*
 *	Parse the network information file, figure out the boot order and
 *	load the network!
 */
	parse();			/* Parse network info file */
	error_scan();			/* Error check and polish node info */
	boot_order();			/* Compute bootstrap ordering */
/*
 *	Reset the transputer and download the bootstrap program.
 */
	open_str = getenv(LINK_NAME);	/* Get desired link name if present */
	LinkId = OpenLink(open_str);	/* Open specified link channel */
	if(open_str == NULL)
		open_str = "(default)";
	if(LinkId < 0)
		fatal("Unable to open link: ",open_str);
	if(ResetLink(LinkId) < 0)
		fatal("Unable to reset link: ",open_str);
	bootstrap();			/* Reset/bootstrap the network */
	download();			/* Download the programs */
	if(CloseLink(LinkId) < 0)
		{
		LinkId = 0;		/* So "fatal" doesn't try again */
		fatal("Unable to close link: ",open_str);
		}
/*
 *	Close the network information file and hand control over to whatever
 *	I/O driver program has been specified (if any).
 *
 *	Note that a dummy slot is put in place in 'argv[0]' since the MSC
 *	'execvp' library routine trashes it anyway!  The handler must get the
 *	desired 'argv[0]' value from the 'argv[1]' slot, etc.
 */
	if(fclose(info_fp) == EOF)
		fatal("Unable to close network information file: ",infobuf);
	if(host_server[0] != '\0')
		{
		if(execvp(host_server,argv) == -1)
			fatal("Unable to EXEC I/O driver: ",host_server);
		}
	else
		warning("No I/O driver installed","");

	exit_close(NOERRORS);
	}

/************************************************************************/
/*									*/
/*	Routine:	add_program					*/
/*									*/
/*	See if a program to be loaded on the network is already on the	*/
/*	load list, if not add it.					*/
/*									*/
/*	Parameters:							*/
/*		s -	Points to the program name to add.		*/
/*									*/
/*	Returns:	A pointer to the "program_info" structure which	*/
/*			describes the program loading details.		*/
/*									*/
/************************************************************************/

struct	program_info	*
add_program(s)
	STRING			s;
	{
	extern	char		*malloc();

	register int		i;
	register struct	program_info	**p;
	register struct program_info	*temp;
/*
 *	See if already on list.
 */
	for(i = 0, p = (struct program_info **) &programs; *p != NULL;
		i++, p = &((*p)->next))
		{
		if(strcmp(s,(*p)->name) == 0)	/* Already on list? */
			return (*p);	/* Yep */
		}
/*
 *	Nope, malloc up a fresh unit.
 */
	if((temp = (struct program_info *)
		malloc(sizeof(struct program_info))) == NULL)
		{
	parse_fatal("Insufficient memory for names of programs being loaded",
			"");
		}
/*
 *	Link it into the programs-to-be-loaded list.
 */
	*p = temp;
	temp->next = NULL;		/* Current end of list */
	temp->prog_num = i;		/* Program # in list */
	strcpy(temp->name,s);		/* Program name */
/*
 *	Open the file to determine the cpu type and other overhead information
 *	(and also to verify that the program exists).
 */
	switch(open_input(s))
		{
		case	TYPE_UNKNOWN:
				temp->cpu_type = BID_T4 | BID_T8;
				break;

		case	TYPE_212:
				temp->cpu_type = BID_T2;
				break;

		case	TYPE_414:
				temp->cpu_type = BID_T4;
				break;

		case	TYPE_800:
				temp->cpu_type = BID_T8;
				break;

		default:
				fatal("Internal error #1","");
		}
	if(in_1b() != T_LOAD)		/* Should be T_LOAD record */
		fatal("Missing T_LOAD record in download input file: ",inbuf);
	in_4b();			/* Toss load address */
	if(in_1b() != T_STACK)		/* Should be T_STACK record */
		fatal("Missing T_STACK record in download input file: ",inbuf);
	temp->stack = in_4b();		/* Initial stack address */
	if(in_1b() != T_ENTRY)		/* Should be T_ENTRY record */
		fatal("Missing T_ENTRY record in download input file: ",inbuf);
	temp->entry = in_4b();		/* Entrypoint */
	close_input();
		
	return (temp);
	}

/************************************************************************/
/*									*/
/*	Routine:	banner						*/
/*									*/
/*	Display sign-on message to user.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
banner()
	{
	fprintf(stderr,
		"LD-NET (Network Loader), %s [Link I/O Driver: '%s']\n",
			VERSION,IODRIVER);
	fprintf(stderr,"%s\n\n",COPYRIGHT);
	}

/************************************************************************/
/*									*/
/*	Routine:	boot_order					*/
/*									*/
/*	Given the network configuration information, perform a in-order	*/
/*	traversal of the sub-tree of the network which will be used	*/
/*	during the bootstrapping operation.  This establishes the node	*/
/*	numbers used during the bootstrapping process and sets the	*/
/*	node depths within the bootstrap tree.  Then we process the	*/
/*	node data to see if any nodes to be booted aren't connected to	*/
/*	the root and convert the level information into timeout values.	*/
/*	Assuming everything is OK, we do a pre-order traversal to	*/
/*	establish a bootstrapping order for the nodes and report the	*/
/*	desired order back (via the "next_boot" linked list through the	*/
/*	"nodes" data structure).					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

/*
 *	Helper for "boot_order".
 *
 *	The in-order traverse routine to set the bootstrap node #'s and
 *	establish the node depths within the tree (and the maximum depth).
 */
int
bo_in_help(node,boot_num,level,max_level)
	int			node;
	int			boot_num;
	int			level;
	int			*max_level;
	{
	if(node == -1)
		return (boot_num);

	nodes[node].timeout = ++level;
	if(level > *max_level)
		*max_level = level;
	boot_num = bo_in_help(nodes[node].system_output,boot_num,level,
		max_level) + 1;
	nodes[node].boot_id = boot_num;
	return (bo_in_help(nodes[node].subsys_output,boot_num,level,
		max_level));
	}

/*
 *	Helper for "boot_order".
 *
 *	A pre-order traverse routine to establish the bootstrapping order.
 */
int
bo_pre_help(new,old)
	int			new;
	int			old;
	{
	if(new == -1)
		return (old);

	nodes[old].next_boot = new;
	old = bo_pre_help(nodes[new].system_output,new);
	return (bo_pre_help(nodes[new].subsys_output,old));
	}

VOID
boot_order()
	{
	BOOL			err_printed;
	int			max_level;
	register int		temp;
/*
 *	First scan through and set the level # of each node and the bootstrap
 *	node numbers.
 */
	max_level = 0;
	(VOID) bo_in_help(root_node,D_MASK,0,&max_level);
/*
 *	Compute the host communication timeout value (tenths of seconds!).
 */
	host_timeout = (((SLONG) MICRO_LOW_TICK) *
			(decode_timeout + (level_timeout *
				((SLONG) max_level + 1L)))) / 100000L;
/*
 *	Now scan through and see if all nodes can be reached (also convert
 *	bootstrap tree node depths to node timeout values).
 */
	for(err_printed = FALSE, temp = 1; temp <= MAX_NODES; temp++)
		{
		if(nodes[temp].prog == NULL)	/* In network? */
			continue;		/* Nope */
		if(nodes[temp].boot_id < 0)	/* Does it have a parent? */
			{			/* Nope */
			if(! err_printed)
				{
fprintf(stderr,
	"FATAL: The following node(s) are unreachable from the root node\n");
				err_printed = TRUE;
				}
			fprintf(stderr,"\t%d\n",temp);
			}
		else			/* In network and has a parent */
			{
			nodes[temp].timeout = decode_timeout + (level_timeout *
				((SLONG) max_level - nodes[temp].timeout));
/*
 *	Ensure that timeout value are non-negative.
 */
			if(nodes[temp].prog->cpu_type == BID_T2)
				{
				if(nodes[temp].timeout & 0xFFFF8000)
					{
					fprintf(stderr,
"WARNING: Node %d timeout too long for T2 CPU, clipping at maximum\n",temp);
					nodes[temp].timeout = 0x00007FFF;
					}
				}
			else
				{
				if(nodes[temp].timeout & 0x80000000)
					{
					fprintf(stderr,
"WARNING: Node %d timeout too long for T4/T8 CPU, clipping at maximum\n",temp);
					nodes[temp].timeout = 0x7FFFFFFF;
					}
				}
			}
		}
	if(err_printed)
		exit_close(ERRORS);
/*
 *	Now do the pre-order traversal to establish a load order.
 */
	temp = bo_pre_help(root_node,0);
	nodes[temp].next_boot = -1;
	}

/************************************************************************/
/*									*/
/*	Routine:	bootstrap					*/
/*									*/
/*	Download the bootstrap programs to the Transputer network and	*/
/*	make sure everything goes OK.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

/*
 *	See the "netboot.tal" file for the source for this bootstrap code.
 */
unsigned char	bootcode[] =
	{
0xF4,0xD0,0x27,0x0C,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x25,0xF7,0x68,0x4E,0x21,0xFB,0x23,
0xFC,0xB7,0xD1,0xD2,0x24,0xF2,0xD1,0x71,0x21,0xFC,0x71,0x21,0xF8,0x71,0x71,
0xE0,0x71,0x71,0x51,0x4A,0x23,0xF4,0x24,0xFA,0x40,0x9E,0x11,0x72,0x41,0xF7,
0x25,0x46,0x21,0xFB,0x72,0x2F,0x4F,0xF7,0x2E,0x05,0x40,0xD2,0x11,0x76,0x44,
0x23,0xF4,0xF4,0x44,0xFB,0x22,0xF0,0x71,0x33,0x21,0xA4,0x18,0x71,0xE5,0x71,
0x33,0x28,0x20,0x80,0x71,0xE4,0x49,0x21,0xFB,0x71,0xE2,0x71,0x53,0x81,0x23,
0xF9,0x22,0xF0,0x72,0x70,0x44,0x23,0xF4,0xF4,0x41,0xFB,0x72,0x31,0x70,0x44,
0x23,0xF4,0xF4,0x72,0x30,0xFB,0x11,0x70,0x44,0xF7,0x21,0xF5,0x63,0x90,0x73,
0x21,0xA5,0x22,0xF9,0x21,0x74,0x77,0x60,0xEF,0x72,0x21,0x75,0x77,0x23,0xFC,
0xF0,0xF6,0x00,0x00,0x00,0x00,0x00,0x00,0xD4,0x14,0x72,0x41,0xF7,0x75,0x72,
0x74,0xF7,0x75,0x84,0x30,0xD1,0x71,0x43,0xF9,0x23,0xAD,0x70,0x71,0xF9,0xA4,
0x1E,0x66,0x9B,0x03,0x18,0x66,0x97,0x22,0xF2,0x21,0x76,0xF5,0xD6,0x10,0x81,
0x23,0xF9,0x21,0xF5,0x7C,0xA3,0x18,0xD1,0x06,0x21,0x72,0x64,0xA4,0x1E,0xD1,
0x22,0xF2,0x76,0xF4,0x40,0xF9,0x61,0xA7,0x71,0x33,0xD3,0x73,0x69,0x96,0x71,
0x34,0x69,0x92,0x14,0x73,0x44,0xF7,0x74,0x6A,0x9B,0x60,0x07,0x1E,0x69,0x92,
0x18,0x6A,0x9F,0x71,0xC0,0xA4,0x41,0xD3,0x63,0x01,0x75,0x30,0x21,0x77,0xF4,
0xC0,0x64,0xA9,0x71,0xC1,0xAC,0x75,0x8C,0x75,0x88,0x30,0x74,0x60,0x84,0x24,
0xFA,0x65,0x0A,0x75,0x88,0x30,0xD1,0x75,0x8C,0x30,0xD3,0x73,0x66,0xAF,0x73,
0x60,0x8F,0xD3,0x40,0x71,0x23,0xFB,0x71,0x81,0xD1,0x60,0x00,0x00,0x00,0x2A,
0x4A,0x26,0x44,0x23,0xF4,0x40,0x21,0x27,0xFC,0xD6,0xD1,0x71,0xA2,0x71,0xD6,
0x76,0x4A,0x22,0xFC,0x87,0xD1,0x44,0x71,0x44,0x24,0xF6,0xC0,0xAB,0x42,0x71,
0x42,0x24,0xF6,0xC0,0xA4,0x29,0xFC,0x48,0x40,0xD3,0x74,0x24,0xF6,0xA9,0x72,
0x75,0xF4,0xAE,0x72,0x22,0x20,0x80,0x04,0x76,0x24,0x20,0x80,0x61,0x2D,0x9E,
0x60,0x0E,0x25,0xF4,0x64,0x4B,0x21,0xFB,0xD5,0x41,0x40,0xE0,0x40,0x44,0xE0,
0x95,0x40,0xE0,0x92,0x6F,0xA7,0x22,0xF2,0x29,0x4C,0xF5,0x22,0xFB,0x40,0x22,
0xF0
	};

VOID
bootstrap()
	{
	register int		i;
	UINT			length;
	int			node;
/*
 *	Handle initial error checking on the bootstrap code.
 */
	length = sizeof(bootcode);	/* Bootstrap length */
	if(length > (MAX_PACKET * 2))
		fatal("Transputer bootstrap code is too long","");
	if((bootcode[0x103] != 0) || (bootcode[0x104] != 0) ||
		(bootcode[0x105] != 0) || (bootcode[0x106] != 0))
		{
fatal("Transputer bootstrap code needs a window from 0x103 through 0x106","");
		}
	for(i = (((LOCAL_OFFSET + LOCALS_STACK) * 4) - 1);
		i < (((LOCAL_OFFSET + LOCALS_STACK + PARAM_LENGTH) * 4) - 1);
			i++)
		{
		if(bootcode[i] != 0)
			{
fatal("Transputer bootstrap code local variables not initialized to zero","");
			}
		}
/*
 *	First load root Transputer with blow-by-blow description (in addition
 *	to root timeout).
 */
	if(TestRead(LinkId))
		fatal("Transputer response available prior to bootstrap","");
	localize_boot(root_node);	/* Initialize node info */

	fprintf(stderr,"Loading first phase of bootstrap to root node %d\n",
		root_node);
	tout_1b((int) MAX_PACKET);
	tout_str((STRING) bootcode,MAX_PACKET);

	fprintf(stderr,
		"Finished loading first phase, awaiting first acknowledge\n");
	do_response(root_node);		/* Handle response */

	fprintf(stderr,"Loading second phase of bootstrap to root node %d\n",
		root_node);
	tout_1b((int) MAX_PACKET);
	tout_str((STRING) (bootcode + MAX_PACKET),MAX_PACKET);

	fprintf(stderr,"Bootstrap loaded, awaiting acknowledge\n");
	do_response(root_node);		/* Handle response */
	fprintf(stderr,"Successfully bootstrapped root node %d\n\n",root_node);
	fprintf(stderr,"Bootstrapping the remainder of the network:\n");
/*
 *	Bootstrap the rest of the network.
 */
	for(node = nodes[root_node].next_boot; node >= 0;
		node = nodes[node].next_boot)
		{
		if(TestRead(LinkId))
			fatal("Transputer response available prior to message",
				"");
		fprintf(stderr,"\tBootstrapping node %d\n",node);
		localize_boot(node);	/* Initialize node info */

		tout_1b((int) MAX_PACKET);
		tout_str((STRING) bootcode,MAX_PACKET);
		do_response(node);	/* Handle response */
		tout_1b((int) MAX_PACKET);
		tout_str((STRING) (bootcode + MAX_PACKET),MAX_PACKET);
		do_response(node);	/* Handle response */
		}
	fprintf(stderr,"Network successfully bootstrapped\n\n");
	}

/************************************************************************/
/*									*/
/*	Routine:	close_input					*/
/*									*/
/*	Close the current download input file.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
close_input()
	{
	if(fclose(in_fp) == EOF)
		fatal("Unable to close download input file: ",inbuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	copy_out					*/
/*									*/
/*	Copy a specified # of bytes from the input load file to the	*/
/*	root transputer.						*/
/*									*/
/*	Parameters:							*/
/*		count -	The number of bytes to copy.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
copy_out(count)
	UINT			count;
	{
	char			buf[MAX_PACKET];
	int			rval;

	if(fread(buf,1,count,in_fp) != count)
		{
		if(ferror(in_fp) || feof(in_fp))
			fatal("Error reading download input file: ",inbuf);
		}
	if((rval = WriteLink(LinkId,buf,count,(int) host_timeout)) != count)
		{
		if(rval < 0)
			fatal("Error writing to link: ",open_str);
		else
			fatal("Timed out waiting to write to link: ",open_str);
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	do_response					*/
/*									*/
/*	Verify that root Transputer gives correct acknowledgements, if	*/
/*	not do appropriate stuff to handle error diagnosis and		*/
/*	reporting.							*/
/*									*/
/*	Parameters:							*/
/*		node -	The node # for which the original message was	*/
/*			intended. 0 if the message is a	broadcast one.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
do_response(node)
	int			node;
	{
	static STRING err1="Unable to continue error traceback any farther\n";
	int			link;
	int			old_node;
	int			result;
	STRING			str;

	result = tin_4b() & 0x7FFF;	/* Only need bottom 15 bits */
	if(result != 0)			/* If some form of error */
		{
		if(node != 0)
			{
fprintf(stderr,"Network error detected while bootstrapping node %d\n",node);
			}
		else
			{
fprintf(stderr,"Network error detected while downloading program(s)");
			}
		fprintf(stderr,"Beginning error traceback:\n");
		old_node = 0;
		node = root_node;
		FOREVER
			{
			if(result & (ERR_BOOT | ERR_CPU | ERR_TIMEOUT))
				break;
			if(nodes[node].prog->cpu_type == BID_T2)
				link = (result - (LINKIN16_ADDR & 0xFF)) / 2;
			else
				link = (result - (LINKIN32_ADDR & 0xFF)) / 4;
			if((link < 0) || (link >= MAX_LINKS))
				{
				fprintf(stderr,err1);
				exit_close(ERRORS);
				}
	fprintf(stderr,"Error received by node %d via channel %d\n",
				node,link);
			
			if(nodes[node].links[link].other_node < 0)
				{
				fprintf(stderr,err1);
				exit_close(ERRORS);
				}
			node = nodes[old_node = node].links[link].other_node;
			result = tin_4b() & 0x7FFF;
			}
		if(result == 0)
			fprintf(stderr,err1);
		else
			{
			if(result & ERR_TIMEOUT)
				node = old_node;	/* Timeout */
	fprintf(stderr,"Error originated at node %d and was caused by ",node);

			if(result & ERR_CPU)
				{
				switch((result & (~ERR_CPU)) / BID_MAX_MREVS)
					{
					case	(BID_T2XX / BID_MAX_MREVS):
						str = "T212/T222";
						break;
					case	(BID_T225 / BID_MAX_MREVS):
						str = "T225";
						break;
					case	(BID_T400 / BID_MAX_MREVS):
						str = "T400";
						break;
					case	(BID_T414 / BID_MAX_MREVS):
						str = "T414";
						break;
					case	(BID_T425 / BID_MAX_MREVS):
						str = "T425";
						break;
					case	(BID_T800 / BID_MAX_MREVS):
						str = "T800";
						break;
					case	(BID_T801 / BID_MAX_MREVS):
						str = "T801";
						break;
					case	(BID_T805 / BID_MAX_MREVS):
						str = "T805";
						break;
					default:
						str = "????";
					}
				fprintf(stderr,"the wrong cpu (was a %s)\n",
					str);
				}
			else
				{
				if(nodes[node].prog->cpu_type == BID_T2)
					{
					link = ((result & 0xFF) -
						(LINKIN16_ADDR & 0xFF)) / 2;
					}
				else
					{
					link = ((result & 0xFF) -
						(LINKIN32_ADDR & 0xFF)) / 4;
					}
				if((link < 0) || (link >= MAX_LINKS))
					fprintf(stderr,"an unknown source\n");
				else if(result & ERR_TIMEOUT)
					{
					fprintf(stderr,
						"a timeout of channel %d\n",
							link);
					}
				else	/* ERR_BOOT */
					{
					fprintf(stderr,
					"the wrong boot channel (was %d)\n",
							link);
					}
				}
			}
		exit_close(ERRORS);
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	download					*/
/*									*/
/*	Download all necessary programs to the Transputer network.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
download()
	{
	SLONG			addr;
	SLONG			big_temp;
	BOOL			finished_file;
	register struct	 program_info	*pinfo;
	register UINT		prog_num;
	register UINT		temp;
/*
 *	Load all the programs into the network, one at a time.
 */
	for(prog_num = 0, pinfo = programs; pinfo != NULL;
		prog_num++, pinfo = pinfo->next)
		{
		(VOID) open_input(pinfo->name);	/* Open file */
		fprintf(stderr,"Downloading program: %s\n",pinfo->name);
/*
 *	Loop through all the records in the load file.  Write those that are
 *	needed to the network, toss the rest.
 */
		for(finished_file = FALSE; ! finished_file; )
			{
			switch(temp = in_1b())
				{
				case T_DATA:	/* Normal data definition */
					in_2b();	/* Toss line # */
					temp = in_2b();	/* Get length */
					while(temp > (buf_size - 12))
						{
						temp -= (buf_size - 12);
						tout_1b((int) buf_size);
						tout_4b((SLONG) prog_num);
						tout_4b((SLONG) D_DATA);
/*lint -e530 thinks addr hasn't been initialized */
						tout_4b((SLONG) addr);
/*lint +e530 thinks addr hasn't been initialized */
						copy_out((UINT)
							(buf_size - 12));
/*lint -e530 thinks addr hasn't been initialized */
						addr += (buf_size - 12);
/*lint +e530 thinks addr hasn't been initialized */
						do_response(0);
						}
					tout_1b((int) (temp + 12));
					tout_4b((SLONG) prog_num);
					tout_4b((SLONG) D_DATA);
/*lint -e530 thinks addr hasn't been initialized */
					tout_4b((SLONG) addr);
/*lint +e530 thinks addr hasn't been initialized */
					copy_out(temp);
/*lint -e530 thinks addr hasn't been initialized */
					addr += temp;
/*lint +e530 thinks addr hasn't been initialized */
					do_response(0);
					break;

				case T_STORAGE:	/* Reserve space pseudo-op */
					tout_1b((int) 16);
					tout_4b((SLONG) prog_num);
					tout_4b((SLONG) D_STORAGE);
					in_2b();	/* Toss line # */
/*lint -e530 thinks addr hasn't been initialized */
					tout_4b((SLONG) addr);
/*lint +e530 thinks addr hasn't been initialized */
					tout_4b((SLONG) (big_temp = in_4b()));
/*lint -e530 thinks addr hasn't been initialized */
					addr += big_temp;
/*lint +e530 thinks addr hasn't been initialized */
					do_response(0);
					break;

				case T_DEBUG_DATA:	/* Sym. debug stuff */
					in_2b();	/* Toss line # */
					in_4b();	/* Toss value */
					temp = in_2b();	/* Get length */
					while(temp--)
						 in_1b(); /* Toss debug data */
					break;

				case T_FILENAME:	/* Filename record */
					in_4b();	/* Toss old/new #'s */
					temp = in_1b();	/* Get length */
					while(temp--)
						in_1b();/* Toss filename */
					break;

				case T_LOAD:	/* Set load address */
					addr = in_4b();
					break;

				case T_STACK:	/* Set stack pointer address */
					in_4b();	/* Ignore */
					break;

				case T_ENTRY:	/* Set entry point address */
					in_4b();	/* Ignore */
					break;

				case T_EOF:	/* Found EOF record */
					finished_file = TRUE;
					break;

				default:
					{
		fatal("Bad record format on download input file: ",inbuf);
					}
				}
			}
		close_input();		/* Close input file */
		}
/*
 *	We've loaded all the programs, jumpstart the network!
 */
	tout_1b((int) 8);
	tout_4b((SLONG) 0);		/* Garbage */
	tout_4b((SLONG) D_DONE);
	do_response(0);
	fprintf(stderr,"Program downloading completed\n\n");
	}

/************************************************************************/
/*									*/
/*	Routine:	error_scan					*/
/*									*/
/*	Scan through the node structure constructed by "parse" and	*/
/*	flesh out the critical information so that the "boot_order"	*/
/*	routine can determine a reasonable bootstrap load order.  Also	*/
/*	perform additional error detection.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
error_scan()
    {
    register int		child_node;
    register int		child_unknown;
    char			errbuf[80];
    register int		known;
    register int		link;
    register int		parent_node;
    register int		parent_unknown;

    for(child_node = 1; child_node <= MAX_NODES; child_node++)
	{
	if((nodes[child_node].prog != NULL) &&
	    ((parent_node = nodes[child_node].parent_num) != 0))
	    {
/*
 *	First, verify no more than one connection to each system/sub-system
 *	reset output.
 */
	    if(nodes[child_node].boot_status & FROM_SYSTEM_RESET)
		{
		if(nodes[parent_node].system_output != -1)
		    {
		    sprintf(errbuf,"%d",parent_node);
fatal("More than one node connected to the system output of node: ",errbuf);
		    }
		nodes[parent_node].system_output = child_node;
		}
	    else if(nodes[child_node].boot_status & FROM_SUBSYS_RESET)
		{
		if(nodes[parent_node].subsys_output != -1)
		    {
		    sprintf(errbuf,"%d",parent_node);
fatal("More than one node connected to the sub-system output of node: ",
					errbuf);
		    }
		nodes[parent_node].subsys_output = child_node;
		}
	    else
		fatal("Internal error #2","");
/*
 *	Second, verify that if more than one connection exists between this
 *	child node and its parent that enough information is available to
 *	determine which links at least one connection has on both ends.
 */

/*
 *	First flesh out links which are fully defined at only one end.
 */
	    for(link = 0; link < MAX_LINKS; link++)
		{
/*
 *	Handle links defined on parent side.
 */
		if((nodes[parent_node].links[link].other_node == child_node) &&
		    ((known = nodes[parent_node].links[link].link_num) >= 0))
		    {
		    child_unknown =
			nodes[child_node].links[known].other_node;
		    if((child_unknown >= 0) && (child_unknown != parent_node))
			{
			sprintf(errbuf,"%d link %d, and node %d",parent_node,
				link,child_node);
			fatal("Conflict in defined connection between node ",
				errbuf);
			}
		    nodes[child_node].links[known].other_node = parent_node;
		    child_unknown =
			nodes[child_node].links[known].link_num;
		    if((child_unknown >= 0) && (child_unknown != link))
			{
			sprintf(errbuf,"%d link %d, and node %d",parent_node,
				link,child_node);
			fatal("Conflict in defined connection between node ",
				errbuf);
			}
		    nodes[child_node].links[known].link_num = link;
		    }
/*
 *	Handle links defined on child side.
 */
		if((nodes[child_node].links[link].other_node == parent_node) &&
		    ((known = nodes[child_node].links[link].link_num) >= 0))
		    {
		    parent_unknown =
			nodes[parent_node].links[known].other_node;
		    if((parent_unknown >= 0) && (parent_unknown != child_node))
			{
			sprintf(errbuf,"%d link %d, and node %d",child_node,
				link,parent_node);
			fatal("Conflict in defined connection between node ",
				errbuf);
			}
		    nodes[parent_node].links[known].other_node = child_node;
		    parent_unknown =
			nodes[parent_node].links[known].link_num;
		    if((parent_unknown >= 0) && (parent_unknown != link))
			{
			sprintf(errbuf,"%d link %d, and node %d",child_node,
				link,parent_node);
			fatal("Conflict in defined connection between node ",
				errbuf);
			}
		    nodes[parent_node].links[known].link_num = link;
		    }
		}
/*
 *	Now go through and figure out how many known and unknown links are
 *	involved so we can make inferences if required.
 */
	    known = child_unknown = parent_unknown = 0;
	    for(link = 0; link < MAX_LINKS; link++)
		{
		if(nodes[parent_node].links[link].other_node == child_node)
		    {
		    if(nodes[parent_node].links[link].link_num >= 0)
			known++;
		    else
			parent_unknown++;
		    }
		if(nodes[child_node].links[link].other_node == parent_node)
		    {
		    if(nodes[child_node].links[link].link_num < 0)
			child_unknown++;
		    }
		}
/*
 *	Now go through and make sure we know enough about the connections
 *	between the two nodes.
 */
	    if((known == 0) && (parent_unknown == 0))
		{
		sprintf(errbuf,"%d to node %d",parent_node,child_node);
		fatal("Missing connection from node ",errbuf);
		}
	    if((known == 0) && ((parent_unknown > 1) || (child_unknown > 1)))
		{
		sprintf(errbuf,"%d to node %d",parent_node,child_node);
		fatal("Unresolved duplicate connections from node ",errbuf);
		}
/*
 *	See if we have no known connections but only one unknown on each end,
 *	if so assume they are connected together.
 */
	    if(known == 0)
		{
		for(link = 0; link < MAX_LINKS; link++)
		    {
		    if(nodes[parent_node].links[link].other_node == child_node)
			parent_unknown = link;
		    if(nodes[child_node].links[link].other_node == parent_node)
			child_unknown = link;
		    }
		nodes[parent_node].links[parent_unknown].link_num =
			child_unknown;
		nodes[child_node].links[child_unknown].link_num =
			parent_unknown;
		}
	    }
	}
    }

/************************************************************************/
/*									*/
/*	Routine:	eval						*/
/*									*/
/*	Evaluate "C" style decimal, hexadecimal and octal numbers.	*/
/*									*/
/*	Parameters:							*/
/*		optional-TRUE if # is optional.				*/
/*									*/
/*	Returns:	The numeric value (-1 if non present and	*/
/*			optional.					*/
/*									*/
/************************************************************************/

SLONG
eval(optional)
	BOOL			optional;
	{
	register int		c;
	SLONG			val;

	skip_space();			/* Skip whitespace */
	if(isdigit(c = *parse_ptr++))	/* Is it a number? */
		{
		val = 0L;
/*
 *	See if it is a octal or hex # or if it is a decimal #.
 */
		if(c == '0')
			{
			c = *parse_ptr++;
/*
 *	Its either octal or hex, figure out which.
 */
			if((c == 'x') || (c == 'X'))
				{
/*
 *	Its a hexadecimal #.
 */
				while(ishex(c = *parse_ptr++))
					{
/*	Detect overflow		*/	if(val & 0xF0000000)
						{
			parse_fatal("Hex constant too large","");
						}
					val = (val << 4) + hexbin(c);
					}
				parse_ptr--;
				}
			else
/*
 *	Its an octal #.
 */
				{
				while(isoct(c))
					{
/*	Detect overflow		*/	if(val & 0xE0000000)
						{
			parse_fatal("Octal constant too large","");
						}
					val = (val << 3) + c - '0';
					c = *parse_ptr++;
					}
				parse_ptr--;
				}
			}
		else
/*
 *	Its a decimal #.
 */
			{
			val = (val * 10) + c - '0';
			while(isdigit(c = *parse_ptr++))
				{
				val = (val * 10) + c - '0';
				}
			parse_ptr--;
			}
		}
	else
/*
 *	Wasn't any # we could recognize!
 */
		{
		parse_ptr--;
		val = -1L;
		if(! optional)
			parse_fatal("Illegal numeric constant","");
		}
/*
 *	Return the result.
 */
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	exit_close					*/
/*									*/
/*	Calls "exit" and closes the link I/O driver interface if	*/
/*	needed.								*/
/*									*/
/*	Parameters:							*/
/*		status-	Desired "exit" return value.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
exit_close(status)
	int			status;
	{
	if(LinkId > 0)
		(VOID) CloseLink(LinkId);
	exit(status);
	}

/************************************************************************/
/*									*/
/*	Routine:	fatal						*/
/*									*/
/*	Print out a "FATAL" error message and quit.			*/
/*									*/
/*	Parameters:							*/
/*		sptr1 -	A pointer to the first of the error message.	*/
/*		sptr2 -	A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fatal(sptr1,sptr2)
	STRING			sptr1;
	STRING			sptr2;
	{
	fprintf(stderr,"FATAL: %s%s\n",sptr1,sptr2);	/* Print message */
	exit_close(ERRORS);			/* Bye... */
	}

/************************************************************************/
/*									*/
/*	Routine:	get_fname					*/
/*									*/
/*	Get a file or pathname from the parse buffer.			*/
/*									*/
/*	Parameters:							*/
/*		buf -	A pointer to where to put the result.		*/
/*									*/
/*	Returns:	TRUE if an error was detected.			*/
/*									*/
/************************************************************************/

BOOL
get_fname(buf)
	char			buf[];
	{
	register UINT		i;
	char			tbuf[FNSIZE + 1];

	skip_space();			/* Skip whitespace */
	for(i = 0; i < FNSIZE; i++)
		{
		if(not_fname(*parse_ptr))
			break;	/* End of name */
		tbuf[i] = *parse_ptr++;
		}
	if(i == FNSIZE)
		return (TRUE);
	tbuf[i] = '\0';
	strcpy(buf,tbuf);		/* Update with new value */
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	hexbin						*/
/*									*/
/*	Convert a hexadecimal digit to binary.				*/
/*									*/
/*	Parameters:							*/
/*		c -	The digit to convert.				*/
/*									*/
/*	Returns:	The resulting binary value.			*/
/*									*/
/************************************************************************/

SLONG
hexbin(c)
	int			c;
	{
	if(isdigit(c))
		return ((SLONG) (c - '0'));
	else
		return ((SLONG) ((isupper(c) ? tolower(c) : c) - 'a' + 10));
	}

/************************************************************************/
/*									*/
/*	Routine:	in_1b,in_2b,in_4b				*/
/*									*/
/*	Read 1, 2 or 4 bytes from the current download input file.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The 1, 2 or 4 byte value read.			*/
/*									*/
/************************************************************************/

int
in_1b()
	{
	int			c;

	if((c = fgetc(in_fp)) == EOF)
		{
		if(ferror(in_fp) || feof(in_fp))
			fatal("Error reading download input file: ",inbuf);
		}
	return (c & 0xFF);
	}

int
in_2b()
	{
	int			rval;

	rval = in_1b();
	rval |= (in_1b() << 8);
	return (rval & 0xFFFF);
	}

SLONG
in_4b()
	{
	SLONG			rval;

	rval = (SLONG) in_1b();
	rval |= (((SLONG) in_1b()) << 8);
	rval |= (((SLONG) in_1b()) << 16);
	rval |= (((SLONG) in_1b()) << 24);
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	ishex						*/
/*									*/
/*	Test a digit to see if it could be part of a hexadecimal #.	*/
/*									*/
/*	Parameters:							*/
/*		c -	The digit to test.				*/
/*									*/
/*	Returns:	TRUE if so, FALSE if not.			*/
/*									*/
/************************************************************************/

BOOL
ishex(c)
	int			c;
	{
	return (isdigit(c) ||
		((c >= 'a') && (c <= 'f')) || ((c >= 'A') && (c <= 'F')));
	}

/************************************************************************/
/*									*/
/*	Routine:	isoct						*/
/*									*/
/*	Test a digit to see if it could be part of a octal #.		*/
/*									*/
/*	Parameters:							*/
/*		c -	The digit to test.				*/
/*									*/
/*	Returns:	TRUE if so, FALSE if not.			*/
/*									*/
/************************************************************************/

BOOL
isoct(c)
	int			c;
	{
	return ((c >= '0') && (c <= '7'));
	}

/************************************************************************/
/*									*/
/*	Routine:	localize_boot					*/
/*									*/
/*	Set the node specific features of the bootstrap.		*/
/*									*/
/*	Parameters:							*/
/*		node -	The node # for which the bootstrap is intended.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/*	WARNING:	ANY CHANGES TO THE LOCAL TEMPLATE IN "tload.h"	*/
/*			OR "netboot.tal" WILL PROBABLY REQUIRE		*/
/*			MATCHING CHANGES to BE MADE HERE!		*/
/*									*/
/************************************************************************/

VOID
localize_boot(node)
	int			node;
	{
	register int		i;
	register int		word_size;

	if(nodes[node].prog->cpu_type == BID_T2)
		word_size = 2;
	else
		word_size = 4;		/* Assumed to be T4/T8 */
/*
 *	Set node address in both parts of the bootstrap and in the local data.
 *	Note that the node addresses in both halves of the bootstrap must be
 *	4 bytes long since they are interpreted by both 16 and 32 bit CPU's.
 */
	local_store(4,-LOCALS_STACK,(SLONG) (i = nodes[node].boot_id));
	bootcode[0x103] = i & 0xFF;
	bootcode[0x104] = (i >> 8) & 0xFF;
	bootcode[0x105] = 0;		/* Max. of 32K nodes */
	bootcode[0x106] = 0;
	local_store(word_size,OUR_ADDRESS,(SLONG) i);
/*
 *	Put channel addresses in bootstrap.  Note that we use the third
 *	parameter to "pick_ichan" to determine which node (source or
 *	destination), to search first when hunting for a link between the
 *	current node and the associated parent or child nodes.  This ensures
 *	that if two or more links connect the same two nodes we will always
 *	pick the same link between the two.  The rule is that we always pick
 *	the first connected link by checking from link zero to the maximum
 *	number of links in the connection list for the child.
 */
	i = nodes[node].parent_num;
	local_store(word_size,IUP_IN,(SLONG) pick_ichan(node,i,FALSE));
	if((i = nodes[node].system_output) < 0)
		local_store(word_size,LEFT+C_ICHAN,(SLONG) 0);
	else
		local_store(word_size,LEFT+C_ICHAN,pick_ichan(node,i,TRUE));
	if((i = nodes[node].subsys_output) < 0)
		local_store(word_size,RIGHT+C_ICHAN,(SLONG) 0);
	else
		local_store(word_size,RIGHT+C_ICHAN,pick_ichan(node,i,TRUE));
/*
 *	Put remainder of node specific information in bootstrap.
 */
	local_store(word_size,USER_NODE,(SLONG) node);
	local_store(word_size,INIT_TIMEOUT,(SLONG) nodes[node].timeout);
/*
 *	Find program specific information and put in bootstrap.
 */
	local_store(word_size,PROG_NUM,(SLONG) nodes[node].prog->prog_num);
	local_store(word_size,ICPU_TYPE,(SLONG) nodes[node].prog->cpu_type);
	local_store(word_size,STACK,(SLONG) nodes[node].prog->stack);
	local_store(word_size,ENTRY,(SLONG) nodes[node].prog->entry);
	}

/************************************************************************/
/*									*/
/*	Routine:	local_store					*/
/*									*/
/*	Set a specific word in the bootstrap to a specified value.	*/
/*									*/
/*	Parameters:							*/
/*		size -	The size of the word "offset" is measured in.	*/
/*		offset-	The relative word offset within bootstrap local	*/
/*			storage.					*/
/*		value -	The value to store.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/*	WARNING:	ANY CHANGES TO THE LOCAL TEMPLATE IN "tload.h"	*/
/*			OR "netboot.tal" WILL PROBABLY REQUIRE		*/
/*			MATCHING CHANGES BE MADE HERE!			*/
/*									*/
/************************************************************************/

VOID
local_store(size,offset,value)
	int			size;
	int			offset;
	SLONG			value;
	{
	register UCHAR		*cptr;

	cptr = bootcode + (LOCAL_OFFSET * 4) +
		((LOCALS_STACK + offset) * size);
	*cptr++ = value;
	*cptr++ = (value >>= 8);
	if(size == 4)
		{
		*cptr++ = (value >>= 8);
		*cptr = (value >> 8);
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	makefn						*/
/*									*/
/*	Expand a skeleton filename by adding a default extension if	*/
/*	none is supplied.						*/
/*									*/
/*	Parameters:							*/
/*		sptr -	A pointer to the skeleton filename.		*/
/*		bptr -	A pointer to the buffer the result goes in.	*/
/*		ext -	A pointer to the default extension to use.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
makefn(sptr,bptr,ext)
	STRING			sptr;
	STRING			bptr;
	STRING			ext;
	{
	extern	STRING		strchr();
	extern	STRING		strrchr();

	STRING			s;

	if(strlen(sptr) > FNSIZE) usagerr();
	strcpy(bptr,sptr);
	if((s = strrchr(bptr,'.')) != NULL)
		{
		if((strchr(s,'/') == NULL) && (strchr(s,'\\') == NULL)) return;
		}
	if(strlen(bptr) > (FNSIZE - 4)) usagerr();
	strcat(bptr,ext);
	}

/************************************************************************/
/*									*/
/*	Routine:	match						*/
/*									*/
/*	Used by the input parsing routines to check and see if the next	*/
/*	characters pointed at by "parse_ptr" match the string passed to	*/
/*	"match".  If so "parse_ptr" is updated to point beyond the	*/
/*	matched item.  Preceeding whitespace is skipped	("parse_ptr"),	*/
/*	prior to performing the match.  Note that the match is case	*/
/*	independent (the parameter to "match" should be upper case).	*/
/*									*/
/*	Parameters:							*/
/*		str -	A pointer to the string being compared.		*/
/*									*/
/*	Returns:	TRUE if successfully matched.			*/
/*									*/
/************************************************************************/

BOOL
match(str)
	STRING			str;
	{
	register STRING		temp;

	skip_space();
	temp = parse_ptr;
	while(*str)
		{
		if(*str++ != (islower(*temp) ? toupper(*temp) : *temp))
			return (FALSE);
		else temp++;
		}
	parse_ptr = temp;		/* Point past matched string */
	return (TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	match_first					*/
/*									*/
/*	Like "match", but also slurps any alphanumerics which		*/
/*	immediately follow the matched string.				*/
/*									*/
/*	Parameters:							*/
/*		str -	A pointer to the string being compared.		*/
/*									*/
/*	Returns:	TRUE if successfully matched.			*/
/*									*/
/************************************************************************/

BOOL
match_first(str)
	STRING			str;
	{
	if(! match(str))
		return (FALSE);
	while(isalnum(*parse_ptr) || (*parse_ptr == '_'))
		parse_ptr++;
	return (TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	not_fname					*/
/*									*/
/*	Return TRUE if the character parameter could not be part of a	*/
/*	filename.							*/
/*									*/
/*	Parameters:							*/
/*		c -	The character being tested.			*/
/*									*/
/*	Returns:	TRUE if not part of a legal filename.		*/
/*									*/
/************************************************************************/

BOOL
not_fname(c)
	int			c;
	{
	if(isspace(c) || (c == ',') || (c == ';') || (c == '\n') ||
		(c == '\0')) return (TRUE);
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	open_input					*/
/*									*/
/*	Open the specified input file, verify that its in load format	*/
/*	and return the intended CPU type.				*/
/*	the network information file.					*/
/*									*/
/*	Parameters:							*/
/*		name -	The filename to open.				*/
/*									*/
/*	Returns:	The intended processor type.			*/
/*									*/
/************************************************************************/

int
open_input(name)
	STRING			name;
	{
	int			cpu;

	strcpy(inbuf,name);		/* Copy of the name */
/*
 *	Open file to be downloaded.
 */
#if	CRLF
	if((in_fp = (STREAM) fopen(inbuf,"rb")) == NULL)
#else
	if((in_fp = (STREAM) fopen(inbuf,"r")) == NULL)
#endif
		{
		fatal("Unable to open download input file: ",inbuf);
		}
/*
 *	Get file type and CPU type.
 */
	if(in_1b() != T_LD_FILE)
		fatal("Download input file isn't in load format: ",inbuf);
	cpu = in_1b();			/* Get load CPU type */
	if((cpu != TYPE_UNKNOWN) && (cpu != TYPE_212) && (cpu != TYPE_414) &&
		(cpu != TYPE_800))
		{
	fatal("Transputer CPU type in download file is not recognizable","");
		}
	return (cpu);
	}

/************************************************************************/
/*									*/
/*	Routine:	parse						*/
/*									*/
/*	Figure out what the user wants done by reading commands from	*/
/*	the network information file.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
parse()
	{
	static	STRING		err = "Expected ';'";
	BOOL			had_buf;
	BOOL			had_dtime;
	BOOL			had_ltime;
	BOOL			had_serv;

	had_buf = had_serv = had_dtime = had_ltime = FALSE;

	while(fgets(parse_buf,(MAXLNSIZE + 1),info_fp) != NULL)
		{
		++line;
		parse_ptr = parse_buf;
		if(match_first("BUFFER"))
			{
			if(had_buf)
				{
			parse_warning("Already had a BUFFER_SIZE command",
					"");
				}
			buf_size = eval(FALSE);
			if((buf_size < 16) || (buf_size > MAX_PACKET))
				{
		parse_fatal("BUFFER_SIZE must be between 16 and 255 bytes","");
				}
			if(! match(";"))
				parse_fatal(err,"");
			had_buf = TRUE;
			}
		else if(match_first("DECODE"))
			{
			if(had_dtime)
				{
			parse_warning("Already had a DECODE_TIMEOUT command",
					"");
				}
			decode_timeout = eval(FALSE);
			if((decode_timeout < MIN_DEC_TOUT) ||
				(decode_timeout > MAX_DEC_TOUT))
				{
		parse_fatal("DECODE_TIMEOUT too large or too small","");
				}
			decode_timeout = (decode_timeout * 1000L) /
				MICRO_LOW_TICK;
			if(! match(";"))
				parse_fatal(err,"");
			
			had_dtime = TRUE;
			}
		else if(match_first("HOST"))
			{
			if(had_serv)
				{
			parse_warning("Already had a HOST_SERVER command",
					"");
				}
			if(get_fname(host_server))
				parse_fatal("Host server name too long","");
			if(! match(";"))
				parse_fatal(err,"");
			had_serv = TRUE;
			}
		else if(match_first("LEVEL"))
			{
			if(had_ltime)
				{
			parse_warning("Already had a LEVEL_TIMEOUT command",
					"");
				}
			level_timeout = eval(FALSE);
			if((level_timeout < MIN_LVL_TOUT) ||
				(level_timeout > MAX_LVL_TOUT))
				{
		parse_fatal("LEVEL_TIMEOUT too large or too small","");
				}
			level_timeout = (level_timeout * 1000L) /
				MICRO_LOW_TICK;
			if(! match(";"))
				parse_fatal(err,"");
			had_ltime = TRUE;
			}
		else if(isdigit(*parse_ptr))
			{
			parse_node();
			if(! match(";"))
				parse_fatal(err,"");
			}
		else if(isalpha(*parse_ptr))
			parse_fatal("Unrecognized command","");
		}
	if(ferror(info_fp))		/* If error */
		fatal("Error reading network information file: ",infobuf);
	if(root_node == -1)
		fatal("No root node specified: ",infobuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	parse_fatal					*/
/*									*/
/*	Display a message to the user in response to a fatal error	*/
/*	detected while the parsing the network information file.	*/
/*									*/
/*	Parameters:							*/
/*		s1 -	A pointer to the first string to display.	*/
/*		s2 -	A pointer to the second string to display.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
parse_fatal(s1,s2)
	STRING			s1;
	STRING			s2;
	{
	STRING			tptr;

	fprintf(stderr,"<%s> @ %u: FATAL: %s%s\n",infobuf,line,s1,s2);
	fprintf(stderr,"%s",parse_buf);
	for(tptr = parse_buf; tptr != parse_ptr; tptr++)
		{
		if(*tptr == '\t')
			putc('\t',stderr);
		else
			putc(' ',stderr);
		}
	putc('^',stderr);		/* Show where error occurred */
	exit_close(ERRORS);			/* Bye... */
	}

/************************************************************************/
/*									*/
/*	Routine:	parse_node					*/
/*									*/
/*	Parse the information pertaining to a single node.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
parse_node()
	{
	extern	STRING		strchr();
	extern	STRING		strrchr();

	static	STRING		err1 = "Expected ','";
	static	STRING		err2 = "Program name too long";
	register int		link;
	char			name[FNSIZE + 1];
	int			node;
	int			parent_node;
	BOOL			root_link;
	register STRING		s;
	register int		temp1;
	register int		temp2;
/*
 *	Get node #.
 */
	node = eval(FALSE);
	if(node > MAX_NODES)
		parse_fatal("Node # too large","");
	if(node == 0)
		parse_fatal("Node 0 is reserved for the host interface","");
	if(nodes[node].prog != NULL)
		parse_fatal("Node # already used","");

	if(! match(","))
		parse_fatal(err1,"");
/*
 *	Get name of program to load onto that node.
 */
	if(get_fname(name))		/* Get desired program */
		parse_fatal(err2,"");
	if(name[0] == '\0')
		parse_fatal("Program name required","");
/*
 *	Add appropriate extension if none supplied.
 */
	if(! (((s = strrchr(name,'.')) != NULL) && (strchr(s,'/') == NULL) &&
		(strchr(s,'\\') == NULL)))
		{
		if(strlen(name) > (FNSIZE - 4))
			parse_fatal(err2,"");
		else
			strcat(name,TLNK_OUTPUT_EXT);
		}
	nodes[node].prog = add_program(name);

	if(! match(","))
		parse_fatal(err1,"");
/*
 *	Parse the reset information.
 */
	if(match("R"))
		{
		nodes[node].boot_status = FROM_SYSTEM_RESET;
		parent_node = eval(FALSE);
		}
	else if(match("S"))
		{
		nodes[node].boot_status = FROM_SUBSYS_RESET;
		parent_node = eval(FALSE);
		if(parent_node == 0)
			{
	parse_fatal("Host interface doesn't have sub-system connection","");
			}
		}
	else
		parse_fatal("Expected 'R' or 'S'","");
	if(parent_node > MAX_NODES)
		parse_fatal("Node # too large","");
	nodes[node].parent_num = parent_node;
	if(parent_node == 0)
		{
		if(root_node != -1)
			parse_fatal("Duplicate root node","");
		root_node = node;
		root_link = FALSE;	/* No root<->host link found yet */
		}
/*
 *	Parse the link information.
 */
	for(link = 0; link < MAX_LINKS; link++)
		{
		if(! match(","))
			parse_fatal(err1,"");
		temp1 = eval(TRUE);
		if(temp1 < 0)
			{
			nodes[node].links[link].other_node = -1;
			nodes[node].links[link].link_num = -1;
			}
		else if(temp1 > MAX_NODES)
			parse_fatal("Node # too large","");
		else
			{
			nodes[node].links[link].other_node = temp1;
			if(temp1 == 0)
				{
				if(node != root_node)
					{
	parse_fatal("Host interface is only connected to root node","");
					}
				else
					{
					if(root_link)
						{
parse_fatal("Only one link allowed between host interface and root node","");
						}
					root_link = TRUE;
					}
				}
			}
		if(match("["))		/* Link # for other end of link? */
			{
			temp2 = eval(TRUE);
			if(temp2 >= 0)
				{
				if(temp2 >= MAX_LINKS)
					parse_fatal("Link # too large","");
				nodes[node].links[link].link_num = temp2;
				}
			else
				nodes[node].links[link].link_num = -1;
			if(! match("]"))
				parse_fatal("Expected ']'","");
			}
		else
			nodes[node].links[link].link_num = -1;
/*
 *	If the link is to the host we arbitrarily decree it to connect to link
 *	zero on the host to facilitate later error checking.
 */
		if(temp1 == 0)
			nodes[node].links[link].link_num = 0;
		}
/*
 *	Verify that one of the links is connected to the node which boots
 *	this node.
 */
	for(link = 0; link < MAX_LINKS; link++)
		{
		if(nodes[node].links[link].other_node ==
			nodes[node].parent_num)
			break;
		}
	if(link == MAX_LINKS)
		{
	parse_fatal("A link MUST exist to the node which resets this node","");
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	parse_warning					*/
/*									*/
/*	Display a message to the user in response to a non-fatal error	*/
/*	detected while the parsing the network information file.	*/
/*									*/
/*	Parameters:							*/
/*		s1 -	A pointer to the first string to display.	*/
/*		s2 -	A pointer to the second string to display.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
parse_warning(s1,s2)
	STRING			s1;
	STRING			s2;
	{
	fprintf(stderr,"<%s> @ %u: WARNING: %s%s\n",infobuf,line,s1,s2);
	}

/************************************************************************/
/*									*/
/*	Routine:	pick_ichan					*/
/*									*/
/*	Given a source and destination node pick a fully defined input	*/
/*	channel to use for communication (input with respect to the	*/
/*	source).							*/
/*									*/
/*	Parameters:							*/
/*		src -	The source node #.				*/
/*		dst -	The destination node #.				*/
/*		order -	Allows for resolving which link to use when	*/
/*			more than one link connects two nodes.  Set to	*/
/*			TRUE to search "dst" links left to right, FALSE	*/
/*			to search "src" links left to right to choose	*/
/*			the link to use.				*/
/*									*/
/*	Returns:	An appropriate input channel.			*/
/*									*/
/************************************************************************/

SLONG
pick_ichan(src,dst,order)
	register int		src;
	register int		dst;
	int			order;
	{
	register int		link;

	for(link = 0; link < MAX_LINKS; link++)
		{
		if(order)
			{
			if((nodes[dst].links[link].other_node == src) &&
				(nodes[dst].links[link].link_num >= 0))
				{
				link = nodes[dst].links[link].link_num;
				break;
				}
			}
		else
			{
			if((nodes[src].links[link].other_node == dst) &&
				(nodes[src].links[link].link_num >= 0))
				{
				break;
				}
			}
		}
	if(link == MAX_LINKS)
		fatal("Internal error #3","");

	if(nodes[src].prog->cpu_type == BID_T2)
		return (LINKIN16_ADDR + (link * 2));
	else
		return (LINKIN32_ADDR + (link * 4));
	}

/************************************************************************/
/*									*/
/*	Routine:	skip_space					*/
/*									*/
/*	Skip whitespace in the parse buffer.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
skip_space()
	{
	while(isspace(*parse_ptr))
		parse_ptr++;
	}

/************************************************************************/
/*									*/
/*	Routine:	tin_4b						*/
/*									*/
/*	Read 4 bytes from the root Transputer link.			*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value read.					*/
/*									*/
/************************************************************************/

SLONG
tin_4b()
	{
	UCHAR			buf[4];
	SLONG			rval;
	int			temp;

	if((temp = ReadLink(LinkId,buf,4,(int) host_timeout)) <= 0)
		{
		if(temp < 0)
			fatal("Error reading from link: ",open_str);
		else
			{
			fatal("Timed out waiting to read from link: ",
				open_str);
			}
		}
	rval = (SLONG) buf[0];
	rval |= (((SLONG) buf[1]) << 8);
	rval |= (((SLONG) buf[2]) << 16);
	rval |= (((SLONG) buf[3]) << 24);
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	tout_str					*/
/*									*/
/*	Copy a specified # of bytes from a buffer to the Transputer.	*/
/*									*/
/*	Parameters:							*/
/*		buf -	The buffer to copy from.			*/
/*		count -	The number of bytes to copy.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
tout_str(buf,count)
	STRING			buf;
	UINT			count;
	{
	int			rval;

	if((rval = WriteLink(LinkId,buf,count,(int) host_timeout)) != count)
		{
		if(rval < 0)
			fatal("Error writing to link: ",open_str);
		else
			fatal("Timed out waiting to write to link: ",open_str);
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	tout_1b,tout_4b					*/
/*									*/
/*	Send 1 or 4 bytes out the root Transputer link.			*/
/*									*/
/*	Parameters:							*/
/*		value -	The 1 or 4 byte value to write.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
tout_1b(value)
	int			value;
	{
	char			c;
	int			rval;

	c = value;
	if((rval = WriteLink(LinkId,&c,1,(int) host_timeout)) <= 0)
		{
		if(rval < 0)
			fatal("Error writing to link: ",open_str);
		else
			{
			fatal("Timed out waiting to write to link: ",
				open_str);
			}
		}
	}

VOID
tout_4b(value)
	SLONG			value;
	{
	tout_1b((int) value);
	tout_1b((int) (value >>= 8));
	tout_1b((int) (value >>= 8));
	tout_1b((int) (value >> 8));
	}

/************************************************************************/
/*									*/
/*	Routine:	usagerr						*/
/*									*/
/*	Report incorrect loader invocation to user with a suggestion	*/
/*	of how it should be done.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
usagerr()
	{
	fatal("Usage is 'ld-net network_info_file [driver_parameters]*'","");
	}

/************************************************************************/
/*									*/
/*	Routine:	warning						*/
/*									*/
/*	Print out a "WARNING" error message.				*/
/*									*/
/*	Parameters:							*/
/*		sptr1 -	A pointer to the first of the error message.	*/
/*		sptr2 -	A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
warning(sptr1,sptr2)
	STRING			sptr1;
	STRING			sptr2;
	{
	fprintf(stderr,"WARNING: %s%s\n",sptr1,sptr2);	/* Print message */
	}
