/************************************************************************/
/*									*/
/*	File:	ld-one.c						*/
/*									*/
/*	Transputer program downloader for use with INMOS Transputer	*/
/*	boards.  This utility takes ".tld" files created by TLNK and	*/
/*	downloads them for execution on a single Transputer.  When	*/
/*	finished loading a "exec" is done to whatever I/O driver the	*/
/*	user has specified.						*/
/*									*/
/*	This program must be linked with a version of the INMOS		*/
/*	"link.c" I/O interface to customize it to the associated	*/
/*	hardware!							*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				12/19/89				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<string.h>		/* String functions */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tload.h"		/* Transputer downloading info */

#define	MAX_PACKET		255	/* Longest packet with this driver */

/************************************************************************/
/*									*/
/*			Global Data Definitions				*/
/*									*/
/************************************************************************/

/*
 *	I/O definitions.
 */

char	inbuf[FNSIZE + 1] =	{'\0'};	/* Input filename buffer */
STREAM	in_fp		=	NULL;	/* Input file pointer */

/*
 *	General definitions.
 */

STRING	cpu_name	=	NULL;	/* Processor name (see "cpu_type") */
UINT	cpu_type	=	0;	/* Processor type */
SLONG	entry		=	0L;	/* Initial entry point address */
int	LinkId		=	0;	/* LinkId from "OpenLink" */
STRING	open_str	=	NULL;	/* Name to use for "OpenLink" call */
SLONG	stack		=	0L;	/* Initial stack pointer address */

/************************************************************************/
/*									*/
/*			Function Type Declarations			*/
/*									*/
/************************************************************************/

VOID	banner();
VOID	bootstrap();
VOID	copy_out();
VOID	download();
VOID	exit();
VOID	fatal();
VOID	main();
VOID	makefn();
SLONG	tin_4b();
VOID	tout_1b();
VOID	tout_4b();
VOID	usagerr();
VOID	warning();

/*
 *	The following functions may be called only for side effects.
 */
/*lint +fvr */

int	in_1b();
int	in_2b();
SLONG	in_4b();

/*lint -fvr */

/************************************************************************/
/*									*/
/*	Routine:	main						*/
/*									*/
/*	Resets the Transputer, loads the ".tld" file into it and	*/
/*	"exec"s a I/O driver process.					*/
/*									*/
/*	Parameters:							*/
/*		"argc" An integer count of the command line arguments.	*/
/*		"argv" A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
main(argc,argv)
	int			argc;
	char			*argv[];
	{
	extern	STRING		getenv();

	STRING			temp;

	banner();			/* Tell user who we are */
	if(argc < 2) usagerr();		/* Must have at least a loadfile */
/*
 *	Open file to be downloaded.
 */
	makefn(argv[1],inbuf,TLNK_OUTPUT_EXT);	/* Make complete filename */
#if	CRLF
	if((in_fp = (STREAM) fopen(inbuf,"rb")) == NULL)
#else
	if((in_fp = (STREAM) fopen(inbuf,"r")) == NULL)
#endif
		{
		fatal("Unable to open download input file: ",inbuf);
		}
/*
 *	Get file type and CPU type.
 */
	if(in_1b() != T_LD_FILE)
		fatal("Specified download input file isn't in load format","");
	cpu_type = in_1b();		/* Get load CPU type */
	switch((int) cpu_type)
		{
		case	TYPE_UNKNOWN:
			cpu_name = "UNKNOWN 32 BIT";
			break;

		case	TYPE_414:
			cpu_name = "T4";
			break;

		case	TYPE_800:
			cpu_name = "T8";
			break;

		case	TYPE_212:
			cpu_name = "T2";
			break;

		default:
	fatal("Transputer CPU type in download file is not recognizable","");
		}
/*
 *	Reset the transputer and download the bootstrap program.
 */
	open_str = getenv(LINK_NAME);	/* Get desired link name if present */
	LinkId = OpenLink(open_str);	/* Open specified link channel */
	if(open_str == NULL)
		open_str = "(default)";
	if(LinkId < 0)
		fatal("Unable to open link: ",open_str);
	if(ResetLink(LinkId) < 0)
		fatal("Unable to reset link: ",open_str);
	bootstrap();			/* Download the bootstrap program */
	download();			/* Download the program */
	if(CloseLink(LinkId) < 0)
		{
		LinkId = 0;		/* So "fatal" doesn't try again */
		fatal("Unable to close link: ",open_str);
		}
/*
 *	Close the input file and hand control over to whatever I/O driver
 *	program the user has specified (if any).
 */
	if(fclose(in_fp) == EOF)
		{
		fatal("Unable to close download input file: ",inbuf);
		}

	if(argc >= 3)
		{
/*
 *	Put program name in what will be argv[0] for the handler.
 *
 *	Note that a dummy slot is put in place in 'argv[0]' since the MSC
 *	'execvp' library routine trashes it anyway!  The handler must get the
 *	desired 'argv[0]' value from the 'argv[1]' slot, etc.
 */
		temp = argv[2];
		argv[2] = argv[1];
		if(execvp(temp,&argv[1]) == -1)
			fatal("Unable to EXEC I/O driver: ",temp);
		}
	else	warning("No I/O driver installed","");

	exit(NOERRORS);
	}

/************************************************************************/
/*									*/
/*	Routine:	banner						*/
/*									*/
/*	Display sign-on message to user.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
banner()
	{
	fprintf(stderr,
		"LD-ONE (Single CPU Loader), %s [Link I/O Driver: '%s']\n",
			VERSION,IODRIVER);
	fprintf(stderr,"%s\n\n",COPYRIGHT);
	}

/************************************************************************/
/*									*/
/*	Routine:	bootstrap					*/
/*									*/
/*	Download the bootstrap program to the transputer and make sure	*/
/*	it responds correctly.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

/*
 *	See the "boot.tal" file for the source for this bootstrap code.
 */
char	bootcode[] =
	{
0xB7,0xF4,0xD0,0xD0,0x25,0xF7,0x24,0xF2,0xD1,0x71,0x21,0xF8,0x71,0x21,0xFC,
0x4B,0xD2,0x72,0x60,0x8F,0xD2,0x71,0x71,0x72,0x23,0xF4,0xF5,0xE0,0x72,0xC0,
0x60,0xA1,0x40,0x25,0xF4,0x2A,0x4A,0x26,0x44,0x23,0xF4,0x40,0x21,0x27,0xFC,
0xD2,0xD1,0x71,0xA2,0x71,0xD2,0x72,0x4A,0x22,0xFC,0x87,0xD1,0x71,0x44,0x24,
0xF6,0xC0,0xA8,0x71,0x42,0x24,0xF6,0xC0,0xA2,0x29,0xFC,0x12,0x70,0x44,0x23,
0xF4,0xF4,0x44,0xFB,0x21,0x9D,0x23,0xAA,0x72,0x60,0x8F,0xAC,0x72,0x60,0x8E,
0x21,0xAB,0x72,0x60,0x8D,0x21,0xAB,0x60,0x0E,0x9A,0x71,0x70,0x72,0xF7,0x71,
0x72,0xF5,0xD1,0x61,0x01,0x40,0xD6,0x16,0x74,0x41,0xF7,0x76,0x22,0xF0,0x22,
0x94,0xD1,0x62,0x03,0x21,0x9F,0x63,0xAF,0x40,0x71,0x23,0xFB,0x11,0x48,0x22,
0xF1,0x71,0x81,0xD1,0x63,0x02,0x22,0xF9,0x9D,0xD1,0x41,0x71,0x60,0xEF,0x97,
0x70,0xF0,0x71,0x23,0xFC,0xF0,0xF6,0x16,0x74,0x44,0xF7,0x76,0x22,0xF0
	};

VOID
bootstrap()
	{
	UINT			actual_cpu;
	UINT			cpu_class;
	UINT			length;
	STRING			s;

	length = sizeof(bootcode);	/* Bootstrap length */
	if(length > 255)
		fatal("Transputer bootstrap code is too long","");
	tout_1b((int) length);
	s = bootcode;
	fprintf(stderr,
		"Loading %Xh bytes of Transputer bootstrap code (%s CPU)\n",
			length,cpu_name);
	while(length--)
		tout_1b((int) *s++);
	fprintf(stderr,"Bootstrap loaded, awaiting reply\n");
/*
 *	Read the actual CPU type from the bootstrap.  Since the processor could
 *	be a 16 bit CPU all type information is returned in the bottom 16
 *	bits.  Although the bootstrap returns the mask revision level for CPU
 *	types which support the "lddevid" instruction, we are currently
 *	ignoring it.
 */
	actual_cpu = (tin_4b() & 0xFFFF) / BID_MAX_MREVS;
	if((actual_cpu + BID_TRANS) & BID_T4)
		cpu_class = TYPE_414;
	else if((actual_cpu + BID_TRANS) & BID_T2)
		cpu_class = TYPE_212;
	else
		cpu_class = TYPE_800;
	if(cpu_class != cpu_type)
		{
		if(cpu_type != TYPE_UNKNOWN)
			{
			switch(actual_cpu)
				{
				case	(BID_T2XX / BID_MAX_MREVS):
fatal("Incorrect bootstrap reply received, CPU appears to be a T212/T222","");
					break;

				case	(BID_T225 / BID_MAX_MREVS):
fatal("Incorrect bootstrap reply received, CPU appears to be a T225","");
					break;

				case	(BID_T400 / BID_MAX_MREVS):
fatal("Incorrect bootstrap reply received, CPU appears to be a T400","");
					break;

				case	(BID_T414 / BID_MAX_MREVS):
fatal("Incorrect bootstrap reply received, CPU appears to be a T414","");
					break;

				case	(BID_T425 / BID_MAX_MREVS):
fatal("Incorrect bootstrap reply received, CPU appears to be a T425","");
					break;

				case	(BID_T800 / BID_MAX_MREVS):
fatal("Incorrect bootstrap reply received, CPU appears to be a T800","");
					break;

				case	(BID_T801 / BID_MAX_MREVS):
fatal("Incorrect bootstrap reply received, CPU appears to be a T801","");
					break;

				case	(BID_T805 / BID_MAX_MREVS):
fatal("Incorrect bootstrap reply received, CPU appears to be a T805","");
					break;

				default:
fatal("Incorrect bootstrap reply received","");
				}
			}
		else
			{
			switch(actual_cpu)
				{
				case	(BID_T2XX / BID_MAX_MREVS):
fatal("Incorrect bootstrap reply received, CPU appears to be a T212/T222","");
					break;

				case	(BID_T225 / BID_MAX_MREVS):
fatal("Incorrect bootstrap reply received, CPU appears to be a T225","");
					break;

				case	(BID_T400 / BID_MAX_MREVS):
				case	(BID_T414 / BID_MAX_MREVS):
				case	(BID_T425 / BID_MAX_MREVS):
				case	(BID_T800 / BID_MAX_MREVS):
				case	(BID_T801 / BID_MAX_MREVS):
				case	(BID_T805 / BID_MAX_MREVS):
					break;

				default:
fatal("Incorrect bootstrap reply received","");
				}
			}
		}
	fprintf(stderr,
		"Bootstrap replied OK, downloading program '%s'\n",inbuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	copy_out					*/
/*									*/
/*	Copy a specified # of bytes from the input load file to the	*/
/*	transputer.							*/
/*									*/
/*	Parameters:							*/
/*		"count"	The number of bytes to copy.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
copy_out(count)
	UINT			count;
	{
	char			buf[MAX_PACKET];
	int			rval;

	if(fread(buf,1,count,in_fp) != count)
		{
		if(ferror(in_fp) || feof(in_fp))
			{
			fatal("Error reading download input file: ",inbuf);
			}
		}
	if((rval = WriteLink(LinkId,buf,count,100)) != count)
		{
		if(rval < 0)
			fatal("Error writing to link: ",open_str);
		else
			fatal("Timed out waiting to write to link: ",open_str);
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	download					*/
/*									*/
/*	Download the actual program to the Transputer.			*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
download()
	{
	register UINT		temp;
/*
 *	Loop through all the records in load file.  Write those that are
 *	needed to the transputer, toss the rest.
 */
	FOREVER
		{
		switch(temp = in_1b())
			{
			case T_DATA:	/* Normal data definition */
				in_2b();	/* Toss line # */
				temp = in_2b();	/* Get length */
				while(temp > MAX_PACKET)
					{
					temp -= MAX_PACKET;
					tout_1b((int) D_DATA);
					tout_1b((int) MAX_PACKET);
					copy_out(MAX_PACKET);
					}
				tout_1b((int) D_DATA);
				tout_1b((int) temp);
				copy_out(temp);
				break;

			case T_STORAGE:	/* Reserve space pseudo-op */
				tout_1b((int) D_STORAGE);
				in_2b();	/* Toss line # */
				tout_4b((SLONG) in_4b());
				break;

			case T_FILENAME:	/* Filename record */
				in_4b();	/* Toss old & new ln #'s */
				temp = in_1b();	/* Get length */
				while(temp--)
					 in_1b();	/* Toss filename */
				break;

			case T_LOAD:	/* Set load address instruction */
				tout_1b((int) D_LOAD);
				tout_4b((SLONG) in_4b());
				break;

			case T_DEBUG_DATA:	/* Sym. debug stuff */
				in_2b();	/* Toss line # */
				in_2b();	/* Toss lower part of value */
				in_2b();	/* Toss upper part of value */
				temp = in_2b();	/* Get length */
				while(temp--)
					 in_1b();	/* Toss debug data */
				break;

			case T_STACK:	/* Set stack pointer address */
				stack = in_4b();
				break;

			case T_ENTRY:	/* Set entry point address */
				entry = in_4b();
				break;

			case T_EOF:	/* Found EOF record */
				tout_1b((int) D_DONE);
				tout_4b((SLONG) stack);
				tout_4b((SLONG) entry);
				fprintf(stderr,
					"Program downloading finished\n\n");
				return;

			default:
				fatal("Bad record format on load file: ",
					inbuf);
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	fatal						*/
/*									*/
/*	Print out a "FATAL" error message and quit.			*/
/*									*/
/*	Parameters:							*/
/*		"sptr1"	A pointer to the first of the error message.	*/
/*		"sptr2" A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fatal(sptr1,sptr2)
	STRING			sptr1;
	STRING			sptr2;
	{
	fprintf(stderr,"FATAL: %s%s\n",sptr1,sptr2);	/* Print message */
	if(LinkId > 0)
		(VOID) CloseLink(LinkId);
	exit(ERRORS);			/* Bye... */
	}

/************************************************************************/
/*									*/
/*	Routine:	in_1b,in_2b,in_4b				*/
/*									*/
/*	Read 1, 2 or 4 bytes from the download input file.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The 1, 2 or 4 byte value read.			*/
/*									*/
/************************************************************************/

int
in_1b()
	{
	int			c;

	if((c = fgetc(in_fp)) == EOF)
		{
		if(ferror(in_fp) || feof(in_fp))
			{
			fatal("Error reading download input file: ",inbuf);
			}
		}
	return (c & 0xFF);
	}

int
in_2b()
	{
	int			rval;

	rval = in_1b();
	rval |= (in_1b() << 8);
	return (rval & 0xFFFF);
	}

SLONG
in_4b()
	{
	SLONG			rval;

	rval = (SLONG) in_1b();
	rval |= (((SLONG) in_1b()) << 8);
	rval |= (((SLONG) in_1b()) << 16);
	rval |= (((SLONG) in_1b()) << 24);
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	makefn						*/
/*									*/
/*	Expand a skeleton filename by adding a default extension if	*/
/*	none is supplied.						*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to the skeleton filename.		*/
/*		"bptr"	A pointer to the buffer the result goes in.	*/
/*		"ext"	A pointer to the default extension to use.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
makefn(sptr,bptr,ext)
	STRING			sptr;
	STRING			bptr;
	STRING			ext;
	{
	STRING			s;

	if(strlen(sptr) > FNSIZE) usagerr();
	strcpy(bptr,sptr);
	if((s = strrchr(bptr,'.')) != NULL)
		{
		if((strchr(s,'/') == NULL) && (strchr(s,'\\') == NULL)) return;
		}
	if(strlen(bptr) > (FNSIZE - 4)) usagerr();
	strcat(bptr,ext);
	}

/************************************************************************/
/*									*/
/*	Routine:	tin_4b						*/
/*									*/
/*	Read 4 bytes from the transputer link.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value read.					*/
/*									*/
/************************************************************************/

SLONG
tin_4b()
	{
	UCHAR			buf[4];
	SLONG			rval;
	int			temp;

	if((temp = ReadLink(LinkId,buf,4,100)) <= 0)
		{
		if(temp < 0)
			fatal("Error reading from link: ",open_str);
		else
			{
			fatal("Timed out waiting to read from link: ",
				open_str);
			}
		}
	rval = (SLONG) buf[0];
	rval |= (((SLONG) buf[1]) << 8);
	rval |= (((SLONG) buf[2]) << 16);
	rval |= (((SLONG) buf[3]) << 24);
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	tout_1b,tout_4b					*/
/*									*/
/*	Send 1 or 4 bytes out the transputer link.			*/
/*									*/
/*	Parameters:							*/
/*		"value"	The 1 or 4 byte value to write.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
tout_1b(value)
	int			value;
	{
	char			c;
	int			rval;

	c = value;
	if((rval = WriteLink(LinkId,&c,1,100)) <= 0)
		{
		if(rval < 0)
			fatal("Error writing to link: ",open_str);
		else
			{
			fatal("Timed out waiting to write to link: ",
				open_str);
			}
		}
	}

VOID
tout_4b(value)
	SLONG			value;
	{
	tout_1b((int) value);
	tout_1b((int) (value >>= 8));
	tout_1b((int) (value >>= 8));
	tout_1b((int) (value >> 8));
	}

/************************************************************************/
/*									*/
/*	Routine:	usagerr						*/
/*									*/
/*	Report incorrect loader invocation to user with a suggestion	*/
/*	of how it should be done.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
usagerr()
	{
	fatal("Usage is 'ld-one load_file [driver] [driver_parameters]*'","");
	}

/************************************************************************/
/*									*/
/*	Routine:	warning						*/
/*									*/
/*	Print out a "WARNING" error message.				*/
/*									*/
/*	Parameters:							*/
/*		"sptr1"	A pointer to the first of the error message.	*/
/*		"sptr2" A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
warning(sptr1,sptr2)
	STRING			sptr1;
	STRING			sptr2;
	{
	fprintf(stderr,"WARNING: %s%s\n",sptr1,sptr2);	/* Print message */
	}
