;************************************************************************
;*									*
;*	File:	lspcdma.asm						*
;*									*
;*	An implementation of the INMOS "link.c" link driver for use	*
;*	under MS-DOS with either a INMOS B004 (or compatible), polling	*
;*	interface or a MicroWay, CSA or INMOS B008 DMA interface for	*
;*	PC compatible Transputer boards.  Conforms to version 0.3	*
;*	(DRAFT 2) of "link.c".						*
;*									*
;*	Designed to be assembled with Microsofts' MASM 5.1 and linked	*
;*	with other modules written using Microsofts' "C" 5.X.		*
;*									*
;*		Written by Gary Oliver and Kirk Bailey.			*
;*									*
;*		Modifications for multiple memory model support by	*
;*		Jeff Urlwin of Compusearch Corporation (Thanks Jeff!)	*
;*									*
;*		Copyright 1988-1989 by Logical Systems.			*
;*									*
;*				10/3/89					*
;*									*
;*	WARNING:  THIS PROGRAM TAKES OVER THE CLOCK INTERRUPT VECTOR	*
;*	AND (OPTIONALLY), CHANNEL #1 OF THE DMA CONTROLLER!  YOU MUST	*
;*	CALL "CloseLink" BEFORE PROGRAM TERMINATION TO REMOVE THE	*
;*	LINKAGE!  THE DRIVER INTERNALLY TRAPS CONTROL-C CALLS AND	*
;*	SAFEGUARDS THAT LOOPHOLE...					*
;*									*
;************************************************************************

		title   LSPCDMA.ASM

;************************************************************************
;*									*
;*			Global Definitions				*
;*									*
;************************************************************************

;
;	Define memory model: WARNING!! These must be consistent with selected
;	MSC compiler model.
;
	DOSSEG
%	.MODEL	memmodel,lang
;
;	Define B004 (and compatible) control/link interface details.
;
HOST_BASE0	equ	0150H		;Base address of #0 host I/O ports
HOST_BASE1	equ	0170H		;Base address of #1 host I/O ports
HOST_BASE2	equ	0300H		;Base address of #2 host I/O ports

HOST_ANALYZE	equ	17		;Offset from base for host analyze
HOST_CSA_INEN	equ	18		;Offset for CSA DMA input enable
HOST_CSA_OTEN	equ	HOST_CSA_INEN+1	;Offset for CSA DMA output enable
HOST_DMA_CNTL	equ	18		;INMOS B008 DMA control port
HOST_IN		equ	0		;Offset from base for input data port
HOST_OUT	equ	1		;Offset from base for output data port
HOST_RESERR	equ	16		;Offset from base for host reset/error
HOST_STATUS	equ	2		;Offset from data port to status port

HOST_ANALYZE_MASK equ	001H		;Link analyze bit mask
HOST_ERROR_MASK	equ	001H		;Link error bit mask
HOST_RBF_MASK	equ	001H		;Link read-ready bit mask
HOST_RESET_MASK	equ	001H		;Initiate reset bit mask
HOST_TBF_MASK	equ	001H		;Link transmit-ready bit mask

HOST_DMA_ADDR	equ	2		;DMA channel #1 address register
HOST_DMA_BYTEF	equ	12		;DMA byte flip/flop control register
HOST_DMA_COUNT	equ	3		;DMA channel #1 count register
HOST_DMA_MASK	equ	10		;DMA enable control register
HOST_DMA_MODE	equ	11		;DMA mode control register
HOST_DMA_PREG	equ	083H		;DMA channel #1 page register
HOST_DMA_STAT	equ	8		;DMA status register

DMA_OFF		equ	05h		;Disable ch 1 mask
DMA_ON		equ	01H		;Enable ch 1 mask
DMA_READ	equ	45H		;Read "mode" for ch 1
DMA_TC_FLAG	equ	02H		;DMA channel 1 TC status bit
DMA_WRITE	equ	49H		;Write "mode" for ch 1

MIN_DMA_COUNT	equ	30		;Min # of bytes before DMA is faster
;
;	Configure time and delay values.
;
CLOCK_TICK	equ	55		;55 milliseconds/interrupt clock tick

ANALYZE_DELAY	equ	((500/CLOCK_TICK) + 1)	;At least 500 milliseconds
DMA_TIME_CHECK	equ	255		;# of polls between DMA ^C/timeout test
TIME_CHECK	equ	1000		;# of polls between ^C/timeout test
RESET_DELAY	equ	((10/CLOCK_TICK) + 1)	;At least 10 milliseconds

;************************************************************************
;*									*
;*		Global Data Definitions	(CODE SEGMENT)			*
;*									*
;************************************************************************

.CODE
dma_enabled	db	?		;1 if DMA use is enabled
hst_analyze	dw	?		;Host analyze control port
hst_in_stat	dw	?		;Host link input status port
hst_out_stat	dw	?		;Host link output status port
hst_reserr	dw	?		;Host reset/error control/status port
in_use		db	0		;1 if "OpenLink" has been called
inmos_dma	db	?		;1 if INMOS B008 board in use
old_cntlc_ivec	dd	?		;Original control-c interrupt vector
old_timer_ivec	dd	?		;Original timer interrupt vector
timeout_left	dw	?		;# of clock ticks remaining
using_timeout	db	0		;1 if a timeout has been requested


;************************************************************************
;*									*
;*	Externally Visible Code (INMOS "link.c" entry points)		*
;*									*
;************************************************************************

.CODE
;
;	int	OpenLink(char *Name)
;
;	Ready the link associated with "Name".
;
;	COMMENTS:
;		1.	If "Name" is NULL or "" then any free link can be used.
;		2.	This call does not reset the subsystem.
;
;	RETURNS:
;		Any positive integer as a link id or a negative value if the
;		open fails.
;
;	NOTES:
;		If the first character in the "Name" string is 'D' then DMA
;		transfers using DMA channel #1 on the PC are supported for
;		either the MicroWay or CSA DMA interfaces.  If a 'I' is the
;		first character then DMA for the INMOS B008 using channel #1 is
;		supported.  Either option MUST be used only with the correct
;		board which has been properly configured.
;
;		After the optional 'D'/'I' the input "Name" may contain a "0",
;		"1" or "2" to select one of three different host link I/O base
;		addresses:
;
;		"0" -	Base address of host I/O ports is 0150H (also the
;			default if "", "D", "I" or NULL is used).
;		"1" -	Base address of host I/O ports is 0170H
;		"2" -	Base address of host I/O ports is 0300H
;
;		If "Name" isn't one of these three strings optionally preceeded
;		by 'D'/'I' (or a link has already been opened), a -1 is
;		returned.  On success 1 is returned (arbitrary since only one
;		link may be opened at a time with this driver anyway).
;
OpenLink	PROC	USES es si, Lname:PTR BYTE

		cmp	byte ptr cs:[in_use],0	;See if opened
		je	ol1		;Jump if not already opened
		mov	ax,-1		;Already opened, return error
		jmp	ol9
ol1:
;
;	Setup default port address variables and dma control information.
;
		mov	word ptr cs:[hst_analyze],HOST_BASE0+HOST_ANALYZE
		mov	word ptr cs:[hst_in_stat],HOST_BASE0+HOST_IN+HOST_STATUS
		mov	word ptr cs:[hst_out_stat],HOST_BASE0+HOST_OUT+HOST_STATUS
		mov	word ptr cs:[hst_reserr],HOST_BASE0+HOST_RESERR
		mov	byte ptr cs:[dma_enabled],0	;No DMA desired
		mov	byte ptr cs:[inmos_dma],0	;Not INMOS B008 DMA
;
;	Determine which port address to use.
;
	IF	@DataSize
		les	si,Lname
		mov	ax,es
		cmp	ax,0
		jne	ol2		;Jump if non-NULL segment
		jmp	ol7		;NULL segment, accept default address
	ELSE
		mov	si,Lname
		cmp	si,0		;See if NULL pointer
		je	ol7		;NULL "Name" pointer, accept default
		push	ds		;Put ds into es
		pop	es
	ENDIF
ol2:
;
;	We have a non-null pointer, see if *Name = 'D'
;
		mov	al,byte ptr es:[si]	;Get *Name
		cmp	al,'D'
		jne	ol3
		mov	byte ptr cs:[dma_enabled],1	;Indicate DMA desired
		inc	si		;Name++
		jmp	ol4
;
;	See if *Name = 'I' (INMOS B008 DMA)
;
ol3:
		cmp	al,'I'		;INMOS B008 DMA desired?
		jne	ol4		;Jump if not
		mov	byte ptr cs:[dma_enabled],1	;Indicate DMA desired
		mov	byte ptr cs:[inmos_dma],1	;For INMOS board
		inc	si		;Name++
;
;	See if "*Name" = "\0","0","1" or "2".
;
ol4:
		mov	al,byte ptr es:[si]	;Get *Name
		cmp	al,0		;"\0"?
		je	ol7		;Jump if so
		cmp	al,'0'		;"0"?
		je	ol7		;Jump if so
		cmp	al,'1'		;"1"?
		je	ol6
		cmp	al,'2'		;"2"?
		je	ol5		;Jump if so
		mov	ax,-1		;Otherwise, return error
		jmp	ol9
ol5:
		mov	word ptr cs:[hst_analyze],HOST_BASE2+HOST_ANALYZE
		mov	word ptr cs:[hst_in_stat],HOST_BASE2+HOST_IN+HOST_STATUS
		mov	word ptr cs:[hst_out_stat],HOST_BASE2+HOST_OUT+HOST_STATUS
		mov	word ptr cs:[hst_reserr],HOST_BASE2+HOST_RESERR
		jmp	ol7
ol6:
		mov	word ptr cs:[hst_analyze],HOST_BASE1+HOST_ANALYZE
		mov	word ptr cs:[hst_in_stat],HOST_BASE1+HOST_IN+HOST_STATUS
		mov	word ptr cs:[hst_out_stat],HOST_BASE1+HOST_OUT+HOST_STATUS
		mov	word ptr cs:[hst_reserr],HOST_BASE1+HOST_RESERR
;
;	Port address is configured, install control-C and timer interrupt
;	routines and enable DMA if desired.
;
ol7:
		xor	al,al
		mov	dx,word ptr cs:[hst_in_stat]
		out	dx,al		;Clear C012 input status registers
		mov	dx,word ptr cs:[hst_out_stat]
		out	dx,al		;Clear C012 output status registers
;
;	Install the timer and control-c tasks now.
;
;	Install timer interrupt routine
;
		call	NEAR PTR install_timeout 
;
;	Initialize DMA stuff if needed.
;
		cmp	byte ptr cs:[dma_enabled],0
		je	ol8		;Jump if not needed
		call	NEAR PTR enable_dma
;
;	Indicate opened and return success!
;
ol8:
		mov	byte ptr cs:[in_use],1	;Indicate opened
		mov	ax,1		;Dummy link id
ol9:
		ret

OpenLink	endp

;
;	int	CloseLink(int LinkId)
;
;	Close the active link "LinkId".
;
;	COMMENTS:
;		None.
;
;	RETURNS:
;		1 on success or a negative value if the	close fails.
;
;	NOTES:
;		For the purposes of this driver (under MS-DOS), the "LinkId"
;		is ignored and the value returned is always 1 (unless
;		the link wasn't open).
;
CloseLink	PROC	LinkID:WORD

		cmp	byte ptr cs:[in_use],0	;See if opened
		jne	cl1		;Jump if opened
		mov	ax,-1		;Not opened, return error
		jmp	cl2
cl1:
		mov	byte ptr cs:[in_use],0	;Indicate closed
		call	NEAR PTR remove_timeout	;Remove timer interrupt routine
		call	NEAR PTR disable_dma	;Remove DMA linkages if needed
		mov	ax,1		;No error
cl2:
		ret

CloseLink	endp

;
;	int	ReadLink(int LinkId,char *Buffer,int Count,int Timeout)
;
;	Read 'Count' bytes into 'Buffer' from the specified link.
;
;	COMMENTS:
;		1.	"LinkId" is a valid link identifier (from "OpenLink").
;		2.	"Timeout" is a non negative integer representing tenths
;			of a second.  A "Timeout" of zero is an infinite
;			timeout.  If "Timeout" is positive then "ReadLink" may
;			return having read less than the number of bytes asked
;			for.
;		3.	"Count" must be non-zero.
;
;	RETURNS:
;		The number of bytes placed in "Buffer" (which may be zero), or
;		negative to indicate an error.
;
;	NOTES:
;		For the purposes of this driver (under MS-DOS), the "LinkId"
;		is ignored and the value returned is always greater than or
;		equal to zero (unless the link wasn't open).  Also the "Buffer"
;		must be completely located in a single segment.
;
ReadLink	PROC	USES es di, \
		LinkId:WORD,Buffer:PTR BYTE,Count:WORD,Timeout:WORD
		cmp	byte ptr cs:[in_use],0	;See if link opened
		jne	rl1		;Jump if opened
		mov	ax,-1		;Not opened, return error
		jmp	rl17
rl1:
		cmp	Timeout,0	;Timeout needed?
		je	rl2		;Jump if not
		mov	ax,Timeout	;Get timeout value
;
;	The following calculation sets the timeout interval to a multiple of
;	.11 seconds (not the ideal .1 seconds).  We also allow an additional
;	+0/-.055 second inaccuracy for the entire period due to uncertainty of
;	the amount of time left in the current clock period.  We accept this
;	10% or so error in favor of a much shorter, more portable, driver (no
;	need to recognize the computer type and reprogram the hardware timer).
;
		sal	ax,1		;Make into 50 MS units (55 MS really)
		mov	word ptr cs:[timeout_left],ax	;Set value
rl2:
	IF	@DataSize
		les	di,Buffer
	ELSE
		mov	di,Buffer
		push	ds
		pop	es		;Put ds into es
	ENDIF
		mov	cx,Count	;Get length of transfer
		cmp	cx,MIN_DMA_COUNT	;Worthwhile considering DMA?
		jl	rl3		;Jump if too small
		cmp	byte ptr cs:[dma_enabled],0
		jne	rl8		;Jump if DMA allowed	
rl3:
		mov	bx,TIME_CHECK	;Delay between CNTL-C/timeout tests
		cld			;Make sure incremental direction
		mov	dx,word ptr cs:[hst_in_stat]	;Addr of receive flags
rl4:
		in	al,dx		;Read flags
		test	al,HOST_RBF_MASK	;See if ready for data
		jnz	rl6		;Read a byte
		dec	bx		;Count down delay
		jnz	rl4		;Test again
;
;	See if we have timed out or user typed control-c.
;	
		cmp	Timeout,0	;Timeout in use?
		je	rl5		;Jump if not
		cmp	word ptr cs:[timeout_left],0	;Timed out?
		je	rl7		;Jump if timed out
rl5:
		mov	ah,11
		int	21h		;Check keyboard status (allow ^C)
		mov	bx,TIME_CHECK	;Count until next status check
		jmp	rl4
rl6:
		dec	dx
		dec	dx
		in	al,dx		;Read data
		inc	dx
		inc	dx
		stosb			;Store byte into buffer
		loop	rl4		;Read more bytes
rl7:
		mov	ax,Count	;Get length of transfer
		sub	ax,cx		;Compute # bytes read
		jmp	rl17
;
;	Enable CSA board DMA input channel (unless INMOS B008).
;
rl8:
		cmp	byte ptr cs:[inmos_dma],0	;INMOS B008 DMA in use?
		jne	rl9		;Jump if so
		mov	dx,word ptr cs:[hst_in_stat]
		add	dx,(-HOST_STATUS + HOST_CSA_INEN)
		mov	al,1
		out	dx,al
;
;	Enable C011/C012 interrupts/DMA activity.
;
rl9:
		mov	dx,word ptr cs:[hst_in_stat]
		mov	al,2
		out	dx,al
;
;	Setup address and count for DMA transfer.
;
		mov	ax,es		;Get segment
		rol	ax,1		;Rotate left
		rol	ax,1		;Rotate left
		rol	ax,1		;Rotate left
		rol	ax,1		;Rotate left
		mov	dh,al		;Get highest nibble of es to dh
		and	al,0F0h		;Zero low nibble from segment
		add	ax,di		;Test for carry from add
		jnc	rl10
		inc	dh		;Carry means high 4 bits must be inc
rl10:
;
;	Figure out if one DMA transfer is enough (data doesn't cross physical
;	64K page boundary).  If not compute size of first and second transfers.
;
		mov	bx,cx		;Get a copy of transfer length
		add	bx,ax		;Set carry if transfer crosses 64K page
		mov	bx,0		;Assume it doesn't: no second transfer
		jnc	rl11		;Jump if a single DMA transfer will do
		mov	bx,cx		;Save total transfer length
		mov	cx,0		;Compute length of first transfer
		sub	cx,ax
		sub	bx,cx		;Compute length of second transfer
rl11:
		cli			;Disable interrupts for a bit
		out	HOST_DMA_BYTEF,al	;Clear byte order flip-flop
		out	HOST_DMA_ADDR,al	;Output low address
		mov	al,ah
		out	HOST_DMA_ADDR,al	;Output high address
		dec	cx		;Compensate for DMA count offset
		mov	al,cl
		out	HOST_DMA_COUNT,al	;Low count byte
		mov	al,ch
		out	HOST_DMA_COUNT,al	;High count byte
		mov	al,dh		;Get high 4 bits of addr
		and	al,0FH
		out	HOST_DMA_PREG,al	;Output high 4 bits to page reg
		mov	al,DMA_READ
		out	HOST_DMA_MODE,al	;Set read mode
		sti			;Enable interrupts
		mov	dl,DMA_TIME_CHECK
		mov	al,DMA_ON	;Channel #1 enable code
		out	HOST_DMA_MASK,al	;Enable 8237 channel
		cmp	byte ptr cs:[inmos_dma],0	;INMOS B008 DMA in use?
		je	rl12		;Jump if not
;
;	Enable INMOS B008 DMA read transfer.
;
		push	dx
		mov	dx,word ptr cs:[hst_in_stat]
		add	dx,(-HOST_STATUS + HOST_DMA_CNTL)
		mov	al,1
		out	dx,al
		pop	dx
;
;	Poll DMA controller for read completion.
;
rl12:
		in	al,HOST_DMA_STAT
		test	al,DMA_TC_FLAG	;See if done
		jnz	rl14		;Jump if so
		cmp	Timeout,0	;Timeout in use?
		je	rl13		;Jump if not
		cmp	word ptr cs:[timeout_left],0	;Timed out?
		je	rl15		;Jump if timed out
rl13:
		dec	dl		;Time for ^C check?
		jnz	rl12		;Test again
		mov	dl,DMA_TIME_CHECK
		mov	ah,11
		int	21h		;Check keyboard status (allow ^C)
		jmp	rl12
;
;	We've finished the first DMA transfer, see if another needed.
;
rl14:
		cmp	bx,0		;Anything left to transfer
		je	rl15		;Jump if not
		mov	ax,0		;Start at beginning of next 64K page
		inc	dh		;Update page address
		mov	cx,bx		;Length of second transfer
		mov	bx,0		;No more needed after this
		jmp	rl11		;Start second transfer
;
;	Disable C011/C012 interrupts/DMA activity.
;
rl15:
		mov	dx,word ptr cs:[hst_in_stat]
		xor	al,al
		out	dx,al
;
;	Disable CSA board DMA input channel (unless INMOS B008).
;
		cmp	byte ptr cs:[inmos_dma],0	;INMOS B008 DMA in use?
		jne	rl16		;Jump if so
		add	dx,(-HOST_STATUS + HOST_CSA_INEN)
		out	dx,al
;
;	Disable 8237 DMA channel.
;
rl16:
		mov	al,DMA_OFF	;Channel #1 disable code
		out	HOST_DMA_MASK,al	;Disable 8237 channel
;
;	Figure out how many bytes we have left (in case of timeout).
;
		cli			;Disable interrupts for a bit
		out	HOST_DMA_BYTEF,al	;Clear byte order flip-flop
		in	al,HOST_DMA_COUNT	;Low count byte
		mov	cl,al
		in	al,HOST_DMA_COUNT	;High count byte
		sti			;Enable interrupts
		mov	ch,al
		inc	cx		;Compensate for DMA count offset
		add	cx,bx		;# left in possible second DMA transfer
		mov	ax,Count	;Get length of transfer
		sub	ax,cx		;Compute # bytes read
rl17:
		ret

ReadLink	endp

;
;	int	WriteLink(int LinkId,char *Buffer,int Count,int Timeout)
;
;	Write "Count" bytes from "Buffer" to the specified link.
;
;	COMMENTS:
;		1.	"LinkId" is a valid link identifier (from "OpenLink").
;		2.	"Timeout" is a non negative integer representing tenths
;			of a second.  A "Timeout" of zero is an infinite
;			timeout.  If "Timeout" is positive then "WriteLink" may
;			return having written less that the number of bytes
;			asked for.
;		3.	"Count" must be non-zero.
;
;	RETURNS:
;		The number of chars actually written (which may	be zero), or
;		negative to indicate an error.
;
;	NOTES:
;		For the purposes of this driver (under MS-DOS), the "LinkId"
;		is ignored and the value returned is always greater than or
;		equal to zero (unless the link wasn't open).  Also the "Buffer"
;		must be completely located in a single segment.
;
WriteLink	PROC	USES es si, \
		LinkId:WORD,Buffer:PTR BYTE,Count:WORD,Timeout:WORD

		cmp	byte ptr cs:[in_use],0	;See if link opened
		jne	wl1		;Jump if opened
		mov	ax,-1		;Not opened, return error
		jmp	wl17
wl1:
		cmp	Timeout,0	;Timeout needed?
		je	wl2		;Jump if not
		mov	ax,Timeout	;Get timeout value
;
;	The following calculation sets the timeout interval to a multiple of
;	.11 seconds (not the ideal .1 seconds).  We also allow an additional
;	+0/-.055 second inaccuracy for the entire period due to uncertainty of
;	the amount of time left in the current clock period.  We accept this
;	10% or so error in favor of a much shorter, more portable, driver (no
;	need to recognize the computer type and reprogram the hardware timer).
;
		sal	ax,1		;Make into 50 MS units (55 MS really)
		mov	word ptr cs:[timeout_left],ax	;Set value
wl2:
	IF	@DataSize
		les	si,Buffer
	ELSE
		mov	si,Buffer
		push	ds
		pop	es		;Put ds into es
	ENDIF
		mov	cx,Count	;Get length of transfer
		cmp	cx,MIN_DMA_COUNT	;Worthwhile considering DMA?
		jl	wl3		;Jump if too small
		cmp	byte ptr cs:[dma_enabled],0
		jne	wl8		;Jump if DMA allowed
wl3:
		mov	bx,TIME_CHECK	;Delay between CNTL-C/timeout tests
		cld			;Make sure incremental direction
		mov	dx,word ptr cs:[hst_out_stat]	;Addr of transmit flags
wl4:
		in	al,dx		;Read flags
		test	al,HOST_TBF_MASK	;See if ready for data
		jnz	wl6		;Write a byte
		dec	bx		;Count down delay
		jnz	wl4		;Test again
;
;	See if we have timed out or user typed control-c.
;	
		cmp	Timeout,0	;Timeout in use?
		je	wl5		;Jump if not
		cmp	word ptr cs:[timeout_left],0	;Timed out?
		je	wl7		;Jump if timed out
wl5:
		mov	ah,11
		int	21h		;Check keyboard status (allow ^C)
		mov	bx,TIME_CHECK	;Count until next status check
		jmp	wl4
wl6:
		lods	BYTE PTR es:0	;Get byte (note segment override!)
		dec	dx
		dec	dx
		out	dx,al		;Write data
		inc	dx
		inc	dx
		loop	wl4		;Read more bytes
wl7:
		mov	ax,Count	;Get length of transfer
		sub	ax,cx		;Compute # bytes read
		jmp	wl17
;
;	Enable CSA board DMA output channel (unless INMOS B008).
;
wl8:
		cmp	byte ptr cs:[inmos_dma],0	;INMOS B008 DMA in use?
		jne	wl9		;Jump if so
		mov	dx,word ptr cs:[hst_in_stat]
		add	dx,(-HOST_STATUS + HOST_CSA_OTEN)
		mov	al,1
		out	dx,al
;
;	Enable C011/C012 interrupts/DMA activity.
;
wl9:
		mov	dx,word ptr cs:[hst_out_stat]
		mov	al,2
		out	dx,al
;
;	Setup address and count for DMA transfer.
;
		mov	ax,es		;Get segment
		rol	ax,1		;Rotate left
		rol	ax,1		;Rotate left
		rol	ax,1		;Rotate left
		rol	ax,1		;Rotate left
		mov	dh,al		;Get highest nibble of es to dh
		and	al,0F0h		;Zero low nibble from segment
		add	ax,si		;Test for carry from add
		jnc	wl10
		inc	dh		;Carry means high 4 bits must be inc
wl10:
;
;	Figure out if one DMA transfer is enough (data doesn't cross physical
;	64K page boundary).  If not compute size of first and second transfers.
;
		mov	bx,cx		;Get a copy of transfer length
		add	bx,ax		;Set carry if transfer crosses 64K page
		mov	bx,0		;Assume it doesn't: no second transfer
		jnc	wl11		;Jump if a single DMA transfer will do
		mov	bx,cx		;Save total transfer length
		mov	cx,0		;Compute length of first transfer
		sub	cx,ax
		sub	bx,cx		;Compute length of second transfer
wl11:
		cli			;Disable interrupts for a bit
		out	HOST_DMA_BYTEF,al	;Clear byte order flip-flop
		out	HOST_DMA_ADDR,al	;Output low address
		mov	al,ah
		out	HOST_DMA_ADDR,al	;Output high address
		dec	cx		;Compensate for DMA count offset
		mov	al,cl
		out	HOST_DMA_COUNT,al	;Low count byte
		mov	al,ch
		out	HOST_DMA_COUNT,al	;High count byte
		mov	al,dh		;Get high 4 bits of addr
		and	al,0FH
		out	HOST_DMA_PREG,al	;Output high 4 bits to page reg
		mov	al,DMA_WRITE
		out	HOST_DMA_MODE,al	;Set write mode
		sti			;Enable interrupts
		mov	dl,DMA_TIME_CHECK
		mov	al,DMA_ON	;Channel #1 enable code
		out	HOST_DMA_MASK,al	;Enable 8237 channel
		cmp	byte ptr cs:[inmos_dma],0	;INMOS B008 DMA in use?
		je	wl12		;Jump if not
;
;	Enable INMOS B008 DMA write transfer.
;
		push	dx
		mov	dx,word ptr cs:[hst_in_stat]
		add	dx,(-HOST_STATUS + HOST_DMA_CNTL)
		xor	al,al
		out	dx,al
		pop	dx
;
;	Poll DMA controller for write completion.
;
wl12:
		in	al,HOST_DMA_STAT
		test	al,DMA_TC_FLAG	;See if done
		jnz	wl14		;Jump if so
		cmp	Timeout,0	;Timeout in use?
		je	wl13		;Jump if not
		cmp	word ptr cs:[timeout_left],0	;Timed out?
		je	wl15		;Jump if timed out
wl13:
		dec	dl		;Time for ^C check?
		jnz	wl12		;Test again
		mov	dl,DMA_TIME_CHECK
		mov	ah,11
		int	21h		;Check keyboard status (allow ^C)
		jmp	wl12
;
;	We've finished the first DMA transfer, see if another needed.
;
wl14:
		cmp	bx,0		;Anything left to transfer
		je	wl15		;Jump if not
		mov	ax,0		;Start at beginning of next 64K page
		inc	dh		;Update page address
		mov	cx,bx		;Length of second transfer
		mov	bx,0		;No more needed after this
		jmp	wl11		;Start second transfer
;
;	Disable C011/C012 interrupts/DMA activity.
;
wl15:
		mov	dx,word ptr cs:[hst_out_stat]
		xor	al,al
		out	dx,al
;
;	Disable CSA board DMA output channel (unless INMOS B008).
;
		cmp	byte ptr cs:[inmos_dma],0	;INMOS B008 DMA in use?
		jne	wl16		;Jump if so
		mov	dx,word ptr cs:[hst_in_stat]
		add	dx,(-HOST_STATUS + HOST_CSA_OTEN)
		out	dx,al
;
;	Disable 8237 DMA channel.
;
wl16:
		mov	al,DMA_OFF	;Channel #1 disable code
		out	HOST_DMA_MASK,al	;Disable 8237 channel
;
;	Figure out how many bytes we have left (in case of timeout).
;
		cli			;Disable interrupts for a bit
		out	HOST_DMA_BYTEF,al	;Clear byte order flip-flop
		in	al,HOST_DMA_COUNT	;Low count byte
		mov	cl,al
		in	al,HOST_DMA_COUNT	;High count byte
		sti			;Enable interrupts
		mov	ch,al
		inc	cx		;Compensate for DMA count offset
		add	cx,bx		;# left in possible second DMA transfer
		mov	ax,Count	;Get length of transfer
		sub	ax,cx		;Compute # bytes written
wl17:
		ret

WriteLink	endp

;
;	int	ResetLink(int LinkId)
;
;	Reset the subsystem associated with the specified link.
;
;	COMMENTS:
;		None.
;
;	RETURNS:
;		1 if the reset is successful, negative otherwise.
;
;	NOTES:
;		For the purposes of this driver (under MS-DOS), the "LinkId"
;		is ignored and the value returned is always greater than or
;		equal to zero (unless the link wasn't open).
;
ResetLink	PROC	LinkID:WORD

		cmp	byte ptr cs:[in_use],0	;See if link opened
		jne	rs1		;Jump if opened
		mov	ax,-1		;Not opened, return error
		jmp	rs2
rs1:
		mov	ax,RESET_DELAY
		call	NEAR PTR tick_wait
		mov	dx,word ptr cs:[hst_analyze]
		mov	al,0
		out	dx,al		;Lower ANALYZE
		mov	dx,word ptr cs:[hst_reserr]
		out	dx,al		;Lower RESET
		mov	ax,RESET_DELAY
		call	NEAR PTR tick_wait
		mov	dx,word ptr cs:[hst_reserr]
		mov	al,HOST_RESET_MASK
		out	dx,al		;Assert RESET
		mov	ax,RESET_DELAY
		call	NEAR PTR tick_wait
		mov	dx,word ptr cs:[hst_reserr]
		mov	al,0
		out	dx,al		;Lower RESET
		mov	ax,RESET_DELAY
		call	NEAR PTR tick_wait
		mov	ax,1		;Everything OK
rs2:
		ret

ResetLink	endp

;
;	int	AnalyzeLink(int LinkId)
;
;	Analyze the subsystem associated with the specified link.
;
;	COMMENTS:
;		None.
;
;	RETURNS:
;		1 if the analyze is successful, negative otherwise.
;
;	NOTES:
;		For the purposes of this driver (under MS-DOS), the "LinkId"
;		is ignored and the value returned is always greater than or
;		equal to zero (unless the link wasn't open).
;
AnalyzeLink	PROC	LinkID:WORD

		cmp	byte ptr cs:[in_use],0	;See if link opened
		jne	al1		;Jump if opened
		mov	ax,-1		;Not opened, return error
		jmp	al2
al1:
		mov	ax,RESET_DELAY
		call	NEAR PTR tick_wait
		mov	dx,word ptr cs:[hst_analyze]
		mov	al,0
		out	dx,al		;Lower ANALYZE
		mov	dx,word ptr cs:[hst_reserr]
		out	dx,al		;Lower RESET
		mov	ax,RESET_DELAY
		call	NEAR PTR tick_wait
		mov	dx,word ptr cs:[hst_analyze]
		mov	al,HOST_ANALYZE_MASK
		out	dx,al		;Assert ANALYZE
		mov	ax,ANALYZE_DELAY
		call	NEAR PTR tick_wait
		mov	dx,word ptr cs:[hst_reserr]
		mov	al,HOST_RESET_MASK
		out	dx,al		;Assert RESET
		mov	ax,RESET_DELAY
		call	NEAR PTR tick_wait
		mov	dx,word ptr cs:[hst_reserr]
		mov	al,0
		out	dx,al		;Lower RESET
		mov	dx,word ptr cs:[hst_analyze]
		out	dx,al		;Lower ANALYZE
		mov	ax,RESET_DELAY
		call	NEAR PTR tick_wait
		mov	ax,1		;Everything OK
al2:
		ret

AnalyzeLink	endp

;
;	int	TestError(int LinkId)
;
;	Test the error status associated with the specified link.
;
;	COMMENTS:
;		None.
;
;	RETURNS:
;		1 if error is set, 0 if it is not and negative to indicate an
;		error.
;
;	NOTES:
;		For the purposes of this driver (under MS-DOS), the "LinkId"
;		is ignored and the value returned is always greater than or
;		equal to zero (even if the link wasn't open).
;
TestError	PROC	LinkId:WORD

		mov	dx,word ptr cs:[hst_reserr]
		in	al,dx
		and	ax,HOST_ERROR_MASK
		xor	ax,HOST_ERROR_MASK	;Negative logic
		ret

TestError	endp

;
;	int	TestRead(int LinkId)
;
;	Test input status of the link.
;
;	COMMENTS:
;		None.
;
;	RETURNS:
;		Returns 1 if "ReadLink" will return one byte without timeout,
;		0 if it may not, and negative to indicate an error.
;
;	NOTES:
;		For the purposes of this driver (under MS-DOS), the "LinkId"
;		is ignored and the value returned is always 0 or 1.
;
TestRead	PROC	LinkID:WORD

		mov	dx,word ptr cs:[hst_in_stat]
		in	al,dx
		and	ax,HOST_RBF_MASK
		ret

TestRead	endp

;
;	int	TestWrite(int LinkId)
;
;	Test output status of the link.
;
;	COMMENTS:
;		None.
;
;	RETURNS:
;		Returns 1 if "WriteLink" can write one byte without timeout,
;		0 if it may not, and negative to indicate an error.
;
;	NOTES:
;		For the purposes of this driver (under MS-DOS), the "LinkId"
;		is ignored and the value returned is always 0 or 1.
;
TestWrite	PROC	LinkId:WORD

		mov	dx,word ptr cs:[hst_out_stat]
		in	al,dx
		and	ax,HOST_TBF_MASK
		ret

TestWrite	endp

;************************************************************************
;*									*
;*			Internal Helper Functions			*
;*									*
;************************************************************************

;
;	catch_cntlc-	Control "C" interrupt catcher.  Removes timer
;			interrupt linkage.
;
catch_cntlc	proc	far

		sti			;Enable interrupts
		push	ax		;Save registers
		push	bx
		push	cx
		push	dx
		push	si
		push	di
		push	bp
		push	ds
		push	es
		call	NEAR PTR remove_timeout	;Remove timer int. 
		call	NEAR PTR disable_dma	;Remove DMA linkages if present
		pop	es		;Restore registers
		pop	ds
		pop	bp
		pop	di
		pop	si
		pop	dx
		pop	cx
		pop	bx
		pop	ax		
		stc			;Tell MS-DOS to abort us
		ret	far

catch_cntlc	endp

;
;	disable_dma -	Remove/disable DMA linkages on channel #1.
;
disable_dma	PROC	NEAR

		cmp	byte ptr cs:[dma_enabled],0	;DMA in use?
		je	dd2		;Jump if not
		mov	al,DMA_OFF	;Channel #1 disable code
		out	HOST_DMA_MASK,al	;Disable 8237 channel
;
;	CSA specific disable board level DMA.
;
		mov	dx,word ptr cs:[hst_out_stat]
		add	dx,-HOST_STATUS	;Address of output data port
		in	al,dx		;Disable
;
;	This code disables interrupts for CSA, disables interrupts/DMA for MW
;
		mov	dx,word ptr cs:[hst_in_stat]
		xor	al,al
		out	dx,al		;Disable CSA interrupts/MW DMA
;
;	Disable CSA board DMA input and output channels.  This code MUST not
;	be executed with the INMOS B008!
;
		cmp	byte ptr cs:[inmos_dma],0	;INMOS B008 DMA in use?
		jne	dd1		;Jump if so
		add	dx,(-HOST_STATUS + HOST_CSA_INEN)
		out	dx,al		;Disable board input DMA channel
		inc	dx
		out	dx,al		;Disable board output DMA channel
dd1:
		mov	byte ptr cs:[dma_enabled],0	;DMA not in use now
dd2:
		ret

disable_dma	endp

;
;	enable_dma -	Install/enable DMA linkages on channel #1.
;
enable_dma	PROC	NEAR
		mov	al,DMA_OFF	;Channel #1 disable code
		out	HOST_DMA_MASK,al	;Disable 8237 channel
		in	al,HOST_DMA_STAT	;Clear status register
;
;	This code disables interrupts for CSA, disables interrupts/DMA for MW
;
		mov	dx,word ptr cs:[hst_in_stat]
		xor	al,al
		out	dx,al		;Disable CSA interrupts/MW DMA
;
;	Disable CSA board DMA input and output channels.  The following code
;	MUST not be executed when using the INMOS B008!
;
		cmp	byte ptr cs:[inmos_dma],0	;INMOS B008 DMA in use?
		jne	ed1		;Jump if so
		add	dx,(-HOST_STATUS + HOST_CSA_INEN)
		out	dx,al		;Disable board input DMA channel
		inc	dx
		out	dx,al		;Disable board output DMA channel
;
;	CSA specific enable board level DMA.  This code shouldn't bother the
;	B008, but why bother...
;
		mov	dx,word ptr cs:[hst_in_stat]
		add	dx,-HOST_STATUS	;Address of data port
		out	dx,al		;Enable
ed1:
		ret

enable_dma	endp

;
;
;	install_timeout-Install timeout and control-c interrupt routines.
;
install_timeout	PROC 	NEAR

		cmp	byte ptr cs:[using_timeout],0	;Timeout in use?
		jne	it1		;Jump if so

		push	es		;Save it
;
;	Save existing timer interrupt vector.
;
		mov	ax,3508h	;Get existing contents of int. vector
		int	21h
		mov	word ptr cs:[old_timer_ivec],bx
		mov	word ptr cs:[old_timer_ivec + 2],es
;
;	Save existing control-c interrupt vector.
;
		mov	ax,3523h	;Get existing contents of int. vector
		int	21h
		mov	word ptr cs:[old_cntlc_ivec],bx
		mov	word ptr cs:[old_cntlc_ivec + 2],es
;
;	Install new timer interrupt vector.
;
		mov	byte ptr cs:[using_timeout],1	;Timeout IN use
		mov	dx,offset cs:new_timer_int	;Point to timeout code
		mov	ax,2508h	;Update vector
		push	ds		;Save old data segment
		push	cs
		pop	ds		;Transfer cs into ds
		int	21h
		pop	ds		;Restore old data segment
;
;	Install new control-c interrupt vector.
;
		mov	dx,offset cs:catch_cntlc	;Point to ctrl-c code
		mov	ax,2523h	;Update vector
		push	ds		;Save old data segment
		push	cs
		pop	ds		;Transfer cs into ds
		int	21h
		pop	ds		;Restore old data segment
;
;	Restore ES.
;
		pop	es		;Restore for string operations
it1:
		ret

install_timeout	endp

;
;	new_timer_int-	Replacement timer interrupt to handle timeouts
;
new_timer_int	proc	far

		pushf			;Push flags to simulate interrupt call
		call	cs:old_timer_ivec	;Chain through to next routine
		cmp	cs:timeout_left,0	;Any counts left in timeout?
		je	nti1		;If not just quit
		dec	cs:timeout_left	;Decrement count
nti1:
		iret			;Done with interrupt

new_timer_int	endp

;
;	remove_timeout-	Remove timeout and control-c interrupt routines.
;
remove_timeout	PROC	NEAR

		cmp	byte ptr cs:[using_timeout],0	;Timeout in use?
		je	rt1		;Jump if not
		push	ds		;Save data segment
		lds	dx,cs:old_timer_ivec
		mov	ax,2508h	;Restore timer vector command
		int	21h
		lds	dx,cs:old_cntlc_ivec
		mov	ax,2523h	;Restore control-c vector command
		int	21h
		pop	ds		;Restore data segment
		mov	byte ptr cs:[using_timeout],0	;Timeout NOT in use
rt1:
		ret

remove_timeout	endp

;
;	tick_wait-	Waits the appropriate # of CLOCK_TICK intervals.
;			Assumes the timeout interrupt routine is installed.
;
tick_wait	PROC	NEAR

		mov	word ptr cs:[timeout_left],ax	;Timeout to use
w1:
		cmp	cs:timeout_left,0	;Are we done?
		jne	w1		;Jump if not
		ret

tick_wait	endp

;************************************************************************
;*									*
;*			End of Code/Data				*
;*									*
;************************************************************************

		end
