/************************************************************************/
/*									*/
/*	File:	tasm4.c							*/
/*									*/
/*	Transputer Assembler (TASM), source file 4.  This file contains	*/
/*	the code which reads an internal temp file and handles whatever	*/
/*	PC relative references can be resolved.  It also contains the	*/
/*	code which converts from the internal temp file format to the	*/
/*	output format used as input to the linker (TLNK).		*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<string.h>		/* String function include file */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tasmdef.h"		/* Assembler definitions */
#include	"tasmtyp.h"		/* Assembler data declarations */
#include	"tasmext.h"		/* Assembler externals */

/************************************************************************/
/*									*/
/*	Routine:	adj_max_loc					*/
/*									*/
/*	This routine is a helping function for routines which are	*/
/*	called by "pass1".  It figures out if the specified symbol	*/
/*	definition is a forward reference from the current location	*/
/*	counter.  If not, the "max_value" for the symbol is returned.	*/
/*	If so, the "max_value" is adjusted with the current value of	*/
/*	"max_loc_adj" to compensate for the change which has occurred	*/
/*	in the maximum location counter since the last time the symbol	*/
/*	definition value was updated (the previous pass of "pass1").	*/
/*									*/
/*	Parameters:							*/
/*		"index"	A symbol table index value.			*/
/*									*/
/*	Returns:	The correct "max_value" to use for the symbol.	*/
/*									*/
/************************************************************************/

SLONG
adj_max_loc(index)
	UINT			index;
	{
/*
 *	Figure out if it is a forward reference.
 */
	if((st[index].max_value - st[index].min_value) >
		(max_loc[module] - min_loc[module]))
		{
		return(st[index].max_value - max_loc_adj[module]);
		}
	return (st[index].max_value);	/* Nope */
	}

/************************************************************************/
/*									*/
/*	Routine:	adj_min_loc					*/
/*									*/
/*	This routine is a helping function for routines which are	*/
/*	called by "pass1" and "pass2".  It figures out if the specified	*/
/*	symbol definition is a forward reference from the current	*/
/*	location counter.  If not, the "min_value" for the symbol is	*/
/*	returned.  If so, the "min_value" is adjusted with the current	*/
/*	value of "min_loc_adj" to compensate for the change which has	*/
/*	occurred in the minimum location counter since the last time	*/
/*	the symbol definition value was updated (the previous pass of	*/
/*	"pass1").							*/
/*									*/
/*	Parameters:							*/
/*		"index"	A symbol table index value.			*/
/*									*/
/*	Returns:	The correct "min_value" to use for the symbol.	*/
/*									*/
/************************************************************************/

SLONG
adj_min_loc(index)
	UINT			index;
	{
/*
 *	Figure out if it is a forward reference.
 */
	if((st[index].max_value - st[index].min_value) >
		(max_loc[module] - min_loc[module]))
		{
		return(st[index].min_value + min_loc_adj[module]);
		}
	return (st[index].min_value);	/* Nope */
	}

/************************************************************************/
/*									*/
/*	Routine:	bind_constant					*/
/*									*/
/*	Takes a PC relative constant which represents the offset from	*/
/*	the current location counter and figures out how long the	*/
/*	corresponding instruction needs to be taking into account that	*/
/*	the Transputer increments the PC to point just beyond the	*/
/*	instruction BEFORE it processes the PC offset, thus the length	*/
/*	of the instruction affects the needed offset AND the offset in	*/
/*	turn affects the needed prefix string length for the		*/
/*	instruction!							*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The desired opcode.			*/
/*		"value"	-	The desired PC offset from the start of	*/
/*				the instruction/prefix string.		*/
/*									*/
/*	Returns:	The instruction length needed.			*/
/*									*/
/************************************************************************/

UINT
bind_constant(function,value)
	UINT			function;
	SLONG			value;
	{
	UINT			length1;
	UINT			length2;
/*
 *	First we calculate the needed instruction length ignoring the effect
 *	of the offset provided by the instruction length itself.
 */
	length1 = i_size(function,value);
/*
 *	Now we take a first cut at the instruction length including the self
 *	referential aspects.
 */
	length2 = i_size(function,((SLONG) (value - length1)));
/*
 *	Finally, we do the calculation again to ensure that we aren't
 *	oscillating across a instruction length straddle point.  The two
 *	iterations of the length calculation are sufficent since we can be
 *	straddling at most one boundary.
 */
	length1 = i_size(function,((SLONG) (value - length2)));
/*
 *	Now take the "max" of the above two calculated lengths to ensure that
 *	we cover the worst case.
 */
	if(length1 > length2)
		return (length1);
	else
		return (length2);
	}

/************************************************************************/
/*									*/
/*	Routine:	check_def					*/
/*									*/
/*	Reads a symbol table entry number from the current temporary	*/
/*	file and verifies that the symbol has been defined.  The symbol	*/
/*	number is also returned.  This routine is a helping function	*/
/*	for the various routiens called by "pass1".			*/
/*									*/
/*	Parameters:							*/
/*		"index"	A UINT pointer to where the result goes.	*/
/*		"pull"	The number of bytes to pull from the temporary	*/
/*			file if the symbol was undefined.		*/
/*									*/
/*	Returns:	TRUE if symbol was undefined, FALSE otherwise.	*/
/*									*/
/************************************************************************/

BOOL
check_def(index,pull)
	UINT			*index;
	UINT			pull;
	{
	char			buf[SYMNMESIZE + 1];

	*index = get_2b();		/* ST index # */
	if(! (st[*index].type & (S_DEFINED | S_EXTERN)))	/* Defined? */
		{
		while(pull--)
			get_1b();	/* Nope, pull rest of record */
		string_copy(buf,st[*index].id.name,SYMNMESIZE);
		buf[SYMNMESIZE] = '\0';	/* Terminate */
		post_error(ERR_UDEF,buf);	/* Replace with err. mess. */
		return (TRUE);
		}
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	copy						*/
/*									*/
/*	Copy a specified # of bytes from the "read" temp file to the	*/
/*	"write" temp file.						*/
/*									*/
/*	Parameters:							*/
/*		"count"	The number of bytes to copy.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
copy(count)
	register	UINT	count;
	{
	while(count--)
		put_1b(get_1b());
	}

/************************************************************************/
/*									*/
/*	Routine:	copy_out					*/
/*									*/
/*	Copy a specified # of bytes from the "read" temp file to the	*/
/*	output file.							*/
/*									*/
/*	Parameters:							*/
/*		"count"	The number of bytes to copy.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
copy_out(count)
	register	UINT	count;
	{
	while(count--)
		out_1b(get_1b());
	}

/************************************************************************/
/*									*/
/*	Routine:	do_addr						*/
/*									*/
/*	A helping function for "pass1" which handles the "T_ADDR_DATA"	*/
/*	records.							*/
/*									*/
/*	Parameters:							*/
/*		"check"	Set to TRUE if we are recording symbols which	*/
/*			the linker will need (they are referenced).	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
do_addr(check)
	BOOL			check;
	{
	UINT			index;

	number = get_2b();		/* Input line # */
	if(check_def(&index,4))
		return;			/* Get ST entry # and ensure defined */
/*
 *	See if symbol was a constant which allows the record to be converted
 *	into a T_DATA record!
 */
	if(st[index].type & S_CONSTANT)
		{
		put_1b((int) T_DATA);	/* Record type */
		put_2b((int) number);	/* Line # */
		put_2b((int) word_size);	/* # of data bytes */
		if(word_size == 2)
			put_2b((int) (get_4b() + st[index].min_value));
		else
			put_4b((SLONG) (get_4b() + st[index].min_value));
		min_loc[module] += word_size;
		max_loc[module] += word_size;
		return;
		}
/*
 *	Symbol was defined, but not a constant.  Copy reference to the output
 *	file.
 */
	if(check)
		st[index].type |= S_KEEP;	/* Indicate linker will need */

	put_1b((int) T_ADDR_DATA);	/* Record type */
	put_2b((int) number);		/* Line # */
	put_2b((int) index);		/* ST index # */
	copy(4);			/* Copy reference offset */
/*
 *	Update location counters.
 */
	min_loc[module] += word_size;
	max_loc[module] += word_size;
	}

/************************************************************************/
/*									*/
/*	Routine:	do_diff						*/
/*									*/
/*	A helping function for "pass1" which handles the "rel-rel"	*/
/*	data reference (T_RELREL_DATA), and tries to completely bind	*/
/*	the addresses involved so that the reference can be converted	*/
/*	into a normal data record.					*/
/*									*/
/*	Parameters:							*/
/*		"check"	Set to TRUE if we are recording symbols which	*/
/*			the linker will need (they are referenced).	*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
do_diff(check)
	BOOL			check;
	{
	UINT			index1;
	UINT			index2;

	number = get_2b();		/* Input line # */
	if(check_def(&index1,6))
		return (FALSE);		/* Get "left" ST # */
	if(check_def(&index2,4))
		return (FALSE);		/* Get "right" ST # */
/*
 *	If symbols were external, not to be used at assembly time, or defined
 *	in different modules, copy verbatum to the temp file since no address
 *	binding can occur.  First though, check to see if both symbols were
 *	constants, and thus the record will be converted by the address binding
 *	code later on.
 */
	if(! ((st[index1].type & S_CONSTANT) &&
		(st[index2].type & S_CONSTANT)))
		{
		if((st[index1].type & (S_EXTERN | S_CONSTANT)) ||
			(st[index2].type & (S_EXTERN | S_CONSTANT)) ||
				(st[index1].module != st[index2].module))
			{
			if(check)
				st[index1].type |= S_KEEP;
			if(check)
				st[index2].type |= S_KEEP;

			put_1b((int) T_RELREL_DATA);	/* Record type field */
			put_2b((int) number);	/* Line # */
			put_2b((int) index1);	/* "left" ST index # */
			put_2b((int) index2);	/* "right" ST index # */
			copy(4);
			min_loc[module] += word_size;
			max_loc[module] += word_size;
			return (FALSE);
			}
		}
/*
 *	It appears we can attempt to do address binding, first see if we now
 *	know enough to convert the reference to a normal data record.
 */
	if((st[index1].max_value - st[index2].max_value) ==
		(st[index1].min_value - st[index2].min_value))
		{
		put_1b((int) T_DATA);	/* Absolute data record type */
		put_2b((int) number);	/* Line # */
		put_2b((int) word_size);	/* Data reference size */
		if(word_size == 2)
			{
			put_2b((int) (get_4b() + st[index1].max_value -
				st[index2].max_value));
			}
		else
			{
			put_4b((SLONG) (get_4b() + st[index1].max_value -
			st[index2].max_value));
			}
		min_loc[module] += word_size;
		max_loc[module] += word_size;
		return (TRUE);		/* We've performed a conversion */
		}
/*
 *	We can't finish the binding, copy the reference to the temp file so
 *	that it may get	converted in some future pass.
 */
	if(check)
		st[index1].type |= S_KEEP;	/* Indicate linker will need */
	if(check)
		st[index2].type |= S_KEEP;	/* Indicate linker will need */

	put_1b((int) T_RELREL_DATA);	/* Record type */
	put_2b((int) number);		/* Line # */
	put_2b((int) index1);		/* "left" ST index # */
	put_2b((int) index2);		/* "right" ST index # */
	copy(4);			/* Copy offset */
	min_loc[module] += word_size;
	max_loc[module] += word_size;
	return (FALSE);			/* No conversion performed */
	}

/************************************************************************/
/*									*/
/*	Routine:	do_i_addr					*/
/*									*/
/*	A helping function for "pass1" which handles the "T_ADDR_OP"	*/
/*	records.							*/
/*									*/
/*	Parameters:							*/
/*		"check"	Set to TRUE if we are recording symbols which	*/
/*			the linker will need (they are referenced).	*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
do_i_addr(check)
	BOOL			check;
	{
	SLONG			con;
	UINT			index;
	UINT			length;
	UINT			min_length;
	UINT			function;

	number = get_2b();		/* Input line # */
	min_length = get_1b();		/* Get minimum instruction length */
	if(check_def(&index,5))
		return (FALSE);	/* Get ST entry # */
/*
 *	See if symbol was a constant, and we can thus convert the record into
 *	a T_DATA record.
 */
	if(st[index].type & S_CONSTANT)
		{
		put_1b((int) T_DATA);	/* Record type */
		put_2b((int) number);
		con = get_4b() + st[index].min_value;
		function = get_1b();
		length = i_size((UINT) function,(SLONG) con);
		if(length < min_length)
			length = min_length;
		put_2b((int) length);	/* # of data bytes */
		build_i((UINT) length,(UINT) function,(SLONG) con);
		min_loc[module] += length;
		max_loc[module] += length;
		return (TRUE);
		}
/*
 *	Symbol was defined, but wasn't a constant.  Copy reference to the
 *	output file.
 */
	if(check)
		st[index].type |= S_KEEP;	/* Indicate linker will need */

	put_1b((int) T_ADDR_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	put_1b((int) min_length);	/* Minimum instruction length */
	put_2b((int) index);		/* ST index # */
	copy(4);			/* Copy reference offset */
	put_1b((int) (function = get_1b()));	/* Copy opcode */
/*
 *	Update location counters.
 */
	min_loc[module] += min_length;
	if(function == DOTLDCR)
		max_loc[module] += (max_i_size + LDPI_SIZE);
	else
		max_loc[module] += max_i_size;
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	do_i_diff					*/
/*									*/
/*	A helping function for "pass1" which handles the "rel-rel"	*/
/*	instruction reference (T_RELREL_OP), and tries to completely	*/
/*	bind the addresses involved so that the reference can be	*/
/*	converted into a normal data record.				*/
/*									*/
/*	Parameters:							*/
/*		"check"	Set to TRUE if we are recording symbols which	*/
/*			the linker will need (they are referenced).	*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
do_i_diff(check)
	BOOL			check;
	{
	SLONG			con;
	UINT			function;
	UINT			index1;
	UINT			index2;
	UINT			length;
	UINT			min_length;

	number = get_2b();		/* Input line # */
	min_length = get_1b();		/* Get minimum instruction length */
	if(check_def(&index1,7))
		return (FALSE);	/* Get "left" ST # */
	if(check_def(&index2,5))
		return (FALSE);	/* Get "right" ST # */
/*
 *	If symbols were external, not to be used at assembly time, or defined
 *	in different modules, copy verbatum to the temp file since no address
 *	binding can occur.  First though, check to see if both symbols were
 *	constants, and thus the record will be converted by the address binding
 *	code later on.
 */
	if(! ((st[index1].type & S_CONSTANT) &&
		(st[index2].type & S_CONSTANT)))
		{
		if((st[index1].type & (S_EXTERN | S_CONSTANT)) ||
			(st[index2].type & (S_EXTERN | S_CONSTANT)) ||
				(st[index1].module != st[index2].module))
			{
			if(check)
				st[index1].type |= S_KEEP;
			if(check)
				st[index2].type |= S_KEEP;

			put_1b((int) T_RELREL_OP);	/* Record type field */
			put_2b((int) number);	/* Line # */
			put_1b((int) min_length);	/* Min. instr. len. */
			put_2b((int) index1);	/* "left" ST index # */
			put_2b((int) index2);	/* "right" ST index # */
			copy(5);
			min_loc[module] += min_length;
			max_loc[module] += max_i_size;
			return (FALSE);
			}
		}
/*
 *	It appears we can attempt to do address binding, first see if we now
 *	know enough to convert the reference to a normal data record.
 */
	if((adj_max_loc(index1) - adj_max_loc(index2)) ==
		(adj_min_loc(index1) - adj_min_loc(index2)))
		{
		put_1b((int) T_DATA);	/* Absolute data record type */
		put_2b((int) number);	/* Line # */
		con = get_4b() + adj_max_loc(index1) -
			adj_max_loc(index2);
		function = get_1b();
		length = i_size((UINT) function,(SLONG) con);
		if((min_length != 1) && (length > min_length))
			fatali("0");
		if(length < min_length)
			length = min_length;
		put_2b((int) length);	/* Data reference size */
		build_i(length,(UINT) function,(SLONG) con);
		min_loc[module] += length;
		max_loc[module] += length;
		return (TRUE);		/* We've performed a conversion */
		}
/*
 *	We can't finish the binding, copy the reference to the temp file so
 *	that it may get	converted in some future pass.
 */
	if(check)
		st[index1].type |= S_KEEP;	/* Indicate linker will need */
	if(check)
		st[index2].type |= S_KEEP;	/* Indicate linker will need */

	put_1b((int) T_RELREL_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	put_1b((int) min_length);	/* Minimum allowable length */
	put_2b((int) index1);		/* "left" ST index # */
	put_2b((int) index2);		/* "right" ST index # */
	put_4b((SLONG) (con = get_4b()));	/* Copy offset */
	put_1b((int) (function = get_1b()));
/*
 *	Figure out how to update the location counters.
 */
	updat_loc((UINT) function,
		(SLONG) (con + adj_max_loc(index1) -
				adj_max_loc(index2)),
		(SLONG) (con + adj_min_loc(index1) -
				adj_min_loc(index2)),
		(UINT) min_length, (BOOL) FALSE);
	return (FALSE);			/* No conversion performed */
	}

/************************************************************************/
/*									*/
/*	Routine:	do_i_wdiff					*/
/*									*/
/*	A helping function for "pass1" which handles the word size	*/
/*	scaled "rel-rel" instruction reference (T_WRELREL_OP), and	*/
/*	tries to completely bind the addresses involved so that the	*/
/*	reference can be converted into a normal data record.		*/
/*									*/
/*	Parameters:							*/
/*		"check"	Set to TRUE if we are recording symbols which	*/
/*			the linker will need (they are referenced).	*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
do_i_wdiff(check)
	BOOL			check;
	{
	SLONG			con;
	UINT			function;
	UINT			index1;
	UINT			index2;
	UINT			length;
	UINT			min_length;

	number = get_2b();		/* Input line # */
	min_length = get_1b();		/* Get minimum instruction length */
	if(check_def(&index1,7))
		return (FALSE);	/* Get "left" ST # */
	if(check_def(&index2,5))
		return (FALSE);	/* Get "right" ST # */
/*
 *	If symbols were external, not to be used at assembly time, or defined
 *	in different modules, copy verbatum to the temp file since no address
 *	binding can occur.  First though, check to see if both symbols were
 *	constants, and thus the record will be converted by the address binding
 *	code later on.
 */
	if(! ((st[index1].type & S_CONSTANT) &&
		(st[index2].type & S_CONSTANT)))
		{
		if((st[index1].type & (S_EXTERN | S_CONSTANT)) ||
			(st[index2].type & (S_EXTERN | S_CONSTANT)) ||
				(st[index1].module != st[index2].module))
			{
			if(check)
				st[index1].type |= S_KEEP;
			if(check)
				st[index2].type |= S_KEEP;

			put_1b((int) T_WRELREL_OP);	/* Record type field */
			put_2b((int) number);	/* Line # */
			put_1b((int) min_length);	/* Min. instr. len. */
			put_2b((int) index1);	/* "left" ST index # */
			put_2b((int) index2);	/* "right" ST index # */
			copy(5);
			min_loc[module] += min_length;
			max_loc[module] += max_i_size;
			return (FALSE);
			}
		}
/*
 *	It appears we can attempt to do address binding, first see if we now
 *	know enough to convert the reference to a normal data record.
 */
	if((adj_max_loc(index1) - adj_max_loc(index2)) ==
		(adj_min_loc(index1) - adj_min_loc(index2)))
		{
		put_1b((int) T_DATA);	/* Absolute data record type */
		put_2b((int) number);	/* Line # */
		con = (get_4b() + adj_max_loc(index1) -
			adj_max_loc(index2)) / (SLONG) word_size;
		function = get_1b();
		length = i_size((UINT) function,(SLONG) con);
		if((min_length != 1) && (length > min_length))
			fatali("0");
		if(length < min_length)
			length = min_length;
		put_2b((int) length);	/* Data reference size */
		build_i(length,(UINT) function,(SLONG) con);
		min_loc[module] += length;
		max_loc[module] += length;
		return (TRUE);		/* We've performed a conversion */
		}
/*
 *	We can't finish the binding, copy the reference to the temp file so
 *	that it may get	converted in some future pass.
 */
	if(check)
		st[index1].type |= S_KEEP;	/* Indicate linker will need */
	if(check)
		st[index2].type |= S_KEEP;	/* Indicate linker will need */

	put_1b((int) T_WRELREL_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	put_1b((int) min_length);	/* Minimum allowable length */
	put_2b((int) index1);		/* "left" ST index # */
	put_2b((int) index2);		/* "right" ST index # */
	put_4b((SLONG) (con = get_4b()));	/* Copy offset */
	put_1b((int) (function = get_1b()));
/*
 *	Figure out how to update the location counters.
 */
	updat_loc((UINT) function,
		(SLONG) (con + adj_max_loc(index1) -
				adj_max_loc(index2)) / (SLONG) word_size,
		(SLONG) (con + adj_min_loc(index1) -
				adj_min_loc(index2)) / (SLONG) word_size,
		(UINT) min_length, (BOOL) FALSE);
	return (FALSE);			/* No conversion performed */
	}

/************************************************************************/
/*									*/
/*	Routine:	dor_i_ref					*/
/*									*/
/*	A helping function for "pass1" which handles the symbolic PC	*/
/*	relative instructions (T_RELSYM_OP), and tries to completely	*/
/*	bind the addresses involved so that the reference can be	*/
/*	converted into a normal data reference.				*/
/*									*/
/*	Parameters:							*/
/*		"check"	Set to TRUE if we are recording symbols which	*/
/*			the linker will need (they are referenced).	*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
dor_i_ref(check)
	BOOL			check;
	{
	SLONG			con;
	UINT			function;
	UINT			index;
	UINT			length;
	UINT			min_length;

	number = get_2b();		/* Input line # */
	min_length = get_1b();		/* Get minimum instruction length */
	if(check_def(&index,5))
		return (FALSE);	/* Get ST entry # */
/*
 *	If the symbol was a constant we can convert the T_RELSYM_OP record
 *	into a T_REL_OP one, and eliminate a symbolic reference.
 */
	if(st[index].type & S_CONSTANT)
		{
		put_1b((int) T_REL_OP);	/* Record type */
		put_2b((int) number);	/* Line # */
		put_1b((int) min_length);	/* Minimum instruction len. */
		put_4b((SLONG) (get_4b() + st[index].min_value));
		copy(1);		/* Copy basic opcode also */
		min_loc[module] += min_length;
		max_loc[module] += max_i_size;
		return (FALSE);
		}
/*
 *	If symbol was external or defined in a different module copy it
 *	verbatum to the temp file since no address binding can occur.
 */
	if((st[index].type & S_EXTERN) || (st[index].module != module))
		{
		if(check)
			st[index].type |= S_KEEP;	/* Linker will need */

		put_1b((int) T_RELSYM_OP);	/* Record type field */
		put_2b((int) number);	/* Line # */
		put_1b((int) min_length);	/* Minimum inst. length */
		put_2b((int) index);	/* ST index # */
		copy(5);
		min_loc[module] += min_length;
		max_loc[module] += max_i_size;
		return (FALSE);
		}
/*
 *	It appears we can attempt to do address binding, first see if we now
 *	know enough to convert the reference to a normal data record.
 */
	if((adj_max_loc(index) - max_loc[module]) ==
		(adj_min_loc(index) - min_loc[module]))
		{
		put_1b((int) T_DATA);	/* Absolute data record type */
		put_2b((int) number);	/* Line # */
		con = get_4b() + adj_max_loc(index) - max_loc[module];
		function = get_1b();
		length = bind_constant((UINT) function,(SLONG) con);
		if((min_length != 1) && (length > min_length))
			fatali("1");
		if(length < min_length)
			length = min_length;	/* Minimum */
		put_2b((int) length);	/* Instr. size */
		build_i(length,(UINT) function,(SLONG) (con - length));
		min_loc[module] += length;
		max_loc[module] += length;
		return (TRUE);		/* We've performed a conversion */
		}
/*
 *	We can't finish the binding, determine the best and worst case for
 *	instruction length and update the location counters appropriately.
 *	We also need to copy the reference to the temp file so that it may get
 *	converted in some future pass.
 */
	if(check)
		st[index].type |= S_KEEP;	/* Indicate linker will need */

	put_1b((int) T_RELSYM_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	put_1b((int) min_length);	/* Minimum allowable length */
	put_2b((int) index);		/* ST index # */
	put_4b((SLONG) (con = get_4b()));	/* Copy offset */
	put_1b((int) (function = get_1b()));
/*
 *	Figure out how to update the location counters.
 */
	updat_loc((UINT) function,
		(SLONG) (con + adj_max_loc(index) - max_loc[module]),
		(SLONG) (con + adj_min_loc(index) - min_loc[module]),
			(UINT) min_length, (BOOL) TRUE);
	return (FALSE);			/* No conversion performed */
	}

/************************************************************************/
/*									*/
/*	Routine:	dor_ref						*/
/*									*/
/*	A helping function for "pass1" which handles the symbolic PC	*/
/*	relative data references (T_RELSYM_DATA), and tries to		*/
/*	completely bind the addresses involved so that the reference	*/
/*	can be converted into a normal data reference.			*/
/*									*/
/*	Parameters:							*/
/*		"check"	Set to TRUE if we are recording symbols which	*/
/*			the linker will need (they are referenced).	*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
dor_ref(check)
	BOOL			check;
	{
	UINT			index;

	number = get_2b();		/* Input line # */
	if(check_def(&index,4))
		return (FALSE);		/* Get ST entry # */
/*
 *	If the symbol was a constant we can convert the T_RELSYM_DATA record
 *	into a T_REL_DATA one, and eliminate a symbolic reference.
 */
	if(st[index].type & S_CONSTANT)
		{
		put_1b((int) T_REL_DATA);	/* Record type */
		put_2b((int) number);	/* Line # */
		put_4b((SLONG) (get_4b() + st[index].min_value));
		min_loc[module] += word_size;
		max_loc[module] += word_size;
		return (FALSE);
		}
/*
 *	If symbol was external or defined in a different module copy it
 *	verbatum to the temp file since no address binding can occur.
 */
	if((st[index].type & S_EXTERN) || (st[index].module != module))
		{
		if(check)
			st[index].type |= S_KEEP;	/* Linker will need */

		put_1b((int) T_RELSYM_DATA);	/* Record type field */
		put_2b((int) number);	/* Line # */
		put_2b((int) index);	/* ST index # */
		copy(4);
		min_loc[module] += word_size;
		max_loc[module] += word_size;
		return (FALSE);
		}
/*
 *	It appears we can attempt to do address binding, first see if we now
 *	know enough to convert the reference to a normal data record.
 */
	if((adj_max_loc(index) - max_loc[module]) ==
		(adj_min_loc(index) - min_loc[module]))
		{
		put_1b((int) T_DATA);	/* Absolute data record type */
		put_2b((int) number);	/* Line # */
		put_2b((int) word_size);	/* Data reference size */
		if(word_size == 2)
			{
			put_2b((int) (get_4b() + adj_max_loc(index) -
				max_loc[module]));
			}
		else
			{
			put_4b((SLONG) (get_4b() + adj_max_loc(index) -
				max_loc[module]));
			}
		min_loc[module] += word_size;
		max_loc[module] += word_size;
		return (TRUE);		/* We've performed a conversion */
		}
/*
 *	We can't finish the binding, copy the reference to the temp file so
 *	that it may get	converted in some future pass.
 */
	if(check)
		st[index].type |= S_KEEP;	/* Indicate linker will need */

	put_1b((int) T_RELSYM_DATA);	/* Record type */
	put_2b((int) number);		/* Line # */
	put_2b((int) index);		/* ST index # */
	copy(4);			/* Copy offset */
	min_loc[module] += word_size;
	max_loc[module] += word_size;
	return (FALSE);			/* No conversion performed */
	}

/************************************************************************/
/*									*/
/*	Routine:	grow_diff					*/
/*									*/
/*	A helping function for "pass2" which increments the minimum	*/
/*	instruction length for "rel-rel" instructions (T_RELREL_OP),	*/
/*	which are not external or inter-module, and are not at their	*/
/*	maximum length already.  Note that this routine needs to be	*/
/*	called at most once during the assembly process for each	*/
/*	"rel-rel" reference.  This is true since the spread between the	*/
/*	minimum and maximum instruction length is 1 to 8, while the	*/
/*	range of an additional prefix instruction is *16, thus a	*/
/*	reference can straddle only one prefix string length breakpoint	*/
/*	even in the worst case.  Of course, if INMOS ever develops a	*/
/*	Transputer with a 64 bit address, requiring 16 prefixes worst	*/
/*	case, all bets are off!						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
grow_diff()
	{
	SLONG			con;
	UINT			function;
	register	UINT	index1;
	register	UINT	index2;
	UINT			new_min_length;
	UINT			old_min_length;

	number = get_2b();		/* Input line # */
	old_min_length = get_1b();	/* Get minimum instruction length */
	index1 = get_2b();		/* "left" ST entry # */
	index2 = get_2b();		/* "right" ST entry # */
/*
 *	If symbols were external, not to be used at assembly time, or defined
 *	in different modules, copy the instruction verbatum to the temp file
 *	since no address binding can occur.
 */
	if((st[index1].type & (S_EXTERN | S_CONSTANT)) ||
		(st[index2].type & (S_EXTERN | S_CONSTANT)) ||
			(st[index1].module != st[index2].module))
		{
		put_1b((int) T_RELREL_OP);	/* Record type field */
		put_2b((int) number);	/* Line # */
		put_1b((int) old_min_length);	/* Minimum inst. length */
		put_2b((int) index1);	/* "left" ST index # */
		put_2b((int) index2);	/* "right" ST index # */
		copy(5);
		min_loc[module] += old_min_length;
		max_loc[module] += max_i_size;
		return;
		}
/*
 *	Compute the best and worst cases for the instruction lengths.  If they
 *	are the same we don't need to adjust the minimum length.  Otherwise we
 *	make the minimum length be the worst case length (which will be 1
 *	greater than the best case length).  We also need to copy the
 *	instruction through to the other temp file so that a future pass of
 *	"pass1" can continue on with the binding process.
 */
	con = get_4b();			/* Offset */
	function = get_1b();
	new_min_length = updat_length((UINT) function,
		(SLONG) (con + adj_max_loc(index1)-adj_max_loc(index2)),
			(SLONG) (con + adj_min_loc(index1) -
				adj_min_loc(index2)),
					(UINT) old_min_length, (BOOL) FALSE);
	put_1b((int) T_RELREL_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	put_1b((int) new_min_length);	/* Minimum allowable length */
	put_2b((int) index1);		/* "left" ST index # */
	put_2b((int) index2);		/* "right" ST index # */
	put_4b((SLONG) con);		/* Copy offset */
	put_1b((int) function);		/* Copy basic opcode */
	return;
	}

/************************************************************************/
/*									*/
/*	Routine:	grow_ref					*/
/*									*/
/*	A helping function for "pass2" which increments the minimum	*/
/*	instruction length for symbolic PC relative instructions	*/
/*	(T_RELSYM_OP), which are not external or inter-module, and are	*/
/*	not at their maximum length already.  Note that this routine	*/
/*	needs to be called at most once during the assembly process for	*/
/*	each PC relative reference.  This is true since the spread	*/
/*	between the minimum and maximum instruction length is 1 to 8,	*/
/*	while the range of an additional prefix instruction is *16,	*/
/*	thus a reference can straddle only one prefix string length	*/
/*	breakpoint even in the worst case.  Of course if INMOS ever	*/
/*	develops a Transputer with a 64 bit address, requiring 16	*/
/*	prefixes worst case, all bets are off! 				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
grow_ref()
	{
	SLONG			con;
	UINT			function;
	register	UINT	index;
	UINT			new_min_length;
	UINT			old_min_length;

	number = get_2b();		/* Input line # */
	old_min_length = get_1b();	/* Get minimum instruction length */
	index = get_2b();		/* ST entry # */
/*
 *	If symbol was external or defined in a different module copy it
 *	verbatum to the temp file since no address binding can occur.
 */
	if((st[index].type & S_EXTERN) || (st[index].module != module))
		{
		put_1b((int) T_RELSYM_OP);	/* Record type field */
		put_2b((int) number);	/* Line # */
		put_1b((int) old_min_length);	/* Minimum inst. length */
		put_2b((int) index);	/* ST index # */
		copy(5);
		min_loc[module] += old_min_length;
		max_loc[module] += max_i_size;
		return;
		}
/*
 *	Compute the best and worst cases for the instruction lengths.  If they
 *	are the same we don't need to adjust the minimum length.  Otherwise we
 *	make the minimum length be the worst case length (which will be 1
 *	greater than the best case length).  We also need to copy the reference
 *	through to the other temp file so that future passes of "pass1" can
 *	continue on with the binding process.
 */
	con = get_4b();			/* Offset */
	function = get_1b();
	new_min_length = updat_length((UINT) function,
		(SLONG) (con + adj_max_loc(index) - max_loc[module]),
			(SLONG) (con + adj_min_loc(index) -
				min_loc[module]),
				(UINT) old_min_length, (BOOL) TRUE);
	put_1b((int) T_RELSYM_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	put_1b((int) new_min_length);	/* Minimum allowable length */
	put_2b((int) index);		/* ST index # */
	put_4b((SLONG) con);		/* Copy offset */
	put_1b((int) function);		/* Copy basic opcode */
	return;
	}

/************************************************************************/
/*									*/
/*	Routine:	grow_wdiff					*/
/*									*/
/*	A helping function for "pass2" which increments the minimum	*/
/*	instruction length for word size scaled "rel-rel" instructions	*/
/*	(T_WRELREL_OP), which are not external or inter-module, and are	*/
/*	not at their maximum length already.  Note that this routine	*/
/*	needs to be called at most once during the assembly process for	*/
/*	each reference.  This is true since the spread between the	*/
/*	minimum and maximum instruction length is 1 to 8, while the	*/
/*	range of an additional prefix instruction is *32 (T2 CPU's), or	*/
/*	*64 (T4/T8 CPU's), thus a reference can straddle only one	*/
/*	prefix string length breakpoint even in the worst case.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
grow_wdiff()
	{
	SLONG			con;
	UINT			function;
	register	UINT	index1;
	register	UINT	index2;
	UINT			new_min_length;
	UINT			old_min_length;

	number = get_2b();		/* Input line # */
	old_min_length = get_1b();	/* Get minimum instruction length */
	index1 = get_2b();		/* "left" ST entry # */
	index2 = get_2b();		/* "right" ST entry # */
/*
 *	If symbols were external, not to be used at assembly time, or defined
 *	in different modules, copy the instruction verbatum to the temp file
 *	since no address binding can occur.
 */
	if((st[index1].type & (S_EXTERN | S_CONSTANT)) ||
		(st[index2].type & (S_EXTERN | S_CONSTANT)) ||
			(st[index1].module != st[index2].module))
		{
		put_1b((int) T_WRELREL_OP);	/* Record type field */
		put_2b((int) number);	/* Line # */
		put_1b((int) old_min_length);	/* Minimum inst. length */
		put_2b((int) index1);	/* "left" ST index # */
		put_2b((int) index2);	/* "right" ST index # */
		copy(5);
		min_loc[module] += old_min_length;
		max_loc[module] += max_i_size;
		return;
		}
/*
 *	Compute the best and worst cases for the instruction lengths.  If they
 *	are the same we don't need to adjust the minimum length.  Otherwise we
 *	make the minimum length be the worst case length (which will be 1
 *	greater than the best case length).  We also need to copy the
 *	instruction through to the other temp file so that a future pass of
 *	"pass1" can continue on with the binding process.
 */
	con = get_4b();			/* Offset */
	function = get_1b();
	new_min_length = updat_length((UINT) function,
		(SLONG) (con + adj_max_loc(index1) - adj_max_loc(index2)) /
			(SLONG) word_size,
		(SLONG) (con + adj_min_loc(index1) - adj_min_loc(index2)) /
			(SLONG) word_size,
		(UINT) old_min_length, (BOOL) FALSE);
	put_1b((int) T_WRELREL_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	put_1b((int) new_min_length);	/* Minimum allowable length */
	put_2b((int) index1);		/* "left" ST index # */
	put_2b((int) index2);		/* "right" ST index # */
	put_4b((SLONG) con);		/* Copy offset */
	put_1b((int) function);		/* Copy basic opcode */
	return;
	}

/************************************************************************/
/*									*/
/*	Routine:	loc_pad						*/
/*									*/
/*	Updates the minimum and maximum location counters as a result	*/
/*	of encountering a ".align" pseudo-op.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
loc_pad()
	{
	int			temp;

	if(max_loc[module] == min_loc[module])
		{
		temp = (word_size - (min_loc[module] & (word_size - 1))) &
			(word_size - 1);
		min_loc[module] += temp;
		max_loc[module] += temp;
		}
	else
		max_loc[module] += (word_size - 1);
	}

/************************************************************************/
/*									*/
/*	Routine:	out_rec						*/
/*									*/
/*	A helping function for "pass3" which is used to flush the	*/
/*	T_DATA/T_STORAGE accumulators when the output file is being	*/
/*	compressed.							*/
/*									*/
/*	Parameters:							*/
/*		"op"	The record type being output.  T_RESERVED is	*/
/*			used as a flag to indicate the accumulators	*/
/*			should be flushed.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
out_rec(op)
	UINT			op;
	{
	if(acc_dtasize > 0)
		{
		out_1b((int) T_DATA);
		out_2b((int) 0);
		out_2b((int) acc_dtasize);
		out_write((char *) acc_dta,acc_dtasize);
		acc_dtasize = 0;
		}
	if(acc_storesize > 0)
		{
		out_1b((int) T_STORAGE);
		out_2b((int) 0);
		out_4b((SLONG) acc_storesize);
		acc_storesize = 0;
		}
	if(op == T_RESERVED)
		return;
	out_1b((int) op);
	}

/************************************************************************/
/*									*/
/*	Routine:	pass1						*/
/*									*/
/*	Makes a pass through the temp files converting whatever PC	*/
/*	relative references which can be bound.  It updates the symbol	*/
/*	table location counters as the actual instruction lengths are	*/
/*	determined more accurately.					*/
/*									*/
/*	Parameters:							*/
/*		"check"	TRUE if "pass1" should record which symbols are	*/
/*			referenced on this pass through the temp file.	*/
/*			The idea is to call "pass1" with check set only	*/
/*			the last time.  This allows the symbol table to	*/
/*			accumulate which local symbol definitions need	*/
/*			to be passed to the linker.			*/
/*									*/
/*	Returns:	TRUE if no additional reference records could	*/
/*			be converted in this pass, otherwise FALSE (so	*/
/*			that "pass1" will be called again to iterate	*/
/*			towards the final result).			*/
/*									*/
/************************************************************************/

BOOL
pass1(check)
	BOOL			check;
	{
	SLONG			big_temp;
	register	UINT	length;
	BOOL			min_size;
	register	UINT	temp;
	UINT			temp2;

#if	DEBUG
	putchar('1');			/* Indicate doing pass 1 */
#endif
/*
 *	Initialize for this pass.
 */
	for(temp = module = 0; temp < NUMMODULES; temp++)
		{
		max_loc_adj[temp] = min_loc_adj[temp] = min_loc[temp] =
			max_loc[temp] = 0;
		}
	min_size = TRUE;		/* FALSE if we reduce something */
/*
 *	Loop through all the records in the temp file.  Do whatever is needed
 *	for each different record type.
 */
	FOREVER
		{
		switch(temp = get_1b())
			{
			case T_DATA:	/* Normal data definition */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_2b((int) (temp = get_2b()));
				copy(temp);	/* Actual data */
				min_loc[module] += temp;
				max_loc[module] += temp;
				break;

			case T_RELSYM_OP:	/* Symbolic PC rel. instr. */
				if(dor_i_ref(check))
					min_size = FALSE;
				break;

			case T_DEF:	/* Label definition */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_2b((int) (temp = get_2b()));
				if((st[temp].min_value != min_loc[module]) ||
					(st[temp].max_value !=
						max_loc[module]))
					{
					min_size = FALSE;
					min_loc_adj[module] =
						min_loc[module] - 
							st[temp].min_value;
					max_loc_adj[module] =
						st[temp].max_value -
							max_loc[module];
					st[temp].min_value = min_loc[module];
					st[temp].max_value = max_loc[module];
					}
				break;

			case T_ALIGN:	/* Word alignment pseudo-op */
				put_1b((int) temp);
				copy(2);	/* Line # */
				loc_pad();	/* Update location counters */
				break;

			case T_STORAGE:	/* Reserve space pseudo-op */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_4b((SLONG) (big_temp = get_4b()));
				min_loc[module] += big_temp;
				max_loc[module] += big_temp;
				break;

			case T_RELREL_OP:	/* "rel-rel" ref. instr. */
				if(do_i_diff(check))
					min_size = FALSE;
				break;

			case T_WRELREL_OP:	/* Scaled "rel-rel" instr. */
				if(do_i_wdiff(check))
					min_size = FALSE;
				break;

			case T_SET:	/* Constant equate operation */
				put_1b((int) temp);
				copy(8);	/* Ln #, sym #, value */
				break;

			case T_ADDR_DATA:	/* Address ref. in data */
				do_addr(check);
				break;

			case T_MODULE:	/* Module # definition */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_1b((int) (temp = get_1b()));
				module = temp;
				break;

			case T_ADDR_OP:	/* Address reference in instruction */
				if(do_i_addr(check))
					min_size = FALSE;
				break;

			case T_RELREL_DATA:	/* "rel-rel" ref. in data */
				if(do_diff(check))
					min_size = FALSE;
				break;

			case T_REL_OP:	/* Non-symbolic PC relative instr. */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_1b((int) (temp = get_1b()));
				copy(5);	/* addr, opcode */
				min_loc[module] += temp;
				max_loc[module] += max_i_size;
				break;

			case T_FILENAME:	/* Filename record */
				put_1b((int) temp);
				copy(4);	/* Old & new line #'s */
				put_1b((int) (length = get_1b()));
				for(temp = 0; temp < length; temp++)
					{
					put_1b((int)
						(lifnbuf[temp] = get_1b()));
					}
				lifnbuf[temp] = '\0';
				break;

			case T_RELSYM_DATA:	/* Sym. PC rel. ref. in data */
				if(dor_ref(check))
					min_size = FALSE;
				break;

			case T_REL_DATA:	/* !sym. PC rel. ref. in dta */
				put_1b((int) temp);
				copy(6);	/* Line #, address */
				min_loc[module] += word_size;
				max_loc[module] += word_size;
				break;

			case T_RESERVED:	/* Found error record */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_1b((int) (temp = get_1b()));
				copy(1);	/* Error code */
				min_loc[module] += temp;
				max_loc[module] += temp;
				break;

			case T_DEBUG_DATA:	/* Debug data record */
				put_1b((int) temp);
				copy(6);	/* Line # & value */
				put_2b((int) (length = get_2b()));
				copy(length);
				break;

			case T_DEBUGSYM_DATA:	/* Symbolic debug data rec */
				number = get_2b();	/* Line # */
				big_temp = get_4b();	/* Offset */
				length = get_2b();	/* Size of data */
				if(check_def(&temp2,length))
					break;	/* Error */
				else
					{
					put_1b((int) temp);
					put_2b((int) number);	/* Line # */
					put_4b((SLONG) big_temp); /* Offset */
					put_2b((int) length);	/* Data len */
					put_2b((int) temp2);	/* ST index */
					copy(length);	/* Data */
					if(check)
						st[temp2].type |= S_KEEP;
					}
				break;

			case T_EOF:	/* Found EOF record */
				put_1b((int) temp);
				copy(2);	/* Line # */
				return (min_size);

			default:
				fatal("Corrupted temp file: ",
				(read_temp == temp1_fp) ? temp1buf : temp2buf);
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	pass2						*/
/*									*/
/*	Makes a pass through the temp files increasing the minimum size	*/
/*	of all symbolic PC relative references which aren't external or	*/
/*	inter-module, and which aren't currently listed as being of	*/
/*	maximum instruction length.  This routine is called once during	*/
/*	the assembly process to resolve mutually dependent PC relative	*/
/*	instructions by increasing the size of ALL the boundary cases	*/
/*	(which means all the cases left when "pass2" is called since	*/
/*	the others were taken care of by iterating "pass1").  In the	*/
/*	best of all possible worlds this process would only increase	*/
/*	the size of one reference at a time and examine the effects,	*/
/*	however this could potentially cause the temp files to be	*/
/*	processed once for each unbound reference (or even worse if the	*/
/*	optimum reference to be grown is chosen).  In the interest of	*/
/*	reasonable execution speed all references are handled together	*/
/*	with the understanding that a minor amount of unnecessary code	*/
/*	might result (prefix instructions execute rapidly).		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
pass2()
	{
	SLONG			big_temp;
	register	UINT	temp;

#if	DEBUG
	putchar('2');			/* Indicate doing pass 2 */
#endif
/*
 *	Initialize for this pass.
 */
	for(temp = module = 0; temp < NUMMODULES; temp++)
		{
		max_loc_adj[temp] = min_loc_adj[temp] = min_loc[temp] =
			max_loc[temp] = 0;
		}
/*
 *	Loop through all the records in the temp file.  Do whatever is needed
 *	for each different record type.
 */
	FOREVER
		{
		switch(temp = get_1b())
			{
			case T_DATA:	/* Normal data definition */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_2b((int) (temp = get_2b()));
				copy(temp);	/* Actual data */
				min_loc[module] += temp;
				max_loc[module] += temp;
				break;

			case T_RELSYM_OP:	/* Symbolic PC rel. instr. */
				grow_ref();
				break;

			case T_DEF:	/* Label definition */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_2b((int) (temp = get_2b()));
				if((st[temp].min_value != min_loc[module]) ||
					(st[temp].max_value !=
						max_loc[module]))
					{
					min_loc_adj[module] =
						min_loc[module] - 
							st[temp].min_value;
					max_loc_adj[module] =
						st[temp].max_value -
							max_loc[module];
					st[temp].min_value = min_loc[module];
					st[temp].max_value = max_loc[module];
					}
				break;

			case T_ALIGN:	/* Word alignment pseudo-op */
				put_1b((int) temp);
				copy(2);	/* Line # */
				loc_pad();	/* Update location counters */
				break;

			case T_STORAGE:	/* Reserve space pseudo-op */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_4b((SLONG) (big_temp = get_4b()));
				min_loc[module] += big_temp;
				max_loc[module] += big_temp;
				break;

			case T_RELREL_OP:	/* "rel-rel" ref. instr. */
				grow_diff();
				break;

			case T_WRELREL_OP:	/* Scaled "rel-rel" instr. */
				grow_wdiff();
				break;

			case T_SET:	/* Constant equate operation */
				put_1b((int) temp);
				copy(8);	/* Ln #, sym #, value */
				break;

			case T_ADDR_DATA:	/* Address reference in data */
				put_1b((int) temp);
				copy(8);	/* Line #, ST entry, offset */
				min_loc[module] += word_size;
				max_loc[module] += word_size;
				break;

			case T_MODULE:	/* Module # definition */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_1b((int) (temp = get_1b()));
				module = temp;
				break;

			case T_ADDR_OP:	/* Address reference in instruction */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_1b((int) (temp = get_1b()));
				min_loc[module] += temp;
				copy(6);	/* ST #, offset */
				put_1b((int) (temp = get_1b())); /* Opcode */
				if(temp == DOTLDCR)
					max_loc[module] +=
						(max_i_size + LDPI_SIZE);
				else
					max_loc[module] += max_i_size;
				break;

			case T_RELREL_DATA:	/* "rel-rel" ref. in data */
				put_1b((int) temp);
				copy(10);	/* Line#, ST entries, offset */
				min_loc[module] += word_size;
				max_loc[module] += word_size;
				break;

			case T_REL_OP:	/* Non-symbolic PC relative instr. */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_1b((int) (temp = get_1b()));
				copy(5);	/* addr, opcode */
				min_loc[module] += temp;
				max_loc[module] += max_i_size;
				break;

			case T_FILENAME:	/* Filename record */
				put_1b((int) temp);
				copy(4);	/* Old & new line #'s */
				put_1b((int) (temp = get_1b()));
				copy(temp);	/* ASCII representation */
				break;

			case T_RELSYM_DATA:	/* Sym. PC rel. ref. in data */
				put_1b((int) temp);
				copy(8);	/* Line #, ST entry, offset */
				min_loc[module] += word_size;
				max_loc[module] += word_size;
				break;

			case T_REL_DATA:	/* !sym. PC rel. ref. in dta */
				put_1b((int) temp);
				copy(6);	/* Line #, address */
				min_loc[module] += word_size;
				max_loc[module] += word_size;
				break;

			case T_RESERVED:	/* Found error record */
				put_1b((int) temp);
				copy(2);	/* Line # */
				put_1b((int) (temp = get_1b()));
				copy(1);	/* Error code */
				min_loc[module] += temp;
				max_loc[module] += temp;
				break;

			case T_DEBUG_DATA:	/* Debug data record */
				put_1b((int) temp);
				copy(6);	/* Line # & value */
				put_2b((int) (temp = get_2b()));
				copy(temp);
				break;

			case T_DEBUGSYM_DATA:	/* Symbolic debug data rec */
				put_1b((int) temp);
				copy(6);	/* Line # & value */
				put_2b((int) (temp = get_2b()));
				copy(temp + 2);	/* Data & ST index */
				break;

			case T_EOF:	/* Found EOF record */
				put_1b((int) temp);
				copy(2);	/* Line # */
				return;

			default:
				fatal("Corrupted temp file: ",
				(read_temp == temp1_fp) ? temp1buf : temp2buf);
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	pass3						*/
/*									*/
/*	Reads the final temp file and produces the relocatable output	*/
/*	file.  This pass makes use of the symbol table information when	*/
/*	determining which symbols need to be written out as part of the	*/
/*	symbol section at the front of the relocatable file.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
pass3()
	{
	SLONG			big_temp;
	char			buf[SYMNMESIZE + 1];
	UINT			ext_def;
	UINT			ext_ref;
	UINT			ext_sym;
	UINT			index;
	UINT			loc_sym;
	register struct symtab	*ptr;
	register	STRING	sptr;
	UINT			st_usage;
	register	UINT	temp;
	UINT			tot_loc;

#if	DEBUG
	putchar('3');			/* Indicate doing pass 3 */
#endif
	acc_dtasize = 0;		/* Size of accumulated T_DATA */
	acc_storesize = 0;		/* Size of accumulated T_STORAGE */
/*
 *	Indicate this will be a relocatable file and indicate what type of
 *	target processor was specified.
 */
	out_1b((int) T_REL_FILE);	/* Indicate normal relocatable file */
	if(cpu_type & T212)
		out_1b((int) TYPE_212);	/* T212/T222/T225 target */
	else if(cpu_type & T414)
		out_1b((int) TYPE_414);	/* T414/T425 target */
	else if(cpu_type & T800)
		out_1b((int) TYPE_800);	/* T800/T801/T805 */
	else
		out_1b((int) TYPE_UNKNOWN);	/* Otherwise unknown */
/*
 *	Now we go through the symbol table and figure out which symbols need
 *	to be included in the relocatable output file as ASCII strings.  This
 *	consists of all symbols marked by ".ext" which are actually referenced,
 *	and all symbols marked ".pub" which are actually defined.  Each of the
 *	symbols is written to the output file as a "T_SYMBOL" record and the
 *	relative order within the "T_SYMBOL" records is remembered so that the
 *	actually references to the symbols can be converted to references to a
 *	symbol # instead of a name (since we aren't sending the entire symbol
 *	table to TLNK).
 *		As mentioned elsewhere, some local symbol references must also
 *	be resolved by TLNK.  For these, the symbol table reference is also
 *	replaced with symbol #'s, starting just past the end of the numbers
 *	assigned to symbols output as "T_SYMBOL" records.  Note that since we
 *	don't know, as we go through the symbol table, how many external
 *	symbols we will have, we don't know where to start the numbering for
 *	local symbols.  This is resolved by starting the local symbol
 *	numbering at a point beyond where the external symbols could extend
 *	and computing a correction factor for local symbol #'s after we finish
 *	processing the entire symbol table.
 */
	ext_sym = 0;			/* External symbol # counter */
	loc_sym = SYMTABSIZE;		/* Local symbol # counter */

	ext_def = 0;			/* # of external definitions */
	ext_ref = 0;			/* # of external references */
	st_usage = 0;			/* # of symbol table slots used */
	tot_loc = 0;			/* Total # of local symbols */

	for(ptr = st; ptr < &st[SYMTABSIZE]; ptr++)
		{
		if(ptr->id.name == NULL)
			continue;	/* If slot unused */
		st_usage++;		/* This slot used */
		if(! (ptr->type & (S_PUBLIC | S_EXTERN)))
			{
			tot_loc++;	/* Was a local symbol */
			if(ptr->type & S_KEEP)
				ptr->id.number = loc_sym++;
			continue;
			}
		if((ptr->type & S_PUBLIC) && (! (ptr->type & S_DEFINED)))
			continue;
		if((ptr->type & S_EXTERN) && (! (ptr->type & S_KEEP)))
			continue;
		out_1b((int) T_SYMBOL);	/* Otherwise put in rel file */
		if(ptr->type & S_PUBLIC)
			{
			out_1b((int) TRUE);
			ext_def++;	/* Was a definition */
			}
		else
			{
			out_1b((int) FALSE);
			ext_ref++;	/* Was a reference */
			}
		string_copy(buf,ptr->id.name,SYMNMESIZE);
		buf[SYMNMESIZE] = '\0';
		out_1b((int) strlen(buf));
		for(sptr = buf; *sptr; )
			out_1b((int) *sptr++);
		ptr->id.number = ext_sym++;
		}
/*
 *	Print out symbol table usage statistics (if desired).
 */
	if(list_fp != NULL)
		{
		fprintf(list_fp,"\nExternal symbols defined: %u",ext_def);
		fprintf(list_fp,"\tExternal symbols referenced: %u\n",ext_ref);
		fprintf(list_fp,"Local symbols defined:    %u",tot_loc);
		fprintf(list_fp,"\tLocal symbols exported:      %u\n",
			loc_sym - SYMTABSIZE);
		fprintf(list_fp,"Symbol table usage: %u%%\n",
			(UINT) (((SLONG) st_usage * (SLONG) 100) / (SLONG)
				SYMTABSIZE));
		}
	if(verbose)
		{
		printf("\nExternal symbols defined: %u",ext_def);
		printf("\tExternal symbols referenced: %u\n",ext_ref);
		printf("Local symbols defined:    %u",tot_loc);
		printf("\tLocal symbols exported:      %u\n",
			loc_sym - SYMTABSIZE);
		printf("Symbol table usage: %u%%\n",
			(st_usage * 100) / SYMTABSIZE);
		}
	
	loc_sym = SYMTABSIZE - ext_sym;	/* Local symbol # correction factor */
/*
 *	Loop through all the records in the temp file.  Figure out whether to
 *	just pass the record through, or if some conversion needs to be done to
 *	make the record compatible with the relocatable file format.
 */
	FOREVER
		{
		switch(temp = get_1b())
			{
			case T_DATA:	/* Normal data definition */
				number = get_2b();	/* Line # */
				temp = get_2b();	/* Data size */
				if(compress)
					{
					if(acc_storesize)
						out_rec(T_RESERVED);
					while(temp > 0)
						{
						if(acc_dtasize >=
							ACCDATASIZE)
							{
							out_rec(T_RESERVED);
							}
						acc_dta[acc_dtasize++] =
							get_1b();
						temp--;
						}
					}
				else
					{
					out_1b((int) T_DATA);
					out_2b((int) number);	/* Line # */
					out_2b((int) temp);
					copy_out(temp);	/* Actual data */
					}
				break;

			case T_RELSYM_OP:	/* Symbolic PC rel. instr. */
				out_rec(temp);
				copy_out(3);	/* Line # & min. length */
				temp = st[get_2b()].id.number;
				if(temp >= SYMTABSIZE)
					temp -= loc_sym;
				out_2b((int) temp);
				copy_out(5);	/* Offset and opcode */
				break;

			case T_DEF:	/* Label definition */
				temp = get_2b();	/* Line # */
				index = get_2b();	/* ST entry # */
				if(st[index].type & (S_KEEP | S_PUBLIC))
					{
					out_rec(T_DEF);
					out_2b((int) temp);
					temp = st[index].id.number;
					if(temp >= SYMTABSIZE)
						temp -= loc_sym;
					out_2b((int) temp);
					}
				break;

			case T_ALIGN:	/* Word alignment pseudo-op */
				out_rec(temp);
				copy_out(2);	/* Line # */
				break;

			case T_STORAGE:	/* Reserve space pseudo-op */
				temp = get_2b();	/* Line # */
				big_temp = get_4b();	/* Storage size */
				if(compress)
					{
					if(acc_dtasize > 0)
						out_rec(T_RESERVED);
					acc_storesize += big_temp;
					}
				else
					{
					out_1b((int) T_STORAGE);
					out_2b((int) temp);	/* Line # */
					out_4b((SLONG) big_temp);
					}
				break;

			case T_RELREL_OP:	/* "rel-rel" ref. instr. */
			case T_WRELREL_OP:	/* Scaled "rel-rel" instr. */
				out_rec(temp);
				copy_out(3);	/* Line # & min. length */
				temp = st[get_2b()].id.number;
				if(temp >= SYMTABSIZE)
					temp -= loc_sym;
				out_2b((int) temp);	/* Left symbol */
				temp = st[get_2b()].id.number;
				if(temp >= SYMTABSIZE)
					temp -= loc_sym;
				out_2b((int) temp);	/* Right symbol */
				copy_out(5);	/* Offset and opcode */
				break;

			case T_SET:	/* Constant equate operation */
				temp = get_2b();	/* Line # */
				index = get_2b();	/* ST entry # */
				if(st[index].type & (S_KEEP | S_PUBLIC))
					{
					out_rec(T_SET);
					out_2b((int) temp);
					temp = st[index].id.number;
					if(temp >= SYMTABSIZE)
						temp -= loc_sym;
					out_2b((int) temp);
					copy_out(4);	/* Value */
					}
				else
					get_4b();	/* Toss value */
				break;

			case T_ADDR_DATA:	/* Address reference in data */
				out_rec(temp);
				copy_out(2);	/* Line # */
				temp = st[get_2b()].id.number;
				if(temp >= SYMTABSIZE)
					temp -= loc_sym;
				out_2b((int) temp);
				copy_out(4);	/* Offset */
				break;

			case T_MODULE:	/* Module # definition */
				out_rec(temp);
				copy_out(3);	/* Line #, module# */
				break;

			case T_ADDR_OP:	/* Address reference in instruction */
				out_rec(temp);
				copy_out(3);	/* Line #, minimum length */
				temp = st[get_2b()].id.number;
				if(temp >= SYMTABSIZE)
					temp -= loc_sym;
				out_2b((int) temp);
				copy_out(5);	/* Offset and opcode */
				break;

			case T_RELREL_DATA:	/* "rel-rel" ref. in data */
				out_rec(temp);
				copy_out(2);	/* Line # */
				temp = st[get_2b()].id.number;
				if(temp >= SYMTABSIZE)
					temp -= loc_sym;
				out_2b((int) temp);	/* Left symbol */
				temp = st[get_2b()].id.number;
				if(temp >= SYMTABSIZE)
					temp -= loc_sym;
				out_2b((int) temp);	/* Right symbol */
				copy_out(4);	/* Offset */
				break;

			case T_REL_OP:	/* Non-symbolic PC reference instr. */
				out_rec(temp);
				copy_out(8);	/* Ln #, size, addr, opcode */
				break;

			case T_FILENAME:	/* Filename record */
				if(compress)
					{
					(VOID) get_4b();	/* Toss #'s */
					temp = get_1b();	/* Length */
					while(temp-- > 0)
						(VOID) get_1b();
					}
				else
					{
					out_1b((int) temp);
					copy_out(4);	/* Old/new line #'s */
					out_1b((int) (temp = get_1b()));
					copy_out(temp);	/* ASCII name */
					}
				break;

			case T_RELSYM_DATA:	/* Sym. PC rel. ref. in data */
				out_rec(temp);
				copy_out(2);	/* Line # */
				temp = st[get_2b()].id.number;
				if(temp >= SYMTABSIZE)
					temp -= loc_sym;
				out_2b((int) temp);
				copy_out(4);	/* Offset */
				break;

			case T_REL_DATA:	/* !sym. PC rel. ref. in dta */
				out_rec(temp);
				copy_out(6);	/* Ln #, addr */
				break;

			case T_DEBUG_DATA:	/* Debug data record */
				out_rec(temp);
				copy_out(6);	/* Line # & value */
				out_2b((int) (temp = get_2b()));
				copy_out(temp);
				break;

			case T_DEBUGSYM_DATA:	/* Symbolic debug data rec */
				out_rec(temp);
				copy_out(6);	/* Line # & value */
				out_2b((int) (temp = get_2b()));
				index = st[get_2b()].id.number;
				if(index >= SYMTABSIZE)
					index -= loc_sym;
				out_2b((int) index);
				copy_out(temp);	/* Data */
				break;

			case T_EOF:	/* Found EOF record */
				out_rec(temp);
				copy_out(2);	/* Line # */
				return;

			default:
				fatal("Corrupted temp file: ",
				(read_temp == temp1_fp) ? temp1buf : temp2buf);
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	updat_length					*/
/*									*/
/*	Takes two possible immediate values for instructions which have	*/
/*	immediate data, computes the length of an instruction resulting	*/
/*	from each and updates the "min_loc" and "max_loc" location	*/
/*	counters appropriately.  The worst case of the two instruction	*/
/*	lengths is returned (which will be 1 greater than the best	*/
/*	case).  This routine is called from the various	routines which	*/
/*	"pass2" calls to "grow" the minimum instruction	lengths where	*/
/*	required.  Note that a running total of the amount we have	*/
/*	adjusted the "min_loc[module]" location	counter is kept to	*/
/*	ensure that the definitions of forward referenced symbols are	*/
/*	adjusted (via "adj_min_loc"), enough to prevent the		*/
/*	"min_loc-min_loc" term of the definition/reference difference	*/
/*	from changing sign.						*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The basic opcode.			*/
/*		"data1"	-	A SLONG which is used to compute the	*/
/*				first possible instruction length.	*/
/*		"data2"	-	A SLONG which is used to compute the	*/
/*				second possible instruction length.	*/
/*		"min_length" -	The minimum instruction length to	*/
/*				generate.				*/
/*		"rel" -		TRUE if the immediate data is relative	*/
/*				to the PC of the instruction being	*/
/*				generated (i.e. does the instruction	*/
/*				length effect the data).		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

UINT
updat_length(function,data1,data2,min_length,rel)
	UINT			function;
	SLONG			data1;
	SLONG			data2;
	UINT			min_length;
	BOOL			rel;
	{
	UINT			length1;
	UINT			length2;

	if(rel)
		{
		length1 = bind_constant((UINT) function,(SLONG) data1);
		length2 = bind_constant((UINT) function,(SLONG) data2);
/*
 *	See if possibly either a forward or backward reference.
 */
		if(((data1 - (SLONG) length1) ^ (data2 - (SLONG) length2)) < 0)
			{
			if((length1 == length2) && (min_length < length1))
				min_loc_adj[module]++;
			}
		}
	else
		{
		length1 = i_size((UINT) function,(SLONG) data1);
		length2 = i_size((UINT) function,(SLONG) data2);
/*
 *	See if instruction immediate data could be either positive or negative.
 */
		if((data1 ^ data2) < 0)
			{
			if((length1 == length2) && (min_length < length1))
				min_loc_adj[module]++;
			}
		}
	if(length1 > length2)
		{
		if(length1 > min_length)
			{
			min_length = length1;
			min_loc_adj[module]++;	/* Adj for fwd definitions */
			}
		min_loc[module] += min_length;
		max_loc[module] += min_length;
		}
	else
		{
		if(length2 > min_length)
			{
			min_length = length2;
			if(length1 < length2)
				min_loc_adj[module]++;
			}
		min_loc[module] += min_length;
		max_loc[module] += min_length;
		}
	return (min_length);
	}

/************************************************************************/
/*									*/
/*	Routine:	updat_loc					*/
/*									*/
/*	Takes two possible immediate values for instructions which have	*/
/*	immediate data, computes the length of an instruction resulting	*/
/*	from each and update the "min_loc" and "max_loc" location	*/
/*	counters appropriately.						*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The opcode of the instruction.		*/
/*		"data1"	-	A SLONG which is used to compute the	*/
/*				first possible instruction length.	*/
/*		"data2"	-	A SLONG which is used to compute the	*/
/*				second possible instruction length.	*/
/*		"min_length"	The minimum instruction length to	*/
/*				generate.				*/
/*		"rel" -		TRUE if the immediate data is relative	*/
/*				to the PC of the instruction being	*/
/*				generated (i.e. does the instruction	*/
/*				length effect the data).		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
updat_loc(function,data1,data2,min_length,rel)
	UINT			function;
	SLONG			data1;
	SLONG			data2;
	UINT			min_length;
	BOOL			rel;
	{
	UINT			length1;
	UINT			length2;

	if(rel)
		{
		length1 = bind_constant((UINT) function,(SLONG) data1);
		length2 = bind_constant((UINT) function,(SLONG) data2);
/*
 *	See if possibly either a forward or backward reference.
 */
		if(((data1 - (SLONG) length1) ^ (data2 - (SLONG) length2)) < 0)
			{
			min_loc[module] += min_length;
			if(length1 > length2)
				{
				if(length1 < min_length)
					max_loc[module] += min_length;
				else
					max_loc[module] += length1;
				}
			else
				{
				if(length2 < min_length)
					max_loc[module] += min_length;
				else
					max_loc[module] += length2;
				}
			return;
			}
		}
	else
		{
		length1 = i_size((UINT) function,(SLONG) data1);
		length2 = i_size((UINT) function,(SLONG) data2);
/*
 *	See if instruction immediate data could be either positive or negative.
 */
		if((data1 ^ data2) < 0)
			{
			min_loc[module] += min_length;
			if(length1 > length2)
				{
				if(length1 < min_length)
					max_loc[module] += min_length;
				else
					max_loc[module] += length1;
				}
			else
				{
				if(length2 < min_length)
					max_loc[module] += min_length;
				else
					max_loc[module] += length2;
				}
			return;
			}
		}
/*
 *	Otherwise we know the direction of the reference, or the sign of the
 *	immediate data.
 */
	if(length1 > length2)
		{
		if(length2 < min_length)
			min_loc[module] += min_length;
		else
			min_loc[module] += length2;
		if(length1 < min_length)
			max_loc[module] += min_length;
		else
			max_loc[module] += length1;
		}
	else
		{
		if(length1 < min_length)
			min_loc[module] += min_length;
		else
			min_loc[module] += length1;
		if(length2 < min_length)
			max_loc[module] += min_length;
		else
			max_loc[module] += length2;
		}
	}
