/************************************************************************/
/*									*/
/*	File:	tasm5.c							*/
/*									*/
/*	Transputer Assembler (TASM), source file 5.  This file contains	*/
/*	the code which generates a listing file.			*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<string.h>		/* String include file */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tasmdef.h"		/* Assembler definitions */
#include	"tasmtyp.h"		/* Assembler data declarations */
#include	"tasmext.h"		/* Assembler externals */

/************************************************************************/
/*									*/
/*	Routine:	list_data					*/
/*									*/
/*	Read the specified # of data bytes from the input temp file.	*/
/*	Store the first eight bytes or less in the array which gets	*/
/*	passed to this routine, updating the first element to reflect	*/
/*	the actual # of bytes stored (assuming the array is empty,	*/
/*	otherwise leave the existing contents alone).			*/
/*									*/
/*	Parameters:							*/
/*		"data"		A pointer to a UINT array to hold	*/
/*				the data read.  The first element of	*/
/*				the array holds the # of data elements	*/
/*				currently in use.			*/
/*		"new_data"	This holds the number of new bytes of	*/
/*				data to read.  If "data[0] == 0" this	*/
/*				new data will update the previous	*/
/*				contents of "data".			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
list_data(data,new_data)
	UINT			data[];
	UINT			new_data;
	{
	register	UINT	i;

	if(data[0])
		{
		for(i = 0; i < new_data; i++)
			get_1b();
		}
	else
		{
		for(i = 1; i < 9; i++)
			{
			if(i > new_data)
				{
				data[0] = new_data;
				return;
				}
			data[i] = get_1b();
			}
		for(; i <= new_data; i++)
			get_1b();	/* Slurp rest */
		data[0] = 8;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	list_hex					*/
/*									*/
/*	Display a two digit hex # on the listing file.			*/
/*									*/
/*	Parameters:							*/
/*		"value"	The number to display.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
list_hex(value)
	UINT			value;
	{
	char			buf[3];

#if	H_OPSYS == BSD4_2
	sprintf(buf,"%02X",value);
#else	/* BSD4_2 */
	sprintf(buf,"%.2X",value);
#endif	/* BSD4_2 */
	list_write(buf);
	}

/************************************************************************/
/*									*/
/*	Routine:	list_lnum					*/
/*									*/
/*	Write the current input source file line # on the listing file.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
list_lnum()
	{
	char			buf[MAXLNSIZE + 1];

#if	H_OPSYS == BSD4_2
	sprintf(buf,"%04u ",number);
#else	/* BSD4_2 */
	sprintf(buf,"%5.4u ",number);
#endif	/* BSD4_2 */
	list_write(buf);
	}

/************************************************************************/
/*									*/
/*	Routine:	list_pass					*/
/*									*/
/*	Read the the "final" internal temp file and generate a listing	*/
/*	file.								*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/*	NOTE:	This routine is probably the most complicated in TASM.	*/
/*		Be very careful that you understand what you are doing	*/
/*		if you make changes and be sure to do lots of testing	*/
/*		afterwards!						*/
/*									*/
/************************************************************************/

VOID
list_pass()
    {
    char			buf[MAXLNSIZE + 1];
    UINT			data[9];
    UINT			error;
    BOOL			had_type;
    BOOL			holding;
    register	UINT		length;
    SLONG			location;
    UINT			new_number;
    register	UINT		temp;
    UINT			type;

#if DEBUG
    putchar('L');			/* Indicate doing listing pass */
#endif
/*
 *	Initialize for this pass.
 */
    for(temp = module = 0; temp < NUMMODULES; temp++)
	min_loc[temp] = 0;
	
    data[0] = 0;			/* No data to start with */
    error = ' ';			/* No errors to start with */
    location = 0;			/* Loc counter for start of src line */
    number = 1;				/* Initialize line # */
    had_type = FALSE;			/* TRUE if "type" field already read */
    holding = FALSE;			/* TRUE if we've a line tucked away */
/*
 *	Loop through all the records in the temp file.  Output the line #,
 *	location counter (if useful), and any generated code.  The actual
 *	source code is appended to the right side of the line.  Note that the
 *	records in the temp file don't reflect all the original source code
 *	lines, just those that caused TASM to do something.  The following
 *	code interpolates between the mentioned source lines to include the
 *	comment and blank lines in the listing also.
 */
    FOREVER
	{
	if(had_type)
		had_type = FALSE;
	else
		type = get_1b();	/* Read type field if needed */
/*
 *	The easiest type of record to handle is the T_MODULE one which
 *	indicates that the module # has changed.  From the point of the view
 *	of the listing this requires no work other than recording the new
 *	module #.
 */
	if(type == T_MODULE)
		{
		get_2b();		/* Toss line # */
		module = get_1b();	/* New module # */
		continue;		/* Do next record */
		}

	new_number = get_2b();		/* Get new line # */
/*
 *	Now we get to something with some difficulty, the T_FILENAME record.
 *	The occurence of this means that there was a "#line" directive in the
 *	input file which contained an explicit filename definition (and thus
 *	indicates the listing source will have to be read from a new file).
 *	The basic scheme represented by the following code is:
 *
 *	1.	If multiple T_FILENAME records appear in a row, DO NOT open or
 *		read from any except the last in a sequence.  This commonly
 *		occurs when nested include files are unwound by the
 *		preprocessor.
 *	2.	If case #1 happens, but the trailing line # field in one record
 *		is not JUST beyond the leading line # field in the previous
 *		record, assume that some comment text and/or blank lines were
 *		on the file listed in the previous record and list them out.
 *	3.	If any text is being "saved" prior to the first T_FILENAME
 *		record being encountered, list it out as well as any trailing
 *		comment/blank lines leading up to where the trailing line #
 *		field in the record indicates the "#line" was encountered.
 *	4.	At the end of any sequence of T_FILENAME records with no
 *		intervening text displayed, spit out a page-eject/title-bar to
 *		let everyone know which source file is active.  A title-bar is
 *		forced out at the beginning of operation by the setting of the
 *		initial values for the line number variables.
 */
	if(type == T_FILENAME)
		{
		FOREVER
			{
			if(holding)
/* Display "held" line */	{
				holding = FALSE;
				list_put((int) error);
				list_lnum();	/* Show line # */
#if	H_OPSYS == BSD4_2
				sprintf(buf,"%08lX ",location);
#else	/* BSD4_2 */
				sprintf(buf,"%.8lX ",location);
#endif	/* BSD4_2 */
				list_write(buf);
/* Show data (if any) */	if(data[0])
					{
					for(temp = 1; temp < 9; temp++)
						{
						if(temp <= data[0])
							{
							list_hex(data[temp]);
							if((temp < (9 - 1)) &&
							(temp < data[0]))
								list_put(':');
							else
								list_put(' ');
							}
						else
							list_write("   ");
						}
					}
				else
					list_write("                        ");
/* Show source code */		list_src();
/* Reset "held" info */		error = ' ';
				location = min_loc[module];
				data[0] = 0;
/* Show trailing lines */	list_range(new_number + 1);
				}
			else
				{
/*
 *	This entire "else" clause could be replaced by moving the "list_range"
 *	function call in the above "if" clause to here and eliminating the
 *	"else" altogether.  This hasn't been done because the following code
 *	ensures that the location counter is spit out for the first line in a
 *	source file, EVEN if the line contained nothing noteworthy.
 */
				if(number < new_number + 1)
					{
					list_put((int) error);
					list_lnum();	/* Show line # */
#if	H_OPSYS == BSD4_2
					sprintf(buf,"%08lX ",location);
#else	/* BSD4_2 */
					sprintf(buf,"%.8lX ",location);
#endif	/* BSD4_2 */
					list_write(buf);
					list_write("                        ");
					list_src();
					}
/* Show trailing lines */	list_range(new_number + 1);
				}
			FOREVER
				{
/*
 *	This code handles the "lookahead" operation for T_FILENAME records.
 */
/* "new" line # */		number = get_2b();	/* Set ln # */
/* Filename length */		length = get_1b();
				for(temp = 0; temp < length; temp++)
/* Update filename */			lifnbuf[temp] = get_1b();
				lifnbuf[temp] = '\0';
/* Next record T_FILENAME? */	if((type = get_1b()) != T_FILENAME)
					{
					had_type = TRUE;
					break;
					}
				if((new_number = get_2b()) != (number - 1))
					break;
				}
/*
 *	Spit out any required title bar and terminate the T_FILENAME processing
 *	if a non-T_FILENAME record has been encountered.
 */
			list_title();
			if(had_type)
				break;
			}
		continue;
		}
/*
 *	This code fragment handles the display of "held" lines for cases other
 *	than the flushing performed above for "#line" records.  The basic idea
 *	is to display the held line as soon as a record is encountered with a
 *	different (larger) line #.  This allows multi-line macros etc. to work
 *	cleanly.  Once the held line is displayed any following comment/blank
 *	lines are written to bring the listing to just before the "new" held
 *	line.
 */
	if(new_number > number)
		{
		list_put((int) error);
		list_lnum();		/* Show line # */
#if	H_OPSYS == BSD4_2
		sprintf(buf,"%08lX ",location);
#else	/* BSD4_2 */
		sprintf(buf,"%.8lX ",location);
#endif	/* BSD4_2 */
		list_write(buf);
		if(data[0])
			{
			for(temp = 1; temp < 9; temp++)
				{
				if(temp <= data[0])
					{
					list_hex(data[temp]);
					if((temp < (9 - 1)) &&
						(temp < data[0]))
							list_put(':');
					else
						list_put(' ');
					}
				else
					list_write("   ");
				}
			}
		else
			list_write("                        ");
		list_src();
		list_range(new_number);	/* Display following comment lines */
		error = ' ';		/* Reset "held" information */
		location = min_loc[module];
		data[0] = 0;
		}

	if(type == T_EOF)
		break;

	holding = TRUE;			/* Indicate we will have "held" data */
/*
 *	The following code fragment handles all the normal TASM temp file
 *	records.  The basic idea is to accumulate the information to display
 *	on the listing (when the line number changes).  For multiple records
 *	on the same source file line (such as a macro might generate), the
 *	following rules hold:
 *
 *	1.	The first record with an error in it is used to determine the
 *		beginning of line error flag on the listing.  If no error is
 *		is present then any record with a '.' error flag indicating
 *		an unresolved reference is displayed.  If there are no
 *		unresolved references display a space to indicate everything is
 *		fine.
 *	2.	The first record which contains code/data to show in the "data"
 *		field of the listing is used (the assembled code section).
 *	3.	The location counter value at the start of the first record is
 *		displayed.
 */
	switch(type)
		{
		case T_DATA:		/* Normal data definition */
			length = get_2b();
			list_data(data,length);
			min_loc[module] += length;
			break;

		case T_RELSYM_OP:	/* Symbolic PC relative instruction */
			length = get_1b();	/* Minimum length */
			get_2b();	/* Toss ST entry # */
			if(error == ' ')
				error = '.';
			list_data(data,4);
			get_1b();	/* Toss opcode coding */
			min_loc[module] += length;
			break;

		case T_DEF:		/* Label definition */
			get_2b();	/* Toss ST entry # */
			break;

		case T_ALIGN:		/* Word alignment pseudo-op */
/*
 *	***##@! MSC5.1 OPTIMIZER BUGS!!!!  I sure get tired of finding these...
 */
#ifdef	DAMN_MSC
			length = (word_size - (min_loc[module] &
					(word_size - 1))) &
						(word_size - 1);
#else	/* !DAMN_MSC */
			length = min_loc[module] & (SLONG) (word_size - 1);
			length = (word_size - length) & (word_size - 1);
#endif	/* DAMN_MSC */
			min_loc[module] += length;
			break;

		case T_STORAGE:		/* Reserve space pseudo-op */
			min_loc[module] += get_4b();
			break;

		case T_RELREL_OP:	/* "rel-rel" reference instruction */
			length = get_1b();	/* Minimum length */
			get_4b();	/* Toss left and right ST entry #'s */
			if(error == ' ')
				error = '.';
			list_data(data,4);
			get_1b();	/* Toss opcode coding */
			min_loc[module] += length;
			break;

		case T_WRELREL_OP:	/* Scaled "rel-rel" ref. instruction */
			length = get_1b();	/* Minimum length */
			get_4b();	/* Toss left and right ST entry #'s */
			if(error == ' ')
				error = '.';
			list_data(data,4);
			get_1b();	/* Toss opcode coding */
			min_loc[module] += length;
			break;

		case T_SET:		/* Constant equate operation */
			get_2b();	/* Toss ST entry # */
			list_data(data,4);	/* Show value */
			break;

		case T_ADDR_DATA:	/* Address reference in data */
			get_2b();	/* Toss ST entry # */
			if(error == ' ')
				error = '.';
			list_data(data,4);
			min_loc[module] += word_size;
			break;

		case T_ADDR_OP:		/* Address reference in instruction */
			length = get_1b();	/* Minimum length */
			get_2b();	/* Toss ST entry # */
			if(error == ' ')
				error = '.';
			list_data(data,4);
			get_1b();	/* Toss opcode coding */
			min_loc[module] += length;
			break;

		case T_RELREL_DATA:	/* "rel-rel" reference in data */
			get_4b();	/* Toss left & right ST entry #'s */
			if(error == ' ')
				error = '.';
			list_data(data,4);
			min_loc[module] += word_size;
			break;

		case T_REL_OP:		/* Non-symbolic PC relative instr. */
			length = get_1b();	/* Minimum length */
			if(error == ' ')
				error = '.';
			list_data(data,4);
			get_1b();	/* Toss opcode coding */
			min_loc[module] += length;
			break;

		case T_RELSYM_DATA:	/* Symbolic PC relative ref. in data */
			get_2b();	/* Toss ST entry # */
			if(error == ' ')
				error = '.';
			list_data(data,4);
			min_loc[module] += word_size;
			break;

		case T_REL_DATA:	/* Non-symbolic PC rel. ref. in data */
			if(error == ' ')
				error = '.';
			list_data(data,4);
			min_loc[module] += word_size;
			break;

		case T_RESERVED:	/* Found error record */
			length = get_1b();	/* Assumed # bytes */
			if((error == ' ') || (error == '.'))
				error = get_1b();
			else
				get_1b();
			min_loc[module] += length;
			break;

		case T_DEBUG_DATA:	/* Debug data record */
			list_data(data,4);
			length = get_2b();
			while(length--)
				get_1b();	/* Toss data */
			break;

		case T_DEBUGSYM_DATA:	/* Symbolic debug data rec */
			if(error == ' ')
				error = '.';
			list_data(data,4);
			length = get_2b();
			get_2b();	/* Toss ST entry # */
			while(length--)
				get_1b();	/* Toss data */
			break;

		default:
			fatal("Corrupted temp file: ",
				(read_temp == temp1_fp) ? temp1buf : temp2buf);
		}
	}
    list_range(new_number);		/* Finish off EOF tail end */
    }

/************************************************************************/
/*									*/
/*	Routine:	list_put					*/
/*									*/
/*	Write a byte to the listing file.				*/
/*									*/
/*	Parameters:							*/
/*		"value"	The byte to write.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
list_put(value)
	int			value;
	{
	if(putc(value,list_fp) == EOF)
		{
		if(ferror(list_fp))
			fatal("Error writing listing file: ",listbuf);
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	list_range					*/
/*									*/
/*	Copy all the source file lines from the point where we last	*/
/*	displayed one (number-1), up to the newly interesting line #.	*/
/*									*/
/*	Parameters:							*/
/*		"line"	The newly interesting line #.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
list_range(line)
	UINT			line;
	{
	while(number < line)		/* Copy up to just before needed one */
		{
		list_put(' ');		/* No errors on these lines */
		list_lnum();		/* Show line # */
		list_write("                                 ");
		list_src();		/* Show source text */
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	list_src					*/
/*									*/
/*	Write the next source text line to the listing file.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
list_src()
	{
	char			buf[MAXLNSIZE + 1];

	if(fgets(buf,(MAXLNSIZE + 1),in_fp) == NULL)
		{
		if(ferror(in_fp))
			fatal("Error reading input file: ",ifnbuf);
		if(feof(in_fp))
			fatal("Unexpected EOF in input file: ",ifnbuf);
		}
	list_write(buf);
	number++;			/* Indicate line has been displayed */
	}

/************************************************************************/
/*									*/
/*	Routine:	list_title					*/
/*									*/
/*	Close the existing input file, open the new one and skip	*/
/*	forward the appropriate # of spots.  Generate a formfeed if	*/
/*	this is not the first page and display a title bar on the	*/
/*	listing which shows the newly active source filename.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
list_title()
	{
	register	UINT	temp;
	char			buf[MAXLNSIZE + 1];
/*
 *	Close old source file and open new one.
 */
	if(fclose(in_fp) == EOF)
		fatal("Unable to close input file: ",ifnbuf);
	strcpy(ifnbuf,lifnbuf);		/* Update with new filename */

#if	CRLF
	if((in_fp = (STREAM) fopen(ifnbuf,"rt")) == NULL)
#else
	if((in_fp = (STREAM) fopen(ifnbuf,"r")) == NULL)
#endif
		{
		fatal("Unable to open input file: ",ifnbuf);
		}
/*
 *	Skip forward in the source file to the #line specified spot.
 */
	for(temp = 1; temp < number; temp++)
		{
		if(fgets(buf,(MAXLNSIZE + 1),in_fp) == NULL)
			{
			if(ferror(in_fp))
				fatal("Error reading input file: ",ifnbuf);
			if(feof(in_fp))
				fatal("Unexpected EOF in input file: ",ifnbuf);
			}
		}
/*
 *	Generate eject if needed and display title bar.
 */
	if(eject)
		list_put('\f');
	else
		eject = TRUE;
	list_write(ASM_NAME);
	list_write(", ");
	list_write(VERSION);
	list_write("\t\tSource filename: ");
	list_write(ifnbuf);
	list_write("\n\n");
	}

/************************************************************************/
/*									*/
/*	Routine:	list_write					*/
/*									*/
/*	Write a string to the listing file.				*/
/*									*/
/*	Parameters:							*/
/*		"s"	A pointer to the string to write.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
list_write(s)
	register	STRING	s;
	{
	while(*s)
		list_put(*s++);
	}
