/************************************************************************/
/*									*/
/*	File:	tlnk2.c							*/
/*									*/
/*	Transputer Linker (TLNK), source file 2.  This file contains	*/
/*	the code which parses the operator input and gets all the	*/
/*	information needed for the actual assemble/link process.  This	*/
/*	also includes reading the input files.				*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<ctype.h>		/* Character type functions */
#include	<string.h>		/* String functions */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tlnkdef.h"		/* Linker definitions */
#include	"tlnktyp.h"		/* Linker data declarations */
#include	"tlnkext.h"		/* Linker externals */

/************************************************************************/
/*									*/
/*	Routine:	add_def_lst					*/
/*									*/
/*	Convert the incoming local/global symbol definition offset #	*/
/*	(for the current input file), into the appropriate offset for	*/
/*	all the input files (ie. the local symbol table entry #).  Also	*/
/*	update the local and global symbol tables to reflect the fact	*/
/*	that this is the "definition" for the symbol, not merely a	*/
/*	"reference" of it.						*/
/*									*/
/*	Parameters:							*/
/*		"index"	Either a legal local symbol table index, or	*/
/*			MAX_SYMBOLS if the index should be read from	*/
/*			the current input/library file.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
add_def_lst(index)
	UINT			index;
	{
	char			buf[SYMNMESIZE + 1];
	register	UINT	gindex;
	register	UINT	lindex;

	if(index < MAX_SYMBOLS)
		lindex = index;
	else
		{
		if((lindex = in_2b() + base_symbol) >= MAX_SYMBOLS)
			fatal("Too many total symbols","");
		}
	if(lindex >= next_symbol)
		next_symbol = lindex + 1;
	if((gindex = locals[lindex].gst_num) < MAX_GLOBALS)	/* "PUBLIC"? */
		{			/* Yes */
		if(globals[gindex].lst_num < MAX_SYMBOLS)
			{		/* It is already defined! */
			string_copy(buf,globals[gindex].name,SYMNMESIZE);
			buf[SYMNMESIZE] = '\0';
			non_fatal("Duplicate external definition: ",buf);
			return;
			}
/*
 *	Otherwise indicate the "PUBLIC" symbol is defined.
 */
		else
			globals[gindex].lst_num = lindex;
		}
	put_rec(T_DEF);			/* T_DEF header and line # */
	put_2b((int) lindex);		/* Local symbol table entry # */
	}

/************************************************************************/
/*									*/
/*	Routine:	add_lst						*/
/*									*/
/*	Convert the incoming local symbol # offset (relative to the	*/
/*	current input file), into the appropriate offset for all the	*/
/*	input files (ie. the local symbol table entry #).		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
add_lst()
	{
	UINT			lindex;

	if((lindex = in_2b() + base_symbol) >= MAX_SYMBOLS)
		fatal("Too many total symbols","");
	if(lindex >= next_symbol)
		next_symbol = lindex + 1;
	put_2b((int) lindex);
	}

/************************************************************************/
/*									*/
/*	Routine:	add_set_lst					*/
/*									*/
/*	Convert the incoming local/global equate symbol definition	*/
/*	offset # (for the current input file), into the appropriate	*/
/*	offset for all the input files (ie. the local symbol table	*/
/*	entry #).  Also update the local and global symbol tables to	*/
/*	reflect the fact that this symbol has a constant "definition"	*/
/*	which will not need to be modified by later passes of TLNK.	*/
/*	This is accomplished by removing the T_SET record from the	*/
/*	input stream and setting the "min_addr" and "max_addr" fields	*/
/*	in the local symbol table entry (normally these fields aren't	*/
/*	updated until "pass0" starts the binding process).		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
add_set_lst()
	{
	char			buf[SYMNMESIZE + 1];
	register	UINT	gindex;
	register	UINT	lindex;
	SLONG			value;

	if((lindex = in_2b() + base_symbol) >= MAX_SYMBOLS)
		fatal("Too many total symbols","");
	value = in_4b();		/* Get symbol's value */
	if(lindex >= next_symbol)
		next_symbol = lindex + 1;
	if((gindex = locals[lindex].gst_num) < MAX_GLOBALS)	/* "PUBLIC"? */
		{			/* Yes */
		if(globals[gindex].lst_num < MAX_SYMBOLS)
			{		/* It is already defined! */
			string_copy(buf,globals[gindex].name,
					SYMNMESIZE);
			buf[SYMNMESIZE] = '\0';
			non_fatal("Duplicate external definition: ",buf);
			return;
			}
/*
 *	Otherwise indicate the "PUBLIC" symbol is defined.
 */
		else
			globals[gindex].lst_num = lindex;
		}
	locals[lindex].min_addr = locals[lindex].max_addr = value;
	}

/************************************************************************/
/*									*/
/*	Routine:	add_symbol					*/
/*									*/
/*	Add a symbol name to the global symbol table.  Add a reference	*/
/*	to it to the local symbol table.  This is a helping function	*/
/*	which "read_file" uses when it is doing T_SYMBOL records.  Also	*/
/*	used by "init_slb" to initialize the static link base symbol.	*/
/*									*/
/*	Parameters:							*/
/*		"sname"	A pointer to the symbol name to add.  If a NULL	*/
/*			pointer is passed the name and type are read	*/
/*			from the current input/library file.		*/
/*		"stype"	If "sname" is non-NULL this specifies the type	*/
/*			of the external ("1" for def, "0" for ref).	*/
/*									*/
/*	Returns:	The resulting local symbol table index.		*/
/*									*/
/************************************************************************/

UINT
add_symbol(sname,stype)
	STRING			sname;
	BOOL			stype;
	{
	char			buf[SYMNMESIZE + 1];
	register	UINT	i;
	register	UINT	length;
	BOOL			type;

	if(sname == NULL)
		{
		type = IN_1B();		/* Get external public status */
		length = IN_1B();	/* Get symbol length */
		for(i = 0; i < length; i++)
			buf[i] = IN_1B();	/* Get characters in symbol */
		buf[i] = '\0';
		}
	else
		{
		type = stype;		/* 0 for ref, 1 for def */
		strcpy(buf,sname);
		}
	i = lookup(buf);		/* Find/add to global symbol table */
	if(next_symbol >= MAX_SYMBOLS)
		fatal("Too many total symbols","");
	if(list_fp != NULL)
		{
		if(type)
			fprintf(list_fp,"%s",buf);
		else
			fprintf(list_fp,"*%s",buf);
		if(++sym_col >= SYM_LIST_COLS)
			{
			sym_col = 0;
			putc('\n',list_fp);
			}
		else
			putc('\t',list_fp);
		}
	locals[next_symbol].gst_num = i;	/* Add to LST */
	return (next_symbol++);
	}

/************************************************************************/
/*									*/
/*	Routine:	bad_command					*/
/*									*/
/*	Display a message to the user in response to an interactive	*/
/*	input or command file error.  In the event of a command file	*/
/*	error the command file name and line # are displayed.		*/
/*									*/
/*	Parameters:							*/
/*		"s1" -	A pointer to the first string to display.	*/
/*		"s2" -	A pointer to the second string to display.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
bad_command(s1,s2)
	STRING			s1;
	STRING			s2;
	{
	if(cfile)			/* If using a command file */
		{
		printf("<%s> @ %u: %s%s\n",cmdfnbuf,cmd_number,s1,s2);
		if(list_fp != NULL)
			{
			fprintf(list_fp,"<%s> @ %u: %s%s\n",cmdfnbuf,
				cmd_number,s1,s2);
			}
		}
	else				/* If running in "interactive" mode */
		printf("*** %s%s\n",s1,s2);
	errors++;			/* Remember we had an error */
	}

/************************************************************************/
/*									*/
/*	Routine:	bad_order					*/
/*									*/
/*	A helping function for "parse" which makes sure that commands	*/
/*	used in command files are correctly ordered.			*/
/*									*/
/*	Parameters:							*/
/*		"current" -	A pointer to a UINT which holds the	*/
/*				current ordering value (starts at 0,	*/
/*				and is increased as successive commands	*/
/*				are processed.				*/
/*		"desired" -	A UINT which holds the ordering value	*/
/*				for this command.			*/
/*									*/
/*	Returns:	TRUE if the ordering is incorrect, FALSE if	*/
/*			everything is OK.				*/
/*									*/
/************************************************************************/

BOOL
bad_order(current,desired)
	UINT			*current;
	UINT			desired;
	{
	if(*current > desired)
		{
		bad_command("Bad command ordering","");
		return (TRUE);
		}
	else
		*current = desired;
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	get_entry					*/
/*									*/
/*	A helping function for "parse" which specifies an entrypoint.	*/
/*									*/
/*	Parameters:							*/
/*		"dflt"	-	TRUE if the default entrypoint name	*/
/*				should be used unconditionally.  FALSE	*/
/*				If the entrypoint needs to be derived	*/
/*				from the command line in the parse	*/
/*				buffer.					*/
/*									*/
/*	Returns:	TRUE if an error was encountered, FALSE		*/
/*			otherwise.					*/
/*									*/
/************************************************************************/

BOOL
get_entry(dflt)
	BOOL			dflt;
	{
	extern	STRING		strchr();

	UINT			gi;
	STRING			s;

	if(entry_type != NO_ENTRY)
		{
		bad_command("Duplicate entry point","");
		return (TRUE);
		}
	if(! dflt)
		skip_space();	/* Skip leading whitespace */
	if((! dflt) && isdigit(*parse_ptr))
		{
		entry_type = NUM_ENTRY;
		if(get_num(&entry_n))
			return (TRUE);
		}
	else
		{
		entry_type = SYM_ENTRY;
		if(dflt || (*parse_ptr == '\0'))
			gi = lookup(ENTRYPOINT);	/* Default */
		else
			{
			if((s = strchr(parse_ptr,'\n')) != NULL)
				*s-- = '\0';
			while((s != NULL) && (s > parse_ptr))
				{
				if(! isspace(*s))
					break;
				s--;
				}
			if(s != NULL)
				*++s = '\0';	/* Prune trailing ' ' */
			gi = lookup(parse_ptr);
			}
		if((entry_s = next_symbol++) >= MAX_SYMBOLS)
			fatal("Too many total symbols","");
		locals[entry_s].gst_num = gi;	/* Remember where name is */
		}
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	get_flags					*/
/*									*/
/*	Look for linker flags.  The possible flags are:			*/
/*									*/
/*		a|A	Sort symbol listing in map by address instead	*/
/*			of by name.					*/
/*									*/
/*		c|C	Combine adjacent T_DATA records and adjacent	*/
/*			T_STORAGE records.  This speeds up linking	*/
/*			and reduces the size the .TLD file.  The line	*/
/*			number information for each of these records is */
/*			lost.						*/
/*									*/
/*		s|S	Process code for use with "static link" global	*/
/*			data model.  Causes all T_ADDR_DATA records to	*/
/*			be passed straight through.  It is assumed that	*/
/*			the addresses involved will be bound later	*/
/*			during program loading and initialization.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if an error was encountered, FALSE		*/
/*			otherwise.					*/
/*									*/
/************************************************************************/

BOOL
get_flags()
	{
	FOREVER
		{
		skip_space();
		switch (*parse_ptr++)
			{
			case '\0':
			case '\n':
				return (FALSE);

			case 'a':
			case 'A':
				byaddress = TRUE;
				break;

			case 'c':
			case 'C':
				compress = TRUE;
				break;

			case 's':
			case 'S':
				static_link = TRUE;
				break;

			default:
				*parse_ptr = 0;
				maybe_fatal("Invalid flag character: ",
					parse_ptr-1);
				return (TRUE);
			}
		}
/*lint -e533 lint doesn't understand about functions which never return */
	}
/*lint +e533 lint doesn't understand about functions which never return */

/************************************************************************/
/*									*/
/*	Routine:	init_slb					*/
/*									*/
/*	Defines the "SLBNAME" symbol in module "SLBMODULE" when the	*/
/*	"static link" global data access model has been selected.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
init_slb()
	{
	if(! static_link)
		return;

	open_temp((UINT) SLBMODULE);	/* Open temp file for reserved mod # */
	PUT_1B((int) T_LOAD);		/* Make it start at address 0 */
	put_4b((SLONG) 0);
/*
 *	Add symbol to symbol table and put T_DEF record in module temp file
 *	to provide a logical base address.
 */
	used_modules[SLBMODULE] = TRUE;	/* Inhibit bogus filename */
	add_def_lst(add_symbol(SLBNAME,(BOOL) 1));
	}

/************************************************************************/
/*									*/
/*	Routine:	input_files					*/
/*									*/
/*	Gets a series of input files to be read from the parse buffer,	*/
/*	verifies that the files exist, and opens and reads the contents	*/
/*	if everything looks OK.  This routine is a helping function for	*/
/*	"parse".							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if a trailing comma was detected, FALSE	*/
/*			otherwise.					*/
/*									*/
/************************************************************************/

BOOL
input_files()
	{
	extern	STRING		strrchr();

	SLONG			big_temp;
	BOOL			had_comma;

	had_comma = FALSE;
/*
 *	Go through the parse buffer and generate T_LOAD records for module
 *	load addresses and read input files.
 */
	FOREVER
		{
		skip_space();		/* Skip whitespace */
/*
 *	First figure out if it is a load address specification or a input
 *	filename.
 */
		if(isdigit(*parse_ptr))	/* Load address specification? */
			{		/* Yep */
			if(get_num(&big_temp))
				return (FALSE);
			if((big_temp >= NUMMODULES) || (big_temp < 0) ||
				(static_link && (big_temp == SLBMODULE)))
				{
				bad_command("Illegal module number","");
				return (FALSE);
				}
			open_temp((UINT) big_temp);	/* Module # */
			if(! match(":"))
				{
				bad_command("Missing ':'","");
				return (FALSE);
				}
			if(get_num(&big_temp))
				return (FALSE);
			if(big_temp & ((SLONG) word_size - 1))
				{
				bad_command("Load address not word aligned",
						"");
				return (FALSE);
				}
			PUT_1B((int) T_LOAD);
			put_4b((SLONG) big_temp);	/* Load address */
			aligned[module] = TRUE;
			}
		else
/*
 *	Must have been an input filename (or trailing comma...).
 */
			{
			if(had_comma && (*parse_ptr == '\0'))
				return (TRUE);
			if(*parse_ptr == '\0')
				return (FALSE);
			if(get_fname(fnbuf))
				{
				bad_command("Illegal input filename","");
				return (FALSE);
				}
			if(makefn(fnbuf,ifnbuf,TLNK_INPUT_EXT))	/* Full name */
				{
				bad_command("Input filename too long","");
				return (FALSE);
				}
			if(first_ifile)	/* First input filename? */
				{
/*
 *	Create default output filename based on first input filename.
 */
				strcpy(ofnbuf,ifnbuf);
				strcpy(strrchr(ofnbuf,'.'),TLNK_OUTPUT_EXT);
				}
/*
 *	The following code initializes the stuff used by "update_fname" to
 *	make sure that at least one T_FILENAME record is around to used for
 *	error message generation and debugging.
 */
			strcpy(fnbuf,ifnbuf);	/* Remember name */
			fnsize = strlen(fnbuf);	/* And length */
			old_ln = 0;
			new_ln = 1;
#if	CRLF
			if((in_fp = (STREAM) fopen(ifnbuf,"rb")) == NULL)
#else
			if((in_fp = (STREAM) fopen(ifnbuf,"r")) == NULL)
#endif
				{
				maybe_fatal("Unable to open input file: ",
						ifnbuf);
				return (FALSE);
				}
			if(list_fp != NULL)
				{
				if(sym_col != 0)	/* Cleanup */
					{
					sym_col = 0;
					putc('\n',list_fp);
					}
				fprintf(list_fp,"\nReading input file: %s\n",
					ifnbuf);
				}
			is_lib = FALSE;
			get_buf.count = 0;	/* Init "in" local buffer */
			if(read_file())
				return (FALSE);	/* Read input file */
			if(fclose(in_fp) == EOF)
				fatal("Unable to close input file: ",ifnbuf);
/*
 *	Note that we have now read the first input file.
 */
			first_ifile = FALSE;
			}
		if(match(","))
			had_comma = TRUE;
		else
			had_comma = FALSE;
		}
/*lint -e533 lint doesn't understand about functions which never return */
	}
/*lint +e533 lint doesn't understand about functions which never return */

/************************************************************************/
/*									*/
/*	Routine:	interactive					*/
/*									*/
/*	Called by "parse" to get the command information from the user	*/
/*	interactively.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
interactive()
	{
	BOOL			more;
/*
 *	Ask user for linkage flags.
 */
	FOREVER
		{
		errors = 0;
		compress = FALSE;
		static_link = FALSE;
		printf("Flags (A|a|C|c|S|s)[none]: ");
		if(fgets(parse_buf,(MAXLNSIZE + 1),stdin) != NULL)
			{
			parse_ptr = parse_buf;
			if(get_flags())
				continue;
			}
		break;
		}
/*
 *	Get possible temp file pathname from user.
 */
	printf("Temp file path [.]\t : ");
	if(fgets(parse_buf,(MAXLNSIZE + 1),stdin) != NULL)
		{
		parse_ptr = parse_buf;
		get_fname(temp_path);	/* Get user supplied directory */
		}
/*
 *	Get optional listing filename from user.
 */
	printf("Listing file [NUL]\t : ");
	if(fgets(parse_buf,(MAXLNSIZE + 1),stdin) != NULL)
		{
		parse_ptr = parse_buf;
		get_fname(listbuf);
		}
/*
 *	Get input filenames from user.
 */
	FOREVER
		{
		init_input();		/* Initialize for input files */
		errors = 0;		/* None to start with */
		open_listing();		/* Open the listing file */
		printf("Input file(s) [%s]\t : ",TLNK_INPUT_EXT);
		if(fgets(parse_buf,(MAXLNSIZE + 1),stdin) == NULL)
			{
			bad_command("At least one input file is required","");
			continue;
			}
		parse_ptr = parse_buf;
		more = input_files();
		if(errors)
			continue;
		if(first_ifile)
			{
			bad_command("At least one input file is required","");
			continue;
			}
/*
 *	Set default load and stack addresses.
 */
		switch(cpu_type)
			{
			case	TYPE_212:
				load_addr = T212_LOAD;
				stack_addr = T212_STACK;
				break;

			case	TYPE_800:
				load_addr = T800_LOAD;
				stack_addr = T800_STACK;
				break;

			default:

			case	TYPE_414:
				load_addr = T414_LOAD;
				stack_addr = T414_STACK;
				break;
			}
/*
 *	See if another line of input files is being requested.
 */
		while(more)
			{
			printf("Input file(s) [%s]\t : ",TLNK_INPUT_EXT);
			if(fgets(parse_buf,(MAXLNSIZE + 1),stdin) == NULL)
				break;
			parse_ptr = parse_buf;
			more = input_files();
			if(errors)
				break;
			}
		if(! errors)
			break;
		}
/*
 *	Get optional entrypoint from user.
 */
	FOREVER
		{
		errors = 0;
		printf("Entry point [%s]\t : ",ENTRYPOINT);
		if(fgets(parse_buf,(MAXLNSIZE + 1),stdin) != NULL)
			{
			parse_ptr = parse_buf;
			if(get_entry(FALSE))
				continue;
			}
		else
			{
			if(get_entry(TRUE))
				continue;
			}
		break;
		}
/*
 *	Get optional library filenames from user.
 */
	FOREVER
		{
		printf("Library file(s) [%s]\t : ",TLNK_LIB_EXT);
		if(fgets(parse_buf,(MAXLNSIZE + 1),stdin) != NULL)
			{
			parse_ptr = parse_buf;
			more = library_files();
			if(errors)
				terminate(ERRORS);	/* No recovery */
			}
		if(! more)
			break;
		}
/*
 *	Get output filename from user.
 */
	FOREVER
		{
		printf("Output file [%s]\t : ",ofnbuf);
		if(fgets(parse_buf,(MAXLNSIZE + 1),stdin) != NULL)
			{
			parse_ptr = parse_buf;
			get_fname(ofnbuf);
			}
#if	CRLF
		if((out_fp = (STREAM) fopen(ofnbuf,"wb")) == NULL)
#else
		if((out_fp = (STREAM) fopen(ofnbuf,"w")) == NULL)
#endif
			{
			bad_command("Unable to open output file: ",ofnbuf);
			errors = 0;	/* Reset error counter */
			continue;
			}
		break;
		}
/*
 *	Get optional load address from user.
 */
	FOREVER
		{
		printf("Load address [0x%08lX]: ",load_addr);
		if(fgets(parse_buf,(MAXLNSIZE + 1),stdin) != NULL)
			{
			parse_ptr = parse_buf;
			if(*parse_ptr == '\n')
				break;
			if(get_num(&load_addr))
				{
				errors = 0;	/* Reset error counter */
				continue;	/* Try again */
				}
			}
		break;
		}
/*
 *	Get optional stack address from user.
 */
	FOREVER
		{
		printf("Stack address[0x%08lX]: ",stack_addr);
		if(fgets(parse_buf,(MAXLNSIZE + 1),stdin) != NULL)
			{
			parse_ptr = parse_buf;
			if(*parse_ptr == '\n')
				break;
			if(get_num(&stack_addr))
				{
				errors = 0;	/* Reset error counter */
				continue;	/* Try again */
				}
			}
		break;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	lib_file					*/
/*									*/
/*	Helping function for "library_files".  This routine reads and	*/
/*	processes a single library file.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if an error was detected or no undefined	*/
/*			symbols remain, FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
lib_file()
	{
	STRING			brf;
	char			buf[SYMNMESIZE + 1];
	STRING			cptr;
	register	UINT	length;
	char			libname[FNSIZE + 1];
	register	UINT	num;
	BOOL			needed;
	STRING			path;
	register struct	gst	*ptr;
	register	int	temp;
	char			temp_buf[MAXLNSIZE + 1];
/*
 *	See if any externals are currently undefined.  If not skip processing
 *	the library.
 */
	for(needed = FALSE, length = 0; length < MAX_GLOBALS; length++)
		{
/* Used? */	if((globals[length].name != NULL) &&
/* Undefined? */	(globals[length].lst_num >= MAX_SYMBOLS))
			{
			needed = TRUE;
			}
		}
	if(! needed)
		return (TRUE);		/* No undefined symbols */

	brf = "Bad record format on library file: ";
/*
 *	Open the library file (check current directory first).  If the current
 *	directory doesn't contain the file, check the LIB_PATH environment
 *	variable for a set of optional alternate directories.
 */
#if	CRLF
	if((in_fp = (STREAM) fopen(ifnbuf,"rb")) == NULL)
#else
	if((in_fp = (STREAM) fopen(ifnbuf,"r")) == NULL)
#endif
		{
		if((path = getenv(LIB_PATH)) == NULL)
			fatal("Unable to open library file: ",ifnbuf);
		if(strlen(path) >= MAXLNSIZE)
			fatal("Library path filename too long: ",ifnbuf);
		path = strcpy(temp_buf,path);	/* Local copy of env. path */

/*lint -e530 lint isn't smart enough to see that "cptr" is initialized */
		for(; *path != '\0'; path = cptr)
/*lint +e530 lint isn't smart enough to see that "cptr" is initialized */
			{
			if((cptr = strchr(path,PATH_PUNC)) != NULL)
				*cptr++ = '\0';
			else
				cptr = "\0";

			if((strlen(path) + strlen(ifnbuf)) >= FNSIZE)
				{
				fatal("Library path filename too long: ",
					ifnbuf);
				}
			strcpy(libname,path);	/* Copy environment path */
#if	H_OPSYS == VMS
			strcat(libname,":");	/* Make it a directory */
#else	/* H_OPSYS == VMS */
			strcat(libname,"/");	/* Make it a directory */
#endif	/* H_OPSYS == VMS */
			strcat(libname,ifnbuf);	/* Append library name */
/*
 *	Open the library file using the supplied environment path.
 */
#if	CRLF
			if((in_fp = (STREAM) fopen(libname,"rb")) != NULL)
#else	/* !CRLF */
			if((in_fp = (STREAM) fopen(libname,"r")) != NULL)
#endif	/* CRLF */
				{
				strcpy(ifnbuf,libname);	/* Real name */
				break;
				}
			}
		if(*path == '\0')
			fatal("Unable to open library file: ",ifnbuf);
		}
/*
 *	Make sure the file and (optional) cpu type are correct.
 */
	if(IN_1B() != T_LIB_FILE)
		fatal("Not a library file: ",ifnbuf);
	temp = IN_1B();			/* Figure out what CPU type */
	if(cpu_type == TYPE_UNKNOWN)
		{
		if((cpu_type = temp) == TYPE_212)
			{
			max_i_size = MAX_I_SIZE_16B;
			word_size = WORD_SIZE_16B;
			}
		else
			{
			max_i_size = MAX_I_SIZE_32B;
			word_size = WORD_SIZE_32B;
			}
		}
	else				/* A type has already been selected */
		{
		if((temp != TYPE_UNKNOWN) && (temp != cpu_type))
			fatal("Unable to mix code for different CPU types","");
		}
/*
 *	Go through each relocatable file in the library and link in any
 *	which are required.  If we ever have a condition where no undefined
 *	symbols exist we return and ignore the rest of the library.
 */
	while((temp = IN_1B()) == T_FILENAME)
		{
		in_4b();		/* Toss old/new  ln. # fields */
		fnsize = IN_1B();	/* Filename size for file */
		for(temp = 0; temp < fnsize; temp++)
			{
			fnbuf[temp] = IN_1B();	/* Read filename */
			}
		fnbuf[temp] = '\0';
		old_ln = 0;		/* Default old line # */
		new_ln = 1;		/* Default new line # */
/*
 *	Make sure that undefined external symbols still exist, if not we are
 *	done.
 */
		for(needed = FALSE, length = 0; length < MAX_GLOBALS; length++)
			{
/* Used? */		if((globals[length].name != NULL) &&
/* Undefined? */		(globals[length].lst_num >= MAX_SYMBOLS))
				{
				needed = TRUE;
				}
			}
		if(! needed)
			return (TRUE);	/* No undefined symbols */
/*
 *	The following code examines each external symbol definition in the
 *	library header for a file, if the symbol matches a symbol which
 *	is needed we set "needed" to be TRUE and ignore the rest of the
 *	symbols.  Note that the library header symbol definitions is sorted
 *	although the following code doesn't depend on it.
 */
		needed = FALSE;
		while((temp = IN_1B()) == T_SYMBOL)
			{
			IN_1B();	/* Toss definition flag */
			length = IN_1B();	/* Symbol length */
			for(num = 0, temp = 0; temp < length; temp++)
				num = (num * 16) + (buf[temp] = IN_1B());
			buf[temp] = '\0';
/*
 *	See if the external symbol just read is currently undefined (needed),
 *	for the current link.  To find out, we compute the symbol table entry
 *	number to start scanning at, but we need to convert that to a pointer
 *	first, therefore:
 */
			num &= GLOBALS_MASK;	/* Mask to size of table */
			ptr = globals + num;
/*
 *	Hash scan the external symbol table for exact match or empty slot.
 */
			for(length = 0; length < MAX_GLOBALS; length++, ptr++)
				{
/* Wrap around? */		if(ptr >= &globals[MAX_GLOBALS])
					ptr = globals;
				if(ptr->name != NULL)	/* Entry filled? */
					{
/* Exact match? */			if(! strncmp(buf,ptr->name,SYMNMESIZE))
						{
/* Yes.  Now is it needed? */			if(ptr->lst_num >= MAX_SYMBOLS)
							needed = TRUE;
/* In any event */				break;
						}
/* No, try next */			continue;
					}
/* Empty slot */		break;
				}
			if(needed)
				{
				temp = IN_1B();	/* Get next record type */
				break;
				}
			}
/*
 *	See if any additional T_SYMBOL records need to be ignored.
 */
		while(temp == T_SYMBOL)
			{
			IN_1B();	/* Toss definition status */
			temp = IN_1B();	/* Symbol length */
			while(temp--)
				IN_1B();	/* Toss symbol */
			temp = IN_1B();	/* Get next record header */
			}
/*
 *	See if we needed anything in this file and add it to the temp
 *	files if so.
 */
		if(needed)
			{
			if(temp != T_SIZE)
				fatal(brf,ifnbuf);
			in_4b();	/* Toss file modification time/date */
			in_4b();	/* Toss file size */
/*
 *	Process this library file.
 */
			if(list_fp != NULL)
				{
				if(sym_col != 0)	/* Cleanup */
					{
					sym_col = 0;
					putc('\n',list_fp);
					}
				fprintf(list_fp,
				"\nReading file '%s' from library: %s\n",
					fnbuf,ifnbuf);
				}
			if(read_file())
				return (TRUE);	
			}
/*
 *	If we didn't need any of the symbols defined in the current file,
 *	use the T_SIZE record to skip past it in the library.
 */
		else
			{
			if(temp != T_SIZE)
				fatal(brf,ifnbuf);
			in_4b();	/* Toss file modification time/date */
			in_seek((SLONG) in_4b());
			}
		}
/*
 *	If we didn't get a T_FILENAME, then this should be the end of the
 *	library.  Close the library file.
 */
	if(temp != T_EOF)
		fatal(brf,ifnbuf);
	if(fclose(in_fp) == EOF)
		fatal("Unable to close library file: ",ifnbuf);

	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	library_files					*/
/*									*/
/*	Gets a series of library files to be read from the parse	*/
/*	buffer, figures out which modules need to be loaded and loads	*/
/*	them.  This routine is a helping function for "parse".		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if a trailing comma is allowed and was	*/
/*			detected, FALSE otherwise.			*/
/*									*/
/************************************************************************/

BOOL
library_files()
	{
	SLONG			big_temp;
	BOOL			had_comma;

	had_comma = FALSE;
/*
 *	Go through the parse buffer and generate T_LOAD records for library
 *	module load addresses and read the library files.
 */
	FOREVER
		{
		skip_space();		/* Skip whitespace */
/*
 *	First figure out if it is a load address specification or a library
 *	filename.
 */
		if(isdigit(*parse_ptr))	/* Load address specification? */
			{		/* Yep */
			if(get_num(&big_temp))
				return (FALSE);
			if((big_temp >= NUMMODULES) || (big_temp < 0) ||
				(static_link && (big_temp == SLBMODULE)))
				{
				bad_command("Illegal module number","");
				return (FALSE);
				}
			open_temp((UINT) big_temp);	/* Module # */
			if(! match(":"))
				{
				bad_command("Missing ':'","");
				return (FALSE);
				}
			if(get_num(&big_temp))
				return (FALSE);
			if(big_temp & ((SLONG) word_size - 1))
				{
				bad_command("Load address not word aligned",
						"");
				return (FALSE);
				}
			PUT_1B((int) T_LOAD);
			put_4b((SLONG) big_temp);	/* Load address */
			aligned[module] = TRUE;
			}
		else
/*
 *	Must have been a library filename (or a trailing comma...).
 */
			{
			if(had_comma && (*parse_ptr == '\0'))
				return (TRUE);
			if(*parse_ptr == '\0')
				return (FALSE);
			if(get_fname(fnbuf))
				{
				bad_command("Illegal library filename","");
				return (FALSE);
				}
			if(makefn(fnbuf,ifnbuf,TLNK_LIB_EXT))	/* Full name */
				{
				bad_command("Library filename too long","");
				return (FALSE);
				}
			is_lib = TRUE;
			if(lib_file())
				return (FALSE);
			}
		if(match(","))
			had_comma = TRUE;
		else
			had_comma = FALSE;
		}
/*lint -e533 lint doesn't understand about functions which never return */
	}
/*lint +e533 lint doesn't understand about functions which never return */

/************************************************************************/
/*									*/
/*	Routine:	maybe_fatal					*/
/*									*/
/*	Used for error messages which are "fatal" if getting input	*/
/*	from a command file, and merely "bad_command"s otherwise.	*/
/*									*/
/*	Parameters:							*/
/*		"s1" -	A pointer to the first string to display.	*/
/*		"s2" -	A pointer to the second string to display.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
maybe_fatal(s1,s2)
	STRING			s1;
	STRING			s2;
	{
	if(cfile)
		fatal(s1,s2);		/* If using a command file */
	else
		bad_command(s1,s2);	/* If running in "interactive" mode */
	}

/************************************************************************/
/*									*/
/*	Routine:	open_listing					*/
/*									*/
/*	Open the listing file.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
open_listing()
	{
	if(listbuf[0] != '\0')
		{
#if	CRLF
		if((list_fp = (STREAM) fopen(listbuf,"wt")) == NULL)
#else
		if((list_fp = (STREAM) fopen(listbuf,"w")) == NULL)
#endif
			fatal("Unable to open listing file: ",listbuf);
		lbanner();
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	open_output					*/
/*									*/
/*	Open the output file.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
open_output()
	{
#if	CRLF
	if((out_fp = (STREAM) fopen(ofnbuf,"wb")) == NULL)
#else
	if((out_fp = (STREAM) fopen(ofnbuf,"w")) == NULL)
#endif
		fatal("Unable to open output file: ",ofnbuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	open_temp					*/
/*									*/
/*	Sees if the temp file for a particular module has been opened.	*/
/*	If not it is opened.						*/
/*									*/
/*	Parameters:							*/
/*		"mod_num"	The module # to open a temp file for.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
open_temp(mod_num)
	UINT			mod_num;
	{
	extern	STRING		malloc();

#if	!VMFILE
	char			buf1[FNSIZE + 1];
	char			buf2[FNSIZE + 1];
#endif	/* !VMFILE */

/*
 *	First see if we haven't opened the "extra" temp file.  If not do so.
 */
	if(temp_fp == NULL)
		{
#if	VMFILE
		if ((temp_fp = (VSTREAM)
			malloc(sizeof(struct vstream))) == NULL)
			{
			fatal("Out of vmfile memory", "");
			}
		temp_fp->first = NULL;
		temp_fp->current = NULL;
#else	/* VMFILE */
		maketdir();		/* Finalize temp file directory */
		strcpy(buf1,temp_path);
		strcat(buf1,EXTRA_TEMP);
#if	CRLF
		if((temp_fp = (STREAM) fopen(buf1,"w+b")) == NULL)
#else
		if((temp_fp = (STREAM) fopen(buf1,"w+")) == NULL)
#endif
			fatal("Unable to open temp file: ",buf1);
#if	IOBUFSIZE
		if((temp_fp_buf = (STRING) malloc(IOBUFSIZE)) == NULL)
			{
			fatal("Insufficient stream buffer memory for file: ",
				buf1);
			}
		if(setvbuf(temp_fp,temp_fp_buf,_IOFBF,IOBUFSIZE) != 0)
			fatal("Error setting stream buffer for file: ",buf1);
#endif	/* IOBUFSIZE */
#endif	/* VMFILE */
		}
/*
 *	Now see if the temp file for the particular module has been opened,
 *	if not do so.
 */
	if(mod_fp[module = mod_num] == NULL)
		{
#if	VMFILE
		if ((mod_fp[module] = (VSTREAM)
			malloc(sizeof(struct vstream))) == NULL)
			{
			fatal("Out of vmfile memory", "");
			}
		mod_fp[module]->first = NULL;
		mod_fp[module]->current = NULL;
#else	/* VMFILE */
		strcpy(buf1,temp_path);
		strcat(buf1,TEMP_NAME);
		sprintf(buf2,"%u",module);
		strcat(buf1,buf2);
#if	CRLF
		if((mod_fp[module] = (STREAM) fopen(buf1,"w+b")) == NULL)
#else	/* CRLF */
		if((mod_fp[module] = (STREAM) fopen(buf1,"w+")) == NULL)
#endif	/* CRLF */
			fatal("Unable to open temp file: ",buf1);
#endif	/* VMFILE */
/*
 *	Update local buffering.
 */
		if(acc_dtasize || acc_storesize)
			put_rec(T_RESERVED);
		put_flush();
/*
 *	All modules are aligned at the beginning.
 */
		aligned[module] = old_file[module] = TRUE;
		}
	else
/*
 *	Otherwise module temp file has already been opened.  Verify that the
 *	module temp file has been written into from the current input file.  If
 *	not we need to align the temp file using a T_ALIGN.
 */
		{
/*
 *	Update local buffering.
 */
		if(acc_dtasize || acc_storesize)
			put_rec(T_RESERVED);
		put_flush();

		if(! old_file[module])
			{
			aligned[module] = old_file[module] = TRUE;
			PUT_1B((int) T_ALIGN);	/* Record type */
			put_2b((int) 0);	/* Dummy line # */
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	parse						*/
/*									*/
/*	Figure out what the user wants done by reading commands from	*/
/*	the command file (if one is given), or asking the user		*/
/*	directly.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
parse()
	{
	register	UINT	i;
	UINT			sequence;
/*
 *	Initialize the module file pointers and stream buffer pointers
 *	to NULL.
 */
	for(module = 0; module < NUMMODULES; module++)
		mod_fp[module] = NULL;
/*
 *	Initialize the global symbol table "name" pointers to NULL.
 */
	for(i = 0; i < MAX_GLOBALS; i++)
		globals[i].name = NULL;
/*
 *	Time to read either the command file or standard input to figure out
 *	what the user wants done.
 */
	if(cfile)			/* If input from command file */
		{
		init_input();		/* Initialize for input files */
		sequence = 0;		/* Insure command ordering */
		FOREVER
			{
			cmd_number++;	/* Update command file line # */
			if(fgets(parse_buf,(MAXLNSIZE + 1),cmdf_fp) == NULL)
				{
				if(feof(cmdf_fp))	/* If EOF */
					{
					if(first_ifile)
						{
				bad_command("No input file(s) specified","");
						}
					if(fclose(cmdf_fp) == EOF)
						{
				warning("Unable to close command file: ",
						cmdfnbuf);
						}
					if(entry_type == NO_ENTRY)
						get_entry(TRUE);
					open_output();
					return;
					}
				if(ferror(cmdf_fp))	/* If error */
					{
					fatal("Error reading command file: ",
						cmdfnbuf);
					}
				}
			parse_ptr = parse_buf;
/*
 *	Any line starting with a non-alpha is a comment.
 */
			skip_space();
			if(! isalpha(*parse_ptr))
				continue;
/*
 *	Handle optional "flags"
 */
			if(match_first("FLAG"))
				{
				if(bad_order(&sequence,0))
					continue;
				(VOID) get_flags();
				continue;
				}
/*
 *	Handle "temp" file pathname command.
 */
			if(match_first("TEMP"))
				{
				if(bad_order(&sequence,1))
					continue;
				get_fname(temp_path);
				continue;
				}
/*
 *	Handle "list" file pathname command.
 */
			if(match_first("LIST"))
				{
				if(list_fp != NULL)
					{
	bad_command("'LIST' filename already specified","");
					continue;
					}
				if(bad_order(&sequence,2))
					continue;
				get_fname(listbuf);
				open_listing();
				continue;
				}
/*
 *	Handle "input" file(s) command.
 */
			if(match_first("INPUT"))
				{
				if(bad_order(&sequence,3))
					continue;
				input_files();		/* Read input files */
/*
 *	Set default load and stack addresses.
 */
				if(! first_ifile)
					{
					switch(cpu_type)
						{
						case	TYPE_212:
							load_addr = T212_LOAD;
							stack_addr =T212_STACK;
							break;

						case	TYPE_800:
							load_addr = T800_LOAD;
							stack_addr =T800_STACK;
							break;

						default:

						case	TYPE_414:
							load_addr = T414_LOAD;
							stack_addr =T414_STACK;
							break;
						}
					}
				continue;
				}
/*
 *	Handle "entrypoint" command.
 */
			if(match_first("ENTRY"))
				{
				if(bad_order(&sequence,4))
					continue;
				get_entry(FALSE);
				continue;
				}
/*
 *	Handle "libraries" command.
 */
			if(match_first("LIB"))
				{
				if(bad_order(&sequence,5))
					continue;
				if(entry_type == NO_ENTRY)
					{
					if(get_entry(TRUE))
						continue;
					}
				library_files();
				continue;
				}
/*
 *	Handle "output" file command.
 */
			if(match_first("OUTPUT"))
				{
				if(bad_order(&sequence,6))
					continue;
				get_fname(ofnbuf);
				continue;
				}
/*
 *	Handle "load" address command.
 */
			if(match_first("LOAD"))
				{
				if(bad_order(&sequence,7))
					continue;
				get_num(&load_addr);
				continue;
				}
/*
 *	Handle "stack" address command.
 */
			if(match_first("STACK"))
				{
				if(bad_order(&sequence,8))
					continue;
				get_num(&stack_addr);
				continue;
				}
			}
		}
/*
 *	Otherwise, handle the "interactive" case.
 */
	interactive();
	}

/************************************************************************/
/*									*/
/*	Routine:	put_rec						*/
/*									*/
/*	Outputs the begining of a T_* record (the code and optional	*/
/*	line number).  If there is an accumulated T_DATA or T_STORAGE	*/
/*	block, this is output first.					*/
/*									*/
/*	Parameters:	T_* code.					*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
put_rec(op)
	UINT			op;
	{
	register	UCHAR	*p;

	if(acc_dtasize > 0)
		{
		update_fname();		/* Add T_FILENAME record? */
		PUT_1B((int) T_DATA);
		put_2b((int) 0);
		put_2b((int) acc_dtasize);
		p = acc_dta;
		while(--acc_dtasize >= 0)
			PUT_1B(*p++);
		acc_dtasize = 0;
		}
	if(acc_storesize > 0)
		{
		update_fname();		/* Add T_FILENAME record? */
		PUT_1B((int) T_STORAGE);
		put_2b((int) 0);
		put_4b((SLONG) acc_storesize);
		acc_storesize = 0;
		}
	if(op == T_RESERVED)
		return;
	update_fname();			/* Add T_FILENAME record? */
	PUT_1B((int) op);
	put_2b((int) number);
	}

/************************************************************************/
/*									*/
/*	Routine:	read_file					*/
/*									*/
/*	Read the contents of the input file into the appropriate	*/
/*	"module" temp files.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if an error encountered, FALSE otherwise.	*/
/*									*/
/************************************************************************/

BOOL
read_file()
	{
	SLONG			big_temp;
	register	UINT	temp;

	base_symbol = next_symbol;	/* Base is updated for each new file */
	acc_dtasize = 0;		/* Size of accumulated T_DATA */
	acc_storesize = 0;		/* Size of accumulated T_STORAGE */
/*
 *	If this is the first input file ensure that what ever is required in
 *	conjunction with the static link is done before processing any user
 *	files.
 */
	if(first_ifile)			/* First input filename? */
		{
/*
 *	If the "static link" global data model is being used, emit the initial
 *	symbol definition to specify the base into the reserved module temp
 *	file.
 */
		init_slb();
		base_symbol = next_symbol;	/* Base updated for new file */
		}
/*
 *	Indicate that any new code which is written to already open modules
 *	must be prefixed with a T_ALIGN to ensure inter-file alignment
 *	restrictions.
 */
	for(temp = 0; temp < NUMMODULES; temp++)
		old_file[temp] = FALSE;
	open_temp(0);			/* Open default mod. file if needed */
/*
 *	Read the records in the file into the various "module" temp files,
 *	doing whatever input conversion is required.
 */
	FOREVER
		{
		switch(temp = IN_1B())
			{
			case T_SYMBOL:	/* External symbol text */
				(VOID) add_symbol(NULL,/* DONT CARE */ 0);
				continue;

			case T_REL_FILE:	/* Normal relocatable file */
				temp = IN_1B();	/* CPU type */
				if(cpu_type == TYPE_UNKNOWN)
					{
					if((cpu_type = temp) == TYPE_212)
						{
						max_i_size = MAX_I_SIZE_16B;
						word_size = WORD_SIZE_16B;
						}
					else
						{
						max_i_size = MAX_I_SIZE_32B;
						word_size = WORD_SIZE_32B;
						}
					}
				else
					{
					if((temp != TYPE_UNKNOWN) &&
						(temp != cpu_type))
						{
		maybe_fatal("Unable to mix code for different CPU types","");
						}
					}
				continue;

			case T_LIB_FILE:	/* Library relocatable file */
			maybe_fatal("Libraries aren't legal input files: ",
					ifnbuf);
				if(acc_storesize || acc_dtasize)
					put_rec(T_RESERVED);
				return (TRUE);

			case T_LD_FILE:		/* Linked and loadable file */
			maybe_fatal("Input file has already been linked: ",
					ifnbuf);
				if(acc_storesize || acc_dtasize)
					put_rec(T_RESERVED);
				return (TRUE);

			}
		number = in_2b();
		switch(temp)
			{
			case T_DATA:	/* Normal data definition */
				aligned[module] = FALSE;
				temp = in_2b();
				if(compress)
					{
					if(acc_storesize)
						put_rec(T_RESERVED);
					while(temp > 0)
						{
						if(acc_dtasize >=
							ACCDATASIZE)
							{
							put_rec(T_RESERVED);
							}
						acc_dta[acc_dtasize++] =
							IN_1B();
						temp--;
						}
					break;
					}
				put_rec(T_DATA);
				put_2b((int) temp);
				copy_in(temp);	/* Actual data stream */
				break;

			case T_RELSYM_OP:	/* Symbolic PC rel. instr. */
				put_rec(temp);
				copy_in(1);	/* Minimum length */
				add_lst();	/* Translate ST # */
				copy_in(5);	/* Offset & opcode */
				aligned[module] = FALSE;
				break;

			case T_DEF:	/* Label definition */
				add_def_lst(MAX_SYMBOLS);	/* LST def */
				break;

			case T_ALIGN:	/* Word alignment pseudo-op */
				if(! aligned[module])	/* Only 1 needed */
					{
					put_rec(temp);
					aligned[module] = TRUE;
					}
				break;

			case T_STORAGE:	/* Reserve space pseudo-op */
				big_temp = in_4b();
				aligned[module] = FALSE;
				if(compress)
					{
					if(acc_dtasize > 0)
						put_rec(T_RESERVED);
					acc_storesize += big_temp;
					break;
					}
				put_rec(temp);
				put_4b(big_temp);	/* Actual data size */
				break;

			case T_RELREL_OP:	/* "rel-rel" ref. instr. */
			case T_WRELREL_OP:	/* Scaled "rel-rel" instr. */
				put_rec(temp);
				copy_in(1);	/* Minimum length */
				add_lst();	/* Translate "left" ST # */
				add_lst();	/* Translate "right" ST # */
				copy_in(5);	/* Offset & opcode */
				aligned[module] = FALSE;
				break;

			case T_SET:	/* Constant equate operation */
				add_set_lst();	/* Add LST definition */
				break;

			case T_ADDR_DATA:	/* Address ref. in data */
				put_rec(temp);
				add_lst();	/* Translate ST # */
				copy_in(4);	/* Offset */
				break;

			case T_MODULE:	/* Module # definition */
				open_temp((UINT) IN_1B());
				break;

			case T_ADDR_OP:	/* Address reference in instruction */
				put_rec(temp);
				copy_in(1);	/* Minimum length */
				add_lst();	/* Translate ST # */
				copy_in(5);	/* Offset & opcode */
				aligned[module] = FALSE;
				break;

			case T_RELREL_DATA:	/* "rel-rel" ref. in data */
				put_rec(temp);
				add_lst();	/* Translate "left" ST # */
				add_lst();	/* Translate "right" ST # */
				copy_in(4);	/* Offset */
				break;

			case T_REL_OP:	/* Non-symbolic PC relative instr. */
				put_rec(temp);
				copy_in(6);	/* Offset & opcode */
				aligned[module] = FALSE;
				break;

			case T_FILENAME:	/* Filename record */
				old_ln = number;
				number = new_ln = in_2b();
				fnsize = IN_1B();
				for(temp = 0; temp < fnsize; temp++)
					fnbuf[temp] = IN_1B();
				fnbuf[temp] = '\0';
/*
 *	Make sure that all modules used from now on get the latest filename
 *	information.
 */
				for(temp = 0; temp < NUMMODULES; temp++)
					used_modules[temp] = FALSE;
				break;

			case T_RELSYM_DATA:	/* Sym. PC rel. ref. in data */
				put_rec(temp);
				add_lst();	/* Translate ST # */
				copy_in(4);	/* Offset */
				break;

			case T_REL_DATA:	/* !sym. PC rel. ref. in dta */
				put_rec(temp);
				copy_in(4);	/* Address */
				break;

			case T_DEBUG_DATA:	/* Debug data record */
				put_rec(temp);
				copy_in(4);	/* Value */
				put_2b((int) (temp = in_2b()));
				copy_in(temp);	/* Data */
				break;

			case T_DEBUGSYM_DATA:	/* Symbolic debug data rec */
				put_rec(temp);
				copy_in(4);	/* Offset value */
				put_2b((int) (temp = in_2b()));
				add_lst();	/* Translate ST # */
				copy_in(temp);	/* Data */
				break;

			case T_EOF:	/* Found EOF record */
				if(acc_storesize || acc_dtasize)
					put_rec(T_RESERVED);
				return (FALSE);

			default:
			maybe_fatal("Bad record format on input file: ",
				ifnbuf);
				if(acc_storesize || acc_dtasize)
					put_rec(T_RESERVED);
				return (TRUE);
			}
		}
/*lint -e533 lint doesn't understand about functions which never return */
	}
/*lint +e533 lint doesn't understand about functions which never return */

/************************************************************************/
/*									*/
/*	Routine:	sort_udef					*/
/*									*/
/*	Sort undefined external symbols onto a linked list whose head	*/
/*	is "udef_list".  This is called by "lib_file" when preparing to	*/
/*	determine whether the current library module contains		*/
/*	definitions which are needed.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if no undefined symbols exist, FALSE	*/
/*			otherwise.					*/
/*									*/
/************************************************************************/

BOOL
sort_udef()
	{
	register	UINT	i;
	register	UINT	j;
	register	UINT	k;

	udef_list = 0;			/* List head pointer */
	for(i = 0; i < MAX_GLOBALS; i++)	/* Check each GST entry */
		{
		if(globals[i].name != NULL) 	/* Occupied? */
			{
			if(globals[i].lst_num >= MAX_SYMBOLS)	/* Undef'd? */
				{
/*
 *	We know the "i" global symbol is undefined, now figure out where in the
 *	undefined list we need to add it (the list is sorted).
 */
				if(udef_list == 0)	/* First on list? */
					{
					udef_list = i + (MAX_SYMBOLS + 1);
					globals[i].lst_num = MAX_SYMBOLS;
					}
				else	/* Not first on list, find spot */
					{
					k = 0;	/* Point one behind j */
					for(j = udef_list; j != MAX_SYMBOLS;
						j = globals[j - (MAX_SYMBOLS +
							1)].lst_num)
						{
						if(strncmp(globals[i].name,
							globals[j -
							(MAX_SYMBOLS+1)].name,
							SYMNMESIZE) < 0)
								break;
						k = j;
						}
					if(k == 0)	/* Insert at start? */
						{
						globals[i].lst_num = udef_list;
						udef_list = i +
							(MAX_SYMBOLS + 1);
						}
					else	/* Insert in middle/at end */
						{
						globals[i].lst_num =
							globals[k -
							(MAX_SYMBOLS + 1)]
							.lst_num;
						globals[k - (MAX_SYMBOLS+1)].
							lst_num = i +
							(MAX_SYMBOLS + 1);
						}
					}
				}
			}
		}
	if(udef_list == 0)
		return (TRUE);
	else
		return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	update_fname					*/
/*									*/
/*	Determines whether the current "module" temp file has had the	*/
/*	most recent "T_FILENAME" record written to it.  If not the	*/
/*	deed is done and the "used_module" data structure updated to	*/
/*	show the fact.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
update_fname()
	{
	register	UINT	i;

	if((! used_modules[module]) && (! compress))
		{
		PUT_1B((int) T_FILENAME);
		put_2b((int) old_ln);
		put_2b((int) new_ln);
		PUT_1B((int) fnsize);
		for(i = 0; i < fnsize; i++)
			PUT_1B((int) fnbuf[i]);
		used_modules[module] = TRUE;
		}
	}
