/*
**	TCX Compiler (11 of 11)
**
**	Copyright 1987-1988 by Logical Systems.
**	All rights reserved.
**
**	Published by:	Logical Systems
**			P.O. Box 1702
**			Corvallis, OR 97339 (USA)
**
**	05/7/87
**
**	Symbol table manipulation
*/

#include <stdio.h>
#include <string.h>
#include "tcx.h"
#include "tcxdecl.h"

/*
 *	Data definitions.
 */

PRIVATE int	roving_bucket	=	0;	/* Used for anon symbols */
PRIVATE	struct	id_node *bucket[MAX_SYM_BUCKETS] = { NULL }; /* The buckets */
PRIVATE	int	current_scope = -1;		/* Current nesting level */

PRIVATE	struct	id_node *freenode_list = NULL;	/* The free list for nodes */
#ifdef	DEBUG
PUBLIC int	Alloc_sym	=	0;	/* Total allocated */
PUBLIC int	Nfree_sym	=	0;	/* Number free */
#endif	/* DEBUG */

PRIVATE struct	id_node *label_list = NULL;	/* The "label" scope list  */
PRIVATE struct	id_node *fixed_list = NULL;	/* "Fixed" symbol scope	   */

PRIVATE struct	id_node *global_list = NULL;	/* Global ripup list    */
PRIVATE struct	id_node	*literal_list = NULL;	/* Literal ripup list	*/

/*
 *	The following list pointer holds the head of the current "local" scope
 *	list.  This list pointer is modified as the compiler enters and exits
 *	scopes (copies of the old value of it are recursively saved as new
 *	scopes are entered and restored when scopes are left).
 */
PRIVATE	struct	id_node	*local_list = NULL;	/* The most "local" scope */


/*
 *	Add a fixed symbol (such as a keyword) to the symbol table.
 *	These symbols will never be removed, unlike externs (removed at
 *	the end of file) or autos (removed at the appropriate scoping level).
 */
PUBLIC	struct	id_node *
add_fixed(name, name_space, value)
  char	*name;
  int	name_space;
  INT32	value;			/* Value of the node			*/
    {
    extern struct id_node *add_node();

    return (add_node(name, 0, name_space, value, &fixed_list));
    }


/*
 *	Add a label into the symbol table.
 *
 *	This is used to add a "goto" symbol into the symbol table under
 *	the user's name.
 *
 *	When we get the node back from "add_node", we must set some type
 *	bits.
 *
 *	Note that we cheat a little on the typing of labels.  Watch this
 *	if labels ever get more ornate then they currently are.
 *
 *	Return NULL if this is a redefinition of the label.
 */
PUBLIC struct id_node *
add_label(name, value, definition)
  char	*name;			/* User's name for the label		*/
  LABEL	value;			/* Label # which represents the label	*/
  int	definition;		/* True if this is defining a function	*/
    {
    extern struct id_node *add_node();
    extern struct id_node *lookup_node();

    struct id_node *label;

    /* See if the label is already in the symbol table			*/
    if ((label = lookup_node(name, NS_LABEL)) != NULL)
	{
	if (!definition)
	    return (label);		/* Reference of existing label	*/
	if (label->type & LABEL_DEFINED)
	    return (NULL);		/* Multiply defined		*/
	label->type |= LABEL_DEFINED;
	return (label);
	}

    /* A new symbol.  Create an entry for it.				*/
    label = add_node(name, 0, NS_LABEL, (INT32) value, &label_list);
    label->type |= LABEL_NODE | ((definition) ? LABEL_DEFINED : 0);
    return (label);
    }


/*
 *	add_loc -- Add a local symbol to the symbol table.
 */
PUBLIC struct id_node *
add_loc(name, base_type, storage_class, name_space, type)
  char			*name;
  struct id_node	*base_type;
  int			storage_class;
  int			name_space;
  INT32			type;
    {
    extern struct id_node	*add_node();

    struct id_node		*sym;

    sym = add_node(name, current_scope, name_space, (INT32) 0, &local_list);

    if(sym != NULL)
	{
	sym->type |= (type & ~LAST_NODE);
	sym->type_ptr.base_type = base_type;
	sym->info.storage_class = storage_class;
	}

    return (sym);
    }

    
/*
 *	add_lit -- Add a literal to the symbol table
 */
PUBLIC struct id_node *
add_lit(name, base_type, name_space)
  char			*name;
  struct id_node	*base_type;
  int			name_space;
    {
    extern struct id_node	*add_node();

    struct id_node		*sym;

    sym = add_node(name, 0, name_space, (INT32) 0, &literal_list);

    if(sym != NULL)
	{
	sym->type |= (base_type->type & ~LAST_NODE);
	sym->type_ptr.base_type = base_type;
	sym->info.storage_class = SC_LITERAL;
	}

    return (sym);
    }

    
/*
 *	Add_node -- Insert a new symbol into the symbol table.  "global"
 *	should be TRUE if the node is to be given global scoping.
 *
 *	A NULL return value represents a name collision.
 */
PRIVATE	struct	id_node *
add_node(name, scope, name_space, value, ripup)
  char			*name;		/* Name of symbol		*/
  int			scope;		/* Current scope level		*/
  int			name_space;	/* Hash offset for type of sym	*/
  INT32			value;		/* Value of the node		*/
  struct id_node	**ripup;	/* Pointer to root of symbol list */
    {
    extern struct id_node	*get_node();
    extern int			hash();
    extern char			*malloc();
    extern VOID			mem_err();

    struct id_node		*node, *new;
    int				hashval;

    if (name == NULL)
	{
	/* We have an anonymous symbol.  It will never be looked up,
	 * so we can put it into any bucket.  So as to not load up
	 * any single bucket, we cycle through all of them.
	 */
	hashval = roving_bucket++;
	if (roving_bucket >= MAX_SYM_BUCKETS)
	    roving_bucket = 0;

	new = get_node();		/* Get fresh node		*/
	}
    else
	{
	/*
	 * If we are attempting to add a symbol at the same "scope"
	 * as an existing symbol, within the same namespace, we have
	 * a multiply defined symbol.  Return NULL.
	 */
	hashval = (hash(name, MAX_SYM_BUCKETS) + name_space) % MAX_SYM_BUCKETS;

	for(new = bucket[hashval];
	    new != NULL;
	    new = ((new->type & LAST_NODE) ? NULL : new->bucket.next_ptr))
	    {
	    if ((new->name != NULL) && (strcmp(name, new->name) == 0) &&
				(scope == new->level))
		return (NULL);
	    }

	new = get_node();			/* Get fresh node	*/
	if ((new->name = malloc((unsigned int) (strlen(name) + 1))) == NULL)
	    mem_err();
	strcpy(new->name, name);		/* Name string		*/
	}

    new->val.integral = value;		/* Use users value (first)	*/
    new->level = scope;			/* Define symbol's scope level	*/

    /*
     *	Insert new node at beginning of specified level list.
     */
    new->level_ptr = *ripup;
    *ripup = new;

    /*
     *	Link node in at appropriate spot in hashed symbol table.  First
     *	see if symbol table hash bucket was empty (the simple case).
     */
    if ((node = bucket[hashval]) == NULL)	/* Empty bucket? */
	{
	new->type = LAST_NODE;	/* End of list */
	new->bucket.head_ptr = &bucket[hashval];
	}
    else
	{
	/*
	 *	Symbol table wasn't empty, link the symbol in at the front of
	 *	the bucket list for the current scoping rules.
	 */
	new->bucket.next_ptr = node;
	}
    bucket[hashval] = new;
    return (new);
    }


/* Add a new type to the fixed_type table
 */
PUBLIC	struct	id_node	*
fixed_type(typecode, base, size)
  INT32	typecode;
  struct id_node *base;
  INT32	size;
    {
    register struct id_node	*rval;

    rval = add_node(NULL, 0, 0, size, &fixed_list);
    rval->type |= (typecode & ~LAST_NODE);
    rval->type_ptr.base_type = base;
    return (rval);
    }


/* Add a new type to table
 */
PUBLIC	struct	id_node	*
add_type(typecode, base, size)
  INT32	typecode;
  struct id_node *base;
  INT32	size;
    {
    register struct id_node	*rval;

    rval = add_node(NULL, 0, 0, size, &local_list);
    rval->type |= (typecode & ~LAST_NODE);
    rval->type_ptr.base_type = base;
    return (rval);
    }


/*
 *	Delete all label symbols
 *
 *	This rips up all of the labels for the entire function.  If any
 *	label has not been defined, then print an error message.
 */
PUBLIC	VOID
del_label()
    {
    extern VOID delete_level();

    delete_level(label_list);
    label_list = NULL;
    }


/*
 *	Delete_level -- Delete a level of symbols/types.
 *
 *	The "level" points to a linked list of all symbols defined on a certain
 *	level.  For each symbol, scan to the end of the symbols in the bucket,
 *	and then scan back to find the symbol.  Unlink the symbol and free it.
 */
PRIVATE	VOID
delete_level(level)
	struct id_node *level;
	{
	extern	VOID		errname();
	extern	VOID		free_level();
	extern	VOID		scan_level();
	extern	VOID		xfer_externs();

	register struct id_node	*scan,*node,**last,*prev;

	scan_level(level,xfer_externs);	/* Transfer any externs to global */

	for(scan = level; scan != NULL; scan = scan->level_ptr)
		{
/*
 *	If the node is a label, but never defined, then barf...
 */
		if((scan->type & LABEL_NODE) &&
			(! (scan->type & LABEL_DEFINED)))
			{
			errname("Undefined label",scan->name);
			}
/*
 *	The following code tests to see if we have a struct/union tag which
 *	never got declared.  Note that in the following test the
 *	"type_ptr.base_type" field is actually a union with the head of the
 *	member list and this fact is used!  I.e. if the symbol table element
 *	has a type that implies it is a variable, not a tag, but if it is a
 *	tag then the field should point to the associated member list!  Only
 *	if it is NULL do we have a undeclared tag...
 */
		else if((scan->type & SU_NODE) &&
			(scan->type_ptr.base_type == NULL))
			{
			warnname(1,"Undefined structure/union tag",scan->name);
			}
/*
 *	Find the end of the bucket chain containing this node.
 */
		for(node = scan; !(node->type & LAST_NODE);
			node = node->bucket.next_ptr)
			;
#ifdef	DEBUG
		if((node->bucket.head_ptr < &bucket[0]) ||
			(node->bucket.head_ptr >= &bucket[MAX_SYM_BUCKETS]))
			{
	    fatal("Compiler error: delete_level: node outside bucket");
			}
#endif	/* DEBUG */

/*
 *	Find the bucket node which is just before the one we are removing.
 */
		for(last = node->bucket.head_ptr; *last != scan;
			last = &((*last)->bucket.next_ptr))
			{
			prev = *last;
			}
/*
 *	Remove the target node and update the bucket linked list.
 */
		*last = scan->bucket.next_ptr;	/* Link around current node */
/*
 *	Is the bucket now empty (ie. pointing to itself)?
 */
		if(last == (struct id_node **) *last)
			{
#ifdef	DEBUG
			if((last < &bucket[0]) ||
				(last >= &bucket[MAX_SYM_BUCKETS]))
				{
		fatal("Compiler error: delete_level: last outside bucket");
				}
#endif	/* DEBUG */
			*last = NULL;
			}
		else if(scan->type & LAST_NODE)	/* Or copy the EOL token */
			prev->type |= LAST_NODE;
		}
	free_level(level);
	}


/*
 *	xfer_externs
 *
 *	Called by scan_ripup() from delete_level().  Called once for
 *	each node in current scoping level.  Test the symbol for being
 *	of storage class SC_EXTERN and with the type REFERENCED_NODE
 *	bit set.  If so, transfer to another name_space if not already
 *	there.
 *
 *	Also called from parse (via scan_ripup), for last pass on
 *	global symbols looking for undef'd externals.
 */
PUBLIC VOID
xfer_externs(sym)
  struct id_node		*sym;
    {
    extern struct id_node	*add_node();
    struct id_node		*p;

    if((sym->info.storage_class == SC_EXTERN) && (sym->type & REFERENCED_NODE))
	{
	p = add_node(sym->name, 0, NS_EXTERNS, (INT32) 0, &global_list);
	if(p != NULL)
		p->info.storage_class = SC_EXTERN;
	}
    }


/*
 *	Dumpsymtab -- Print out the symbol table
 *
 *	Used mostly for debugging.
 */
PUBLIC	VOID
dumpsymtab()
    {
    struct id_node *node;
    int bucketnum;

    for (bucketnum = 0; bucketnum < MAX_SYM_BUCKETS; bucketnum++)
	{
	if (bucket[bucketnum] == NULL)
	    continue;
	printf("Bucket %d (%08p):\n", bucketnum, &bucket[bucketnum]);
	for (node = bucket[bucketnum]; ; node = node->bucket.next_ptr)
	    {
	    printf("(%08p) value %ld type %08lx next %08p level %08p %s\n",
		node, node->val.integral, node->type,
		node->bucket.next_ptr, node->level_ptr, node->name);
	    if (node->type & LAST_NODE)
		break;
	    }
	}
    }


/*
 *	Enter a new scoping level.
 *
 *	This routine returns a magic cookie that the caller must remember
 *	and specify as a parameter to "exit_scope" at the end of the scoping
 *	level.
 *
 *	This value turns out to be the rip up chain.
 */
PUBLIC struct id_node *
enter_scope()
    {
    struct id_node *retval;

    current_scope++;
    retval = local_list;
    local_list = NULL;
    return (retval);
    }


/*
 *	Exit a scoping level.
 *
 *	This routine is called with the return value to the preceding
 *	"enter_scope" routine.  The current scope is deleted, and the
 *	old scoping level is reinstated.
 */
PUBLIC VOID
exit_scope(oldscope)
  struct id_node *oldscope;
    {
    delete_level(local_list);
    current_scope--;
    local_list = oldscope;
    }


/*
 *	find_tag
 *
 *	Lookup a structure tag and return its pointer.
 */
PUBLIC struct id_node *
find_tag(name)
  char *name;
    {
    extern struct id_node	*lookup_node();

    return (lookup_node(name, NS_TAG));
    }


/*
 *	Free_level -- Free a list of nodes linked via "level_ptr".  It is
 *	assumed that the nodes to be freed up are already removed from all
 *	all other data structures.
 */
PUBLIC	VOID
free_level(node)
  struct id_node *node;
    {
    extern VOID		free_node();
    struct id_node	*temp;

    while (node != NULL)		/* Free all nodes in level list */
	{
	temp = node;
	node = node->level_ptr;
	free_node(temp);
	}
    }


/*
 *	Free_node -- Free up a symbol/type node.
 */
PRIVATE VOID
free_node(node)
  struct id_node *node;
    {
    extern VOID			dvrt_free();
    extern VOID			free_plist();

    /* Free up any associated name string.				*/
    if (node->name != NULL)
	free(node->name);

    /* Free up any associated diversion */
    if(node->diversion != NULL)
	dvrt_free(node->diversion);

    /* Free any associated parameter prototype list			*/
    free_plist(node->proto_list);

    /* Free the node itself.						*/
    node->level_ptr = freenode_list;
    freenode_list = node;
#ifdef	DEBUG
    ++Nfree_sym;		/* One more free */
#endif	/* DEBUG */
    }


PUBLIC	VOID
free_plist(list)
  struct prototype	*list;
    {
    struct prototype	*next;

    /* Free up the list (only the list) of the prototype chain		*/
    while(list != NULL)
	{
	next = list->next;
	if(list->name != NULL)
	    free(list->name);
	free((char *) list);		/* Should be special purpose */
	list = next;
	}
    }

/*
 *	Get_node -- Get a free symbol/type structure.
 */
PUBLIC	struct	id_node	*
get_node()
    {
    extern char		*malloc();
    extern VOID		mem_err();
    register int	count;
    register struct	id_node	*retval;

    if (freenode_list == NULL)
	{
	/* If the free list is empty get a block of nodes and add them to it */
	if ((retval = (struct id_node *)
			malloc(SYM_ALLOC * sizeof(struct id_node))) == NULL)
	    {
	    mem_err();
	    }
#ifdef	DEBUG
	Alloc_sym += SYM_ALLOC;	/* Total allocated */
#endif	/* DEBUG */
	for (count = 0; count < SYM_ALLOC; count++)
	    {
	    retval->level_ptr = freenode_list;
	    freenode_list = retval++;
#ifdef	DEBUG
	    ++Nfree_sym;
#endif	/* DEBUG */
	    }
	}

    /* Remove first node from the free list and return it.		*/
    retval = freenode_list;
    freenode_list = retval->level_ptr;
#ifdef	DEBUG
    --Nfree_sym;			/* One less free */
#endif	/* DEBUG */

    /* Initialize node contents.					*/
    retval->name = NULL;		/* Name pointer field */
    retval->bucket.next_ptr =
	retval->level_ptr =
	retval->type_ptr.base_type = NULL;
    retval->proto_list = NULL;
    retval->type = 0;
    retval->diversion = NULL;
    retval->info.next_member = NULL;
    retval->val.integral = 0;
    retval->level = 0;
    return (retval);
    }


/*
 *	Compute a hash bucket for a specified symbol name.
 */
PRIVATE	int
hash(name, num_buckets)
  char	*name;
  int	num_buckets;
    {
    int	hash_value;

    /* This hash function is only dependant upon the last 8
     * characters of the symbol.  Not perfect.	DCR	25 May 87	*/
    for (hash_value = 0; *name != '\0';	hash_value = (hash_value * 4) + *name++)
	;

    hash_value &= 0x7FFF;
    return (hash_value % num_buckets);
    }


/*
 *	Initialize the well known types
 */
PUBLIC	VOID
inittypes()
    {
    extern struct id_node	*fixed_type();

    /* Define some oft-used types, and mark them permanent */

    Voidtype   = fixed_type(VOID_NODE,			NULL, (INT32) 0);
    Chartype   = fixed_type(CHAR_NODE | SIGNED_NODE,	NULL, (INT32) SIZEOF_CHAR);
    Uchartype  = fixed_type(CHAR_NODE | UNSIGNED_NODE,	NULL, (INT32) SIZEOF_CHAR);
    Shorttype  = fixed_type(SHORT_NODE | SIGNED_NODE,	NULL, (INT32) Sizeof_short);
    Ushorttype = fixed_type(SHORT_NODE | UNSIGNED_NODE,	NULL, (INT32) Sizeof_short);
    Inttype    = fixed_type(INT_NODE | SIGNED_NODE,	NULL, (INT32) Sizeof_int);
    Uinttype   = fixed_type(INT_NODE | UNSIGNED_NODE,	NULL, (INT32) Sizeof_int);
    Longtype   = fixed_type(LONG_NODE | SIGNED_NODE,	NULL, (INT32) Sizeof_long);
    Ulongtype  = fixed_type(LONG_NODE | UNSIGNED_NODE,	NULL, (INT32) Sizeof_long);
    Floattype  = fixed_type(FLOAT_NODE,			NULL, (INT32) SIZEOF_FLOAT);
    if(Float_mode == FLOAT_SINGLE)
	{
	Doubletype = fixed_type(DOUBLE_NODE,		NULL, (INT32) SIZEOF_S_DBLE);
	Ldoubletype = fixed_type(LDOUBLE_NODE,		NULL, (INT32) SIZEOF_S_LDBLE);
	}
    else
	{
	Doubletype = fixed_type(DOUBLE_NODE,		NULL, (INT32) SIZEOF_L_DBLE);
	Ldoubletype = fixed_type(LDOUBLE_NODE,		NULL, (INT32) SIZEOF_L_LDBLE);
	}
    Ifunctype =  fixed_type(FUNC_NODE,			Inttype,  (INT32) 0);

    /* Define TRUE and FALSE nodes for expression evaluation */

    True = fixed_type(INT_NODE | SIGNED_NODE, Inttype, (INT32) TRUE);
    True->info.storage_class = SC_CONSTANT;

    False = fixed_type(INT_NODE | SIGNED_NODE, Inttype, (INT32) FALSE);
    False->info.storage_class = SC_CONSTANT;
    }


/*
 *	Lookup_node -- Find a symbol/type in the symbol table.
 *
 *	Return NULL if the label isn't found in the appropriate namespace.
 */
PUBLIC	struct	id_node	*
lookup_node(name, name_space)
  char	*name;
  int	name_space;
    {
    extern int			hash();
    register struct id_node	*node;

    node = bucket[(hash(name, MAX_SYM_BUCKETS) + name_space) % MAX_SYM_BUCKETS];
    for (; node != NULL; node = node->bucket.next_ptr)
	{
	if ((node->name != NULL) && (strcmp(name, node->name) == 0))
	    return (node);
	if (node->type & LAST_NODE)
	    break;	/* End node? */
	}
    return (NULL);
    }


/*
 *	New_name_space
 *
 *	Return a new, unused, name space.  This gives an error message (once)
 *	if we ever re-use namespaces.
 *
 *	Reset name space counter by passing a FALSE value.
 */
int
new_name_space(getspace)
  int	getspace;		/* TRUE = Generate new name; FALSE = Reset */
    {
    extern	VOID	error();

    static int curr_name_space = NS_MEMBER;

    if (getspace)
	{
	if (++curr_name_space == MAX_SYM_BUCKETS)
	    error("Too many structure types declared");
	}
    else
	{
	curr_name_space = NS_MEMBER;
	}
    return (curr_name_space);
    }


/*
 *	Scan_level
 *
 *	Scan the specified level of symbol table nodes, and call a
 *	user-supplied function for each occurrance.
 */
PRIVATE	VOID
scan_level(level, func)
  struct id_node	*level;
  VOID			(*func)();
    {
    struct id_node *scan;

    for (scan = level; scan != NULL; scan = scan->level_ptr)
	(*func)(scan);
    }


/*
 *	Scan_ripup
 *
 *	Scan the local level of symbol table nodes, and call a user-supplied
 *	function for each occurrance.
 */
PUBLIC	VOID
scan_ripup(func)
  VOID (*func)();
    {
    extern VOID		scan_level();

    scan_level(local_list, func);
    }


/*
 *	scan_global
 *
 *	Scan the global_list and call a user-supplied function for each node.
 *	When finished, the global_list is deleted.
 */
PUBLIC	VOID
scan_global(func)
  VOID (*func)();
    {
    extern VOID		delete_level();
    extern VOID		scan_level();

    scan_level(global_list, func);
    delete_level(global_list);
    global_list = NULL;
    }


/*
 *	scan_literal
 *
 *	Scan the literal_list and call a user-supplied function for each node.
 */
PUBLIC	VOID
scan_literals(func)
  VOID (*func)();
    {
    extern VOID		scan_level();

    scan_level(literal_list, func);
    }

/*
 *	Emit floating point literals.
 */
PUBLIC VOID
emit_literals(sym)
  struct id_node		*sym;
    {
    extern VOID			align();
    extern VOID			instruction();
    extern struct id_node	*lookup_node();
    extern VOID			printlabel();

    struct id_node		*p;
/*
 *	First see if 32 bit literal.
 */
    if (((p = lookup_node(sym->name, NS_FP32LITS)) != NULL) &&
	(p->info.storage_class == SC_LITERAL)		 &&
	(p->type & REFERENCED_NODE)			 &&
	!(p->type & INITIALIZED_NODE))
	{
	p->type |= INITIALIZED_NODE;	/* Print only once */

	align();
	printlabel(p->address.surrogate);
	instruction(TOP_REAL32, ABS, p->name, NULL, (INT32) 0, NULL);
	}
/*
 *	Otherwise is it a 64 bit literal?
 */
    else if (((p = lookup_node(sym->name, NS_FP64LITS)) != NULL) &&
	(p->info.storage_class == SC_LITERAL)		 &&
	(p->type & REFERENCED_NODE)			 &&
	!(p->type & INITIALIZED_NODE))
	{
	p->type |= INITIALIZED_NODE;	/* Print only once */

	align();
	printlabel(p->address.surrogate);
	instruction(TOP_REAL64, ABS, p->name, NULL, (INT32) 0, NULL);
	}
    }


/*
 *	del_literal()
 *
 *	Rip up the literal list and release it.
 */
PUBLIC VOID
del_literals()
    {
    extern VOID		delete_level();

    delete_level(literal_list);
    literal_list = NULL;
    }

/* End of TCX11.C */
